(***********************************************************************)
(* Lib.v   adpated from SfLib.v and CPDTtactics.v                      *)
(***********************************************************************)
Require Omega.
Require Export Bool List Arith Arith.EqNat.

Definition relation (T : Type) := T -> T -> Prop.

Definition admit {T: Type} : T. Admitted.

Require String. Open Scope string_scope.

Ltac move_to_top x :=  
  match reverse goal with | H : _ |- _ => try move x after H  end.

Tactic Notation "assert_eq" ident(x) constr(v) :=
  let H := fresh in (assert (H : x = v) by reflexivity; clear H).

Tactic Notation "Case_aux" ident(x) constr(name) :=
  first [ set (x := name);  move_to_top x
        | assert_eq x name; move_to_top x
        | fail 1 "[working on a different case]"].

Tactic Notation "Case"        constr(name) := Case_aux Case        name.
Tactic Notation "SCase"       constr(name) := Case_aux SCase       name.
Tactic Notation "SSCase"      constr(name) := Case_aux SSCase      name.
Tactic Notation "SSSCase"     constr(name) := Case_aux SSSCase     name.
Tactic Notation "SSSSCase"    constr(name) := Case_aux SSSSCase    name.
Tactic Notation "SSSSSCase"   constr(name) := Case_aux SSSSSCase   name.
Tactic Notation "SSSSSSCase"  constr(name) := Case_aux SSSSSSCase  name.
Tactic Notation "SSSSSSSCase" constr(name) := Case_aux SSSSSSSCase name.

Inductive id : Type := 
  Id :> nat -> id.

Theorem ex_falso_quodlibet : forall (P:Prop), False -> P.
Proof. intuition. Qed.

Theorem eq_id_dec : forall id1 id2 : id, {id1 = id2} + {id1 <> id2}.
Proof. decide equality. decide equality.
Defined.

Lemma eq_id : forall (T:Type) (x : id) (p q : T), 
              (if eq_id_dec x x then p else q) = p. 
Proof. intros; 
  destruct (eq_id_dec x x); [reflexivity|apply ex_falso_quodlibet;auto].
Qed.

Lemma neq_id : forall (T:Type) (x y : id) (p q:T), 
               x <> y -> 
               (if eq_id_dec x y then p else q) = q. 
Proof.
  intros;
  destruct (eq_id_dec x y); [apply ex_falso_quodlibet;auto|reflexivity].
Qed.

Definition partial_map (A:Type) := id -> option A.

Definition empty {A:Type} : partial_map A := (fun _ => None). 

Notation "'\empty'" := empty.

Definition extend {A:Type} (Gamma : partial_map A) (x:id) (T : A) :=
           fun x' => if eq_id_dec x x' then Some T else Gamma x'.

Lemma extend_eq : forall (A : Type) (ctxt: partial_map A) (x : id) (T : A),
  (extend ctxt x T) x = Some T.
Proof.
  intros. unfold extend. rewrite eq_id; auto. 
Qed.

Lemma extend_neq : forall (A : Type) (ctxt: partial_map A) (x1 : id) (T : A) (x2 : id),
                   x2 <> x1 ->
                   (extend ctxt x2 T) x1 = ctxt x1.
Proof.
  intros. unfold extend. rewrite neq_id; auto. 
Qed.

Lemma extend_shadow : forall (A : Type) (ctxt: partial_map A) (t1 t2 : A) (x1 x2 : id),
      extend (extend ctxt x2 t1) x2 t2 x1 = extend ctxt x2 t2 x1.
Proof with auto.
  intros. unfold extend. destruct (eq_id_dec x2 x1)...
Qed. 


Ltac remove_exts tac :=
  let rec go' :=
      match goal with
        | [ |- (extend _ ?X _ ?X) = _ ] => rewrite extend_eq; auto; go'
        | [ |- (extend _ ?X _ ?Y) = _ ] => rewrite extend_neq; go'
        | [ |- ?X <> ?Y               ] => discriminate; go'
        | _                             => auto
      end
  in match goal with
       | [ |- context[(extend _ _ _ _ = _)] ] => go'
       | _                                    => tac
     end.

Ltac smash2 tac1 tac2 :=
  let rec go' :=
      match goal with
        | [ |- context[(if eq_id_dec ?x ?x then _ else _) = _] ] => rewrite eq_id; eauto; go'
        | [ |- context[(if eq_id_dec ?x ?y then _ else _) = _] ] => rewrite neq_id;eauto; go'
        | [ |- ?x <> ?y ] => discriminate; go'
        | _ => auto
      end
  in (tac1; (match goal with
               | [ |- context[(extend _ _ _ _ = _)] ] => unfold extend; go'
               | _                                    => try (eapply tac2; eauto)
             end)).

Ltac prim_neq_in_hyp H1 x y :=
  let H := fresh in (
    let H' := fresh in (
      assert (H : Id x <> Id y) by (intros H'; inversion H'));
    rewrite (@neq_id _ x y _ _ H) in H1; clear H
  ). 

Ltac prim_neq_in_hyp' H1 x y :=
  let H := fresh in (
      assert (H : x <> y) by discriminate;
      rewrite (@neq_id _ x y _ _ H) in H1; clear H
  ). 

Ltac prim_neq_in_goal x y :=
  let H := fresh in (
    let H' := fresh in (
      assert (H : Id x <> Id y) by (intros H'; inversion H'));
    rewrite (@neq_id _ x y _ _ H); clear H
  ).

Ltac prim_neq_in_goal' x y :=
  let H := fresh in (
      assert (H : x <> y) by discriminate;
      rewrite (@neq_id _ x y _ _ H); clear H
  ).


Ltac remove_id_in_hyps :=
  let rec go' :=
     lazymatch goal with
      | [H: context[(eq_id_dec ?x ?x)] |- _ ] => 
        rewrite eq_id in H; go'
      | [H: context[(eq_id_dec ?x ?y)] |- _ ] => 
        let H' := fresh in (
          assert (H' : x <> y) by discriminate;
          rewrite (@neq_id _ x y _ _ H') in H; clear H');go'
      | _ => idtac                                                                 
      end
    in (lazymatch goal with
       | [H: context[(eq_id_dec _ _)] |- _ ] => go'
       | _ => fail "No clauses for the match"
       end).

Ltac remove_id_in_goal :=
  let rec go' :=
     lazymatch goal with
      | [ |- context[(eq_id_dec ?x ?x)] ] => rewrite eq_id; go'
      | [ |- context[(eq_id_dec ?x ?y)] ] => 
        let H' := fresh in (
          assert (H' : x <> y) by discriminate;
          rewrite (@neq_id _ x y _ _ H'); clear H'); go'
      | _ => idtac                                                                 
      end
    in (lazymatch goal with
       | [ |- context[(eq_id_dec _ _)] ] => go'
       | _ => fail "No clauses for the match"
       end).

Ltac remove_ids :=
  let rec go' :=
      lazymatch goal with
      | [H: context[(eq_id_dec ?x ?x)] |- _ ] => 
        rewrite eq_id in H; go'
      | [H: context[(eq_id_dec ?x ?y)] |- _ ] => 
        let H' := fresh in (
          assert (H' : x <> y) by discriminate;
          rewrite (@neq_id _ x y _ _ H') in H; clear H');go'
      | [ |- context[(eq_id_dec ?x ?x)] ]     => 
        rewrite eq_id; go'
      | [ |- context[(eq_id_dec ?x ?y)] ]     => 
        let H' := fresh in (
          assert (H' : x <> y) by discriminate;
          rewrite (@neq_id _ x y _ _ H'); clear H'); go'
      | _ => idtac                                                                 
      end
   in (lazymatch goal with
      | [                            |- context[(eq_id_dec _ _)] ] => go'
      | [H: context[(eq_id_dec _ _)] |- _                        ] => go'
      | _ => fail "No clauses for the match"
       end).
      

Ltac inv H := inversion H; subst; auto.

Tactic Notation "subst" "++" :=
    repeat (
      match goal with
      | x : _ |- _ => subst x
      end);
    cbv zeta beta in *.

(** ** Hiding hypotheses 
        This is adapted from LibTactics.v in the SF library. *)

Definition ltac_some (T:Type) (x:T) := x.

Notation "'Some_T'" := (@ltac_some _ _).

Lemma ltac_some_Term : forall (x:Type),
                         x = (@ltac_some _ x).
Proof. auto. Qed.

Lemma ltac_some_Hide : forall (x:Type),
                       x -> (@ltac_some _ x).
Proof. auto. Qed.

Lemma ltac_some_Show : forall (x:Type),
                      (@ltac_some _ x) -> x.
Proof. auto. Qed.  

Tactic Notation "hide_term" constr(X) :=
  change X with (@ltac_some _ X).
Tactic Notation "show_term" constr(X) :=
  change (@ltac_some _ X) with X.
Tactic Notation "show_term" :=
  unfold ltac_some.
Tactic Notation "hide_term" constr(X) "in" hyp(H) :=
  change X with (@ltac_some _ X) in H.
Tactic Notation "show_term" constr(X) "in" hyp(H) :=
  change (@ltac_some _ X) with X in H.
Tactic Notation "show_term" "in" hyp(H) :=
  unfold ltac_some in H.
Tactic Notation "hide" "hyp" hyp(H) :=
  apply ltac_some_Hide in H.
Tactic Notation "show" "hyp" hyp(H) :=
  apply ltac_some_Show in H.  

Tactic Notation "show" "all" "hyps" :=
  repeat (match goal with
              H: @ltac_some _ _ |- _ => show hyp H
          end).

Ltac remove_duplicate :=
  match goal with
    | [ H1 : ?A |- _ ] =>
      match goal with
        | [ H2 : ?B |- _ ] =>
          match H1 with
            | H2 => fail 2
            | _  => unify A B ; (clear H2 || clear H1)
          end
      end
  end.

Ltac remove_duplicates :=
     repeat (match goal with
                 | [ H1 : ?A |- _ ] =>
                 match goal with
                     | [ H2 : ?B |- _ ] =>
                     match H1 with
                         | H2 => fail 2
                         | _  => unify A B ; (clear H2 || clear H1)
                     end
                 end
             end).

Set Implicit Arguments.

Ltac gen_dep E :=
     let go E := (generalize dependent E) in
         match E with
         | (?X1, ?X2, ?X3, ?X4, ?X5, ?X6, ?X7, ?X8, ?X9, ?X10) => 
           go X10;go X9; go X8; go X7; go X6; go X5; go X4; go X3; go X2; go X1
         | (?X1, ?X2, ?X3, ?X4, ?X5, ?X6, ?X7, ?X8, ?X9)       => 
           go X9; go X8; go X7; go X6; go X5; go X4; go X3; go X2; go X1
         | (?X1, ?X2, ?X3, ?X4, ?X5, ?X6, ?X7, ?X8)            => 
           go X8; go X7; go X6; go X5; go X4; go X3; go X2; go X1
         | (?X1, ?X2, ?X3, ?X4, ?X5, ?X6, ?X7)                 => 
           go X7; go X6; go X5; go X4; go X3; go X2; go X1
         | (?X1, ?X2, ?X3, ?X4, ?X5, ?X6)                      => 
           go X6; go X5; go X4; go X3; go X2; go X1
         | (?X1, ?X2, ?X3, ?X4, ?X5)                           => 
           go X5; go X4; go X3; go X2; go X1
         | (?X1, ?X2, ?X3, ?X4)                                => 
           go X4; go X3; go X2; go X1
         | (?X1, ?X2, ?X3)                                     => 
           go X3; go X2; go X1
         | (?X1, ?X2)                                          => 
           go X2; go X1
         | ?X1                                                 => 
           go X1
         | _                                                   => idtac
         end. 

Tactic Notation "generalize" "dependent" ident(X1) := 
                 gen_dep X1.
Tactic Notation "generalize" "dependent" ident(X1) ident(X2) :=
                 gen_dep (X1,X2).
Tactic Notation "generalize" "dependent" ident(X1) ident(X2) ident(X3) :=
                 gen_dep (X1,X2,X3).
Tactic Notation "generalize" "dependent" ident(X1) ident(X2) ident(X3) ident(X4) :=
                 gen_dep (X1,X2,X3,X4).
Tactic Notation "generalize" "dependent" ident(X1) ident(X2) ident(X3) ident(X4) ident(X5) :=
                 gen_dep (X1,X2,X3,X4,X5).
Tactic Notation "generalize" "dependent" ident(X1) ident(X2) ident(X3) ident(X4) ident(X5) ident(X6) :=
                 gen_dep (X1,X2,X3,X4,X5,X6).
Tactic Notation "generalize" "dependent" ident(X1) ident(X2) ident(X3) ident(X4) ident(X5) ident(X6) ident(X7) :=
                 gen_dep (X1,X2,X3,X4,X5,X6,X7).
Tactic Notation "generalize" "dependent" ident(X1) ident(X2) ident(X3) ident(X4) ident(X5) ident(X6) ident(X7) ident(X8) :=
                 gen_dep (X1,X2,X3,X4,X5,X6,X7,X8).
Tactic Notation "generalize" "dependent" ident(X1) ident(X2) ident(X3) ident(X4) ident(X5) ident(X6) ident(X7) ident(X8) ident(X9) :=
                 gen_dep (X1,X2,X3,X4,X5,X6,X7,X8,X9).
Tactic Notation "generalize" "dependent" ident(X1) ident(X2) ident(X3) ident(X4) ident(X5) ident(X6) ident(X7) ident(X8) ident(X9) ident(X10) :=
                 gen_dep (X1,X2,X3,X4,X5,X6,X7,X8,X9,X10).

(* ********************************************************** *)


(* Section Relations.
   Require Import Coq.Init.Wf. *)




(*
Lemma acc_irr : forall {T : Type} (S : T -> T -> Prop) {t : T}, 
      Acc S t -> ~ S t t.
Proof.
  intros ? ? ? HAccSt; 
  induction HAccSt as [t _ H]; 
  intro HStt; exact (H t HStt HStt).
Qed.

Corollary wf_irr : forall {T : Type} {S : T -> T -> Prop}, 
 well_founded S -> forall {t : T}, ~ S t t.
Proof. 
  red; intros ? ? Hwf ? HStt; exact (@acc_irr T S t (Hwf t) HStt). 
Qed.

Lemma irr_neq : forall {T : Type} (S : T -> T -> Prop),
 (forall {t : T}, ~ S t t) -> forall {t t' : T}, S t t' -> t <> t'.
Proof. 
  intros ? ? HnotStt ? ? HStt; red; 
  intros; subst; exact (HnotStt t' HStt). 
Qed.

Corollary wf_neq : forall {T : Type} (S : T -> T -> Prop), 
 well_founded S -> forall {t t' : T}, S t t' -> t <> t'.
Proof. eauto using @irr_neq, @wf_irr. Qed. 

End Relations.

*)


(* Copyright (c) 2008-2012, Adam Chlipala
 * 
 * This work is licensed under a
 * Creative Commons Attribution-Noncommercial-No Derivative Works 3.0
 * Unported License.
 * The license text is available at:
 *   http://creativecommons.org/licenses/by-nc-nd/3.0/
 *)

Require Import Eqdep Omega.

Set Implicit Arguments.

Ltac omegaContradiction :=
  cut False; [contradiction|omega].

(** A version of [injection] that does some standard simplifications afterward: clear the hypothesis in question, bring the new facts above the double line, and attempt substitution for known variables. *)
Ltac inject H := injection H; clear H; intros; try subst.

(** Try calling tactic function [f] on all hypotheses, keeping the first application that doesn't fail. *)
Ltac appHyps f :=
  match goal with
    | [ H : _ |- _ ] => f H
  end.

(** Succeed iff [x] is in the list [ls], represented with left-associated nested tuples. *)
Ltac inList x ls :=
  match ls with
    | x => idtac
    | (_, x) => idtac
    | (?LS, _) => inList x LS
  end.

(** Try calling tactic function [f] on every element of tupled list [ls], keeping the first call not to fail. *)
Ltac app f ls :=
  match ls with
    | (?LS, ?X) => f X || app f LS || fail 1
    | _ => f ls
  end.

(** Run [f] on every element of [ls], not just the first that doesn't fail. *)
Ltac all f ls :=
  match ls with
    | (?LS, ?X) => f X; all f LS
    | (_, _) => fail 1
    | _ => f ls
  end.

(** Workhorse tactic to simplify hypotheses for a variety of proofs.
   * Argument [invOne] is a tuple-list of predicates for which we always do inversion automatically. *)
Ltac simplHyp invOne :=
  (** Helper function to do inversion on certain hypotheses, where [H] is the hypothesis and [F] its head symbol *)
  let invert H F :=
    (** We only proceed for those predicates in [invOne]. *)
    inList F invOne;
    (** This case covers an inversion that succeeds immediately, meaning no constructors of [F] applied. *)
      (inversion H; fail)
    (** Otherwise, we only proceed if inversion eliminates all but one constructor case. *)
      || (inversion H; [idtac]; clear H; try subst) in

  match goal with
    (** Eliminate all existential hypotheses. *)
    | [ H : ex _ |- _ ] => destruct H

    (** Find opportunities to take advantage of injectivity of data constructors, for several different arities. *)
    | [ H : ?F ?X = ?F ?Y |- ?G ] =>
      (** This first branch of the [||] fails the whole attempt iff the arguments of the constructor applications are already easy to prove equal. *)
      (assert (X = Y); [ assumption | fail 1 ])
      (** If we pass that filter, then we use injection on [H] and do some simplification as in [inject].
         * The odd-looking check of the goal form is to avoid cases where [injection] gives a more complex result because of dependent typing, which we aren't equipped to handle here. *)
      || (injection H;
        match goal with
          | [ |- X = Y -> G ] =>
            try clear H; intros; try subst
        end)
    | [ H : ?F ?X ?U = ?F ?Y ?V |- ?G ] =>
      (assert (X = Y); [ assumption
        | assert (U = V); [ assumption | fail 1 ] ])
      || (injection H;
        match goal with
          | [ |- U = V -> X = Y -> G ] =>
            try clear H; intros; try subst
        end)

    (** Consider some different arities of a predicate [F] in a hypothesis that we might want to invert. *)
    | [ H : ?F _ |- _ ] => invert H F
    | [ H : ?F _ _ |- _ ] => invert H F
    | [ H : ?F _ _ _ |- _ ] => invert H F
    | [ H : ?F _ _ _ _ |- _ ] => invert H F
    | [ H : ?F _ _ _ _ _ |- _ ] => invert H F

    (** Use an (axiom-dependent!) inversion principle for dependent pairs, from the standard library. *)
    | [ H : existT _ ?T _ = existT _ ?T _ |- _ ] => generalize (inj_pair2 _ _ _ _ _ H); clear H

    (** If we're not ready to use that principle yet, try the standard inversion, which often enables the previous rule. *)
    | [ H : existT _ _ _ = existT _ _ _ |- _ ] => inversion H; clear H

    (** Similar logic to the cases for constructor injectivity above, but specialized to [Some], since the above cases won't deal with polymorphic constructors. *)
    | [ H : Some _ = Some _ |- _ ] => injection H; clear H
  end.

(** Find some hypothesis to rewrite with, ensuring that [auto] proves all of the extra subgoals added by [rewrite]. *)
Ltac rewriteHyp :=
  match goal with
    | [ H : _ |- _ ] => rewrite H by solve [ auto ]
  end.

(** Combine [autorewrite] with automatic hypothesis rewrites. *)
Ltac rewriterP := repeat (rewriteHyp; autorewrite with core in *).
Ltac rewriter := autorewrite with core in *; rewriterP.

(** This one is just so darned useful, let's add it as a hint here. *)
Hint Rewrite app_ass.

(** Devious marker predicate to use for encoding state within proof goals *)
Definition done (T : Type) (x : T) := True.

(** Try a new instantiation of a universally quantified fact, proved by [e].
   * [trace] is an accumulator recording which instantiations we choose. *)
Ltac inster e trace :=
  (** Does [e] have any quantifiers left? *)
  match type of e with
    | forall x : _, _ =>
      (** Yes, so let's pick the first context variable of the right type. *)
      match goal with
        | [ H : _ |- _ ] =>
          inster (e H) (trace, H)
        | _ => fail 2
      end
    | _ =>
      (** No more quantifiers, so now we check if the trace we computed was already used. *)
      match trace with
        | (_, _) =>
          (** We only reach this case if the trace is nonempty, ensuring that [inster] fails if no progress can be made. *)
          match goal with
            | [ H : done (trace, _) |- _ ] =>
              (** Uh oh, found a record of this trace in the context!  Abort to backtrack to try another trace. *)
              fail 1
            | _ =>
              (** What is the type of the proof [e] now? *)
              let T := type of e in
                match type of T with
                  | Prop =>
                    (** [e] should be thought of as a proof, so let's add it to the context, and also add a new marker hypothesis recording our choice of trace. *)
                    generalize e; intro;
                      assert (done (trace, tt)) by constructor
                  | _ =>
                    (** [e] is something beside a proof.  Better make sure no element of our current trace was generated by a previous call to [inster], or we might get stuck in an infinite loop!  (We store previous [inster] terms in second positions of tuples used as arguments to [done] in hypotheses.  Proofs instantiated by [inster] merely use [tt] in such positions.) *)
                    all ltac:(fun X =>
                      match goal with
                        | [ H : done (_, X) |- _ ] => fail 1
                        | _ => idtac
                      end) trace;
                    (** Pick a new name for our new instantiation. *)
                    let i := fresh "i" in (pose (i := e);
                      assert (done (trace, i)) by constructor)
                end
          end
      end
  end.

(** After a round of application with the above, we will have a lot of junk [done] markers to clean up; hence this tactic. *)
Ltac un_done :=
  repeat match goal with
           | [ H : done _ |- _ ] => clear H
         end.

Require Import JMeq.

(** A more parameterized version of the famous [crush].  Extra arguments are:
   * - A tuple-list of lemmas we try [inster]-ing 
   * - A tuple-list of predicates we try inversion for *)
Ltac crush' lemmas invOne :=
  (** A useful combination of standard automation *)
  let sintuition := simpl in *; intuition; try subst;
    repeat (simplHyp invOne; intuition; try subst); try congruence in

  (** A fancier version of [rewriter] from above, which uses [crush'] to discharge side conditions *)
  let rewriter := autorewrite with core in *;
    repeat (match goal with
              | [ H : ?P |- _ ] =>
                match P with
                  | context[JMeq] => fail 1 (** JMeq is too fancy to deal with here. *)
                  | _ => rewrite H by crush' lemmas invOne
                end
            end; autorewrite with core in *) in

  (** Now the main sequence of heuristics: *)
    (sintuition; rewriter;
      match lemmas with
        | false => idtac (** No lemmas?  Nothing to do here *)
        | _ =>
          (** Try a loop of instantiating lemmas... *)
          repeat ((app ltac:(fun L => inster L L) lemmas
          (** ...or instantiating hypotheses... *)
            || appHyps ltac:(fun L => inster L L));
          (** ...and then simplifying hypotheses. *)
          repeat (simplHyp invOne; intuition)); un_done
      end;
      sintuition; rewriter; sintuition;
      (** End with a last attempt to prove an arithmetic fact with [omega], or prove any sort of fact in a context that is contradictory by reasoning that [omega] can do. *)
      try omega; try (elimtype False; omega)).

(** [crush] instantiates [crush'] with the simplest possible parameters. *)
Ltac crush := crush' false fail.

(** * Wrap Program's [dependent destruction] in a slightly more pleasant form *)

Require Import Program.Equality.

(** Run [dependent destruction] on [E] and look for opportunities to simplify the result.
   The weird introduction of [x] helps get around limitations of [dependent destruction], in terms of which sorts of arguments it will accept (e.g., variables bound to hypotheses within Ltac [match]es). *)
Ltac dep_destruct E :=
  let x := fresh "x" in
    remember E as x; simpl in x; dependent destruction x;
      try match goal with
            | [ H : _ = E |- _ ] => try rewrite <- H in *; clear H
          end.

(** Nuke all hypotheses that we can get away with, without invalidating the goal statement. *)
Ltac clear_all :=
  repeat match goal with
           | [ H : _ |- _ ] => clear H
         end.

(** Instantiate a quantifier in a hypothesis [H] with value [v], or, if [v] doesn't have the right type, 
     with a new unification variable.
   * Also prove the lefthand sides of any implications that this exposes,
     simplifying [H] to leave out those implications. *)
Ltac guess v H :=
  repeat match type of H with
           | forall x : ?T, _ =>
             match type of T with
               | Prop =>
                 (let H' := fresh "H'" in
                   assert (H' : T); [
                     solve [ eauto 6 ]
                     | specialize (H H'); clear H' ])
                 || fail 1
               | _ =>
                 specialize (H v)
                 || let x := fresh "x" in
                   evar (x : T);
                   let x' := eval unfold x in x in
                     clear x; specialize (H x')
             end
         end.

(** Version of [guess] that leaves the original [H] intact *)
Ltac guessKeep v H :=
  let H' := fresh "H'" in
    generalize H; intro H'; guess v H'.



Definition iffb (b1 b2 : bool) : bool := 
           match b1,b2 with
           |  true,true     => true
           | false,false    => true
           | _,_            => false
           end.

Definition impb (b1 b2 : bool) : bool :=
           match b1,b2 with
           |   true,false   => false
           |      _,_       => true
           end.

Fact andb_true_ : forall a b,
     a && b = true -> a = true /\ b = true.
Proof.
  intros a b Hab; split; [destruct a|destruct b]; crush.
  destruct a; inversion Hab.
Qed.

Fact andb_true_in : forall a b,
  a = true /\ b = true -> a && b = true.
Proof. intros a b [Hat Hbt]; rewrite Hat; rewrite Hbt; auto. Qed.

Fact andb_false_ : forall a b,
  a && b = false -> a = false \/ b = false.
Proof. intros a b Hab; destruct a; destruct b; destruct Hab; crush. Qed.

Fact orb_true_ : forall a b,
  a || b = true -> a = true \/ b = true.
Proof. intros a b Hab; destruct a; destruct b; destruct Hab; crush. Qed.

Fact orb_false_elim_ : forall a b,
  a || b = false -> a = false /\ b = false.
Proof. intros a b Hab; destruct a; destruct b; destruct Hab; crush. Qed.
 
Fact orb_false_l : forall a b,
  a || b = true /\ a = false -> b = true.
Proof. intros a b [Hab Haf]; destruct a; destruct b; destruct Hab; crush. Qed.

Fact orb_false_r : forall a b,
  a || b = true /\ b = false -> a = true.
Proof. intros a b [Hab Hbf]; destruct a; destruct b; destruct Hab; crush. Qed.

Fact imp_and_ : forall a b,
  negb (a && negb b) = true -> impb a b = true.
Proof. intros a b Hnegb; destruct a; destruct b; crush. Qed.

Fact contrapose_l : forall a b,
     impb a b = true -> impb (negb b) (negb a) = true.
Proof. intros a b Hab; destruct a; destruct b; crush. Qed.

Fact contrapose_r : forall a b,
     impb (negb b) (negb a) = true -> impb a b = true.
Proof. intros a b Hab; destruct a; destruct b; crush. Qed.

Fact negb_true : forall a,
     negb a = true -> a = false.
Proof. intros a Hnegb; destruct a; crush. Qed.  

Fact negb_dec_f : forall {T: Type} (p q : T),
    (if false then p else q) = q.
Proof. crush. Qed.

Fact negb_dec_t : forall {T: Type} (p q : T),
    (if true then p else q) = p.
Proof. crush. Qed.

(** Extra Goodies: Induction on the nats *)


Theorem strong_induction : forall (P : nat -> Prop),
                             (forall n : nat,
                                (forall m, m < n -> P m) -> P n) ->
                             (forall n : nat, P n).
Proof.
 assert (Peano_aux_1 : forall (P : nat -> Prop),
                       (forall n : nat, (forall m : nat, m < n -> P m) -> P n) ->
                                 (forall n : nat,
                                    (forall m : nat, ((m < n) -> P m)))).
        induction n as [ | n' IHn' ]; [intros m H1; inversion H1 |]. 
        intros.
        assert (H_dec: m < n' \/ m = n') by omega.
        inversion H_dec.
        apply IHn'; auto. subst.
        apply (H n'); auto.
 assert (Peano_aux_2 : forall (P : nat -> Prop),
                       (forall n : nat, (forall m : nat, ((m < n) -> P m))) ->
                         (forall n : nat, P n))by (intros P H n; eauto). 
  intros.
  apply Peano_aux_2.
  apply Peano_aux_1; auto. 
Qed.

Lemma strong_induction_alt : forall (P: nat -> Prop),
                               (forall n : nat, (forall m : nat, m < n -> P m) -> P n) ->
                               (forall n : nat, P n).
Proof.
  intros ? H.
  assert (H_aux : forall n : nat, 
                  (forall m : nat, m < n -> P m)) by
           (induction n as [ |n' IHn']; intros; [omegaContradiction
                                                |apply H; intros; apply IHn'; omega]).
  intro n; apply H_aux with (S n); omega.
Qed.

Notation " l /{ x } " := (remove eq_id_dec x l)
                           (at level 59, format "'[hv ' l /{ x } ']'").

Notation "l [<=] m" := (@incl id l m)
                         (at level 70, no associativity, format "l  [<=]  m").

Hint Resolve app_assoc app_assoc_reverse: datatypes.
Hint Resolve app_comm_cons app_cons_not_nil: datatypes.
Hint Immediate app_eq_nil: datatypes.
Hint Resolve app_eq_unit app_inj_tail: datatypes.
Hint Resolve in_eq in_cons in_inv in_nil in_app_or in_or_app: datatypes. 

Theorem id_dec_in `(x : id) (l : list id) :
    {In x l} + {~ In x l}.
Proof. apply (in_dec eq_id_dec x l). Defined.

Hint Resolve id_dec_in : datatypes.

Ltac dauto  := auto with datatypes.
Ltac deauto := eauto with datatypes.
Ltac idauto := try (intuition;dauto;deauto).
Ltac list_smash := 
  first [ intuition;(dauto;deauto)
        | firstorder with datatypes v62
        | try (firstorder;fail)
        | fail 1 "Tactic [list_smash] failed"].

Lemma in_cons_tl : forall (X : Type) (l : list X) (x y : X), 
  x <> y ->
  In y (x::l) ->
  In y l.
Proof.
  intros ? ? ? ? ?; simpl; list_smash. 
Qed.

Hint Resolve in_cons_tl : datatypes.

Lemma in_app_l : forall (X : Type) (x : X) (l1 l2 : list X),
  In x l1 ->
  In x (l1 ++ l2).
Proof. list_smash. Qed.

Lemma in_app_r : forall (X : Type) (x : X) (l1 l2 : list X),
  In x l2 ->
  In x (l1 ++ l2).
Proof. list_smash. Qed.

Hint Resolve in_app_l in_app_r : datatypes.

Lemma notin_app_intro : forall  (X : Type) (x : X) (l1 l2 : list X),
  ~ (In x l1) ->
  ~ (In x l2) ->
  ~ (In x (l1 ++ l2)).
Proof. list_smash. eapply in_app_or in H1 as [H1 | H1];auto.
Qed.

Lemma notin_app_inv : forall  (X : Type) (x : X) (l1 l2 : list X),
 ~ (In x (l1 ++ l2)) ->
 ~ (In x l1) /\ ~ (In x l2).
Proof. list_smash. Qed.

Lemma notin_app_inv_l : forall  (X : Type) (x : X) (l1 l2 : list X),
  ~ (In x (l1 ++ l2)) -> ~ (In x l1).
Proof. list_smash. Qed.

Lemma notin_app_inv_r : forall  (X : Type) (x : X) (l1 l2 : list X),
  ~ (In x (l1 ++ l2)) -> ~ (In x l2).
Proof. list_smash. Qed.

Hint Resolve notin_app_intro notin_app_inv  : datatypes.
Hint Resolve notin_app_inv_l notin_app_inv_r : datatypes. 

Lemma notin_cons_hd : forall  (X : Type) (x y : X) (l : list X),
  ~ (In x (y :: l)) -> x <> y.
Proof. firstorder with datatypes. Qed.

Lemma notin_cons_tl : forall (X : Type) (x y : X) (l : list X),
  ~ (In x (y :: l)) -> ~ (In x l).
Proof. list_smash. Qed.

Lemma notin_cons_intro : forall  (X : Type) (x y : X) (l : list X),
   x <> y -> ~ (In y l) -> ~ (In y (x :: l)).
Proof. list_smash. Qed.

Hint Resolve notin_cons_hd notin_cons_tl notin_cons_intro : datatypes.

Lemma notin_nil : forall  (X : Type) (x : X),
  ~ (In x nil).
Proof. auto. Qed.

Lemma notin_singleton : forall  (X : Type) (x y : X),
  x <> y -> ~ (In y (x :: nil)).
Proof. firstorder. Qed.

Hint Resolve notin_singleton : datatypes.

Lemma notin_app_or : forall  (X : Type) (x : X) (l1 l2 : list X),
  ~ (In x (l1 ++ l2)) -> ~ (In x l1) /\ ~ (In x l2).
Proof. list_smash. Qed.

Lemma notin_or_app : forall  (X : Type) (x : X) (l1 l2 : list X),
  ~ (In x l1) /\ ~ (In x l2) -> ~ (In x (l1 ++ l2)).
Proof. intros ? ? ? ? [H1 H2]; deauto. Qed.

Ltac case_id_dec :=
    let destr_id x y := destruct (eq_id_dec x y); [try subst x | idtac] in
 repeat match goal with
        | [H:  appcontext [(eq_id_dec ?x ?x)] |- _ ] => 
          rewrite eq_id in H;list_smash
        | [ |- appcontext [(eq_id_dec ?x ?x)] ]      => 
          rewrite eq_id; list_smash
        | [ |- appcontext [(eq_id_dec ?x ?y)] ]   => 
          (destr_id x y);list_smash
        | [H:  appcontext [(eq_id_dec ?x ?y)] |- _ ] => 
          (destr_id x y);list_smash
        end.

Lemma in_rem_id : forall (x y : id) (l : list id),
    In x (l/{y}) -> In x l.
Proof.
    induction l; simpl; intros; auto.
    repeat case_id_dec; firstorder.
Qed.

Lemma in_rem_neq : forall (x y : id) (l : list id), 
    x <> y ->
    In x l ->
    In x (l/{y}).
Proof.
    induction l; simpl; intros; auto.
    repeat case_id_dec; firstorder; congruence.
Qed.

Lemma in_rem_ids_neq : forall (x y : id) (l : list id),
    In x (l/{y}) -> x <> y.
Proof.
    induction l; simpl; intros; auto.
    case_id_dec; firstorder; congruence.
Qed.

Hint Resolve in_rem_id in_rem_neq in_rem_ids_neq : datatypes.

Lemma notin_rem_id : forall (x : id) (l : list id),
    ~ In x (l/{x}).
Proof.
    induction l; simpl; intros; try case_id_dec; firstorder.
Qed.

Lemma notin_rem_neq : forall (x y : id) (l : list id),  
    x <> y -> ~ (In x (l/{y})) ->  ~ (In x l).
Proof.
    firstorder using notin_rem_id.
Qed.

Hint Resolve notin_rem_id notin_rem_neq : datatypes.

Lemma notin_rem_eq : forall (x : id) (l : list id),
    ~ (In x l) -> l/{x} = l.
Proof.
    induction l; simpl; try case_id_dec; firstorder. 
    f_equal. auto.
Qed.

Lemma notin_rem_notin : forall (x y : id) (l : list id),
    ~ (In x l) ->  ~ (In x (l/{y})).
Proof.
    induction l; simpl; try case_id_dec; firstorder. 
Qed.

Hint Resolve notin_rem_eq notin_rem_notin : datatypes.

Lemma neq_rem_cons : forall (x y : id) (l : list id),
      x <> y ->
      y :: (l/{x}) = (y :: l)/{x}.
Proof. intros; simpl; 
       destruct (in_dec eq_id_dec x l);list_smash;
       case_id_dec; firstorder.
Qed.

Hint Resolve neq_rem_cons : datatypes.

Lemma rem_app : forall (x : id) (l1 l2 : list id),
      (l1/{x}) ++ (l2/{x}) = (l1 ++ l2)/{x}.
Proof. induction l1; auto; simpl; intros.
       case_id_dec; simpl; congruence.
Qed. 

Lemma rem_eq_drop : forall (x : id) (l : list id),
                         (l/{x})/{x} = l/{x}.
Proof.
  induction l; auto; simpl; intros. 
  case_id_dec; subst; auto. 
Qed.

Lemma rem_neq_swap : forall (x y : id) (l : list id),
                     (l/{x})/{y} = (l/{y})/{x}.
Proof. 
  induction l; auto; simpl; intros.
  repeat (case_id_dec; subst; simpl); intuition congruence.
Qed.

Hint Resolve rem_app rem_eq_drop rem_neq_swap : datatypes.

Hint Unfold incl.

Lemma nil_Incl : forall (l: list id), nil [<=] l. 
Proof.
  red; simpl; intuition.
Qed.

Lemma Incl_nil : forall (l : list id),
                   l [<=] nil -> l = nil.
Proof.
  unfold incl; intros l H.
  induction l as [ | a]; simpl in *; auto. 
  elim (H a); auto. 
Qed.

Hint Resolve nil_Incl Incl_nil : datatypes.

Lemma In_Incl : forall (x : id) (l1 l2 : list id),
  In x l1 -> (l1 [<=] l2) -> In x l2.
Proof. unfold incl. auto. Qed.

Lemma Incl_cons_id : forall  (l1 l2 : list id) (x : id),
                       (x :: l1) [<=] l2 -> In x l2.
Proof. unfold incl. dauto. Qed.

Lemma Incl_cons_list : forall (l1 l2: list id) (x : id),
      (x :: l1) [<=] l2 -> l1 [<=] l2.
Proof. unfold incl. dauto. Qed.

Lemma Incl_cons_In : forall (x : id) (l : list id), 
                       l [<=] (x::l).
Proof. unfold incl. dauto. Qed.

Lemma Incl_cons_add : forall (l1 l2 : list id) (x : id),
      l1 [<=] l2 -> (x::l1) [<=] (x::l2).
Proof. unfold incl; simpl; firstorder. Qed.

Hint Resolve In_Incl Incl_cons_id Incl_cons_list Incl_cons_In Incl_cons_add : datatypes.

Lemma Incl_appL : forall (l1 l2 l3 : list id),
       (l1 ++ l2) [<=] l3 -> l1 [<=] l3.
Proof. unfold incl; deauto. Qed.

Lemma Incl_appR : forall (l1 l2 l3 : list id),
  (l1 ++ l2) [<=] l3 -> l2 [<=] l3.
Proof. unfold incl; deauto. Qed.

Hint Resolve Incl_appL Incl_appR : datatypes.

Lemma rem_Incl : forall (x : id) (l : list id), (l/{x}) [<=] l.
Proof.
  unfold incl; induction l as [|x' l IH ]; simpl; intros; auto.
  destruct (eq_id_dec x x');simpl in *;intuition.
Qed.

Lemma rem_Incl_cons : forall (x : id) (l : list id),
            ((x::l)/{x}) [<=] l.
Proof.
  simpl; intros; rewrite eq_id; auto using rem_Incl.
Qed.

Hint Resolve rem_Incl rem_Incl_cons : datatypes.

Lemma Incl_rem : forall (l1 l2 : list id) (x : id),
                   l1 [<=] l2 -> (l1/{x}) [<=] (l2/{x}).
Proof. unfold incl; 
       induction l1 as [| x' l IH]; [simpl; intuition
                                    |intros].
       list_smash.
Qed.

Lemma Incl_rem_cons : forall (l1 l2 : list id) (x : id),
  (l1/{x}) [<=] l2 ->
  l1 [<=] (x :: l2).
Proof.
  unfold incl; simpl; intros.
  case (eq_id_dec x a); auto with datatypes.
Qed.

Lemma Incl_cons_rem : forall (l1 l2 : list id) (x : id),
      l1 [<=] (x :: l2) ->
      (l1/{x}) [<=] l2.
Proof.
  intros; eapply incl_tran.
  eapply Incl_rem; eassumption.
  dauto. 
Qed.

Lemma Incl_rem_tl : forall (l1 l2 : list id) (x y : id),
      l1 [<=] l2 -> x = y ->
      (y :: l1) [<=] (y :: (l2/{x})).
Proof.
  unfold incl; simpl; intros.
  case (eq_id_dec y a); intros; subst; firstorder. 
Qed.

Hint Resolve Incl_rem Incl_rem_cons Incl_cons_rem Incl_rem_tl : datatypes.

Lemma in_singleton : forall (x y : id),
      In x (y :: nil) -> x = y.
Proof.
  simpl; intuition congruence.
Qed.

Lemma In_Incl_Iff : forall (l : list id) (x : id),
      ((x :: nil) [<=] l) <-> (In x l).
Proof.
  split. dauto. 
  unfold incl; simpl; intros. 
  intuition congruence.
Qed. 

(** We need to be able to pick a new id, given a set of ids. *)

Definition update (E : Type) (f : id -> E) (x : id) (e : E) : id -> E :=
  fun x' => if eq_id_dec x x' then e else f x'.

Theorem update_eq : forall (E : Type) f x (n : E),
  (update f x n) x = n.
Proof.
  intros. unfold update. apply eq_id.
Qed. 

Theorem update_neq : forall (E : Type) f x2 x1 (n : E),
  x2 <> x1 ->
  (update f x2 n) x1 = (f x1).
Proof.
  intros. unfold update. apply neq_id; auto. 
Qed.

Theorem update_shadow : forall (E : Type) f x1 x2 (n1 n2: E),
   (update (update f x2 n1) x2 n2) x1 = (update f x2 n2) x1.
Proof.
  intros. unfold update.
  destruct (eq_id_dec x2 x1); reflexivity.
Qed.

Theorem update_same : forall (E : Type) f x1 x2 (n : E),
  f x1 = n ->
  (update f x1 n) x2 = f x2.
Proof.
  intros.
  unfold update. 
  destruct (eq_id_dec x1 x2); subst; reflexivity.
Qed.

Theorem update_permute : forall (E : Type) f x1 x2 x3 (n1 n2 : E),
  x2 <> x1 ->
  (update (update f x2 n1) x1 n2) x3 = (update (update f x1 n2) x2 n1) x3.
Proof.
  intros. unfold update.
  destruct (eq_id_dec x1 x3); 
    destruct (eq_id_dec x2 x3); subst;
  try (reflexivity).
  unfold not in H. apply ex_falso_quodlibet. apply H. reflexivity.
Qed.

Ltac dec_Eq :=
  match goal with
  | [ |- (_, _) = (_, _) ] =>
      apply injective_projections; unfold fst,snd; try reflexivity
  | [ |- (@Some ?T _ = @Some ?T _) ] =>
      apply (f_equal (@Some T)); try reflexivity
  | [ |- (@cons ?T _ _ = @cons ?T _ _) ] =>
      apply (f_equal2 (@cons T)); try reflexivity
  | [ |- (?X _ _ _ _ _ = ?X _ _ _ _ _) ] =>
      apply (f_equal5 X); try reflexivity
  | [ |- (?X _ _ _ _ = ?X _ _ _ _) ] =>
      apply (f_equal4 X); try reflexivity
  | [ |- (?X _ _ _ = ?X _ _ _) ] =>
      apply (f_equal3 X); try reflexivity
  | [ |- (?X _ _ = ?X _ _) ] =>
      apply (f_equal2 X); try reflexivity
  | [ |- (?X _ = ?X _) ] =>
      apply (f_equal X); try reflexivity
  | [ |- (?X ?A <> ?X ?B) ] =>
      cut (A <> B); [intro; congruence | try discriminate]
  end.

Lemma map_app : forall {A : Type} {B : Type} (f: A -> B) (l1 l2: list A),
  map f (l1 ++ l2) = (map f l1) ++ (map f l2).
Proof.
  induction l1; simpl; intros. auto. dec_Eq; auto.
Qed.



(****
 **** End preliminaries
 ****)
