Require Import Lib.
Require Import ReWire_Base.
Require Import ReWire_Subst.
Require Import ReWire_Step.
Require Import ReWire_Progress.
Require Import ReWire_Preservation.

(*
 *
 * No forbidden updates.
 *
 *)

Inductive same_where_no_write : Mo -> store -> store -> Prop :=
  | SWNW_Identity : same_where_no_write (MNonReact MIdentity) nil nil
  | SWNW_ReactT :
       forall Sto Sto' SM T T',
         same_where_no_write (MNonReact SM) Sto Sto' ->
         same_where_no_write (MReactT T T' SM) Sto Sto'
  | SWNW_StateT_None :
       forall Sto Sto' SM s T,
         same_where_no_write (MNonReact SM) Sto Sto' ->
         same_where_no_write (MNonReact (MStateT T EffNone SM)) (s::Sto) (s::Sto')
  | SWNW_StateT_R :
       forall Sto Sto' SM s T,
         same_where_no_write (MNonReact SM) Sto Sto' ->
         same_where_no_write (MNonReact (MStateT T EffR SM)) (s::Sto) (s::Sto')
  | SWNW_StateT_W :
       forall Sto Sto' SM s s' T,
         same_where_no_write (MNonReact SM) Sto Sto' ->
         same_where_no_write (MNonReact (MStateT T EffW SM)) (s::Sto) (s'::Sto')
  | SWNW_StateT_RW :
       forall Sto Sto' SM s s' T,
         same_where_no_write (MNonReact SM) Sto Sto' ->
         same_where_no_write (MNonReact (MStateT T EffRW SM)) (s::Sto) (s'::Sto').

Hint Constructors same_where_no_write.

Lemma same_where_no_write_refl : forall M Sto,
                                    store_matches_mo Sto M ->
                                    same_where_no_write M Sto Sto
with same_where_no_write_refl_sm : forall SM Sto,
                                      store_matches_mo Sto (MNonReact SM) ->
                                      same_where_no_write (MNonReact SM) Sto Sto.
Proof.
  destruct M; intros Sto Hmatch.
    constructor.
    apply same_where_no_write_refl_sm.
    inversion Hmatch; auto.

    apply same_where_no_write_refl_sm.
    assumption.

  induction SM; intros Sto Hmatch; inversion Hmatch; subst.
    constructor.
    destruct e; constructor; apply IHSM; assumption.
Qed.

Lemma same_where_no_write_trans : forall M Sto1 Sto2 Sto3,
                                    same_where_no_write M Sto1 Sto2 ->
                                    same_where_no_write M Sto2 Sto3 ->
                                    same_where_no_write M Sto1 Sto3 
with same_where_no_write_trans_sm : forall SM Sto1 Sto2 Sto3,
                                      same_where_no_write (MNonReact SM) Sto1 Sto2 ->
                                      same_where_no_write (MNonReact SM) Sto2 Sto3 ->
                                      same_where_no_write (MNonReact SM) Sto1 Sto3.
Proof.
  destruct M; intros.
  inversion H.
  inversion H0.
  subst.
  constructor.
    apply (same_where_no_write_trans_sm _ _ _ _ H6 H12).
    apply (same_where_no_write_trans_sm _ _ _ _ H H0).

 induction SM; intros;
 [inversion H; inversion H0; subst;auto
 |].
 destruct e; [ | | | ]; 
 destruct Sto1; destruct Sto2; destruct Sto3; try (inversion H; fail) ; try (inversion H0; fail);
 inversion H; inversion H0; subst; constructor;
 match goal with
   | [H: same_where_no_write (MNonReact SM) ?S1 ?S2, H': same_where_no_write (MNonReact SM) ?S2 ?S3 |- _ ] =>
     apply (IHSM _ _ _ H H') end.
Qed.

Lemma same_where_no_write_less_permissive :
  forall SM1 SM2 Sto Sto',
    smo_less_permissive SM1 SM2 ->
    same_where_no_write (MNonReact SM1) Sto Sto' ->
    same_where_no_write (MNonReact SM2) Sto Sto'.
Proof.
  intros SM1 SM2 Sto Sto' Hless.
  generalize dependent Sto'.
  generalize dependent Sto.
  induction Hless; intros Sto Sto' Hswnw; inversion Hswnw; subst;
    try eauto;
    try (inversion H; subst; eauto).
Qed.

Theorem no_forbidden_updates :
  forall t Sto M T,
    (t,Sto) |> TMonadic M T ->
  forall t' Sto',
    (t,Sto) ~~~> (t',Sto') ->
    same_where_no_write M Sto Sto'.
Proof.
  induction t;
    try (intros ? ? ? Htco ? ? ?Hstep;
        inversion Hstep; subst;
        apply same_where_no_write_refl;
        inversion Htco; assumption; fail).

  Case "tbind".
    intros Sto M T Htco t' Sto' Hstep.
    inversion Htco; subst.
    inversion H3; subst.
    inversion Hstep; subst;
      try (apply same_where_no_write_refl; auto); try eauto.

  Case "tlift".
    intros Sto M T Htco t' Sto' Hstep.
    inversion Htco; subst.
    inversion H3; subst.
    inversion Hstep; subst; try (apply same_where_no_write_refl; eauto; fail).

    inversion H5; subst.
    destruct E; econstructor; eapply IHt; try econstructor;eauto;
    match goal with
        | [ H: store_all_values (_ :: ?Sto0) |- store_all_values ?Sto0 ] => inv H end.
    inversion Hstep; subst; try (apply same_where_no_write_refl; auto; fail).
    econstructor. eapply IHt;eauto. econstructor; eauto. inversion H5; auto. 

  Case "televate".
    intros Sto M T Htco t' Sto' Hstep.
    inversion Htco; subst.
    inversion H3; subst.
    inversion Hstep; subst; try (apply same_where_no_write_refl; auto; fail).

    eapply same_where_no_write_less_permissive; eauto.
    eapply IHt; eauto.
    econstructor; eauto.
    eapply more_permissive_store_matches; eauto.

  Case "tput".
    intros Sto M T Htco t' Sto' Hstep.
    inversion Htco; subst.
    inversion H3; subst.
    inversion Hstep; subst; try (apply same_where_no_write_refl; auto; fail).

    inversion H5; subst.
    destruct E;
      try (inversion H6; fail);
      try (econstructor; apply same_where_no_write_refl; auto; fail).

  Case "trunst".
    intros Sto M T Htco t' Sto' Hstep.
    inversion Htco; subst.
    inversion H3; subst.
    inversion Hstep; subst; try (apply same_where_no_write_refl; auto; fail).

    assert ((t1,t2::Sto) |> TMonadic (MNonReact (MStateT TS E SM)) T0) by (repeat econstructor;eauto).
    pose proof (IHt1 _ _ _ H _ _ H11).
    inversion H0; auto.

  Case "trunre".
    intros Sto SM T Htco t' Sto' Hstep.
    inversion Htco; subst.
    inversion H3; subst.
    inversion Hstep; subst; try (apply same_where_no_write_refl; auto; fail).
    assert ((t0,Sto) |> TMonadic (MReactT TI TO SM0) t) by (constructor; auto).
    pose proof (IHt _ _ _ H _ _ H9).
    inversion H0; auto.
Qed.




Theorem no_forbidden_updates_star :
  forall t Sto M T,
    (t,Sto) |> TMonadic M T ->
  forall t' Sto',
    (t,Sto) ~~~>* (t',Sto') ->
    same_where_no_write M Sto Sto'.
Proof. intros.
       remember (t,Sto) as co; remember (t',Sto') as co';
       generalize dependent t Sto t' Sto' M T.
       R* induction H0 using trans; intros.
       subst.
       inversion Heqco'.
       eapply same_where_no_write_refl.
       inv H.
       subst.
       eapply no_forbidden_updates;eauto.
       destruct y as (m,Sm).
       preservation.
       subst.
       pose proof (IHrt_closure2 _ _ eq_refl _ _ eq_refl _ _ H0).
       pose proof (IHrt_closure1 _ _ eq_refl _ _ eq_refl _ _ H).
       apply (same_where_no_write_trans M _ _ _ H3 H2).
Qed.  

(*
 *
 * No forbidden reads.
 *
 *)
Inductive same_where_read : Mo -> store -> store -> Prop :=
  | SWR_Identity : same_where_read (MNonReact MIdentity) nil nil
  | SWR_ReactT :
       forall Sto Sto' SM T T',
         same_where_read (MNonReact SM) Sto Sto' ->
         same_where_read (MReactT T T' SM) Sto Sto'
  | SWR_StateT_None :
       forall Sto Sto' SM s s' T,
         same_where_read (MNonReact SM) Sto Sto' ->
         same_where_read (MNonReact (MStateT T EffNone SM)) (s::Sto) (s'::Sto')
  | SWR_StateT_R :
       forall Sto Sto' SM s T,
         same_where_read (MNonReact SM) Sto Sto' ->
         same_where_read (MNonReact (MStateT T EffR SM)) (s::Sto) (s::Sto')
  | SWR_StateT_W :
       forall Sto Sto' SM s s' T,
         same_where_read (MNonReact SM) Sto Sto' ->
         same_where_read (MNonReact (MStateT T EffW SM)) (s::Sto) (s'::Sto')
  | SWR_StateT_RW :
       forall Sto Sto' SM s T,
         same_where_read (MNonReact SM) Sto Sto' ->
         same_where_read (MNonReact (MStateT T EffRW SM)) (s::Sto) (s::Sto').

Inductive write_consistent : (store*store) -> (store*store) -> Prop :=
  | WC_Identity  : write_consistent (nil,nil) (nil,nil)
  | WC_Unchanged : forall Sto1 Sto2 Sto1' Sto2' s1 s2,
                     write_consistent (Sto1,Sto2) (Sto1',Sto2') ->
                     write_consistent (s1::Sto1,s2::Sto2) (s1::Sto1',s2::Sto2')
  | WC_Changed   : forall Sto1 Sto2 Sto1' Sto2' s1 s2 s,
                     write_consistent (Sto1,Sto2) (Sto1',Sto2') ->
                     write_consistent (s1::Sto1,s2::Sto2) (s::Sto1',s::Sto2').

Hint Constructors write_consistent.

Lemma write_consistent_same_length_refl :
  forall Sto1 Sto2,
    same_length Sto1 Sto2 ->
    write_consistent (Sto1,Sto2) (Sto1,Sto2).
Proof.
  intros ? ? Hsl; induction Hsl; auto.
Qed.

Lemma same_monad_same_length :
  forall M Sto1 Sto2,
    store_matches_mo Sto1 M ->
    store_matches_mo Sto2 M ->
    same_length Sto1 Sto2
with same_monad_same_length_sm :
  forall SM Sto1 Sto2,
    store_matches_mo Sto1 (MNonReact SM) ->
    store_matches_mo Sto2 (MNonReact SM) ->
    same_length Sto1 Sto2.
Proof.
  destruct M; intros Sto1 Sto2 Hmatch1 Hmatch2.
    inversion Hmatch1; inversion Hmatch2; subst.
    eapply same_monad_same_length_sm; repeat eassumption.

    eapply same_monad_same_length_sm; repeat eassumption.

  induction SM; intros Sto1 Sto2 Hmatch1 Hmatch2.
    inversion Hmatch1; inversion Hmatch2; subst.
    constructor.

    inversion Hmatch1; inversion Hmatch2; subst.
    constructor.
    eapply IHSM; repeat eassumption.
Qed.

Lemma same_where_read_same_length :
  forall M Sto1 Sto2,
    same_where_read M Sto1 Sto2 ->
    same_length Sto1 Sto2
with same_where_read_same_length_sm :
  forall SM Sto1 Sto2,
    same_where_read (MNonReact SM) Sto1 Sto2 ->
    same_length Sto1 Sto2.
Proof.
  destruct M; intros Sto1 Sto2 Hswr.
    inversion Hswr; subst.
    eapply same_where_read_same_length_sm; eassumption.

    eapply same_where_read_same_length_sm; eassumption.

  induction SM; intros Sto1 Sto2 Hswr.
    inversion Hswr; subst.
    constructor.

    inversion Hswr; subst; constructor; apply IHSM; assumption.
Qed.

Lemma same_where_read_tail :
  forall T E SM s1 Sto1 s2 Sto2,
    same_where_read (MNonReact (MStateT T E SM)) (s1::Sto1) (s2::Sto2) ->
    same_where_read (MNonReact SM) Sto1 Sto2.
Proof.
  intros T E SM s1 Sto1 s2 Sto2 Hswr; inversion Hswr; auto.
Qed.

Lemma same_where_read_head :
  forall T E SM s1 Sto1 s2 Sto2,
    Eff_lt EffR E ->
    same_where_read (MNonReact (MStateT T E SM)) (s1::Sto1) (s2::Sto2) ->
    s1 = s2.
Proof.
  intros T E SM s1 Sto1 s2 Sto2 Hlt Hswr;
    inversion Hlt; subst; inversion Hswr; auto.
Qed.

Lemma same_where_read_less_permissive :
  forall SM1 SM2,
    smo_less_permissive SM1 SM2 ->
  forall Sto1 Sto2,
    same_where_read (MNonReact SM2) Sto1 Sto2 ->
    same_where_read (MNonReact SM1) Sto1 Sto2.
Proof.
  intros SM1 SM2 Hlp.
  induction Hlp; intros.
    inversion H0; subst; inversion H; subst; constructor; auto.

    assumption.
Qed.

Theorem no_forbidden_reads : forall t Sto1 Sto2 T M,
                               (t,Sto1) |> TMonadic M T ->
                               (t,Sto2) |> TMonadic M T ->
                               same_where_read M Sto1 Sto2 ->
                             forall t1' t2' Sto1' Sto2',
                               (t,Sto1) ~~~> (t1',Sto1') ->
                               (t,Sto2) ~~~> (t2',Sto2') ->
                               t1' = t2' /\ write_consistent (Sto1,Sto2) (Sto1',Sto2').
Proof.
  induction t;
   (* knock out things that can't be monadic, by contra *)
   try (intros ? ? ? ? ? ?Htco; inversion Htco;
        match goal with
        [ Hmon : (_ |-- ?t \in TMonadic _ _) |- _ ] => inversion Hmon
        end;
        fail);
   (* knock out things where reduction must be by lc *)
   try (intros ? ? ? ? ? ? ? ? ? ? ? ?Hstepco1 ?Hstepco2; inversion Hstepco1; inversion Hstepco2; subst;
        split;
         [ eapply step_deterministic;
             eassumption;
             assumption
         | apply write_consistent_same_length_refl;
             eapply same_where_read_same_length;
             eassumption];
        fail);
   intros Sto1 Sto2 T M HT1 HT2 Hswr t1' t2' Sto1' Sto2' Hstep1 Hstep2.

  Case "tbind".
    rename t1 into T';
    rename t2 into t1;
    rename t3 into t2.
    inversion HT1; subst.
    inversion H3; subst.
    inversion HT2; subst.
    inversion H7; subst.
    assert (TMonadic M T0 = TMonadic M T1) by (eapply type_unique; eauto).
    inversion H; subst.
    inversion Hstep1; subst.
    SCase "1: lc".
      inversion Hstep2; subst.
      SSCase "2: lc".
        split.
        eapply step_deterministic; eauto.
        apply write_consistent_same_length_refl.
        eapply same_where_read_same_length; eauto.
      SSCase "2: monad step in t1".
        contradict H1; apply value_not_step; repeat constructor; auto.
      SSCase "2: bind-ret".
        contradict H1; apply value_not_step; repeat constructor; auto.
      SSCase "2: bind-pause".
        contradict H1; apply value_not_step; repeat constructor; auto.
    SCase "1: monad step in t1".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H1; apply value_not_step; repeat constructor; auto.
      SSCase "2: monad step in t2".
        assert ((t1,Sto1) |> TMonadic M T1) by (constructor; auto).
        assert ((t1,Sto2) |> TMonadic M T1) by (constructor; auto).
        pose proof (IHt1 _ _ _ _ H0 H1 Hswr _ _ _ _ H18 H21).
        destruct H11; subst; auto.
      SSCase "2: bind-ret".
        contradict H18; apply done_not_step; auto.
      SSCase "2: bind-pause".
        contradict H18; apply done_not_step; auto.
    SCase "1: bind-ret".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H1; apply value_not_step; repeat constructor; auto.
      SSCase "2: monad step in t1".
        contradict H20; apply done_not_step; auto.
      SSCase "2: bind-ret".
        split; auto.
        apply write_consistent_same_length_refl.
        eapply same_where_read_same_length; eauto.
    SCase "1: bind-pause".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H1; apply value_not_step; repeat constructor; auto.
      SSCase "2: monad step in t1".
        contradict H20; apply done_not_step; auto.
      SSCase "2: bind-pause".
        split; auto.
        apply write_consistent_same_length_refl.
        eapply same_where_read_same_length; eauto.

  Case "tlift".
    inversion HT1; subst.
    inversion HT2; subst.
    inversion H3; subst.
    SCase "tlift @StT".
      inversion Hstep1; subst.
      SSCase "1: lc".
        inversion Hstep2; subst.
        SSSCase "2: lc".
          split.
            eapply step_deterministic; eauto.

            apply write_consistent_same_length_refl.
            eapply same_where_read_same_length; eauto.
        SSSCase "2: monad step in t".
          contradict H0; apply value_not_step; repeat constructor; auto.
        SSSCase "2: lift-ret".
          contradict H0; apply value_not_step; repeat constructor; auto.
      SSCase "1: monad step in t".
        inversion Hstep2; subst.
        SSSCase "2: lc".
          contradict H0; apply value_not_step; repeat constructor; auto.
        SSSCase "2: monad step in t".
          inversion H4; subst.
          inversion H6; subst.
          assert ((t,Sto) |> TMonadic (MNonReact SM) T) by (repeat econstructor;eauto;inv H5).
          assert ((t,Sto0) |> TMonadic (MNonReact SM) T) by (repeat econstructor;eauto;[inv H7 | inv H8]).
          destruct (IHt _ _ _ _ H H0 (same_where_read_tail _ _ _ _ _ _ _ Hswr) _ _ _ _ H14 H16); subst;auto.
       SSSCase "2: lift-ret".
          contradict H14; apply done_not_step; auto.
      SSCase "1: lift-ret".
        inversion Hstep2; subst.
        SSSCase "2: lc".
          contradict H2; apply value_not_step; repeat constructor; auto.
        SSSCase "2: monad step in t".
          contradict H15; apply done_not_step; auto.
        SSSCase "2: lift-ret".
          split; auto.
          apply write_consistent_same_length_refl.
          eapply same_where_read_same_length; eauto.

    SCase "tlift @ReT".
      inversion Hswr; subst.
      inversion Hstep1; subst.
      SSCase "1: lc".
        inversion Hstep2; subst.
        SSSCase "2: lc".
          split.
            eapply step_deterministic; eauto.

            apply write_consistent_same_length_refl.
            eapply same_where_read_same_length; eauto.
        SSSCase "2: monad step in t".
          contradict H0; apply value_not_step; repeat constructor; auto.
        SSSCase "2: lift-ret".
          contradict H0; apply value_not_step; repeat constructor; auto.
      SSCase "1: monad step in t".
        inversion Hstep2; subst.
        SSSCase "2: lc".
          contradict H0; apply value_not_step; repeat constructor; auto.
        SSSCase "2: monad step in t".
          inversion H5; subst.
          inversion H8; subst.
          assert ((t,Sto1) |> TMonadic (MNonReact SM) T) by (constructor; auto).
          assert ((t,Sto2) |> TMonadic (MNonReact SM) T) by (constructor; auto).
          destruct (IHt _ _ _ _ H H0 H11 _ _ _ _ H15 H17); subst;auto.
        SSSCase "2: lift-ret".
          contradict H15; apply done_not_step; auto.
      SSCase "1: lift-ret".
        inversion Hstep2; subst.
        SSSCase "2: lc".
          contradict H2; apply value_not_step; repeat constructor; auto.
        SSSCase "2: monad step in t".
          contradict H16; apply done_not_step; auto.
        SSSCase "2: lift-ret".
          split; auto.
          apply write_consistent_same_length_refl.
          eapply same_where_read_same_length; eauto.

  Case "televate".
    inversion Hstep1; subst.
    SCase "1: lc".
      inversion Hstep2; subst.
      SSCase "2: lc".
        split.
          eapply step_deterministic; eauto.

          apply write_consistent_same_length_refl.
          eapply same_where_read_same_length; eauto.
      SSCase "2: monad step in t".
        contradict H0; apply value_not_step; auto.
      SSCase "2: elevate-ret".
        contradict H0; apply value_not_step; auto.
    SCase "1: monad step in t".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H0; apply value_not_step; auto.
      SSCase "2: monad step in t".
        inversion HT1; subst.
        inversion HT2; subst.
        inversion H6; subst.
        pose proof (more_permissive_store_matches _ _ _ H13 H9).
        pose proof (more_permissive_store_matches _ _ _ H13 H12).
        assert ((t,Sto1) |> TMonadic (MNonReact SM) T) by (constructor; auto).
        assert ((t,Sto2) |> TMonadic (MNonReact SM) T) by (constructor; auto).
        destruct (IHt _ _ _ _ H3 H4 (same_where_read_less_permissive _ _ H13 _ _ Hswr) _ _ _ _ H5 H7); subst; split; auto.
      SSCase "2: elevate-ret".
        contradict H5; apply done_not_step; auto.
    SCase "1: elevate-ret".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H1; apply value_not_step; auto.
      SSCase "2: monad step in t".
        contradict H6; apply done_not_step; auto.
      SSCase "2: elevate-ret".
        split; auto.
        apply write_consistent_same_length_refl.
        eapply same_where_read_same_length; eauto.

  Case "tget".
    inversion Hstep1; subst.
    SCase "1: lc".
      contradict H0; apply value_not_step; auto.
    SCase "1: get".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H0; apply value_not_step; auto.
      SSCase "2: get".
        inversion HT1; subst.
        inversion H3; subst.
        pose proof (same_where_read_head _ _ _ _ _ _ _ H2 Hswr); subst.
        split; auto.
        apply write_consistent_same_length_refl.
        eapply same_where_read_same_length; eauto.

  Case "tput".
    inversion Hstep1; subst.
    SCase "1: lc".
      inversion Hstep2; subst.
      SSCase "2: lc".
        split.
          eapply step_deterministic; eauto.

          apply write_consistent_same_length_refl.
          eapply same_where_read_same_length; eauto.
      SSCase "2: put".
        contradict H0; apply value_not_step; auto.
    SCase "1: put".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H1; apply value_not_step; auto.
      SSCase "2: put".
        split; auto.
        constructor.
        apply write_consistent_same_length_refl.
        pose proof (same_where_read_same_length _ _ _ Hswr).
        inversion H; auto.

  Case "trunst".
    inversion Hstep1; subst.
    SCase "1: lc".
      inversion Hstep2; subst.
      SSCase "2: lc".
        split.
          eapply step_deterministic; eauto.

          apply write_consistent_same_length_refl.
          eapply same_where_read_same_length; eauto.
      SSCase "2: monad step in t1".
        contradict H0; apply value_not_step; auto.
      SSCase "2: runst-return".
        contradict H0; apply value_not_step; auto.
    SCase "1: monad step in t1".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H0; apply value_not_step; auto.
      SSCase "2: monad step in t1".
        inversion HT1; subst.
        inversion HT2; subst.
        inversion H7; subst.
        assert ((t1,t2::Sto1) |> TMonadic (MNonReact (MStateT TS E SM)) T0) by (repeat constructor; auto).
        assert ((t1,t2::Sto2) |> TMonadic (MNonReact (MStateT TS E SM)) T0) by (repeat constructor; auto).
        assert (same_where_read (MNonReact (MStateT TS E SM)) (t2::Sto1) (t2::Sto2)).
          destruct E; constructor; auto.
        pose proof (IHt1 _ _ _ _ H H0 H1 _ _ _ _ H6 H9).
        destruct H2; subst.
        inversion H16; subst; auto.
      SSCase "2: runst-return".
        contradict H6; apply done_not_step; auto.
    SCase "1: runst-return".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H0; apply value_not_step; auto.
      SSCase "2: monad step in t1".
        contradict H8; apply done_not_step; auto.
      SSCase "2: runst-return".
        split; auto.
        apply write_consistent_same_length_refl.
        eapply same_where_read_same_length; eauto.

  Case "tunfold".
    rename t1 into T'.
    rename t2 into T''.
    rename t3 into t1.
    rename t4 into t2.
    inversion HT1; subst.
    inversion HT2; subst.
    inversion H3; subst.
    inversion Hstep1; subst.
    SCase "1: lc".
      inversion Hstep2; subst.
      SSCase "2: lc".
        split.
          eapply step_deterministic; eauto.

          apply write_consistent_same_length_refl.
          eapply same_where_read_same_length; eauto.
      SSCase "2: unfold".
        contradict H0; apply value_not_step; auto.
    SCase "1: unfold".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H0; apply value_not_step; auto.
      SSCase "2: unfold".
        split; auto.
        apply write_consistent_same_length_refl.
        eapply same_where_read_same_length; eauto.

  Case "trunre".
    rename t into T'.
    rename t0 into t.
    inversion HT1; subst.
    inversion HT2; subst.
    inversion H3; subst.
    inversion Hstep1; subst.
    SCase "1: lc".
      inversion Hstep2; subst.
      SSCase "2: lc".
        split.
          eapply step_deterministic; eauto.

          apply write_consistent_same_length_refl.
          eapply same_where_read_same_length; eauto.
      SSCase "2: monad step in t".
        contradict H0; apply value_not_step; auto.
      SSCase "2: runre-ret".
        contradict H0; apply value_not_step; auto.
      SSCase "2: runre-pause".
        contradict H0; apply value_not_step; auto.
    SCase "1: monad step in t".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H0; apply value_not_step; auto.
      SSCase "2: monad step in t".
        assert ((t,Sto1) |> TMonadic (MReactT TI TO SM) T') by (repeat constructor; auto).
        assert ((t,Sto2) |> TMonadic (MReactT TI TO SM) T') by (repeat constructor; auto).
        assert (same_where_read (MReactT TI TO SM) Sto1 Sto2) by (constructor; auto).
        destruct (IHt _ _ _ _ H H0 H10 _ _ _ _ H12 H14); subst; auto.
      SSCase "2: runre-ret".
        contradict H12; apply done_not_step; auto.
      SSCase "2: runre-pause".
        contradict H12; apply done_not_step; auto.
    SCase "1: runre-ret".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H2; apply value_not_step; auto.
      SSCase "2: monad step in t".
        contradict H13; apply done_not_step; auto.
      SSCase "2: runre-ret".
        split; auto.

        apply write_consistent_same_length_refl.
        eapply same_where_read_same_length; eauto.
    SCase "1: runre-pause".
      inversion Hstep2; subst.
      SSCase "2: lc".
        contradict H2; apply value_not_step; auto.
      SSCase "2: monad step in t".
        contradict H13; apply done_not_step; auto.
      SSCase "2: runre-pause".
        split; auto.

        apply write_consistent_same_length_refl.
        eapply same_where_read_same_length; eauto.
Qed.

  
(* ###################################################################### *)
(** Read Only Predicate: 
 *)  

Create HintDb ro_hints.

Inductive read_only : SMo -> Prop :=
 | Ronly_ST  : forall T E SM,
                 Eff_lt E EffR     ->
                 read_only SM      ->
                 read_only (MStateT T E SM)
 | Ronly_Id  :  read_only MIdentity.

Hint Constructors read_only.

Theorem read_only_eff' : forall T E SM,
                read_only (MStateT T E SM) ->
                     Eff_lt E EffR.
Proof. intros; inv H. Qed.              

Theorem read_only_smo :  forall T E SM,
                read_only (MStateT T E SM) ->
                    read_only SM.
Proof. intros; inv H. Qed.             


Theorem read_only_smo_lp' : forall SM1 SM2,
                    smo_less_permissive SM1 SM2 ->
                    read_only SM2  ->
                    read_only SM1.
Proof.
  intros. generalize dependent SM1 SM2; induction SM1; intros; eauto.
  inv H.
  inv H0.
  constructor; eauto.
Qed.
  
Theorem read_only_smo_lp_ST : forall T E SM1 SM2,
                    smo_less_permissive SM1 SM2 ->
                    read_only (MStateT T E SM2) ->
                    read_only (MStateT T E SM1).
Proof. 
  intros;constructor; inv H0; eapply read_only_smo_lp'; eauto.
Qed. 

Hint Resolve read_only_smo read_only_smo_lp' read_only_smo_lp_ST : ro_hints.

Theorem read_only_elim_w : forall T SM, read_only (MStateT T EffW SM) -> False.
Proof. intros. inv H. inv H2. Qed.

Theorem read_only_elim_rw : forall T SM, read_only (MStateT T EffRW SM) -> False.
Proof. intros. inv H. inv H2. Qed.

Theorem read_only_elim_base:
  forall SM T E, 
    ~ read_only SM ->
        ~ read_only (MStateT T E SM).
Proof. intros. red;intros. red in H.
       inv H0.
Qed.

Hint Resolve read_only_elim_w read_only_elim_rw read_only_elim_base : ro_hints.

Theorem read_only_st_dec : forall T E SM, 
        read_only (MStateT T E SM) \/ ~ read_only (MStateT T E SM).
Proof.  intros ? ? ?; generalize dependent SM E T.
        induction SM; induction E; intros; 
        try (right; red; eauto with ro_hints;fail).
        left; constructor; eauto.
        left; constructor; eauto.
        destruct (IHSM e t).
                 left. constructor;eauto.
                 right; eauto with ro_hints.
        destruct (IHSM e t).
                 left. constructor;eauto.
                 right; eauto with ro_hints.
Defined.
 
 
Theorem read_only_dec : forall SM, read_only SM \/ ~ read_only SM.
Proof. intros ?; generalize dependent SM; 
       induction SM; [left;constructor|eapply read_only_st_dec;eauto].
Defined.

Lemma same_where_no_write_read_only_eq :
  forall SM,
    read_only SM ->
  forall Sto Sto',
    same_where_no_write (MNonReact SM) Sto Sto' -> Sto = Sto'.
Proof.
  intros SM Hro; induction Hro; intros.
  Case "StateT".
    inversion H0; subst.
    pose proof (IHHro _ _ H6).
    subst; auto.
    pose proof (IHHro _ _ H6).
    subst; auto.
    inversion H.
    inversion H.
  Case "Identity".
    inversion H; auto.
Qed.

Theorem step_read_only_no_change : forall SM,
                                     read_only SM ->
                                     forall t Sto t' Sto' T,
                                       (t,Sto)  ~~~> (t',Sto')               ->
                                       (t,Sto) |> TMonadic (MNonReact SM) T  -> Sto = Sto'.
Proof.
  intros;
    eapply same_where_no_write_read_only_eq; [eauto | eapply no_forbidden_updates; eauto].
Qed.


Theorem step_read_only_no_change_star : forall SM,
                                     read_only SM ->
                                     forall t Sto t' Sto' T,
                                       (t,Sto)  ~~~>* (t',Sto')               ->
                                       (t,Sto) |> TMonadic (MNonReact SM) T  -> Sto = Sto'.
Proof.
  intros ? HRO_SM ? ? ? ? ? HStep HTco.
  remember (t,Sto) as co; remember (t',Sto') as co'.
  generalize dependent t Sto t' Sto' T SM.
  R* induction HStep using steps; intros.
  subst.
  inv Heqco'.
  subst. 
  destruct y as (m,Sm).
  pose proof (IHHStep _ _ eq_refl _ _ eq_refl _ _ HRO_SM HTco).
  subst.
  preservation.
  eapply step_read_only_no_change;eauto.
Qed.
