Require Import Lib.
Require Import ReWire_Base.
Require Import ReWire_Subst.
Require Import ReWire_Step.
Require Import ReWire_Progress.
Require Import ReWire_Preservation.
Require Import ReWire_Normalization.
Require Import ReWire_EqTh.
Require Import ReWire_Effects.

(** ** Monad Laws **)
                   
Theorem left_unit  : forall (M : Mo) (T1 T2 : Ty) (t1 t2 : tm),
        \empty |-- t1 \in T1 ->
        \empty |-- t2 \in TArrow T1 (TMonadic M T2) ->
        (tapp t2 t1) =[TMonadic M T2]= (treturn M t1 >>=[T2] t2).
Proof. 
  intros.
  lc_normalize t1.
  lc_normalize t2.
  assert (tapp t2 t1 ~~>* tapp v0 v) by (rtc_Trans;[eapply Congruence_App1;eauto|eapply Congruence_App2;eauto]).
  assert ((treturn M t1 >>=[T2] t2) ~~>* (treturn M v >>=[T2] v0)).
         rtc_Trans.
            eapply Congruence_Tbind1.
                   eapply Congruence_LC_Ret.
                          eauto.
            eapply Congruence_Tbind2;eauto.
  assert (\empty |-- tapp v0 v \in TMonadic M T2) by (preservation;econstructor;eauto).
  assert ((tapp v0 v) =[TMonadic M T2]= (tapp v0 v)) by (eapply term_eq_refl;eauto).
  repeat split; [repeat econstructor;eauto
                |repeat econstructor;eauto
                |intros;induction M; rename s into SM; [rename t into TI; rename t0 into TO|]].
Case "M is MReactT".
        destruct H4 as [? [? ?]].
        destruct (H7 _ _ H5) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? ?]]]]]]]].
        exists t1', Sto1', t2', Sto2'.
               repeat (split;eauto).
                      rtc_Trans; [eapply Congruence_LC;eauto|auto].
                      rtc_Trans; [eapply Congruence_LC;eauto|rtc_Trans; [eapply Congruence_BindRet;eauto|auto]].
Case "M is MNonReact".
        destruct H4 as [? [? ?]].
        destruct (H7 _ _ H5) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? [? ?]]]]]]]]].
        exists t1', Sto1', t2', Sto2'.
               repeat (split;eauto).
                      rtc_Trans; [eapply Congruence_LC;eauto|auto].
                      rtc_Trans; [eapply Congruence_LC;eauto|rtc_Trans; [eapply Congruence_BindRet;eauto|auto]].
Qed.
           
Lemma right_unit : forall (t : tm) (SM : SMo) (T : Ty) (x : id),
      \empty |-- t \in TMonadic (MNonReact SM) T ->
      t =[(TMonadic (MNonReact SM) T)]= (t >>=[T] \(x:T)(treturn (MNonReact SM) (tvar x))).
Proof.
 intros t SM T x HTt. 
 lc_normalize t.
 eapply eq_square_rev.
        auto.
        repeat econstructor.
        auto.
        unfold extend; rewrite eq_id;auto.
        eauto.
        eapply Congruence_Tbind1;eauto.
 preservation.
 repeat split;[auto|repeat econstructor;eauto; unfold extend;rewrite eq_id; auto|intros ? ? HStoeq].
 destruct  (term_eq_refl _ _ H) as [? [? Heq1]]. 
 destruct (Heq1 _ _ HStoeq)  as [v3 [Sto3 [v3' [Sto3' [? [? [? [? [? ?]]]]]]]]].
 exists v3, Sto3, v3', Sto3'.
        repeat (split;eauto).
        rtc_Trans; [eapply Congruence_Mo_Bind1;eauto
                   |rtc_Trans;[eapply Congruence_BindRet;eauto
                              |rtc_Trans;[eapply Congruence_AppAbs_Mo;eauto|simpl]]].
        simpl;rewrite eq_id;auto.
Qed.

Lemma associativity_of_bind : forall (SM : SMo) (T1 T2 T3 : Ty) (f g h : tm) (x : id),
   \empty |-- f \in TMonadic (MNonReact SM) T1 ->
   \empty |-- g \in TArrow T1 (TMonadic (MNonReact SM) T2) ->
   \empty |-- h \in TArrow T2 (TMonadic (MNonReact SM) T3) ->
   ((f >>=[T2] g) >>=[T3] h) =[TMonadic (MNonReact SM) T3]= (f >>=[T3] (\(x:T1) (tapp g (tvar x) >>=[T3] h))).
Proof. intros.
       lc_normalize f.
       lc_normalize g.
       lc_normalize h.
       split;[econstructor;
               [econstructor;
                 [eauto|eauto]
               |eassumption]
             |split;[repeat econstructor;eauto|intros]].
       eapply typable_empty_everywhere;eauto.
              unfold extend; rewrite eq_id; auto.
       eapply typable_empty_everywhere;eauto.
       destruct  (term_eq_refl _ _ H) as [? [? Heq1]].
       destruct (Heq1 _ _ H2)  as [v3 [Sto3 [v3' [Sto3' [? [? [? [? [? ?]]]]]]]]].
       assert (H5': (v, Sto1) ~~~>* (treturn (MNonReact SM) v3, Sto3)).
              eapply step_same_done.
                     constructor;auto.
                     eapply Congruence_LC.
                            eapply HS.
                            auto.
       assert (HTapp1: \empty |-- tapp v0 v3 \in TMonadic (MNonReact SM) T2).
              econstructor.
              preservation.
              eauto.
              eapply eq_typable_empty_l;eauto.
       assert (HTapp2: \empty |-- tapp v0 v3' \in TMonadic (MNonReact SM) T2).
              econstructor.
              preservation.
              eauto.
              eapply eq_typable_empty_r;eauto.
       lc_normalize (tapp v0 v3).
       lc_normalize (tapp v0 v3').
       assert (H6': (v, Sto2) ~~~>* (treturn (MNonReact SM) v3', Sto3')).
              eapply step_same_done.
                     constructor;auto.
                     eapply Congruence_LC.
                            eapply HS.
                            auto.                     
       assert (((f >>=[T2] g), Sto1) ~~~>* (tapp v0 v3, Sto3)).
              rtc_Trans.
                 eapply Congruence_LC.
                        eapply Congruence_Tbind1.
                                eapply HS.
              rtc_Trans.
                 eapply Congruence_LC.
                        eapply Congruence_Tbind2.
                                      auto.
                                      eapply HS0.
              rtc_Trans.
                 eapply Congruence_Mo_Bind1.
                        auto.
                        eapply H5'.
                        eapply Congruence_BindRet;auto.
       assert (((f >>=[T2] g) >>=[T3] h, Sto1) ~~~>* (v2 >>=[T3] v1,Sto3)).
              rtc_Trans.                 
                 eapply Congruence_LC.
                        eapply Congruence_Tbind1.
                               eapply Congruence_Tbind1.
                                      eapply HS.
              rtc_Trans.
                 eapply Congruence_LC.
                        eapply Congruence_Tbind1.  
                               eapply Congruence_Tbind2.
                                      auto.
                                      eapply HS0.
              rtc_Trans.
                 eapply Congruence_LC.
                        eapply Congruence_Tbind2.
                               constructor;auto.
                               eapply HS1.
              rtc_Trans.
                 eapply Congruence_Mo_Bind1.
                        auto.
                 eapply Congruence_Mo_Bind1.
                        auto.
                 eapply H5'.
              rtc_Trans.
                 eapply Congruence_Mo_Bind1.
                        auto.
                        eapply Congruence_BindRet.
                               auto.
                               auto.
              eapply Congruence_Mo_Bind1.
                     auto.
                     eapply Congruence_LC.
                            auto. 
       assert ((f >>=[T3] (\(x:T1) (tapp g (tvar x) >>=[T3] h)), Sto2) ~~~>* (v4 >>=[T3] v1,Sto3')).
              rtc_Trans.
                 eapply Congruence_Mo_Bind1.
                        constructor.
                        eapply H6.
              rtc_Trans.
                 eapply Congruence_BindRet.
                        auto.
                        constructor.
              rtc_Trans.
                 eapply Congruence_AppAbs_Mo.
                        auto.
                        simpl.
                        rewrite eq_id.
                        rewrite_subst.
              rtc_Trans.
                 eapply Congruence_LC.
                        eapply Congruence_Tbind1.
                               eapply Congruence_App1.
                                      eauto.
              rtc_Trans.
                 eapply Congruence_LC.
                        eapply Congruence_Tbind1.
                               eauto.
              rtc_Trans.
                 eapply Congruence_LC.
                        eapply Congruence_Tbind2.
                               auto.
                               eauto.
              eauto.
       assert (v2 =[TMonadic (MNonReact SM) T2]= v4).
              eapply eq_square.
                     eapply TApp_Term_eq.
                     eapply term_eq_refl.
                     eapply H0.
                     eapply H10.
                     rtc_Trans.
                        eapply Congruence_App1.
                               eapply HS0.
                        auto.
                     rtc_Trans.
                        eapply Congruence_App1.
                               eapply HS0.
                        auto.  
      assert ((v2 >>=[T3] v1) =[(TMonadic (MNonReact SM) T3)]= (v4 >>=[T3] v1)).
             eapply TBind_Term_eq.                     
                    eapply term_eq_refl.
                    preservation;eauto.
                    auto.
      destruct H15 as [? [? ?]].
      destruct (H17 _ _ H9)  as [v5 [Sto5 [v5' [Sto5' [? [? [? [? [? ?]]]]]]]]].
      exists v5, Sto5, v5', Sto5'.
             split.
             rtc_Trans;eauto.
             split.
             rtc_Trans;eauto.
             repeat (split;auto).
Qed.

(* ###################################################################### *)
(** Reactive Monad Axioms *)
 
Axiom RE_right_unit : forall (t : tm) (SM : SMo) (TI TO T : Ty) (x : id),
      \empty |-- t \in TMonadic (MReactT TI TO SM) T ->
      t =[(TMonadic (MReactT TI TO SM) T)]= (t >>=[T] \(x:T)(treturn (MReactT TI TO SM) (tvar x))).

Axiom RE_associativity_of_bind : forall (SM : SMo) (TI TO T1 T2 T3 : Ty) (f g h : tm) (x : id),
   \empty |-- f \in TMonadic (MReactT TI TO SM) T1 ->
   \empty |-- g \in TArrow T1 (TMonadic (MReactT TI TO SM) T2) ->
   \empty |-- h \in TArrow T2 (TMonadic (MReactT TI TO SM) T3) ->
   ((f >>=[T2] g) >>=[T3] h) =[TMonadic (MReactT TI TO SM) T3]= (f >>=[T3] (\(x:T1) (tapp g (tvar x) >>=[T3] h))).

(* ###################################################################### *)
(** ** Monad Transformer Laws *)
(*
    Lift . Return  = Return
    Lift (m >>= k) = (Lift m) >>= (Lift . k)
*)

Lemma lift_return_nonreact : forall t T TS E SM,
      \empty |-- t \in T ->   
      (tlift (MNonReact (MStateT TS E SM)) (treturn (MNonReact SM) t)) 
        =[TMonadic (MNonReact (MStateT TS E SM)) T]= 
       (treturn (MNonReact (MStateT TS E SM)) t).
Proof.
  intros.
  lc_normalize t.
  eapply eq_square_rev;[repeat econstructor;eauto|repeat econstructor;eauto| | | ].
         rtc_Trans.
            eapply Congruence_Lift.
            eapply Congruence_LC_Ret;eauto.
         eauto.
         eapply Congruence_LC_Ret.
         eauto.
         preservation.
         repeat split;[repeat econstructor;eauto|repeat econstructor;eauto|intros;do 4 eexists].
         repeat (split;eauto);eapply term_eq_refl;eauto.
Qed.

  
Lemma lift_return_react : forall t SM T TI TO,
      \empty |-- t \in T ->
     (tlift (MReactT TI TO SM) (treturn (MNonReact SM) t)) =[TMonadic (MReactT TI TO SM) T]= (treturn (MReactT TI TO SM) t).
Proof.
  intros.
  lc_normalize t.
  eapply eq_square_rev.
         repeat econstructor;eauto.
         repeat econstructor;eauto.
         rtc_Trans.
         eapply Congruence_Lift.
         eapply Congruence_LC_Ret.
         eauto.
         eauto.
         rtc_Trans.
         eapply Congruence_LC_Ret.
         eauto.
         eauto.
         preservation.
         split.
         repeat econstructor;eauto.
         split.
         repeat econstructor;eauto.
         intros.
         do 4 eexists.
         split.
         eapply Congruence_LiftRetRe.
                auto.
         split.
         eapply Congruence_LC.
                eapply rtc_refl.
         repeat (split;auto).
         assert ((treturn (MReactT TI TO SM) v) =[TMonadic (MReactT TI TO SM) T]= (treturn (MReactT TI TO SM) v)).
                eapply TReturn_Term_eq.
                eapply term_eq_refl;eauto.
         destruct H2 as [_ [_ HRetvEq]]. 
         destruct (HRetvEq _ _ H1) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? ?]]]]]]]].
         eapply step_preserves_along2_react_star;eauto.
Qed.

Lemma lift_bind_nonreact : forall x t1 t2 T1 T2 TS E SM,
      \empty |-- t1 \in TMonadic (MNonReact SM) T1 ->
      \empty |-- t2 \in TArrow T1 (TMonadic (MNonReact SM) T2) ->
      (tlift (MNonReact (MStateT TS E SM)) (t1 >>=[T2] t2))
         =[TMonadic (MNonReact (MStateT TS E SM)) T2]=
       ((tlift (MNonReact (MStateT TS E SM)) t1) >>=[T2] \(x:T1) (tlift (MNonReact (MStateT TS E SM)) (tapp t2 (tvar x)))).
Proof. 
  intros.
  lc_normalize t1.
  lc_normalize t2.
  preservation.
  assert ((t1 >>=[T2] t2) ~~>* (v >>=[T2] v0)) by (rtc_Trans; [eapply Congruence_Tbind1;eauto|eapply Congruence_Tbind2;eauto]).
  eapply eq_square_rev;[repeat econstructor;eauto
                       |repeat econstructor;eauto;
                       [eapply typable_empty_everywhere;eauto
                       |unfold extend;rewrite eq_id;auto ]
                       |eapply Congruence_Lift;eauto | | ].
         rtc_Trans.
            eapply Congruence_Tbind1.
                   eapply Congruence_Lift;eauto.
         eapply rtc_refl.
  repeat split; [repeat econstructor;eauto
                |repeat econstructor;eauto;
                [eapply typable_empty_everywhere;eauto
                |unfold extend;rewrite eq_id;auto ] |intros].
  generalize H4 as H4';intros H4'.
  destruct H4' as [? [? [? [? ?]]]].
  inversion H5.
  inversion H7.
  assert (store_eq (MNonReact SM) Sto Sto0).
         repeat (split;eauto);[subst;inv H6|subst;inv H8| ].
         destruct H9 as [? [? [? [? [? [? [? ?]]]]]]].
                  subst++.
                  inv H18.
                  inv H12.
  subst.
  destruct (term_eq_refl _ _ H2) as [_ [_ ?]]. 
  destruct (H10 _ _ H22) as [v1 [Stov1 [v1' [Stov1' [? [? [? [? [? ?]]]]]]]]].
  assert (\empty |-- v0 $ v1  \in TMonadic (MNonReact SM) T2) by (econstructor;eauto;eapply eq_typable_empty_l;eauto).
  assert (\empty |-- v0 $ v1' \in TMonadic (MNonReact SM) T2) by (econstructor;eauto;eapply eq_typable_empty_r;eauto).
  lc_normalize (v0 $ v1).
  lc_normalize (v0 $ v1').
  assert ((tlift (MNonReact (MStateT TS E SM)) (v >>=[T2] v0), t :: Sto) ~~~>*
                                                                         (tlift (MNonReact (MStateT TS E SM)) v2, t :: Stov1)).
  eapply Congruence_LiftSt.
         constructor;eauto.
  rtc_Trans.
     eapply Congruence_Mo_Bind1;eauto.
  rtc_Trans.
     eapply Congruence_BindRet; eauto.
            eapply Congruence_LC;eauto.
  assert (
     (tlift (MNonReact (MStateT TS E SM)) v >>=[T2] (\(x:T1) tlift (MNonReact (MStateT TS E SM)) (t2 $ tvar x)), t0 :: Sto0) ~~~>*
       (tlift (MNonReact (MStateT TS E SM)) v3, t0 :: Stov1')).
     rtc_Trans.
        eapply Congruence_Mo_Bind1.
               constructor.
        eapply Congruence_LiftSt;eauto.
     rtc_Trans.
        eapply Congruence_Mo_Bind1.
               constructor.
        eapply Congruence_LiftRetSt.
               auto.
    rtc_Trans.
       eapply Congruence_BindRet.
              auto.
              constructor.
    rtc_Trans.
       eapply Congruence_AppAbs_Mo.
              auto.
              simpl.
              rewrite eq_id.
              rewrite_subst.
    rtc_Trans.
        eapply Congruence_LC.
               eapply Congruence_Lift.
                      eapply Congruence_App1.
                             eauto.
   eapply Congruence_LC.
          eapply Congruence_Lift;eauto.
   assert ((v0 $ v1) =[(TMonadic (MNonReact SM) T2)]= (v0 $ v1')).
           eapply TApp_Term_eq.               
                  eapply term_eq_refl;eauto.
                  auto.
  assert (v2 =[(TMonadic (MNonReact SM) T2)]= v3) by (eapply eq_square;eauto).
  destruct H27 as [? [? ?]].
  destruct (H29 _ _ H17) as [v1'' [Sto1'' [v2'' [Sto2'' [? [? [? [? [? ?]]]]]]]]].
  assert (store_eq (MNonReact (MStateT TS E SM)) (t :: Sto1'') (t0 :: Sto2'')).
         simpl.
         repeat (split;auto).
            constructor;auto.
            eapply store_eq_matches_l;eauto.
            constructor; [inv H6| eapply store_eq_all_values_l;eauto].
            constructor;auto.
            eapply store_eq_matches_r;eauto.
            constructor; [inv H8| eapply store_eq_all_values_r;eauto].
            destruct H9 as [t3 [t4 [Sto3 [Sto4 [? [? [? ?]]]]]]].
            exists t3, t4, Sto1'', Sto2''.
                   inv H37.
                   inv H38.
                   destruct H34 as [? [? [? [? ?]]]].
                   repeat (split;auto).
  exists v1'', (t :: Sto1''), v2'', (t0 :: Sto2'').
     split.  
     rtc_Trans.
        eauto.
        rtc_Trans.
           eapply Congruence_LiftSt;eauto.
           eapply Congruence_LiftRetSt;eauto.
     split.
     rtc_Trans.
        eauto.
        rtc_Trans.
           eapply Congruence_LiftSt;eauto.
           eapply Congruence_LiftRetSt;eauto. 
     repeat (split;auto).
Qed.


Lemma lift_bind_react : forall x t1 t2 T1 T2 TI TO SM,
      \empty |-- t1 \in TMonadic (MNonReact SM) T1 ->
      \empty |-- t2 \in TArrow T1 (TMonadic (MNonReact SM) T2) ->
      (tlift (MReactT TI TO SM) (t1 >>=[T2] t2))
         =[TMonadic (MReactT TI TO SM) T2]=
       ((tlift (MReactT TI TO SM) t1) >>=[T2] \(x:T1) (tlift (MReactT TI TO SM) (tapp t2 (tvar x)))).
Proof. 
  intros.
  lc_normalize t1.
  lc_normalize t2.
  preservation.
  assert ((t1 >>=[T2] t2) ~~>* (v >>=[T2] v0)) by (rtc_Trans; [eapply Congruence_Tbind1;eauto|eapply Congruence_Tbind2;eauto]).
  eapply eq_square_rev;[repeat econstructor;eauto
                       |repeat econstructor;eauto;
                       [eapply typable_empty_everywhere;eauto
                       |unfold extend;rewrite eq_id;auto ]
                       |eapply Congruence_Lift;eauto
                       |eapply Congruence_Tbind1; eapply Congruence_Lift;eauto
                       | ].
  repeat split; [repeat econstructor;eauto
                |repeat econstructor;eauto;
                [eapply typable_empty_everywhere;eauto
                |unfold extend;rewrite eq_id;auto ] |intros].
  assert (store_eq (MNonReact SM) Sto1 Sto2).
         destruct H4 as [? [? [? [? ?]]]].
         repeat (split;auto);[inv H4|inv H6].
  destruct (term_eq_refl _ _ H2) as [_ [_ ?]]. 
  destruct (H6 _ _ H5) as [v1 [Stov1 [v1' [Stov1' [? [? [? [? [? ?]]]]]]]]].
  assert (\empty |-- v0 $ v1  \in TMonadic (MNonReact SM) T2) by (econstructor;eauto;eapply eq_typable_empty_l;eauto).
  assert (\empty |-- v0 $ v1' \in TMonadic (MNonReact SM) T2) by (econstructor;eauto;eapply eq_typable_empty_r;eauto).
  lc_normalize (v0 $ v1).
  lc_normalize (v0 $ v1').
  assert ((tlift (MReactT TI TO SM) (v >>=[T2] v0), Sto1) ~~~>* (tlift (MReactT TI TO SM) v2, Stov1)).
  eapply Congruence_LiftRe.
         constructor;eauto.
  rtc_Trans.
     eapply Congruence_Mo_Bind1;eauto.
  rtc_Trans.
     eapply Congruence_BindRet; eauto.
            eapply Congruence_LC;eauto.
  assert (
     (tlift (MReactT TI TO SM) v >>=[T2] (\(x:T1) tlift (MReactT TI TO SM) (t2 $ tvar x)), Sto2) ~~~>*
       (tlift (MReactT TI TO SM) v3, Stov1')).
     rtc_Trans.
        eapply Congruence_Mo_Bind1.
               constructor.
        eapply Congruence_LiftRe;eauto.
     rtc_Trans.
        eapply Congruence_Mo_Bind1.
               constructor.
        eapply Congruence_LiftRetRe.
               auto.
    rtc_Trans.
       eapply Congruence_BindRet.
              auto.
              constructor.
    rtc_Trans.
       eapply Congruence_AppAbs_Mo.
              auto.
              simpl.
              rewrite eq_id.
              rewrite_subst.
    rtc_Trans.
        eapply Congruence_LC.
               eapply Congruence_Lift.
                      eapply Congruence_App1.
                             eauto.
   eapply Congruence_LC.
          eapply Congruence_Lift;eauto.
  assert ((v0 $ v1) =[(TMonadic (MNonReact SM) T2)]= (v0 $ v1')).
           eapply TApp_Term_eq.               
                  eapply term_eq_refl;eauto.
                  auto.
  assert (v2 =[(TMonadic (MNonReact SM) T2)]= v3) by (eapply eq_square;eauto).
  destruct H18 as [? [? ?]].
  destruct (H20 _ _ H11) as [v1'' [Sto1'' [v2'' [Sto2'' [? [? [? [? [? ?]]]]]]]]].
  assert ((tlift (MReactT TI TO SM) (v >>=[T2] v0), Sto1) ~~~>* (treturn (MReactT TI TO SM) v1'', Sto1'')).
         rtc_Trans.
            eapply H15.
         rtc_Trans.
            eapply Congruence_LiftRe;eauto.
            eapply Congruence_LiftRetRe;eauto.
  assert ((tlift (MReactT TI TO SM) v >>=[T2] (\(x:T1) 
                                              tlift 
                                                (MReactT TI TO SM)
                                                (t2 $ tvar x)), Sto2) ~~~>*  (treturn (MReactT TI TO SM) v2'', Sto2'')).
         rtc_Trans.         
            eapply H16.
         rtc_Trans.
            eapply Congruence_LiftRe;eauto.
            eapply Congruence_LiftRetRe;eauto.
  assert (treturn (MReactT TI TO SM) v1'' =[TMonadic (MReactT TI TO SM) T2 ]= treturn (MReactT TI TO SM) v2'') by (eapply TReturn_Term_eq;eauto).
  destruct H29 as [? [? ?]].
  assert (store_eq (MReactT TI TO SM) Sto1'' Sto2'').
         destruct H25 as [? [? [? [? ?]]]].
         repeat split; auto.
  destruct (H31 _ _ H32) as [v1''' [Sto1''' [v2''' [Sto2''' [? [? [? [? ?]]]]]]]]. 
  do 4 eexists; repeat (split;eauto).
     eapply step_preserves_along2_react_star;eauto.
Qed.

(** ** Null Bind *)

Notation "t1 >>[ T1 , T2 ] t2" := (t1 >>=[T2] \(6:T1) t2)
                             (at level 40, T at level 99, T' at level 99, format "'[hv ' t1  >>[ T1 , T2 ]  t2 ']'").

Theorem generic_null_bind :  forall t t' M T T',  
        \empty |-- t  \in TMonadic M T  ->
        \empty |-- t' \in TMonadic M T' ->
                                  forall b : id,
        (t >>=[T'] (\(b:T) t')) =[TMonadic M T']= (t >>[T,T'] t').
Proof. intros.
       eapply TBind_Term_eq; [|eapply term_eq_refl;eauto].
       repeat split; [repeat econstructor;eauto;eapply typable_empty_everywhere;eauto
                     |repeat econstructor;eauto;eapply typable_empty_everywhere;eauto
                     |intros ].
       pose proof (eq_typable_empty_l H1); lc_normalize t1.
       pose proof (eq_typable_empty_r H1); lc_normalize t1'.
       exists t', t'.
              split.
              rtc_Trans.
                 rtc_Trans.
                 eapply Congruence_App2.
                        constructor.
                        eapply HS.
                 eapply rtc_R.
                        eapply ST_AppAbs.
                               auto.
                               rewrite_subst.
                               eauto.
              split.
              rtc_Trans.
                 rtc_Trans.
                 eapply Congruence_App2.
                        constructor.
                        eapply HS0.
                 eapply rtc_R.
                        eapply ST_AppAbs.
                               auto.
                               rewrite_subst.
                               eauto.
               eapply term_eq_refl;eauto.
Qed.


(* ###################################################################### *)
(** ** Stateful Computations *)
(*: the three properties of get and put, proved hereafter,
                           were taken from Harrison 2006 (page 103) and Gibbons & Hinze 2011 (page 5).

   put_put:    put s >> put s'           = put s'
   put_get:    put s >> get              = put s >> return s
   get_get:    get >>= \s -> get >>= k s = get >>= \s -> k s s   

   the update operator could be defined in terms of either put or get:

   update f  = put (f s)
   update id = get


 *)

Theorem ST_put_put : forall s s' E SM T,
       Eff_lt EffW E ->
       \empty |-- s  \in T ->
       \empty |-- s' \in T ->
                         ((tput  (MStateT T E SM) s) >>[TNil,TNil] (tput (MStateT T E SM) s'))
                         =[TMonadic (MNonReact (MStateT T E SM)) TNil]= tput (MStateT T E SM) s'.
Proof. intros ? ? ? ? ? HEff HTs HTs'.
       repeat split; [smash|smash|intros ? ? HStoeq12];auto.
       lc_normalize s.
       lc_normalize s'.
       edestruct HStoeq12 as [HStoMatch1 [? [HStoMatch2 [? ?]]]]; inv HStoMatch1; inv HStoMatch2.
       preservation.
       exists tunit, (v0 :: Sto), tunit, (v0 :: Sto0).
              repeat (split;auto).
              rtc_Trans.
                 eapply Congruence_Mo_Bind1.
                        constructor.
                        rtc_Trans.
                           eapply Congruence_PutE;eauto.
                        eapply rtc_R_mo.
                        eapply STM_Put. auto.
              rtc_Trans.
                 eapply Congruence_BindRet.
                        auto.
                        constructor.
              rtc_Trans.
                 eapply Congruence_AppAbs_Mo.
                        auto.
                        simpl.
                        rewrite_subst.
              rtc_Trans.
                 eapply Congruence_PutE;eauto.
              eapply rtc_R_mo.
                     eapply STM_Put.
                     auto.
             rtc_Trans.
                eapply Congruence_PutE; eauto.
                eapply rtc_R_mo.
                       eapply STM_Put.
                              auto.
             constructor;auto.
             inv H.
             constructor;auto.
             inv H0.
             simpl.
             exists v0, v0, Sto, Sto0.
             repeat split.
                    eauto.
             eapply store_eq_stack in HStoeq12 as [? [? [? [? ?]]]]; auto.
             constructor.
             constructor.
Qed.

Theorem ST_put_get : forall s T SM,
        \empty |-- s \in T ->
                         (tput (MStateT T EffRW SM) s) >>[TNil,T] (tget (MStateT T EffRW SM))
                         =[TMonadic (MNonReact (MStateT T EffRW SM)) T]=
                         (tput (MStateT T EffRW SM) s) >>[TNil,T] (treturn (MNonReact (MStateT T EffRW SM)) s).
Proof. intros. repeat split; [smash|smash|intros ? ? HStoeq12].
       lc_normalize s.
       edestruct HStoeq12 as [HStoMatch1 [? [HStoMatch2 [? ?]]]]; inv HStoMatch1; inv HStoMatch2.
       preservation.
       do 4 eexists; split.
          rtc_Trans.
               eapply Congruence_Mo_Bind1.
                        constructor.
                        rtc_Trans.
                           eapply Congruence_PutE;eauto.
                        eapply rtc_R_mo.
                        eapply STM_Put. auto.
          rtc_Trans.
             eapply Congruence_BindRet;eauto.
             rtc_Trans.
                eapply Congruence_AppAbs_Mo;auto.
                       simpl.
                       eapply rtc_R_mo.
                              eapply STM_Get.
          split.                    
          rtc_Trans.
             eapply Congruence_Mo_Bind1.
                    constructor.
                    rtc_Trans.
                       eapply Congruence_PutE;eauto.
                    eapply rtc_R_mo.
                    eapply STM_Put;auto.
          rtc_Trans.
             eapply Congruence_BindRet;eauto.
             rtc_Trans.
                eapply Congruence_AppAbs_Mo;auto.
                       simpl.
                rewrite_subst.
                eapply Congruence_LC.
                eapply Congruence_LC_Ret;eauto.
          eapply store_eq_stack in HStoeq12 as [? [? [? [? ?]]]].
          repeat split; eauto.
                 constructor;auto.
                 constructor;auto.
          exists v, v, Sto, Sto0.
          repeat split; auto.
Qed.

Theorem ST_get_get : forall T E SM x y z,
        Eff_lt EffR E ->
        tget (MStateT T E SM) >>=[TProd T T] (\(x:T) (tget (MStateT T E SM) >>=[TProd T T] (\(y:T) (treturn (MNonReact (MStateT T E SM)) (tpair (tvar x) (tvar y))))))
             =[TMonadic (MNonReact (MStateT T E SM)) (TProd T T)]=
        tget (MStateT T E SM) >>=[TProd T T] (\(z:T) (treturn (MNonReact (MStateT T E SM)) (tpair (tvar z) (tvar z)))).
Proof. intros. repeat split; [repeat econstructor
                             |smash|intros ? ? HStoeq12]; auto.
       unfold extend.
       destruct (eq_id_dec y x).
                auto.
       rewrite eq_id.
       auto.
       unfold extend.
       destruct (eq_id_dec y x).
                auto.
       rewrite eq_id.
       auto.
       rewrite eq_id.
       auto.
       edestruct HStoeq12 as [HStoMatch1 [? [HStoMatch2 [? ?]]]]; inv HStoMatch1; inv HStoMatch2.
       lc_normalize t; lc_normalize t0.
       edestruct H2 as [v1 [v2 [Sto1' [Sto2' [? [? [? ?]]]]]]].
       do 4 eexists.
       split. rtc_Trans.
                 eapply Congruence_Mo_Bind1.
                        constructor.
                        rtc_Trans.
                           eapply rtc_R_mo.
                                  eapply STM_Get.
                           eapply Congruence_LC.          
                                  eapply Congruence_LC_Ret;eauto.
                        rtc_Trans.
                           eapply Congruence_BindRet;eauto.
                        rtc_Trans.
                           eapply Congruence_AppAbs_Mo.
                                  auto.
                                  simpl.
                                  destruct (eq_id_dec x y).
                                  subst.
              rtc_Trans.                    
                 eapply Congruence_Mo_Bind1.
                        constructor.
                        rtc_Trans.
                           eapply rtc_R_mo.
                                  eapply STM_Get.
                           eapply Congruence_LC.          
                                  eapply Congruence_LC_Ret;eauto.
                        rtc_Trans.
                           eapply Congruence_BindRet;eauto.
                        rtc_Trans.
                           eapply Congruence_AppAbs_Mo.
                                  auto.
                                  simpl.
                                  rewrite eq_id.
                                  eapply rtc_refl.
              rewrite eq_id.
              rtc_Trans.                    
                 eapply Congruence_Mo_Bind1.
                        constructor.
                        rtc_Trans.
                           eapply rtc_R_mo.
                                  eapply STM_Get.
                           eapply Congruence_LC.          
                                  eapply Congruence_LC_Ret;eauto.
                        rtc_Trans.
                           eapply Congruence_BindRet;eauto.
                        rtc_Trans.
                           eapply Congruence_AppAbs_Mo.
                                  auto.
                                  simpl.
                                  rewrite eq_id.
                                  preservation.
                                  rewrite_subst.
                                  eapply rtc_refl.
          split. rtc_Trans.
                 eapply Congruence_Mo_Bind1.
                        constructor.
                        rtc_Trans.
                           eapply rtc_R_mo.
                                  eapply STM_Get.
                           eapply Congruence_LC.          
                                  eapply Congruence_LC_Ret;eauto.
                        rtc_Trans.
                           eapply Congruence_BindRet;eauto.
                        rtc_Trans.
                           eapply Congruence_AppAbs_Mo.
                                  auto.
                                  simpl.
                                  rewrite eq_id.
                                  eapply rtc_refl.
              preservation.
              repeat (split;auto).                       
              constructor;auto.
              constructor;auto.
              inv H5;inv H9.
              do 4 eexists; repeat split.
              eapply rtc_refl.
              eapply rtc_refl.
              eapply eq_square;eauto.
              eapply eq_square;eauto.
Qed. 



Theorem elevate_absorb : forall t t' SM SM' T T',
        smo_less_permissive SM SM' ->                 
        \empty |-- t  \in TMonadic (MNonReact SM)  T  ->
        \empty |-- t' \in TMonadic (MNonReact SM') T' ->                                    
                                        read_only SM' ->
        ((televate SM' t >>[T,T'] t') =[TMonadic (MNonReact SM') T']= t').
Proof. intros ? ? ? ? ? ? HLP HTt HTt' HRO_SM'.
       assert (HRO_SM: read_only SM) by (eapply read_only_smo_lp';eauto).
       lc_normalize t.
       lc_normalize t'.
       preservation.              
       repeat split; [repeat econstructor;eauto;eapply typable_empty_everywhere;eauto
                     |auto|intros ? ? HStoeq12].
       destruct (term_eq_refl _ _ H0) as [_ [_ HReflt]].
       destruct (HReflt _ _ (store_eq_mp _ _ _ _ HLP HStoeq12)) 
                 as [t1 [Sto1' [t2 [Sto2' [? [? [? [? [? ?]]]]]]]]].
       assert ((televate SM' t >>=[T'] (\(6:T) t'), Sto1) ~~~>* (v0, Sto1')).
              rtc_Trans.
                 eapply Congruence_Mo_Bind1.
                        constructor.
                        eapply Congruence_LC.
                        eapply Congruence_ST_Elevate.
                        eauto.
              rtc_Trans.
                 eapply Congruence_Mo_Bind1.
                        constructor.
                        rtc_Trans; 
                          [eapply Congruence_STM_Elevate;eauto
                          |eapply Congruence_STM_ElevateRet;eauto].
              rtc_Trans.
                 eapply Congruence_BindRet;eauto.
              rtc_Trans.
                 eapply Congruence_AppAbs_Mo;eauto.
                        rewrite_subst.
                        eapply Congruence_LC;eauto.
          destruct (term_eq_refl _ _ H) as [_ [_ HReflt']]. 
          destruct (HReflt' _ _ HStoeq12) 
                 as [t1' [Sto1'' [t2' [Sto2'' [? [? [? [? [? ?]]]]]]]]].
          assert (HSeq1: Sto1 = Sto1').
                 eapply step_read_only_no_change_star. 
                 eapply HRO_SM.
                 eapply H1.
                 repeat constructor;eauto.
                 eapply store_eq_all_values_l; eauto.
                 eapply store_eq_matches_l;eauto.
                 eapply store_eq_mp;eauto.
          assert (HSeq1': Sto1 = Sto1'').
                 eapply step_read_only_no_change_star. 
                 eapply HRO_SM'.
                 eapply H8.
                 repeat constructor;eauto.
                 eapply store_eq_all_values_l; eauto.
                 eapply store_eq_matches_l;eauto.
          assert (HSeq2: Sto2 = Sto2'').
                 eapply step_read_only_no_change_star. 
                 eapply HRO_SM'.
                 eapply H9.
                 repeat constructor;eauto.
                 eapply store_eq_all_values_l; eauto.
                 symmetry.
                 eapply HStoeq12.
                 eapply store_eq_matches_r;eauto. 
         exists t1', Sto1'', t2', Sto2''.
         repeat (split;subst;auto).
                rtc_Trans.
                   eapply Congruence_LC;eauto.
                   auto.
Qed.
