Require Import Lib.
Require Import ReWire_Base.
Require Import ReWire_Subst.
Require Import ReWire_Step.
Require Import ReWire_Progress.
Require Import ReWire_Preservation.
Require Import ReWire_Normalization.
Require Import ReWire_EqTh.
Require Import ReWire_Effects.
Require Import ReWire_MonadLaws.

(** ** Mask **)


Create HintDb arb_mask.

Fixpoint arb (T:Ty) : tm :=
  match T with
  | TArrow T1 T2 => tabs 7 T1 (arb T2)
  | TProd T1 T2  => tpair (arb T1) (arb T2)
  | TSum T1 T2   => tinl T2 (arb T1)
  | TNil         => tunit
  | TMonadic M T => treturn M (arb T)
  end.

Theorem arb_ty_empty : forall T, \empty |-- arb T \in T.
Proof.
  induction T; repeat constructor; auto.
  fold arb.
  weakening.
Qed.

Theorem arb_val : forall T, value (arb T).
Proof.
  induction T; repeat constructor; auto.
Qed.

Theorem arb_ty : forall Gamma T, Gamma |-- arb T \in T.
Proof.
  intros.
  eapply typable_empty_everywhere.
  apply arb_ty_empty; auto.
Qed. 

Hint Resolve arb_ty_empty arb_val arb_ty : arb_mask.

Function mask (SM : SMo) : tm :=
   match SM with
   | MIdentity         => treturn (MNonReact MIdentity) tunit
   | MStateT TS E SM'  => (match E with 
                           |  EffW      => tput SM (arb TS) >>[TNil,TNil] (tlift (MNonReact SM) (mask SM')) 
                           |  EffRW     => tput SM (arb TS) >>[TNil,TNil] (tlift (MNonReact SM) (mask SM'))
                           |  EffR      => tlift (MNonReact SM) (mask SM')
                           |  EffNone   => tlift (MNonReact SM) (mask SM')
                          end)
   end.

Lemma mask_eq : forall SM,
       mask SM =
       match SM with
       | MIdentity              => treturn (MNonReact MIdentity) tunit
       | MStateT TS EffNone SM' => tlift (MNonReact SM) (mask SM')
       | MStateT TS EffR SM'    => tlift (MNonReact SM) (mask SM')
       | MStateT TS EffW SM'    =>
           tput SM (arb TS) >>=[TNil] (\(6:TNil) tlift 
                                                  (MNonReact SM) 
                                                  (mask SM'))
       | MStateT TS EffRW SM'   =>
           tput SM (arb TS) >>=[TNil] (\(6:TNil) tlift 
                                                  (MNonReact SM) 
                                                  (mask SM'))
       end.
Proof. induction SM; unfold mask in *; fold mask in *; intros; auto.
Qed.

Lemma mask_ty_empty : forall SM, \empty |-- mask SM \in TMonadic (MNonReact SM) TNil.
Proof. induction SM; rewrite (@mask_eq _); [|destruct e]; repeat econstructor; eauto with arb_mask; weakening.
Qed.

Lemma mask_val : forall SM, value (mask SM).
Proof. induction SM; rewrite (@mask_eq _); [|destruct e]; repeat constructor;eauto with arb_mask.
Qed.

Hint Resolve mask_val mask_ty_empty : arb_mask.

Lemma mask_ty : forall Gamma SM, Gamma |-- mask SM \in TMonadic (MNonReact SM) TNil.
Proof. intros; eapply typable_empty_everywhere;eauto with arb_mask.
Qed.

Hint Resolve mask_ty : arb_mask.

Lemma tlift_ret_step_inj : forall SM TS E v v' Sto1 Sto2,
            value v ->                 
    (tlift (MNonReact (MStateT TS E SM)) (treturn (MNonReact SM) v),Sto1) ~~~>* (v',Sto2) ->
                           Sto1 = Sto2.
Proof.
  intros SM TS E v v' Sto1 Sto2 Hval Hstep.
  inv Hstep.
  pose proof (step_deterministic_mo _ _ _ H (STM_LiftRetSt v TS E SM Sto1 Hval)); subst.
  eapply step_return_no_change_store.
  eapply H0.
Qed.

  
(**
Lemma tunit_unique : forall v, value v -> \empty |-- v \in TNil -> v = tunit.
Proof. induction v; intros Hval HTv; inv Hval;inv HTv.
Qed.*)


Lemma lift_drop_statet : forall t T E SM T' t' s Sto s' Sto',
                           value t' ->
                           (t,Sto) |> TMonadic (MNonReact SM) T' ->
      (tlift (MNonReact (MStateT T E SM)) t,s::Sto) ~~~>* (treturn (MNonReact (MStateT T E SM)) t',s'::Sto') ->
                           (t,Sto) ~~~>* (treturn (MNonReact SM) t',Sto').
Proof. intros t T E SM T' t' s Sto s' Sto' HVt' HTco HStep.
       mo_normalize t.
       destruct (non_reactive_done_mo _ _ _ _ HTco _ _ HV HS) as [v' ?]; subst.
       inv HTco; lc_normalize t.
       assert (HStptv: (v,Sto) ~~~>* (treturn (MNonReact SM) v', Sto0)).
               eapply step_same_done.
                    auto.
                    eapply Congruence_LC.
                    eassumption.
                    auto.
      assert (HstpLift:  (tlift (MNonReact (MStateT T E SM)) t, s :: Sto) ~~~>*  (tlift (MNonReact (MStateT T E SM)) v, s :: Sto)).
             eapply Congruence_LC; eapply Congruence_Lift; eauto.
       assert (HSt:
                (tlift (MNonReact (MStateT T E SM)) v, s :: Sto) ~~~>*  (treturn (MNonReact (MStateT T E SM)) v', s :: Sto0)
).
      rtc_Trans.
         eapply Congruence_LiftSt.
         auto.
         eapply HStptv.
      rtc_Trans.
         eapply Congruence_LiftRetSt.
         inv HV.
         auto.        
       assert (HStep' :
                (tlift (MNonReact (MStateT T E SM)) v, s :: Sto) ~~~>*
                                                                (treturn (MNonReact (MStateT T E SM)) t', s' :: Sto')).
              eapply step_same_done.
                    auto.
                    eassumption.
                    auto. 
     assert (HD1:
               (treturn (MNonReact (MStateT T E SM)) t', s' :: Sto') =  (treturn (MNonReact (MStateT T E SM)) v', s :: Sto0)).
            eapply done_unique.
                   constructor;auto.
                   constructor;auto.
                   inv HV.
                   eapply HStep'.
                   auto.
                   inv HD1.
Qed.      

Lemma eq_store_eq : forall SM S1 S2,
                      S1 = S2 ->
                      store_matches_mo S1 (MNonReact SM) ->
                      store_all_values S1 ->
                      store_eq (MNonReact SM) S1 S2.
Proof. intros. subst.
       eapply store_eq_refl; eauto.
Qed.

Ltac canonical_tunit_term_eq :=
  match goal with
         | [H0: (term_eq TNil ?x ?y), H1: value ?x, H2: value ?y |- _ ] =>
           let TMP1 := fresh in (
             pose proof (canonical_forms_tunit _ (eq_typable_empty_l H0) H1) as TMP1);
           let TMP2 := fresh in (
             pose proof (canonical_forms_tunit _ (eq_typable_empty_r H0) H2) as TMP2);
             subst;
             clear H0 H1 H2; remove_duplicates
  end.

Ltac canonical_tunit_ty_judgement :=
    repeat match goal with
         | [H0: (empty |-- ?y \in TNil), H1: value ?y |- _ ] =>
           let TMP1 := fresh in (
             pose proof (canonical_forms_tunit _ H0 H1) as TMP1);
             subst;
             clear H0 H1; remove_duplicates
       end.

Tactic Notation "canonical_tunit" := first [canonical_tunit_term_eq
                                         |canonical_tunit_ty_judgement
                                         ].
    

(* This would fail; too, I suspect.
                        
Inductive configs_sww : Mo -> configuration -> configuration -> Prop := 
  | SWW_Id : forall (t1 t2 : tm),
                   configs_sww (MNonReact MIdentity) (t1,nil) (t2,nil)
  | SWW_Re : forall (t1 t2 : tm) (S1 S2 : store) (SM : SMo) (T T' : Ty),
                   configs_sww (MNonReact SM)    (t1,S1) (t2,S2) ->
                   configs_sww (MReactT T T' SM) (t1,S1) (t2,S2)
  | SWW_ST_N : forall (t1 t2 : tm) (S1 S2 : store) (SM : SMo) (T : Ty),
                    configs_sww (MNonReact SM) (t1,S1) (t2,S2) ->
                    configs_sww (MNonReact (MStateT T EffNone SM)) (t1,S1) (t2,S2)
  | SWW_ST_R :  forall (t1 t2 : tm) (S1 S2 : store) (SM : SMo) (T : Ty),
                    configs_sww (MNonReact SM) (t1,S1) (t2,S2) ->
                    configs_sww (MNonReact (MStateT T EffR SM)) (t1,S1) (t2,S2).
  | SWW_ST_W : forall (Sto Sto' : store) (SM : SMo) (s : tm) (T : Ty),
                   same_where_write (MNonReact SM) Sto Sto' ->
                   same_where_write (MNonReact (MStateT T EffW SM)) 
                     (s :: Sto) (s :: Sto')
  | SWW_ST_RW : forall (Sto Sto' : store) (SM : SMo) (s : tm) (T : Ty),
                    same_where_write (MNonReact SM) Sto Sto' ->
                    same_where_write (MNonReact (MStateT T EffRW SM))
                      (s :: Sto) (s :: Sto').

Hint Constructors same_where_write. *)
       
Inductive same_where_write : Mo -> store -> store -> Prop :=
    SWW_Identity : same_where_write (MNonReact MIdentity) nil nil
  | SWW_ReactT : forall (Sto Sto' : store) (SM : SMo) (T T' : Ty),
                 same_where_write (MNonReact SM) Sto Sto' ->
                 same_where_write (MReactT T T' SM) Sto Sto'
  | SWW_StateT_None : forall (Sto Sto' : store) (SM : SMo) 
                        (s s' : tm) (T : Ty),
                      same_where_write (MNonReact SM) Sto Sto' ->
                      same_where_write (MNonReact (MStateT T EffNone SM))
                        (s :: Sto) (s' :: Sto')
  | SWW_StateT_R : forall (Sto Sto' : store) (SM : SMo) (s s': tm) (T : Ty),
                   same_where_write (MNonReact SM) Sto Sto' ->
                   same_where_write (MNonReact (MStateT T EffR SM)) 
                     (s :: Sto) (s' :: Sto')
  | SWW_StateT_W : forall (Sto Sto' : store) (SM : SMo) (s : tm) (T : Ty),
                   same_where_write (MNonReact SM) Sto Sto' ->
                   same_where_write (MNonReact (MStateT T EffW SM)) 
                     (s :: Sto) (s :: Sto')
  | SWW_StateT_RW : forall (Sto Sto' : store) (SM : SMo) (s : tm) (T : Ty),
                    same_where_write (MNonReact SM) Sto Sto' ->
                    same_where_write (MNonReact (MStateT T EffRW SM))
                      (s :: Sto) (s :: Sto').

Hint Constructors same_where_write.

Lemma same_where_write_refl : forall M Sto,
                                    store_matches_mo Sto M ->
                                    same_where_write M Sto Sto
with same_where_write_refl_sm : forall SM Sto,
                                      store_matches_mo Sto (MNonReact SM) ->
                                      same_where_write (MNonReact SM) Sto Sto.
Proof.
  destruct M; intros Sto Hmatch.
    constructor.
    apply same_where_write_refl_sm.
    inversion Hmatch; auto.

    apply same_where_write_refl_sm.
    assumption.

  induction SM; intros Sto Hmatch; inversion Hmatch; subst.
    constructor.
    destruct e; constructor; apply IHSM; assumption.
Qed.


Lemma same_where_write_trans : forall M Sto1 Sto2 Sto3,
                                    same_where_write M Sto1 Sto2 ->
                                    same_where_write M Sto2 Sto3 ->
                                    same_where_write M Sto1 Sto3 
with same_where_write_trans_sm : forall SM Sto1 Sto2 Sto3,
                                      same_where_write (MNonReact SM) Sto1 Sto2 ->
                                      same_where_write (MNonReact SM) Sto2 Sto3 ->
                                      same_where_write (MNonReact SM) Sto1 Sto3.
Proof.
  destruct M; intros.
  inversion H.
  inversion H0.
  subst.
  constructor.
    apply (same_where_write_trans_sm _ _ _ _ H6 H12).
    apply (same_where_write_trans_sm _ _ _ _ H H0).

 induction SM; intros;
 [inversion H; inversion H0; subst;auto
 |].
 destruct e; [ | | | ]; 
 destruct Sto1; destruct Sto2; destruct Sto3; try (inversion H; fail) ; try (inversion H0; fail);
 inversion H; inversion H0; subst; constructor;
 match goal with
   | [H: same_where_write (MNonReact SM) ?S1 ?S2, H': same_where_write (MNonReact SM) ?S2 ?S3 |- _ ] =>
     apply (IHSM _ _ _ H H') end.
Qed.
  
Lemma same_where_write_more_permissive :
  forall SM1 SM2 Sto Sto',
    smo_less_permissive SM1 SM2 ->   
    same_where_write (MNonReact SM2) Sto Sto' ->
      same_where_write (MNonReact SM1) Sto Sto'.
Proof.
  induction SM1; intros ? ??  Hless Hsw; destruct SM2;
  [auto
  |inv Hless
  |inv Hless
  |inv Hless ].
  destruct e; destruct e0; inv H1; try (inv Hsw; econstructor;eauto;fail).
Qed.

Theorem mask_same_where_write : forall SM Sto1 Sto2 Sto1' Sto2',
                             store_all_values Sto1                ->
                             store_all_values Sto2                ->
                             store_matches_mo Sto1 (MNonReact SM) ->
                             store_matches_mo Sto2 (MNonReact SM) ->
        (mask SM,Sto1) ~~~>* (treturn (MNonReact SM) tunit,Sto1') ->
        (mask SM,Sto2) ~~~>* (treturn (MNonReact SM) tunit,Sto2') ->
        same_where_write (MNonReact SM) Sto1' Sto2'.
Proof.
  induction SM; intros ? ? ? ? Hallv1 Hallv2 Hmatch1 Hmatch2 Hstep1 Hstep2.
  Case "MIdentity".
    inv Hmatch1; inv Hmatch2;
    pose proof (step_mo_same_length_star _ _ _ _ Hstep1) as H_Len1;
    pose proof (step_mo_same_length_star _ _ _ _ Hstep2) as H_Len2; inv H_Len1; inv H_Len2.
 Case "MStateT".
   assert (HStoeq1: store_eq (MNonReact (MStateT t e SM)) Sto1 Sto1) by (eapply store_eq_refl;eauto);
   assert (HStoeq2: store_eq (MNonReact (MStateT t e SM)) Sto2 Sto2) by (eapply store_eq_refl;eauto);  
   inv Hmatch1; inv Hmatch2.
   assert (HTmask: (mask SM,Sto) |> TMonadic (MNonReact SM) TNil) by (inv Hallv1; inv Hallv2; repeat constructor; eauto with arb_mask);
   assert (HTmask': (mask SM,Sto0) |> TMonadic (MNonReact SM) TNil) by (inv Hallv1; inv Hallv2; repeat constructor; eauto with arb_mask);
   assert (HStoeq1': store_eq (MNonReact SM) Sto   Sto) by (eapply store_eq_stack;eauto);
   assert (HStoeq2': store_eq (MNonReact SM) Sto0 Sto0) by (eapply store_eq_stack;eauto).
   assert (HTu_val: value tunit) by constructor.
   destruct e;
   simpl in Hstep1; simpl in Hstep2;
   try (pose proof (step_mo_same_length_star _ _ _ _ Hstep1) as H_Len1; inv H_Len1;
        pose proof (step_mo_same_length_star _ _ _ _ Hstep2) as H_Len2; inv H_Len2;
        pose proof (lift_drop_statet _ _ _ _ _ _ _ _ _ _ HTu_val HTmask Hstep1) as HSM1;
        pose proof (lift_drop_statet _ _ _ _ _ _ _ _ _ _ HTu_val HTmask' Hstep2) as HSM2;
   try (econstructor; apply (IHSM _ _ _ _
                                       (store_eq_all_values_l HStoeq1')
                                       (store_eq_all_values_l HStoeq2')
                                       (store_eq_matches_l HStoeq1')
                                       (store_eq_matches_l HStoeq2') HSM1 HSM2); fail));
    destruct (term_eq_refl _ _ (mask_ty_empty SM)) as [_ [_ HRefMask]];
    destruct (HRefMask _ _  HStoeq1') as [a [Stoa [x [Stox [? [? [? [? [? ?]]]]]]]]];
    destruct (HRefMask _ _  HStoeq2') as [b [Stob [y [Stoy [? [? [? [? [? ?]]]]]]]]];
    clear HRefMask; canonical_tunit;
    repeat match goal with
       | [H0: (term_eq TNil ?x ?y), H1: value ?x, H2: value ?y |- _ ] =>
         pose proof (canonical_forms_tunit  _ (eq_typable_empty_l H0) H1) as Hpu;
           pose proof (canonical_forms_tunit _ (eq_typable_empty_r H0) H2) as Hpu'; subst;
           clear H0 H1 H2; remove_duplicates end;
    repeat match goal with
             | [H: context[Stox] |- _ ] => clear H
             | [H: context[Stoy] |- _ ] => clear H
           end.
  SCase "EffW".
  assert (Hstep2' :
            (tput (MStateT t EffW SM) (arb t) >>=[TNil] 
             (\(6:TNil) tlift (MNonReact (MStateT t EffW SM)) (mask SM)),
           t1 :: Sto0) ~~~>*
           (treturn (MNonReact (MStateT t EffW SM)) tunit,(arb t)::Stob)
    ).
     rtc_Trans.
             eapply Congruence_Mo_Bind1.
             constructor.
             eapply rtc_R_mo.
                 eapply STM_Put.
                 eauto with arb_mask.
          rtc_Trans.
             eapply Congruence_BindRet; try constructor.
          rtc_Trans.
             eapply Congruence_AppAbs_Mo.
             auto.
             simpl.
             rewrite_subst.
          rtc_Trans.
             eapply Congruence_LiftSt.
             eauto with arb_mask.
             eapply H9.
             eapply Congruence_LiftRetSt.
             auto.
   assert (Hstep1' : (tput (MStateT t EffW SM) (arb t) >>=[TNil] 
             (\(6:TNil) tlift (MNonReact (MStateT t EffW SM)) (mask SM)),
           t0 :: Sto) ~~~>*
                      (treturn (MNonReact (MStateT t EffW SM)) tunit, (arb t):: Stoa)
          ).
    rtc_Trans.
             eapply Congruence_Mo_Bind1.
             constructor.
             eapply rtc_R_mo.
                 eapply STM_Put.
                 eauto with arb_mask.
          rtc_Trans.
             eapply Congruence_BindRet; try constructor.
          rtc_Trans.
             eapply Congruence_AppAbs_Mo.
             auto.
             simpl.
             rewrite_subst.
          rtc_Trans.
             eapply Congruence_LiftSt.
             eauto with arb_mask.
             eapply H.
             eapply Congruence_LiftRetSt.
             auto.
      assert (HD1: 
                (treturn (MNonReact (MStateT t EffW SM)) tunit, arb t :: Stoa) =
                (treturn (MNonReact (MStateT t EffW SM)) tunit, Sto1')
             ).     
             eapply done_unique.
                    constructor; eauto.
                    constructor; eauto.
                    eassumption.
                    eassumption.
      assert (HD2: 
                (treturn (MNonReact (MStateT t EffW SM)) tunit, arb t :: Stob) =
                (treturn (MNonReact (MStateT t EffW SM)) tunit, Sto2')
             ).     
             eapply done_unique.
                    constructor; eauto.
                    constructor; eauto.
                    eassumption.
                    eassumption.
       inv HD1.             
       inv HD2.
       constructor.
       apply (IHSM _ _ _ _
                   (store_eq_all_values_l HStoeq1')
                   (store_eq_all_values_l HStoeq2')
                   (store_eq_matches_l HStoeq1')
                   (store_eq_matches_l HStoeq2') H H9).
   SCase "EffRW".
    assert (Hstep2' :
            (tput (MStateT t EffRW SM) (arb t) >>=[TNil] 
             (\(6:TNil) tlift (MNonReact (MStateT t EffRW SM)) (mask SM)),
           t1 :: Sto0) ~~~>*
           (treturn (MNonReact (MStateT t EffRW SM)) tunit,(arb t)::Stob)
    ).
     rtc_Trans.
             eapply Congruence_Mo_Bind1.
             constructor.
             eapply rtc_R_mo.
                 eapply STM_Put.
                 eauto with arb_mask.
          rtc_Trans.
             eapply Congruence_BindRet; try constructor.
          rtc_Trans.
             eapply Congruence_AppAbs_Mo.
             auto.
             simpl.
             rewrite_subst.
          rtc_Trans.
             eapply Congruence_LiftSt.
             eauto with arb_mask.
             eapply H9.
             eapply Congruence_LiftRetSt.
             auto.
   assert (Hstep1' : (tput (MStateT t EffRW SM) (arb t) >>=[TNil] 
             (\(6:TNil) tlift (MNonReact (MStateT t EffRW SM)) (mask SM)),
           t0 :: Sto) ~~~>*
                      (treturn (MNonReact (MStateT t EffRW SM)) tunit, (arb t):: Stoa)
          ).
    rtc_Trans.
             eapply Congruence_Mo_Bind1.
             constructor.
             eapply rtc_R_mo.
                 eapply STM_Put.
                 eauto with arb_mask.
          rtc_Trans.
             eapply Congruence_BindRet; try constructor.
          rtc_Trans.
             eapply Congruence_AppAbs_Mo.
             auto.
             simpl.
             rewrite_subst.
          rtc_Trans.
             eapply Congruence_LiftSt.
             eauto with arb_mask.
             eapply H.
             eapply Congruence_LiftRetSt.
             auto.
      assert (HD1: 
                (treturn (MNonReact (MStateT t EffRW SM)) tunit, arb t :: Stoa) =
                (treturn (MNonReact (MStateT t EffRW SM)) tunit, Sto1')
             ).     
             eapply done_unique.
                    constructor; eauto.
                    constructor; eauto.
                    eassumption.
                    eassumption.
      assert (HD2: 
                (treturn (MNonReact (MStateT t EffRW SM)) tunit, arb t :: Stob) =
                (treturn (MNonReact (MStateT t EffRW SM)) tunit, Sto2')
             ).     
             eapply done_unique.
                    constructor; eauto.
                    constructor; eauto.
                    eassumption.
                    eassumption.
       inv HD1.             
       inv HD2.
       constructor.
        apply (IHSM _ _ _ _
                   (store_eq_all_values_l HStoeq1')
                   (store_eq_all_values_l HStoeq2')
                   (store_eq_matches_l HStoeq1')
                   (store_eq_matches_l HStoeq2') H H9).
Qed.

Lemma mask_store_eq : forall SM Sto1 Sto2 Sto1' Sto2',
                        store_eq (MNonReact SM) Sto1 Sto2     ->
    (mask SM,Sto1) ~~~>* (treturn (MNonReact SM) tunit,Sto1') ->
    (mask SM,Sto2) ~~~>* (treturn (MNonReact SM) tunit,Sto2') ->
    store_eq (MNonReact SM) Sto1' Sto2'.
Proof. intros SM S1 S2 S1' S2' HSeq HSt1 HSt2.
       destruct (term_eq_refl _ _ (mask_ty_empty SM)) as [_ [_ HMskeq]].
       destruct (HMskeq _ _ HSeq) as [v1 [Sv1 [v2 [Sv2 [HSt1' [HSt2' [HVal1 [HVal2 [HSeq'' HTnil]]]]]]]]].
       assert (HD: (treturn (MNonReact SM) tunit, S1') = (treturn (MNonReact SM) v1, Sv1)
                   /\
                   (treturn (MNonReact SM) tunit, S2') = (treturn (MNonReact SM) v2, Sv2)).
              split; eapply done_unique;eauto.
       destruct HD as [HD1 HD2]; inversion HD1; inversion HD2; subst; auto.
Qed.       
  


Lemma mask_step_SWW_SWNW :
  forall SM Sto1 Sto2,
    same_where_write (MNonReact SM) Sto1 Sto2 ->
    same_where_no_write (MNonReact SM) Sto1 Sto2 -> Sto1 = Sto2.
Proof. induction SM; intros.
       inv H.
       inv H0; inv H; (f_equal;apply (IHSM _ _ H2 H6)).
Qed.

Lemma mask_step_effr : forall SM t t' T Sto Sto',
                         \empty |-- t \in TMonadic (MNonReact SM) T ->
                                          store_matches_mo Sto (MNonReact SM) ->
                                          store_all_values Sto ->
      ((t >>[T,TNil] mask SM), Sto) ~~~>* (t',Sto') -> same_where_no_write (MNonReact SM) Sto Sto'.
Proof.
  intro SM; functional induction (mask SM); intros ? ? ? ? ? HTt Hmatches Hsallv HStep.
  eapply no_forbidden_updates_star.
  assert  (HTB: \empty |-- t >>=[TNil] (\(6:T) treturn (MNonReact MIdentity) tunit) \in TMonadic (MNonReact MIdentity) TNil).
         repeat econstructor; eauto with arb_mask.
  econstructor. eapply HTB. auto. auto.
  eapply HStep.
  assert (HTB: \empty |-- t >>=[TNil] (\(6:T) (tput (MStateT TS EffW SM') (arb TS) >>=[TNil] 
                                 (\(6:TNil) tlift
                                              (MNonReact
                                                 (MStateT TS EffW SM'))
                                              (mask SM')))) \in  TMonadic (MNonReact (MStateT TS EffW SM')) TNil).
           repeat econstructor; eauto with arb_mask.
   eapply no_forbidden_updates_star.
    econstructor. eapply HTB. auto. auto.
  eapply HStep.        
  assert (HTB: \empty |-- t >>=[TNil] (\(6:T) (tput (MStateT TS EffRW SM') (arb TS) >>=[TNil] 
                                 (\(6:TNil) tlift
                                              (MNonReact
                                                 (MStateT TS EffRW SM'))
                                              (mask SM')))) \in  TMonadic (MNonReact (MStateT TS EffRW SM')) TNil).
           repeat econstructor; eauto with arb_mask.
           eapply no_forbidden_updates_star.
    econstructor. eapply HTB. auto. auto.
    eapply HStep.
    assert (HTB: \empty |-- t >>=[TNil] (\(6:T) tlift (MNonReact (MStateT TS EffR SM'))
                                 (mask SM')) \in  TMonadic (MNonReact (MStateT TS EffR SM')) TNil).
           repeat econstructor; eauto with arb_mask.
           eapply no_forbidden_updates_star.
    econstructor. eapply HTB. auto. auto.
    eapply HStep.
     assert (HTB: \empty |-- t >>=[TNil] (\(6:T) tlift (MNonReact (MStateT TS EffNone SM'))
                                 (mask SM')) \in  TMonadic (MNonReact (MStateT TS EffNone SM')) TNil).
           repeat econstructor; eauto with arb_mask.
           eapply no_forbidden_updates_star.
    econstructor. eapply HTB. auto. auto.
    eapply HStep.
Qed.


Lemma clobber_ID : forall T t,
        \empty |-- t \in TMonadic (MNonReact MIdentity) T ->
        t >>[T,TNil] (mask MIdentity) =[TMonadic (MNonReact MIdentity) TNil]= (mask MIdentity). 
Proof. intros; rewrite (@mask_eq _). 
       repeat split; [repeat econstructor; auto
                     |repeat econstructor; auto
                     |intros ? ? HStoeq12].
       pose proof (store_eq_matches_l HStoeq12) as HSto1M;  
       pose proof (store_eq_matches_r HStoeq12) as HSto2M; inv HSto1M; inv HSto2M.
       exists tunit, nil, tunit, nil.
       destruct (term_eq_refl _ _ H) as [_ [_ HReflt]].
       destruct (HReflt _ _  HStoeq12) 
                 as [t1 [Sto1' [t2 [Sto2' [? [? [? [? [? ?]]]]]]]]]. 
       pose proof (store_eq_matches_l H4) as HSto1M';  
       pose proof (store_eq_matches_r H4) as HSto2M'; inv HSto1M'; inv HSto2M'.
       repeat (split; eauto); try (constructor;auto).
       rtc_Trans.
          eapply Congruence_Mo_Bind1.
                 constructor.
                 eapply H1; eauto.
       rtc_Trans.
          eapply Congruence_BindRet; auto.
       rtc_Trans. 
          eapply Congruence_AppAbs_Mo; eauto.
          simpl. eauto.
Qed.

Hint Resolve clobber_ID : arb_mask.

Lemma return_store_eq : forall t S1 S2 S1' S2' SM,
            empty |-- t \in TMonadic (MNonReact SM) TNil ->
            store_eq (MNonReact SM) S1 S1' ->                                 
            (t,S1 ) ~~~>* (treturn (MNonReact SM) tunit, S2 ) ->
            (t,S1') ~~~>* (treturn (MNonReact SM) tunit, S2') ->
            store_eq (MNonReact SM) S2 S2'.
Proof. intros ? ? ? ? ? ?; generalize t SM.
       induction SM; intros ? ? HTt HStoEq12 HStep1 HStep2;  try (eapply IHSM;eauto).
       destruct (term_eq_refl _ _ HTt) as [_ [_ HReflt]].
       destruct (HReflt _ _  HStoEq12) 
         as [t1' [S3 [t2' [S3' [? [? [? [? [? ?]]]]]]]]].
       canonical_tunit.
       assert (HD1: done_mo ((treturn (MNonReact SM) tunit, S2))) by (constructor;auto).
       assert (HD2: done_mo ((treturn (MNonReact SM) tunit, S2'))) by (constructor;auto).
       assert (HD3: done_mo ((treturn (MNonReact SM) tunit, S3))) by (constructor;auto).
       assert (HD4: done_mo ((treturn (MNonReact SM) tunit, S3'))) by (constructor;auto).
       pose proof (done_unique _ _ _ HD1 HD3 HStep1 H).
       pose proof (done_unique _ _ _ HD2 HD4 HStep2 H0).
       inv H1; inv H2.
Qed.

Lemma store_eq_term_eq_st : forall T E SM t1 S1 t2 S2,
      store_eq (MNonReact (MStateT T E SM)) (t1 :: S1) (t2 :: S2) ->
      t1 =[T]= t2.                        
Proof. intros.
       destruct H as [? [? [? [? ?]]]].
       simpl in H3.
       destruct H3 as [t1' [t2' [s1' [s2' [? [? [? ?]]]]]]].
       inv H5.
       inv H6.
Qed.

Lemma term_eq_tail_eq_st : forall T E SM t1 S1 t2 S2,
                             store_eq (MNonReact SM) S1 S2 ->
                             t1 =[T]= t2 ->
                             value t1    ->
                             value t2    ->
                             store_eq (MNonReact (MStateT T E SM)) (t1 :: S1) (t2 :: S2).
Proof. intros; repeat split;
       [constructor;[eapply eq_typable_empty_l;eauto|eapply store_eq_matches_l; eauto]
       |constructor;[auto|eapply store_eq_all_values_l;eauto]
       |constructor;[eapply eq_typable_empty_r;eauto|eapply store_eq_matches_r; eauto]
       |constructor;[auto|eapply store_eq_all_values_r;eauto]
       | ].
       destruct SM;
         [exists t1, t2, S1, S2; repeat split;auto
         | ].
       edestruct H as [HStoM1 [HAV1 [HStoM2 [HAV2 Hsmeq]]]].
       repeat match goal with
                | [H: store_matches_mo ?S (MNonReact (MStateT _ _ _)) |- _ ] => inv H; clear H
                | [H: store_matches_mo ?S (MNonReact MIdentity)       |- _ ] => inv H; clear H
              end.
       assert (t3 =[t]= t0).
              eapply store_eq_term_eq_st;eauto.
       simpl.
       exists t1, t2, (t3 :: Sto0), (t0 :: Sto); split; auto; split.
Qed.
       
Lemma clobber_st : forall t T1 T2,
     empty |-- t \in TMonadic (MNonReact (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))) TNil ->
      (t >>[TNil,TNil] (mask (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))) =[TMonadic (MNonReact (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))) TNil]= (mask (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))).
Proof. intros ? ? ? HTt;
       repeat split; [ repeat econstructor; eauto; eauto with arb_mask
                   |   repeat econstructor; eauto; eauto with arb_mask
                   |   intros S1 S2 HStoEq12].
       destruct (term_eq_refl _ _ HTt) as [_ [_ HReflt]].
       destruct (HReflt _ _  HStoEq12) 
         as [t1' [S1' [t2' [S2' [HStepa [HStepb [HVt1' [HVt2' [HStoEq12' ?]]]]]]]]].
       canonical_tunit.
       pose proof (store_eq_matches_l HStoEq12) as HSto1M; inv HSto1M. 
       pose proof (store_eq_matches_r HStoEq12) as HSto2M; inv HSto2M.
       inv H4. inv H8. clear H8.
       inv H6. inv H9. clear H9.
       pose proof (store_eq_matches_l HStoEq12') as HSto1M'; inv HSto1M'.  
       pose proof (store_eq_matches_r HStoEq12') as HSto2M'; inv HSto2M'.
       inv H12. inv H14. clear H14.
       inv H10. inv H15. clear H15.
       pose proof (store_eq_stack _ _ _ _ _ _ _ HStoEq12) as HStoStk;
         edestruct HStoStk  as [HTar1 [HTav1 [HTar1' [HTsm1 HTsmeq]]]].
      pose proof (store_eq_stack _ _ _ _ _ _ _ HStoEq12') as HStoStk';
         edestruct HStoStk'  as [HTar2 [HTav2 [HTar2' [HTsm2 HTsmeq']]]]. 
      clear HReflt.
      destruct (term_eq_refl _ _
               (mask_ty_empty (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))))
               as [_ [_ HReflmask]].
      destruct (HReflmask _ _ HStoEq12)
        as [t1'
              [S1'
                 [t2'
                    [S2'
                       [HStepa' [HStepb' [HVt1' [HVt2' [HStoEq12'' ?]]]]]]]]];
        canonical_tunit; clear HReflmask.
      assert (HLiftV:
                value ((\(6:TNil) mask
                               (MStateT T1 EffRW
                                  (MStateT T2 EffNone MIdentity))))
             ) by eauto.
      pose proof (Congruence_Mo_Bind1
                  TNil _ _ _ _ _ HLiftV HStepa) as HG1.
      assert (HTunitV: value tunit) by eauto.
      pose proof (Congruence_BindRet
      TNil _ _
           (MNonReact (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))
           (t4 :: t7 :: nil) HTunitV HLiftV) as HG2.
      pose proof (Congruence_AppAbs_Mo
      6 TNil
        (mask (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))
        tunit
        (t4 :: t7 :: nil) HTunitV) as HG3.
       assert (HRE1: [6 := tunit]mask (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)) = mask (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))) by (
       eapply subst_rewrite; eauto with arb_mask).
       rewrite HRE1 in HG3;clear HRE1.
       pose proof (rtc_Trans_mo _ _ _ HG1
                                (rtc_Trans_mo _ _ _ HG2 HG3)) as HG4.
       (* Next step: reduce the mask expression *)
       pose proof (STM_Put
       _ (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))
       (t7 :: nil) t4 (arb_val T1)) as HG5; eapply rtc_R_mo in HG5.
       assert (HLiftV':
                value (\(6:TNil) tlift
                       (MNonReact
                          (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))
                       (tlift (MNonReact (MStateT T2 EffNone MIdentity))
                          (treturn (MNonReact MIdentity) tunit)))
              ) by eauto.
       pose proof (Congruence_Mo_Bind1
                  TNil _ _ _ _ _ HLiftV' HG5) as HG6. 
       Check Congruence_BindRet.
       pose proof (Congruence_BindRet
       TNil _ _
       (MNonReact (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))
       (arb T1 :: t7 :: nil)
       HTunitV HLiftV') as HG7.
       pose proof (Congruence_AppAbs_Mo
       6 TNil
       (tlift (MNonReact (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))
                      (tlift (MNonReact (MStateT T2 EffNone MIdentity))
                         (treturn (MNonReact MIdentity) tunit)))
        tunit
        (arb T1 :: t7 :: nil) HTunitV) as HG7'.
       simpl in HG7'.
       pose proof (rtc_Trans_mo _ _ _ HG6 
                  (rtc_Trans_mo _ _ _ HG7 HG7')) as HG8.
       assert (HG91: (treturn (MNonReact MIdentity) tunit, nil)
                       ~~~>*
                       (treturn (MNonReact MIdentity) tunit, nil)) by eauto.
       assert (HTunitV': value (treturn (MNonReact MIdentity) tunit)) by eauto.
       pose proof (Congruence_LiftSt _ _ _ _
       t7 T2 EffNone MIdentity HTunitV' HG91) as HG92.
       Check Congruence_LiftRetSt.
       pose proof (Congruence_LiftRetSt
       T2 EffNone MIdentity tunit (t7::nil) HTunitV) as HG93. 
       assert (HLiftV'':
                 value (tlift (MNonReact (MStateT T2 EffNone MIdentity))
              (treturn (MNonReact MIdentity) tunit))
              ) by eauto.
       pose proof (Congruence_LiftSt _ _ _ _ 
       (arb T1) T1 EffRW (MStateT T2 EffNone MIdentity) HLiftV'' HG93) as HG94.          Check Congruence_LiftRetSt.                  
       pose proof (Congruence_LiftRetSt
                     T1 EffRW (MStateT T2 EffNone MIdentity)
                     tunit (arb T1 :: t7 :: nil) HTunitV) as HG95.
       pose proof (rtc_Trans_mo _ _ _ HG94 HG95) as HG9.
       clear HG91 HG92 HG93 HG94 HG95.
       pose proof (rtc_Trans_mo _ _ _ HG8 HG9) as HG10.
       assert (HEQmsk:
                 (mask (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))) =
                 tput (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)) (arb T1) >>=[TNil] 
          (\(6:TNil) tlift
                       (MNonReact
                          (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))
                       (tlift (MNonReact (MStateT T2 EffNone MIdentity))
                          (treturn (MNonReact MIdentity) tunit)))

                 ) by (unfold mask in |-*;auto).
       rewrite <- HEQmsk in HG10; clear HEQmsk.
       pose proof (rtc_Trans_mo _ _ _ HG4 HG10) as HGoal1.
       assert (HSWNW:
         same_where_no_write (MNonReact (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))) (t0 :: t2 :: nil) (arb T1 :: t7 :: nil)).
         eapply mask_step_effr.
         eapply HTt.
         eauto.
         eapply store_eq_all_values_l.
         eapply HStoEq12.
         eapply HGoal1.
       assert (Heq27: t2 = t7). 
              inv HSWNW. inv H0.
              rewrite Heq27 in *.
       clear Heq27 HSWNW HG1 HG2 HG3 HG4 HG5 HG6 HG7 HG7' HG8 HG9 HG10.
       (* The other reduction sequence of mask to return *)
       pose proof (store_eq_matches_l HStoEq12'') as HSto1M''; inv HSto1M''.  
       pose proof (store_eq_matches_r HStoEq12'') as HSto2M''; inv HSto2M''.
       inv H18. inv H20. clear H20.
       inv H16. inv H21. clear H21.
       simpl in HStepb'.
      
       pose proof (STM_Put
       _ (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))
       (t3 :: nil) t1 (arb_val T1)) as HG1; eapply rtc_R_mo in HG1.
       pose proof (Congruence_Mo_Bind1
                  TNil _ _ _ _ _ HLiftV' HG1) as HG2. 
       pose proof (Congruence_BindRet
       TNil _ _
       (MNonReact (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))
       (arb T1 :: t3 :: nil)
       HTunitV HLiftV') as HG3.
       pose proof (Congruence_AppAbs_Mo
       6 TNil
       (tlift (MNonReact (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))
                      (tlift (MNonReact (MStateT T2 EffNone MIdentity))
                         (treturn (MNonReact MIdentity) tunit)))
        tunit
        (arb T1 :: t3 :: nil) HTunitV) as HG4'.
       simpl in HG4'.
       pose proof (rtc_Trans_mo _ _ _ HG2 
                  (rtc_Trans_mo _ _ _ HG3 HG4')) as HG5.
       assert (HG91: (treturn (MNonReact MIdentity) tunit, nil)
                       ~~~>*
                       (treturn (MNonReact MIdentity) tunit, nil)) by eauto.
       pose proof (Congruence_LiftSt _ _ _ _
       t7 T2 EffNone MIdentity HTunitV' HG91) as HG92.
       pose proof (Congruence_LiftRetSt
       T2 EffNone MIdentity tunit (t3::nil) HTunitV) as HG93. 
       pose proof (Congruence_LiftSt _ _ _ _ 
       (arb T1) T1 EffRW (MStateT T2 EffNone MIdentity) HLiftV'' HG93) as HG94.
       pose proof (Congruence_LiftRetSt
                     T1 EffRW (MStateT T2 EffNone MIdentity)
                     tunit (arb T1 :: t3 :: nil) HTunitV) as HG95.
       pose proof (rtc_Trans_mo _ _ _ HG94 HG95) as HG6.
       clear HG91 HG92 HG93 HG94 HG95.
       pose proof (rtc_Trans_mo _ _ _ HG5 HG6) as HGoal2.
       assert (HEQmsk:
                 (mask (MStateT T1 EffRW (MStateT T2 EffNone MIdentity))) =
                 tput (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)) (arb T1) >>=[TNil] 
          (\(6:TNil) tlift
                       (MNonReact
                          (MStateT T1 EffRW (MStateT T2 EffNone MIdentity)))
                       (tlift (MNonReact (MStateT T2 EffNone MIdentity))
                          (treturn (MNonReact MIdentity) tunit)))

                 ) by (unfold mask in |-*;auto).
       rewrite <- HEQmsk in HGoal2; clear HEQmsk.
      clear HG1 HG2 HG3 HG4' HG5 HG6.
      exists tunit, (arb T1 :: t7 :: nil), tunit, (arb T1 :: t3 :: nil).
      split; [auto
             |split;[auto
                    |split;[eauto
                           |split;[eauto
                                  |split;[|eapply term_eq_refl;constructor]]]]].
      repeat split; eauto; repeat constructor;eauto; eauto with arb_mask.
      inv HTav1. inv HTsm1.
      simpl.
      exists (arb T1), (arb T1), (t7 :: nil), (t3 :: nil).
      split;[eapply term_eq_refl;eauto with arb_mask
            |split;[exists t7, t3, nil, nil|]]; repeat split;eauto.
      eapply store_eq_term_eq_st;eauto.
Qed.

Lemma put_rw_arb : forall t T, 
  (mask (MStateT T EffRW MIdentity), t :: nil) ~~~>*
            (treturn (MNonReact (MStateT T EffRW MIdentity)) tunit, arb T :: nil).
Proof. intros.
       simpl.
       rtc_Trans.
       eapply Congruence_Mo_Bind1.
              eauto.
       eapply rtc_R_mo.
              eapply STM_Put.
              eauto with arb_mask.
       rtc_Trans.
       eapply Congruence_BindRet.
              eauto.
              eauto.
       rtc_Trans.
       eapply Congruence_AppAbs_Mo.
              eauto.
              simpl.
       rtc_Trans.
       eapply Congruence_LiftRetSt.
              eauto.
       eauto.
Qed.
              
       
Axiom arb_vals_replace: forall T v,
                          empty |-- v \in T ->
                                          value v ->
                          (arb T) =[T]= v. 

Lemma atomic_noninterference  : forall t T1 T2,
     empty |-- t \in TMonadic (MNonReact (MStateT T2 EffRW MIdentity)) TNil ->
    (tlift (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))) t >>[TNil,TNil] (mask (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))))
                     =[TMonadic (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))) TNil]=
                     ((mask (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))) >>[TNil,TNil]
                     tlift (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))) t).
Proof. intros ? ? ? HTt. 
       repeat split;
         [ repeat econstructor; eauto; try (eapply typable_empty_everywhere; eauto with arb_mask)
         | repeat econstructor; eauto; try (eapply typable_empty_everywhere; eauto with arb_mask)
         |   intros S1 S2 HStoEq12].
       pose proof (store_eq_matches_l HStoEq12) as HSto1M; inv HSto1M. 
       pose proof (store_eq_matches_r HStoEq12) as HSto2M; inv HSto2M.
       repeat match goal with
                | [H: store_matches_mo nil (MNonReact MIdentity) |- _ ] => clear H
                | [H: store_matches_mo ?S (MNonReact (MStateT T2 _ _)) |- _ ] => inv H;clear H
                | [H: store_matches_mo ?S (MNonReact MIdentity) |- _ ] => inv H                                                                            end.
       clear HSto1M HSto2M.
       pose proof (store_eq_stack _ _ _ _ _ _ _ HStoEq12) as HSto12Stk.
       rename H2 into HTt0; rename H3 into HTt1; rename H5 into HTt2; rename H6 into HTt3.
        destruct (term_eq_refl _ _ (mask_ty_empty (MStateT T2 EffRW MIdentity))) as [_ [_ HReflt]].
       destruct (HReflt _ _  HSto12Stk) 
         as [? [S1 [? [S2 [HStepM1 [HStepM2 [? [? [HSto12Stk' ?]]]]]]]]].
       canonical_tunit. clear HReflt.
       pose proof (store_eq_matches_l HSto12Stk') as HSto1M; inv HSto1M. 
       pose proof (store_eq_matches_r HSto12Stk') as HSto2M; inv HSto2M.
       repeat match goal with
                | [H: store_matches_mo nil (MNonReact MIdentity) |- _ ] => clear H
                | [H: store_matches_mo ?S (MNonReact (MStateT T2 _ _)) |- _ ] => inv H;clear H
                | [H: store_matches_mo ?S (MNonReact MIdentity) |- _ ] => inv H                                                                            end.
       remove_duplicates.
       rename H2 into HTt4; rename H3 into HTt5.
       pose proof (put_rw_arb t3 T2) as HStepM1'.
       pose proof (put_rw_arb t2 T2) as HStepM2'.
       assert (HDeq1:
                 (treturn (MNonReact (MStateT T2 EffRW MIdentity)) tunit, arb T2 :: nil) =
                 (treturn (MNonReact (MStateT T2 EffRW MIdentity)) tunit, t4 :: nil)
              ) by (
       eapply done_unique;[constructor;eauto
                          |constructor;eauto| eapply HStepM1'| eapply HStepM1]).
              
       assert (HDeq2:
                 (treturn (MNonReact (MStateT T2 EffRW MIdentity)) tunit, arb T2 :: nil) =
                 (treturn (MNonReact (MStateT T2 EffRW MIdentity)) tunit, t5 :: nil)
              ) by (
       eapply done_unique;[constructor;eauto
                          |constructor;eauto|eapply HStepM2'|eapply HStepM2]).
       inv HDeq1;inv HDeq2; clear HDeq1 HDeq2; remove_duplicates.
       clear HTt4.
       assert (HTLift: empty |-- tlift (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))) t
                   \in TMonadic (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))) TNil)
         by (constructor;eauto). 
       destruct (term_eq_refl _ _ HTLift) as [_ [_ HReflt]].
       destruct (HReflt _ _  HStoEq12) 
         as [t1' [S1' [t2' [S2' [HStepa [HStepb [HVt1' [HVt2' [HStoEq12' ?]]]]]]]]].
       canonical_tunit;clear HReflt.
       assert (HSWNW1: same_where_no_write
                         (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                         (t0 :: t3 :: nil) S1').
       eapply no_forbidden_updates_star.
       econstructor.
       eapply HTLift.
       eapply store_eq_all_values_l;eauto.
       eapply store_eq_matches_l;eauto.
       eapply HStepa.
       assert (HSWNW2: same_where_no_write
                         (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                                             (t1 :: t2 :: nil) S2').
       eapply no_forbidden_updates_star.
       econstructor.
       eapply HTLift.
       eapply store_eq_all_values_r;eauto.
       eapply store_eq_matches_r;eauto.
       eapply HStepb.
       pose proof (store_eq_matches_l HStoEq12') as HSto1M'; inv HSto1M'. 
       pose proof (store_eq_matches_r HStoEq12') as HSto2M'; inv HSto2M'.
       repeat match goal with
                | [H: store_matches_mo nil (MNonReact MIdentity) |- _ ] => clear H
                | [H: store_matches_mo ?S (MNonReact (MStateT T2 _ _)) |- _ ] => inv H;clear H
                | [H: store_matches_mo ?S (MNonReact MIdentity) |- _ ] => inv H                                    end.
       remove_duplicates.
       inv HSWNW1; inv HSWNW2.
       assert (HuV: value tunit) by eauto.
       pose proof (rtc_Trans_mo _ _ _
                    (Congruence_LiftSt
                     _ _
                     (t2 :: nil)
                     (arb T2 :: nil)
                     t5
                     T1
                     EffNone
                     (MStateT T2 EffRW MIdentity)
                     (mask_val (MStateT T2 EffRW MIdentity)) HStepM2')
                    (Congruence_LiftRetSt
                    T1
                    EffNone
                    (MStateT T2 EffRW MIdentity)
                    tunit 
                    (t5 :: arb T2 :: nil)
                    HuV)) as HG1.
       assert (MSKeq:
                 (tlift (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                        (mask (MStateT T2 EffRW MIdentity))) =
                  (mask (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                 ) by (unfold mask; auto).
       rewrite MSKeq in HG1; clear MSKeq.
       assert (HTliftV:
                 value ( (\(6:TNil) tlift
                    (MNonReact
                       (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))) t))
              ) by eauto.
       pose proof (rtc_Trans_mo _ _ _ 
                  (Congruence_Mo_Bind1
                  TNil _ _ _ _ _ HTliftV HG1)
                  (rtc_Trans_mo _ _ _
                  (Congruence_BindRet
                  TNil _ _   (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                  (t5 :: arb T2 :: nil) HuV HTliftV)
                  (Congruence_AppAbs_Mo
                   6 TNil 
                  (tlift (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))) t)
                  tunit
                  (t5 :: arb T2 :: nil) HuV))) as HsubG1.
       rewrite_subst.
       lc_normalize t.
       preservation.
       rename H into HTv.
       Check lift_drop_statet.
       assert (HTcoV:
                (v,arb T2 :: nil) |>
                   TMonadic (MNonReact (MStateT T2 EffRW MIdentity)) TNil                
                 ).
                 constructor;eauto.
                 constructor;eauto with arb_mask.
                 constructor.
                 eapply store_eq_matches_l;eauto.
       assert (Heq54: t4 =[T1]= t5).
              eapply store_eq_term_eq_st in HStoEq12;auto.
       assert (HV4: value t4) by
              (pose proof (store_eq_all_values_l HStoEq12) as HSV; inv HSV).
       assert (HV5: value t5) by
              (pose proof (store_eq_all_values_r HStoEq12) as HSV; inv HSV).
       assert (HStoLift : store_eq
               (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
               (t4 :: arb T2 :: nil) (t5 :: arb T2 :: nil)) by (
               eapply term_eq_tail_eq_st;eauto).
       destruct (term_eq_refl _ _ HTLift) as [_ [_ HReflt]].
       destruct (HReflt _ _  HStoLift) 
         as [t1' [S1' [t2' [S2' [HStepL1 [HStepL2 [HVt1' [HVt2' [HStoLift' ?]]]]]]]]].
       canonical_tunit.
       clear HReflt.
       pose proof (store_eq_matches_l HStoLift') as HStoL1'; inv HStoL1'. 
       pose proof (store_eq_matches_r HStoLift') as HStoL2'; inv HStoL2'.
       repeat match goal with
                | [H: store_matches_mo nil (MNonReact MIdentity) |- _ ] => clear H
                | [H: store_matches_mo ?S (MNonReact (MStateT T2 _ _)) |- _ ] => inv H;clear H
                | [H: store_matches_mo ?S (MNonReact MIdentity) |- _ ] => inv H                                    end.
       remove_duplicates.
       assert (HSWNW3 : same_where_no_write
                          (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                          (t4 :: arb T2 :: nil) (t0 :: t9 :: nil)).
              eapply no_forbidden_updates_star.
              constructor.
              eapply HTLift.
              eapply store_eq_all_values_l;eauto.
              eapply store_eq_matches_l;eauto.
              eapply HStepL1.
      assert (HSWNW4 : same_where_no_write
                          (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                          (t5 :: arb T2 :: nil) (t1 :: t8 :: nil)).
              eapply no_forbidden_updates_star.
              constructor.
              eapply HTLift.
              eapply store_eq_all_values_r;eauto.
              eapply store_eq_matches_r;eauto.
              eapply HStepL2.
       inv HSWNW3; inv HSWNW4; clear HSWNW3 HSWNW4.
       pose proof (rtc_Trans_mo _ _ _ HsubG1 HStepL2) as HGoal1.
       assert (HTmaskV:
                 value (\(6:TNil) mask (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
              ) by eauto.

       
       pose proof (rtc_Trans_mo _ _ _ 
                  (Congruence_Mo_Bind1
                  TNil _ _ _ _ _ HTmaskV HStepa) 
                  (rtc_Trans_mo _ _ _
                  (Congruence_BindRet
                  TNil _ _   (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                  (t0 :: t7 :: nil) HuV HTmaskV)
                  (Congruence_AppAbs_Mo
                   6 TNil 
                  (mask (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                  tunit
                  (t0 :: t7 :: nil) HuV))) as HsubG2.
       assert (HRE1: [6 := tunit]mask (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)) = mask (MStateT T1 EffNone (MStateT T2 EffRW MIdentity))) by (
       eapply subst_rewrite; eauto with arb_mask).
       rewrite HRE1 in HsubG2;clear HRE1.
       pose proof (rtc_Trans_mo _ _ _
                  (Congruence_LiftSt _ _ _ _
                                     t0
                                     T1
                                     EffNone
                                     (MStateT T2 EffRW MIdentity)
                                     (mask_val (MStateT T2 EffRW MIdentity))
                                     (put_rw_arb t7 T2)
                  )
                  (Congruence_LiftRetSt
                   T1 EffNone (MStateT T2 EffRW MIdentity)
                   tunit (t0 :: arb T2 :: nil) HuV)) as HsubG3.   
       assert (HMSKeq:
                  (tlift (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                         (mask (MStateT T2 EffRW MIdentity))) =
                      (mask (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
                 ) by (unfold mask;auto).
       rewrite HMSKeq in HsubG3.
       pose proof (rtc_Trans_mo _ _ _ HsubG2 HsubG3) as HGoal2.
       assert (HsubG4: 
       store_eq
               (MNonReact (MStateT T2 EffRW MIdentity))
               (arb T2 :: nil) (t8 :: nil)).
               eapply term_eq_tail_eq_st.
               repeat split;econstructor;eauto.
               eapply arb_vals_replace;eauto; eauto with arb_mask.
               pose proof (store_eq_all_values_r HStoLift') as HG4; inv HG4.
               inv H14.
               eauto with arb_mask.
               pose proof (store_eq_all_values_r HStoLift') as HG4; inv HG4.
               inv H14.
     assert (HGoal3: 
       store_eq
                (MNonReact (MStateT T1 EffNone (MStateT T2 EffRW MIdentity)))
               (t0 :: arb T2 :: nil) (t1 :: t8 :: nil)).
      eapply term_eq_tail_eq_st;eauto.
      exists tunit,(t0 :: arb T2 :: nil),tunit,(t1 :: t8 :: nil).
             split;auto. split; auto.
             split;eauto.
             split;eauto.
             split;auto.
             eapply term_eq_refl.
             constructor.
Qed.


     
   
   
     
       
       
  


 