Require Import Lib.
Require Import ReWire_Base.
Require Import ReWire_Subst.
Require Import ReWire_Step.
Require Import ReWire_Progress.           

Ltac smashed tac := 
  let rec go' :=
      match goal with
        | [ |- context[(if eq_id_dec ?x ?x then _ else _) = _] ] => rewrite eq_id; eauto; go'
        | [ |- context[(if eq_id_dec ?x ?y then _ else _) = _] ] => rewrite neq_id;eauto; go'
        | [ |- ?x <> ?y ] => discriminate; go' (*let TEMP := fresh in (intros TEMP; inversion TEMP); eauto; go'*)
        | _ => auto                                            
      end
  in (tac; (match goal with
               | [ |- context[(extend _ _ _ _ = _)] ] => unfold extend; go'
               | [ |- (_ |-- _ \in _) ]               => weakening
               | [ |- ?x ]                            => idtac x
            end)).

Ltac smash := smashed ltac:(repeat econstructor).

(* ###################################################################### *)
(** *** Lemmas used to establish preservation of typing over the single and
        multistep reduction relation; for lambda, and monadic expressions. *)

(** *** Preservation *)
Theorem preservation' : forall t T,
                          \empty |-- t \in T ->
                          ((forall t', t ~~> t' -> \empty |-- t' \in T)
                           /\ (forall M Tret,
                                 T = TMonadic M Tret ->
                                 forall t' Sto Sto',
                                   store_all_values Sto ->
                                   store_matches_mo Sto M ->
                                   (t,Sto) ~~~> (t',Sto') ->
                                   (\empty |-- t' \in T /\ store_all_values Sto' /\ store_matches_mo Sto' M))).
Proof with eauto.
  remember (@empty Ty) as Gamma.
  intros t T HTt.
  induction HTt; subst Gamma.

  Case "T_Var".
    split.
    SCase "lc".
      intros ? Hstep; inversion Hstep.
    SCase "mo".
      intros ? ? ? ? ? ? ? ? Hstep; inversion Hstep; subst.
      inversion H4.

  Case "T_Abs".
    split.
    SCase "lc".
      intros ? Hstep; inversion Hstep.
    SCase "mo".
      intros ? ? ? ? ? ? ? ? Hstep; inversion Hstep; subst.
      inversion H3.

  Case "T_App".
    assert (forall t' : tm, tapp t1 t2 ~~> t' -> \empty |-- t' \in T').    
      intros ? Hstep.
      inversion Hstep; subst.
      SCase "t1, t2 values".
        inversion HTt1; subst.
        apply (subst_preserves_typing _ _ _ _ _ _ H1 HTt2).
      SCase "t1 steps".
        pose proof (IHHTt1 eq_refl).
        destruct H.
        econstructor.
          eapply H; eauto.
          assumption.
      SCase "t2 steps".
        pose proof (IHHTt2 eq_refl).
        destruct H.
        econstructor.
          eassumption.
          eapply H; eauto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? ? Hstep; inversion Hstep; subst; auto.

  Case "T_Nil".
    split.
    SCase "lc".
      intros ? Hstep; inversion Hstep.
    SCase "mo".
      intros ? ? ? ? ? ? ? ? Hstep; inversion Hstep; subst.
      inversion H3.

  Case "T_Pair".
    split.
    SCase "lc".
      intros ? Hstep; inversion Hstep; subst.
      SSCase "t1 steps".
        pose proof (IHHTt1 eq_refl).
        destruct H.
        econstructor.
          eapply H; eauto.
          assumption.
      SSCase "t2 steps".
        pose proof (IHHTt2 eq_refl).
        destruct H.
        econstructor.
          assumption.
          eapply H; eauto.
    SCase "mo".
      intros ? ? Hismonadic; inversion Hismonadic.

  Case "T_Proj".
    assert (forall t'' : tm, tproj t t' ~~> t'' -> \empty |-- t'' \in T3).
      intros ? Hstep; inversion Hstep; subst.
      SCase "t steps".
        pose proof (IHHTt1 eq_refl).
        destruct H.
        econstructor.
          eapply H; eauto.
          assumption.
      SCase "t' steps".
        pose proof (IHHTt2 eq_refl).
        destruct H.
        econstructor.
          eassumption.
          eapply H; eauto.
      SCase "reduce".
        inversion HTt1; subst.
        repeat econstructor; eauto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? t'' ? ? Hmatch Hstep; inversion Hstep; subst; auto.

  Case "T_InL".
    split.
    SCase "lc".
      intros ? Hstep; inversion Hstep; subst.
      pose proof (IHHTt eq_refl).
      destruct H.
      constructor; auto.
    SCase "mo".
      intros ? ? Hismonadic; inversion Hismonadic.

  Case "T_InR".
    split.
    SCase "lc".
      intros ? Hstep; inversion Hstep; subst.
      pose proof (IHHTt eq_refl).
      destruct H.
      constructor; auto.
    SCase "mo".
      intros ? ? Hismonadic; inversion Hismonadic.

  Case "T_Case".
    assert (forall t' : tm, tcase ts tl tr ~~> t' -> \empty |-- t' \in Tres).
      intros ? Hstep; inversion Hstep; subst.
      SCase "steps".
         pose proof (IHHTt1 eq_refl).
         destruct H as [L R].
         pose proof (L t1' H3).
         eapply T_Case; eauto.
      SCase "reduces to L val".
         eapply T_App; eauto.
         inv HTt1.
      SCase "reduces to R val".
         eapply T_App; eauto.
         inv HTt1.

    split; [ auto | intros ? ? ? ? ? ? ? ? ?; split; [inv H3|inv H3]].
      
  Case "T_Return".
    assert (forall t' : tm, treturn M t ~~> t' -> \empty |-- t' \in TMonadic M T).
      intros ? Hstep; inversion Hstep; subst.
      pose proof (IHHTt eq_refl).
      destruct H.
      constructor; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? Hmatch Hstep; inversion Hstep; subst; auto.

  Case "T_Bind".
    assert (forall t'' : tm, t >>=[T2] t' ~~> t'' -> \empty |-- t'' \in TMonadic M T2).
      intros ? Hstep; inversion Hstep; subst.
      SCase "t steps".
        pose proof (IHHTt1 eq_refl).
        destruct H.
        econstructor; eauto.
      SCase "t' steps".
        pose proof (IHHTt2 eq_refl).
        destruct H.
        econstructor; eauto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? Hmatch Hstep.
      inversion H0; subst.
      inversion Hstep; subst.
      SSCase "lc-step".
        auto.
    
      SSCase "(t,Sto) steps".
        pose proof (IHHTt1 eq_refl).
        destruct H2.
        pose proof (H3 _ _ eq_refl _ _ _ H1 Hmatch H10).
        destruct H4 as [? [? ?]].
        split; auto.
        econstructor; eauto.
      SSCase "(t,Sto) done".
        split; auto.
        inversion HTt1; subst.
        econstructor; eauto.
      SSCase "t is bind-pause".
        split; auto.
        inversion HTt1; subst.
        constructor.
        econstructor;eauto.
        smash. 
          
  Case "T_LiftSt".
    assert (forall t' : tm,
              tlift (MNonReact (MStateT T E SM)) t ~~> t' ->
              \empty |-- t' \in TMonadic (MNonReact (MStateT T E SM)) T').
      intros ? Hstep.
      inversion Hstep; subst.
      pose proof (IHHTt eq_refl).
      destruct H.
      constructor; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? Hav Hmatch Hstep.
      inversion H0; subst.
      inversion Hstep; subst.
      SSCase "lc-step".
        auto.
      SSCase "(t,Sto) steps".
        pose proof (IHHTt eq_refl).
        destruct H1.
        inversion Hmatch; subst.
        inversion Hav; subst.
        pose proof (H2 _ _ eq_refl _ _ _ H8 H11 H9).
        destruct H4 as [? [? ?]].
        split; repeat constructor; auto.
      SSCase "(t,Sto) done".
        split; auto.
        inversion HTt; subst.
        constructor; auto.

  Case "T_LiftRe".
    assert (forall t' : tm,
             tlift (MReactT TI TO SM) t ~~> t' ->
             \empty |-- t' \in TMonadic (MReactT TI TO SM) T).
      intros ? Hstep.
      inversion Hstep; subst.
      pose proof (IHHTt eq_refl).
      destruct H.
      constructor; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? Hmatch Hstep.
      inversion H0; subst.
      inversion Hstep; subst.
      SSCase "lc-step".
        auto.
      SSCase "(t,Sto) steps".
        pose proof (IHHTt eq_refl).
        destruct H2.
        inversion Hmatch; subst.
        pose proof (H3 _ _ eq_refl _ _ _ H1 H7 H10).
        destruct H5 as [? [? ?]]; split; constructor; auto.
      SSCase "(t,Sto) done".
        split; auto.
        inversion HTt; subst.
        constructor; auto.

  Case "T_Elevate".
    assert (forall t' : tm,
              televate SM' t ~~> t' -> \empty |-- t' \in TMonadic (MNonReact SM') T).
      intros ? Hstep.
      inversion Hstep; subst.
      pose proof (IHHTt eq_refl).
      destruct H0.
      econstructor; eauto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? Hmatch Hstep.
      inversion H1; subst.
      inversion Hstep; subst.
      SSCase "lc-step".
        auto.
      SSCase "(t,Sto) steps".
        pose proof (IHHTt eq_refl).
        destruct H3.
        pose proof (more_permissive_store_matches _ _ _ H Hmatch).
        pose proof (H4 _ _ eq_refl _ _ _ H2 H6 H9).
        destruct H7 as [? [? ?]].
        repeat split.
        SSSCase "ty".
          econstructor; eauto.
        SSSCase "all values".
          auto.
        SSSCase "store matches".
          eapply less_permissive_store_matches; eauto.
      SSCase "(t,Sto) done".
        split; auto.
        inversion HTt; subst.
        constructor; auto.

  Case "T_Get".
    split.
    SCase "lc".
      intros ? Hstep; inversion Hstep.
    SCase "mo".
      intros ? ? ? ? ? ? ? Hmatch Hstep.
      inversion H0; subst.
      inversion Hstep; subst.
      SSCase "lc-step".
        inversion H3;subst.
      SSCase "mo-step".
        split; auto.
        inversion Hmatch; subst.
        econstructor; eauto.

  Case "T_Put".
    assert (forall t' : tm,
              tput (MStateT T E SM) t ~~> t' ->
              \empty |-- t' \in TMonadic (MNonReact (MStateT T E SM)) TNil).
      intros ? Hstep.
      inversion Hstep; subst.
      pose proof (IHHTt eq_refl).
      destruct H0.
      constructor; auto.

    split.

    SCase "mo".
      auto.

    SCase "lc".
      intros ? ? ? ? ? ? ? Hmatch Hstep.
      inversion H1; subst.
      inversion Hstep; subst.
      SSCase "lc-step".
        auto.
      SSCase "mo-step".
        inversion Hmatch; subst.
        inversion H2; subst.
        repeat split; repeat constructor; auto.

  Case "T_RunSt".
    assert (forall t' : tm,
              trunst t1 t2 ~~> t' ->
              \empty |-- t' \in TMonadic (MNonReact SM) (TProd T TS)).
      intros ? Hstep.
      inversion Hstep; subst.
      SCase "t1 steps".
        pose proof (IHHTt1 eq_refl).
        destruct H.
        econstructor; auto.
     
      SCase "t2 steps".
        pose proof (IHHTt2 eq_refl).
        destruct H.
        econstructor; eauto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? Hmatch Hstep.
      inversion H0; subst.
      inversion Hstep; subst.
      SSCase "lc-step".
        auto.
      SSCase "mo-step".
        pose proof (IHHTt1 eq_refl).
        destruct H2.
        assert (store_matches_mo (t2::Sto) (MNonReact (MStateT TS E SM))) by (constructor; auto).
        assert (store_all_values (t2::Sto)) by (constructor; auto).
        pose proof (H3 _ _ eq_refl _ _ _ H5 H4 H9).
        destruct H7 as [? [? ?]].
        inversion H11; subst; repeat split; auto.
        econstructor; eauto.
        inversion H10; auto.
      SSCase "mo-done".
        split; auto.
        inversion HTt1; subst; repeat constructor; auto.

  Case "T_RunId".
    assert (forall t' : tm, trunid t ~~> t' -> \empty |-- t' \in T).
      intros ? Hstep.
      inversion Hstep; subst.
      SCase "t1 lc-steps".
        pose proof (IHHTt eq_refl).
        destruct H.
        constructor; auto.
      SCase "t1 mo-steps".
        pose proof (IHHTt eq_refl).
        destruct H.
        assert (store_matches_mo nil (MNonReact MIdentity)) by constructor.
        assert (store_all_values nil) by constructor.
        pose proof (H2 _ _ eq_refl _ _ _ H4 H3 H1).
        destruct H5 as [? [? ?]].
        constructor; auto.
      SCase "t1 done".
        inversion HTt; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? ? Hstep; inversion Hstep; subst; auto.

  Case "T_Pause".
    assert (forall t' : tm,
              tpause (MReactT TI TO SM) T t ~~> t' ->
              \empty |-- t' \in TMonadic (MReactT TI TO SM) T).
      intros ? Hstep.
      inversion Hstep; subst.
      pose proof (IHHTt eq_refl).
      destruct H; constructor; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? ? Hstep; inversion Hstep; subst; auto.

  Case "T_Unfold".
    assert (forall t' : tm,
              tunfold (MReactT TI TO SM) TA TB t1 t2 ~~> t' ->
              \empty |-- t' \in TMonadic (MReactT TI TO SM) TA).
      intros ? Hstep; inversion Hstep; subst.
      SCase "t1 steps".
        pose proof (IHHTt1 eq_refl).
        destruct H; constructor; auto.
      SCase "t2 steps".
        pose proof (IHHTt2 eq_refl).
        destruct H; constructor; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? Hmatch Hstep.
      inversion H0; subst.
      inversion Hstep; subst.
      SSCase "lc-step".
        auto.
      SSCase "mo-step".
        split; auto.
         econstructor;eauto.
        repeat econstructor; eauto.
        eapply typable_empty_everywhere; eauto.
        smash.

  Case "T_RunRe".
    assert (forall t' : tm,
              trunre T t ~~> t' ->
              \empty |-- t' \in
                TMonadic (MNonReact SM)
                         (TSum T (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))))).
      intros ? Hstep.
      inversion Hstep; subst.
      pose proof (IHHTt eq_refl).
      destruct H; repeat constructor; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? ? ? ? ? Hmatch Hstep.
      inversion H0; subst.
      inversion Hstep; subst.
      SSCase "lc-step".
        auto.
      SSCase "mo-step".
        pose proof (IHHTt eq_refl).
        destruct H2.
        assert (store_matches_mo Sto (MReactT TI TO SM)) by (constructor; auto).
        pose proof (H3 _ _ eq_refl _ _ _ H1 H5 H8).
        destruct H6 as [? [? ?]].
        inversion H9; subst.
        split; auto.
        repeat econstructor; eauto.
      SSCase "t is return".
        split; auto.
        inversion HTt; subst.
        repeat econstructor; eauto.
      SSCase "t is pause".
        split; auto.
        inversion HTt; subst.
        smash.
Qed.

Corollary preservation : forall T t,
           \empty |-- t \in T ->
           forall t', t ~~> t' -> \empty |-- t' \in T.
Proof.
  intros T t; apply (preservation' t T); auto.
Qed.

Corollary preservation_mo : forall T co,
          co |> T ->
          forall co',
          co ~~~> co' -> co' |> T.
Proof.
  intros T co HTco co' Hstep.
  destruct co as (t,Sto).
  destruct co' as (t',Sto').
  inversion HTco; subst.
  pose proof (preservation' _ _ H1).
  destruct H.
  pose proof (H0 _ _ eq_refl _ _ _ H2 H4 Hstep).
  destruct H3 as [? [? ?]].
  constructor; auto.
Qed.

Corollary preservation_star : forall T t t',
               t ~~>* t'  ->
      \empty |-- t  \in T ->
      \empty |-- t' \in T.
Proof.
  intros T t t' Hstep HTt.
  generalize dependent T.
  induction Hstep; try auto.
  intros T HTx.
  apply IHHstep. eapply preservation; eauto.
Qed.

Corollary preservation_mo_star : forall T t Sto t' Sto',
      (t,Sto) ~~~>* (t',Sto') ->
                 (t,Sto) |> T ->
               (t',Sto') |> T.
Proof.
  intros T t Sto t' Sto' Hstep HTco.
  generalize dependent T.
  induction Hstep; try auto.
  intros T HTx.
  apply IHHstep. destruct x, y; subst. eapply preservation_mo; eauto.
Qed.

Corollary preservation_mo_star_co : forall T co co',
      co ~~~>* co' ->
           co  |> T ->
           co' |> T.
Proof.
 intros T co co' Hstep HTco.
 destruct co  as (t,Sto).
 destruct co' as (t',Sto').
 eapply preservation_mo_star;eauto.
Qed. 


(* ###################################################################### *)
(** *** A custom tactic to generate all typing judgements that follow from the information
        immediately available in the context. *)

Ltac preservation' tac :=
     tac;
     repeat (match goal with 
     | [HT: (has_type ?G ?t ?T) |- _ ] =>
       repeat (match goal with 
               | [HS : (t ~~>  ?t') |- _ ] => let H := fresh in (pose proof (preservation T t HT t' HS) as H); hide hyp HS
               end); hide hyp HT 
             end); show all hyps;
     repeat (match goal with 
     | [HT: (has_type ?G ?t ?T) |- _ ] =>
       repeat (match goal with 
               | [HS : (t ~~>*  ?t') |- _ ] => let H := fresh in (pose proof (preservation_star T t t' HS HT) as H); hide hyp HS
               end); hide hyp HT 
             end); show all hyps;
     repeat (match goal with 
     | [HT: (configuration_has_type ?t ?T) |- _ ] =>
       repeat (match goal with 
               | [HS : (t ~~~>  ?t') |- _ ] => let H := fresh in (pose proof (preservation_mo T t HT t' HS) as H); hide hyp HS
               end); hide hyp HT 
             end); show all hyps;
     repeat (match goal with 
     | [HT: (configuration_has_type ?t ?T) |- _ ] =>
       repeat (match goal with 
               | [HS : (t ~~~>*  ?t') |- _ ] => let H := fresh in (pose proof (preservation_mo_star_co T t t' HS HT) as H); hide hyp HS
               end); hide hyp HT 
             end); show all hyps.

Ltac preservation := preservation' idtac.