Require Import Lib.
Require Import ReWire_Base.
Require Import ReWire_Subst.
Require Import ReWire_Step.

Theorem progress' : forall t T,
                      \empty |-- t \in T ->
                          (value t \/ exists t', t ~~> t') /\
                          (forall M Tret,
                             T = TMonadic M Tret ->
                             forall Sto,
                               store_matches_mo Sto M ->
                               (done_mo (t,Sto) \/ exists t' Sto', ((t,Sto) ~~~> (t',Sto') /\ same_length Sto Sto'))).
Proof with eauto.
  intros t T Ht.
  remember (@empty Ty) as Gamma.
  induction Ht;
      try (split; [left; constructor | intros ? ? Hismonadic; inversion Hismonadic]; fail);
      try (split; [| intros ? ? Hismonadic; inversion Hismonadic; fail]);
      subst Gamma; try (inversion H; fail).
  Case "T_App".
    assert (exists t' : tm, tapp t1 t2 ~~> t').
    SCase "lc".
    destruct IHHt1 as [[? | ?]]; eauto.
      SSCase "value t1".
      destruct IHHt2 as [[? | ?]]; eauto.
        SSSCase "t2 is a value".
        pose proof (canonical_forms_fun _ _ _ Ht1 H).
        destruct H3 as [x [u Heq]]. subst ++; eauto.
        SSSCase "t2 steps".
        inversion H1 as [t2' Hstp].
        exists (tapp t1 t2'); eauto.
      SSCase "t1 steps".
      inversion H as [t1' Hstp].
      exists (tapp t1' t2); eauto.

      split; auto.

      SCase "mo".
       intros; subst; destruct H; right; repeat eexists; [eapply STM_LC; eauto|eauto].
   Case "T_Pair".
 (*  assert (exists t'0 : tm, tpair t1 t2 ~~> t'0).*)
    destruct IHHt1 as [[? | ?]]...
    SCase "t1 is a value".
      destruct IHHt2 as [[? | ?]]...
      SSCase "t2 steps".
        right. inversion H1 as [t2' Hstp].
        pose proof (ST_Pair2 _ _ _ H Hstp).
        exists (tpair t1 t2'); eauto. 
    SCase "t1 steps".
       right. inversion H as [t1' Hstp].
       pose proof (ST_Pair1 _ _ t2 Hstp).
       exists (tpair t1' t2); auto.
  Case "T_Proj".
  assert (exists t'0 : tm, tproj t t' ~~> t'0).
  SCase "lc".
  destruct IHHt1 as [[? | ?]];eauto.
   SSCase "t is a value".
      destruct IHHt2 as [[? | ?]];eauto.
      SSSCase "t' is a value".
      pose proof (canonical_forms_pair _ _ _ Ht1 H).
      destruct H3 as [x [u Heq]]. subst ++; eauto.
      inv H.
      pose proof (ST_Proj _ _ _ H5 H6 H1).
      exists (tapp (tapp t' x) u); auto.
      SSSCase "t' steps".
      inversion H1 as [t'' Hstp].
      pose proof (ST_Proj2 _ _ _ H Hstp).
      exists ((tproj t t'')); auto.
   SSCase "t steps".
     inversion H as [tt Hstp].
     exists (tproj tt t'); eauto.

     split; auto.

  SCase "mo". 
   intros; subst; destruct H; right; repeat eexists; [eapply STM_LC; eauto | eauto].

  Case "T_Inl".
    destruct IHHt as [[? | ?]]...
    SCase "t steps".
      right. inversion H as [t' Hstp]. 
      exists (tinl T2 t'); eauto.

  Case "T_Inr".
    destruct IHHt as [[? | ?]]...
    SCase "t steps".
      right. inversion H as [t' Hstp]. 
      exists (tinr T1 t'); eauto.

  Case "T_Case".
    assert (exists t' : tm, tcase ts tl tr ~~> t').
  SCase "Lambda Calculus".
  destruct IHHt1 as [[? | ?]]; eauto.
    SSCase "ts is a value".
    pose proof (canonical_forms_copair _ _ _ Ht1 H).
    destruct H1 as [H1|H1]; destruct H1 as [x Heq]; subst; inversion H.
    SSSCase "Tl".
      eexists; eapply ST_CaseL; eauto.
    SSSCase "Tr".
      eexists; eapply ST_CaseR; eauto.
    SSCase "ts steps".
      inversion H as [t' Hstp].
      pose proof (ST_Case _ _ tl tr Hstp).
      exists (tcase t' tl tr); eauto.

      split; auto.
    
  SCase "mo". 
   intros; subst; destruct H; right; repeat eexists; [eapply STM_LC; eauto | eauto].


  Case "T_Return".
    split.
    SCase "lc".
      destruct IHHt as [[? | ?]]...
      SSCase "t steps".
        right. inversion H as [t' Hstp].
        exists (treturn M t'); eauto.
    SCase "mo".
      intros.
      destruct IHHt as [[? | ?]]...
      SSCase "t steps".
        right. inversion H1 as [t' Hstp]; eexists; eexists; split.
        SSSCase "step".
          eapply STM_LC; eapply ST_Return; eauto.
        SSSCase "shape".
          apply same_length_refl.

  Case "T_Bind".
    split.
    SCase "lc".
      destruct IHHt1 as [[? | ?]]...
      SSCase "t is a value".
        destruct IHHt2 as [[? | ?]]...
        SSSCase "t' steps".
          right. inversion H1 as [t'0 Hstp].
          exists (t >>=[T2] t'0); eauto. 
      SSCase "t steps".
        right. inversion H as [t0 Hstp].
        exists (t0 >>=[T2] t'); eauto.
    SCase "mo".
      intros.
      destruct IHHt1 as [[? | ?]]...
      SSCase "t is a value".
        destruct IHHt2 as [[? | ?]]...
        SSSCase "t' is a value".
          inversion H; subst.
          destruct (H2 M0 T1 eq_refl Sto H0).
          SSSSCase "(t,Sto) is done".
            inversion H5; subst.
            SSSSSCase "t is return".
              right. 
              do 2 eexists; split.
              SSSSSSCase "step".
                eapply STM_BindRet; eauto.
              SSSSSSCase "shape".
                apply same_length_refl.
              SSSSSCase "t is pause".
                right.
                inversion Ht1; subst.
                do 2 eexists; split.
              SSSSSSCase "step".
              eapply STM_PauseBind; eauto.
              SSSSSSCase "shape".
                apply same_length_refl.
          SSSSCase "(t,Sto) steps".
            right; destruct H5 as [? [? [? ?]]]; eexists; eexists; split; [eapply STM_Bind1; eauto | auto].
        SSSCase "t' steps".
          right; destruct H3; eexists; eexists; split; [eapply STM_LC; eapply ST_Bind2; eauto | apply same_length_refl].
      SSCase "t' steps".
        right; destruct H1; eexists; eexists; split; [eapply STM_LC; eapply ST_Bind1; eauto | apply same_length_refl].

  Case "T_LiftSt".
    split.
    SCase "lc".
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t steps".
        right. inversion H as [t' Hstp].
        eexists; econstructor; eauto.
    SCase "mo".
      intros.
      right.
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t is a value".
        inversion H; subst.
        inversion H0; subst.
        destruct (H2 (MNonReact SM) Tret eq_refl _ H8).
        SSSCase "(t,Sto0) done".
          inversion H3; subst.
          SSSSCase "t is return".
            inversion Ht; subst.
            eexists; eexists; split; [eapply STM_LiftRetSt; eauto | apply same_length_refl].
          SSSSCase "t is signal".
            inversion Ht.
        SSSCase "(t,Sto0) steps".
          destruct H3 as [? [? [? ?]]]; eexists; eexists; split; [eapply STM_LiftSt; eauto | constructor; auto].
      SSCase "t steps".
        destruct H1; eexists; eexists; split; [eapply STM_LC; econstructor; eauto | apply same_length_refl].

  Case "T_LiftRe".
    split.
    SCase "lc".
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t steps".
        right.
        destruct H as [t' Hstp].
        eexists; econstructor; eauto.
    SCase "mo".
      intros.
      right.
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t is a value".
        inversion H; subst.
        inversion H0; subst.
        destruct (H2 (MNonReact SM) Tret eq_refl _ H5).
        SSSCase "(t,Sto) done".
          inversion H3; subst.
          SSSSCase "t is return".
            inversion Ht; subst.
            eexists; eexists; split; [eapply STM_LiftRetRe; eauto | apply same_length_refl].
          SSSSCase "t is pause".
            inversion Ht.
        SSSCase "(t,Sto) steps".
          destruct H3 as [? [? [? ?]]]; eexists; eexists; split; [eapply STM_LiftRe; eauto | auto].
      SSCase "t steps".
        destruct H1; eexists; eexists; split; [eapply STM_LC; econstructor; eauto | apply same_length_refl].

  Case "T_Elevate".
    split.
    SCase "lc".
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t steps".
        right.
        destruct H0 as [t' Hstp].
        eexists; econstructor; eauto.
    SCase "mo".
      intros.
      right.
      inversion H0; subst.
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t is a value".
        pose proof (more_permissive_store_matches _ _ _ H H1).
        destruct (H3 _ _ eq_refl Sto H4).
        SSSCase "(t,Sto) done".
          inversion H5; subst.
          SSSSCase "t is treturn".
            inversion Ht; subst.
            eexists; eexists; split; [eapply STM_ElevateRet; eauto | apply same_length_refl].
          SSSSCase "t is pause".
            inversion Ht.
        SSSCase "(t,Sto) steps".
          destruct H5 as [? [? [? ?]]]; eexists; eexists; split; [eapply STM_Elevate; eauto | auto].
      SSCase "t steps".
        destruct H2; eexists; eexists; split; [eapply STM_LC; econstructor; eauto | apply same_length_refl].

  Case "T_Get".
    split.
    SCase "lc".
      left; constructor.
    SCase "mo".
      intros.
      right.
      inversion H0; subst.
      inversion H1; subst.
      eexists; eexists; split; [eapply STM_Get; eauto | apply same_length_refl].

  Case "T_Put".
    split.
    SCase "lc".
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t steps".
        right. destruct H0 as [? ?]; eexists; eapply ST_Put; eauto.
    SCase "mo".
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t is a value".
        intros.
        right.
        inversion H2; subst.
        inversion H3; subst.
        eexists; eexists; split; [eapply STM_Put; eauto | constructor; apply same_length_refl].
      SSCase "t steps".
        intros; right; destruct H0 as [? ?]; eexists; eexists; split; [eapply STM_LC; eapply ST_Put; eauto | apply same_length_refl].

  Case "T_RunSt".
    split.
    SCase "lc".
      destruct IHHt1 as [[? | ?]]; eauto.
      SSCase "t1 is a value".
        destruct IHHt2 as [[? | ?]]; eauto.
        SSSCase "t2 steps".
          right; destruct H1; eexists; eapply ST_RunSt2; eauto.
      SSCase "t1 steps".
        right; destruct H; eexists; eapply ST_RunSt1; eauto.
    SCase "mo".
      intros.
      destruct IHHt1 as [[? | ?]]; eauto.
      SSCase "t1 is a value".
        destruct IHHt2 as [[? | ?]]; eauto.
        SSSCase "t2 is a value".
          right.
          inversion H; subst.
          assert (store_matches_mo (t2::Sto) (MNonReact (MStateT TS E SM))) by (constructor; auto).
          destruct (H2 _ _ eq_refl _ H5).
          SSSSCase "(t1,t2::Sto) done".
            inversion H6; subst.
            SSSSSCase "t1 is return".
              inversion Ht1; subst.
              eexists; eexists; split; [eapply STM_RunStRet; eauto | apply same_length_refl].
            SSSSSCase "t1 is pause".
              inversion Ht1.
          SSSSCase "(t1,t2::Sto) steps".
            destruct H6 as [t1' [Sto' [? ?]]].
            inversion H7; subst.
            eexists; eexists; split; [eapply STM_RunSt; eauto | auto].
        SSSCase "t2 steps".
          right; destruct H3; eexists; eexists; split; [eapply STM_LC; eapply ST_RunSt2; eauto | apply same_length_refl].
      SSSCase "t1 steps".
        right; destruct H1; eexists; eexists; split; [eapply STM_LC; eapply ST_RunSt1; eauto | apply same_length_refl].

  Case "T_RunId".
    assert (exists t' : tm, trunid t ~~> t').
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t is a value".
        assert (store_matches_mo nil (MNonReact MIdentity)) by constructor.
        destruct (H0 _ _ eq_refl _ H1).
        SSSCase "(t,nil) done".
          inversion H2; subst.
          SSSSCase "t is return".
            eexists; apply ST_RunIdRet; auto.
          SSSSCase "t is pause".
            inversion Ht.
        SSSCase "(t,nil) steps".
          destruct H2 as [t' [? [? ?]]].
          inversion H3; subst.
          eexists; eapply ST_RunIdMo; eauto.
      SSCase "t steps".
        destruct H; eexists; eapply ST_RunId; eauto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros; subst.
      right.
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t is a value".
        assert (store_matches_mo nil (MNonReact MIdentity)) by constructor.
        destruct (H2 _ _ eq_refl _ H3).
        SSSCase "(t,nil) done".
          inversion H4; subst.
          SSSSCase "t is return".
            destruct H;
            eexists; eexists; split;
              [eapply STM_LC; eauto
              | apply same_length_refl].
          SSSSCase "t is pause".
            inversion Ht.
        SSSCase "(t,nil) steps".
          destruct H4 as [t' [? [? ?]]].
          inversion H5; subst.
          destruct H;
          eexists; eexists; split; [eapply STM_LC; eauto | apply same_length_refl].
      SSCase "t steps".
        destruct H; eexists; eexists; split; [eapply STM_LC; eauto | apply same_length_refl].

  Case "T_Pause".
    split.
    SCase "lc".
      destruct IHHt as [[? | ?]]...
      SSCase "t steps".
        right; destruct H; eexists; eapply ST_Pause; eauto.
    SCase "mo".
      intros.
      destruct IHHt as [[? | ?]]...
      SSCase "t steps".
        right; destruct H1; eexists; eexists; split; [eapply STM_LC; eapply ST_Pause; eauto | apply same_length_refl].

  Case "T_Unfold".
    split.
    SCase "lc".
      destruct IHHt1 as [[? | ?]]...
      SSCase "t1 is a value".
        destruct IHHt2 as [[? | ?]]...
        SSSCase "t2 steps".
          right; destruct H1; eexists; eapply ST_Unfold2; eauto.
      SSCase "t1 steps".
        right; destruct H; eexists; eapply ST_Unfold1; eauto.

    SCase "mo".
      intros.
      right.
      inversion H; subst.
      inversion H0; subst.
      destruct IHHt1 as [[? | ?]]...
      SSCase "t1 is a value".
      destruct IHHt2 as [[? | ?]]...
      SSSCase "t2 steps".
          destruct H4; eexists; eexists; split; [eapply STM_LC; eapply ST_Unfold2; eauto | apply same_length_refl].      
      SSCase "t1 steps".
        destruct H1; eexists; eexists; split; [eapply STM_LC; eapply ST_Unfold1; eauto | apply same_length_refl].

  Case "T_RunRe".
    split.
    SCase "lc".
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t steps".
        right; destruct H; eexists; eapply ST_RunRe; eauto.
    SCase "mo".
      intros.
      right.
      destruct IHHt as [[? | ?]]; eauto.
      SSCase "t is a value".
        inversion H; subst.
        assert (store_matches_mo Sto (MReactT TI TO SM)) by (constructor; auto).
        destruct (H2 _ _ eq_refl _ H3).
        SSSCase "(t,Sto) done".
          inversion H4; subst.
          SSSSCase "t is return".
            inversion Ht; subst.
            eexists; eexists; split; [eapply STM_RunReRet; eauto | apply same_length_refl].
          SSSSCase "t is pause".
            inversion Ht; subst.
            eexists; eexists; split; [eapply STM_RunRePause; eauto | apply same_length_refl].
        SSSCase "(t,Sto) steps".
          destruct H4 as [? [? [? ?]]]; eexists; eexists; split; [eapply STM_RunRe; eauto | auto].
      SSCase "t steps".
        destruct H1; eexists; eexists; split; [eapply STM_LC; eapply ST_RunRe; eauto | apply same_length_refl].
Qed.

Corollary progress : forall t T,
                       \empty |-- t \in T ->
                       (value t \/ exists t', t ~~> t').
Proof.
  intros ? ? H; destruct (progress' _ _ H); auto.
Qed.

Corollary progress_mo : forall co T,
                          co |> T ->
                          (done_mo co \/ exists co', co ~~~> co').
Proof.
  intros ? ? Hty.
  destruct Hty.
  pose proof (progress' _ _ H).
  destruct H2.
  pose proof (H3 _ _ eq_refl _ H1).
  destruct H4; auto.
  right.
  destruct H4 as [? [? [? ?]]].
  eexists; eauto.
Qed.

Corollary not_value_step : forall t T,
                             \empty |-- t \in T ->
                             ~ value t ->
                             exists t', t ~~> t'.
Proof.
  intros t T Ht Hnv.
  destruct (progress t T Ht); [contradiction | assumption].
Qed.

Corollary not_value_step_mo : forall t T Sto,
                           \empty |-- t \in T ->
                           ~ value t ->
                           exists co', (t,Sto) ~~~> co'.
Proof.
  intros t T Sto Ht Hnv.
  assert (exists t', t ~~> t') as [t' Hstep].
    eapply not_value_step; eassumption; assumption.
  exists (t',Sto); apply STM_LC; assumption.
Qed.

Corollary not_done_step_mo : forall co T,
                               co |> T ->
                               ~ done_mo co ->
                               exists co', co ~~~> co'.
Proof.
  intros co T HT Hnd.
  destruct (progress_mo co T HT); [contradiction | assumption].
Qed.
