Require Import Lib.
Require Import ReWire_Base.
Require Import ReWire_Subst.

(** * Reduction *)

(** ** Lambda-calculus and monadic reduction relations *)
Reserved Notation "t1 '~~>' t2" (at level 40).
Reserved Notation "t1 '~~~>' t2" (at level 40).

Inductive step : tm -> tm -> Prop :=
  | ST_AppAbs : forall x T t12 v2,
                  value v2 ->
                  (tapp (tabs x T t12) v2) ~~> [x:=v2]t12
  | ST_App1   : forall t1 t1' t2,
                  t1 ~~> t1' ->
                  tapp t1 t2 ~~> tapp t1' t2
  | ST_App2   : forall v1 t2 t2',
                  value v1 ->
                  t2 ~~> t2' ->
                  tapp v1 t2 ~~> tapp v1 t2'
  | ST_Pair1  : forall t1 t1' t2,
                  t1 ~~> t1' ->
                  tpair t1 t2 ~~> tpair t1' t2
  | ST_Pair2  : forall v1 t2 t2',
                  value v1 ->
                  t2 ~~> t2' ->
                  tpair v1 t2 ~~> tpair v1 t2'
  | ST_Proj1  : forall t1 t1' t2,
                  t1 ~~> t1' ->
                  tproj t1 t2 ~~> tproj t1' t2
  | ST_Proj2  : forall v1 t2 t2',
                  value v1 ->
                  t2 ~~> t2' ->
                  tproj v1 t2 ~~> tproj v1 t2'
  | ST_Proj   : forall v1 v2 v3,
                  value v1 ->
                  value v2 ->
                  value v3 ->
                  tproj (tpair v1 v2) v3 ~~> tapp (tapp v3 v1) v2
  | ST_InL    : forall T t1 t1',
                  t1 ~~> t1' ->
                  tinl T t1 ~~> tinl T t1'
  | ST_InR    : forall T t1 t1',
                  t1 ~~> t1' ->
                  tinr T t1 ~~> tinr T t1'
  | ST_Case   : forall t1 t1' t2 t3,
                  t1 ~~> t1' ->
                  tcase t1 t2 t3 ~~> tcase t1' t2 t3
  | ST_CaseL  : forall v1 T t2 t3,
                  value v1 ->
                  tcase (tinl T v1) t2 t3 ~~> tapp t2 v1
  | ST_CaseR  : forall v1 T t2 t3,
                  value v1 ->
                  tcase (tinr T v1) t2 t3 ~~> tapp t3 v1
  | ST_Return : forall M t1 t1',
                  t1 ~~> t1' ->
                  treturn M t1 ~~> treturn M t1'
  | ST_Bind1  : forall T t1 t1' t2,
                  t1 ~~> t1' ->
                  tbind T t1 t2 ~~> tbind T t1' t2
  | ST_Bind2  : forall T t1 t2 t2',
                  value t1 ->
                  t2 ~~> t2' ->
                  tbind T t1 t2 ~~> tbind T t1 t2'
  | ST_Lift   : forall M t1 t1',
                  t1 ~~> t1' ->
                  tlift M t1 ~~> tlift M t1'
  | ST_Elevate  : forall SM t1 t1',
                    t1 ~~> t1' ->
                    televate SM t1 ~~> televate SM t1'
  | ST_Put      : forall SM t1 t1',
                    t1 ~~> t1' ->
                    tput SM t1 ~~> tput SM t1'
  | ST_RunSt1   : forall t1 t1' t2,
                    t1 ~~> t1' ->
                    trunst t1 t2 ~~> trunst t1' t2
  | ST_RunSt2   : forall t1 t2 t2',
                    value t1 ->
                    t2 ~~> t2' ->
                    trunst t1 t2 ~~> trunst t1 t2'
  | ST_RunId    : forall t1 t1',
                    t1 ~~> t1' ->
                    trunid t1 ~~> trunid t1'
  | ST_RunIdMo  : forall t1 t1',
                    value t1 ->
                    (t1,nil) ~~~> (t1',nil) ->
                    trunid t1 ~~> trunid t1'
  | ST_RunIdRet : forall M v,
                    value v ->
                    trunid (treturn M v) ~~> v
  | ST_Pause   : forall M T t1 t1',
                   t1 ~~> t1' ->
                   tpause M T t1 ~~> tpause M T t1'
  | ST_Unfold1 : forall M TA TB t1 t1' t2,
                   t1 ~~> t1' ->
                   tunfold M TA TB t1 t2 ~~> tunfold M TA TB t1' t2
  | ST_Unfold2 : forall M TA TB t1 t2 t2',
                   value t1 ->
                   t2 ~~> t2' ->
                   tunfold M TA TB t1 t2 ~~> tunfold M TA TB t1 t2'
  | ST_RunRe   : forall T t1 t1',
                   t1 ~~> t1' ->
                   trunre T t1 ~~> trunre T t1'

with step_mo : configuration -> configuration -> Prop :=
  | STM_LC        : forall t t' Sto,
                      t ~~> t' ->
                      (t,Sto) ~~~> (t',Sto)   
  | STM_Bind1     : forall T t1 t1' t2 Sto Sto',
                      value t1 ->
                      value t2 ->
                      (t1,Sto) ~~~> (t1',Sto') ->
                      (tbind T t1 t2,Sto) ~~~> (tbind T t1' t2,Sto')
  | STM_BindRet   : forall v1 v2 T M Sto,
                      value v1 ->
                      value v2 ->
                      (tbind T (treturn M v1) v2,Sto) ~~~> (tapp v2 v1,Sto)
  | STM_LiftSt    : forall t t' Sto Sto' s TS b SM,
                      value t ->
                      (t,Sto) ~~~> (t',Sto') ->
                      (tlift (MNonReact (MStateT TS b SM)) t,s::Sto) ~~~>
                        (tlift (MNonReact (MStateT TS b SM)) t',s::Sto')
  | STM_LiftRe    : forall t t' Sto Sto' TI TO SM,
                      value t ->
                      (t,Sto) ~~~> (t',Sto') ->
                      (tlift (MReactT TI TO SM) t,Sto) ~~~>
                        (tlift (MReactT TI TO SM) t',Sto')
  | STM_LiftRetSt : forall v TS b SM Sto,
                      value v ->
                      (tlift (MNonReact (MStateT TS b SM)) (treturn (MNonReact SM) v),Sto)
                         ~~~> (treturn (MNonReact (MStateT TS b SM)) v,Sto)
  | STM_LiftRetRe : forall v TI TO SM Sto,
                      value v ->
                      (tlift (MReactT TI TO SM) (treturn (MNonReact SM) v),Sto)
                         ~~~> (treturn (MReactT TI TO SM) v,Sto)
  | STM_Get       : forall SM Sto s,
                      (tget SM,s::Sto)
                         ~~~> (treturn (MNonReact SM) s,s::Sto)
  | STM_Put       : forall v SM Sto s,
                      value v ->
                      (tput SM v,s::Sto)
                         ~~~> (treturn (MNonReact SM) tunit,v::Sto)
  | STM_Elevate   : forall SM t1 t1' Sto Sto',
                      value t1 ->
                      (t1,Sto) ~~~> (t1',Sto') ->
                      (televate SM t1,Sto)
                         ~~~> (televate SM t1',Sto')
  | STM_ElevateRet : forall SM SM' v Sto,
                      value v ->
                      (televate SM' (treturn (MNonReact SM) v),Sto)
                         ~~~> (treturn (MNonReact SM') v,Sto)
  | STM_RunSt      : forall t1 t1' s s' Sto Sto',
                       value t1 ->
                       value s ->
                       (t1,s::Sto) ~~~> (t1',s'::Sto') ->
                       (trunst t1 s,Sto) ~~~> (trunst t1' s',Sto')
  | STM_RunStRet   : forall t1 s Sto TS b SM,
                       value t1 ->
                       value s ->
                       (trunst (treturn (MNonReact (MStateT TS b SM)) t1) s,Sto)
                         ~~~> (treturn (MNonReact SM) (tpair t1 s),Sto)
  | STM_Unfold     : forall t1 t2 Sto TI TO SM TA TB,
                       value t1 ->
                       value t2 ->
                       (tunfold (MReactT TI TO SM) TA TB t1 t2,Sto) ~~~>
                       (tbind TA
                              (tlift (MReactT TI TO SM) (tapp t2 t1))
                              (tabs 0 (TSum TA (TProd TO (TArrow TI TB)))
                                (tcase (tvar 0)
                                   (tabs 1 TA (treturn (MReactT TI TO SM) (tvar 1)))
                                   (tabs 1 (TProd TO (TArrow TI TB))
                                      (tproj
                                         (tvar 1)
                                         (tabs 2 TO
                                           (tabs 3 (TArrow TI TB)
                                             (tpause (MReactT TI TO SM) TA
                                                  (treturn (MNonReact SM)
                                                    (tpair (tvar 2) 
                                                           (tabs 4 TI
                                                             (tunfold (MReactT TI TO SM) TA TB
                                                               (tapp (tvar 3) (tvar 4))
                                                               t2))))))))))),Sto)
  | STM_PauseBind  : forall t1 t2 Sto TI TO SM T1 T2,
                       value t1 ->
                       value t2 ->
                       (tbind
                          T2
                          (tpause
                             (MReactT TI TO SM)
                             T1
                             t1)
                          t2,Sto) ~~~>
                         (tpause
                            (MReactT TI TO SM)
                            T2
                            (tbind
                               (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T2)))
                               t1
                               (tabs 0 (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T1)))
                                 (tproj
                                   (tvar 0)
                                   (tabs 1 TO
                                     (tabs 2 (TArrow TI (TMonadic (MReactT TI TO SM) T1))
                                       (treturn (MNonReact SM)
                                          (tpair
                                            (tvar 1)
                                            (tabs 3 TI
                                              (tbind
                                                T2
                                                (tapp (tvar 2) (tvar 3))
                                                t2))))))))),Sto)
  | STM_RunRe      : forall T t1 t1' Sto Sto',
                       value t1 ->
                       (t1,Sto) ~~~> (t1',Sto') ->
                       (trunre T t1,Sto) ~~~> (trunre T t1',Sto')
  | STM_RunReRet   : forall v TI TO SM TA Sto,
                       value v ->
                       (trunre TA (treturn (MReactT TI TO SM) v),Sto) ~~~>
                          (treturn
                             (MNonReact SM)
                             (tinl (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) TA))) v),Sto)
  | STM_RunRePause : forall v TI TO SM TA Sto,
                       value v ->
                       (trunre TA (tpause (MReactT TI TO SM) TA v),Sto) ~~~>
                          (tbind (TSum TA (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) TA))))
                                 v
                                 (tabs 0 (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) TA)))
                                       (treturn (MNonReact SM) (tinr TA (tvar 0)))),Sto)
where "k1 '~~>' k2"  := (step k1 k2)
                          and "k1 '~~~>' k2" := (step_mo k1 k2).

Hint Constructors step.
Hint Constructors step_mo. 

Section Reflexive_Transitive_Closure.
Variables (X : Type) (R : relation X).

Inductive rt_closure : relation X :=
    | rtc_refl  : forall (x : X), 
                  rt_closure x x
    | rtc_step : forall (x y z : X),
                  R x y ->  
                  rt_closure y z ->
                  rt_closure x z.

Hint Resolve rtc_refl.
Hint Constructors rt_closure.

Theorem rtc_R' : forall (x y : X), 
        R x y -> rt_closure x y.
Proof.
  intros.
  apply rtc_step with y. 
  apply H. apply rtc_refl.    
Qed.

Hint Resolve rtc_R'.

(* the reflexive_transitive_closure is _transitive_ *)
Theorem rtc_Trans' :forall (x y z : X),
                    rt_closure x y  ->
                    rt_closure y z  ->
                    rt_closure x z.
Proof.
  intros.
  induction H. assumption. 
  apply rtc_step with y. assumption.
  apply IHrt_closure. assumption.
Qed.    

Lemma rtc_Last : forall (x y z : X),
      rt_closure x y -> R y z -> rt_closure x z.
Proof. 
  intros.
  induction H.  
  apply rtc_step with z. assumption.
  apply rtc_refl. 
  apply rtc_step with y. assumption.
  apply (IHrt_closure H0).
Qed.     


(** ** Induction Principles *)

Lemma rtc_ind_with_trans : forall (P : X -> X -> Prop),
  (forall x : X, P x x) ->
  (forall x y : X, R x y -> P x y) ->
  (forall y x z : X, rt_closure x y -> P x y -> rt_closure y z -> P y z -> P x z) ->
   forall x y : X, rt_closure x y -> P x y.
Proof.
  intros ? ? ? ?. induction 1.
                  apply H. 
                  apply (@H1 y);auto.
Qed.

Lemma rtc_ind_steps_last : forall (P : X -> X -> Prop),
  (forall x : X, P x x) ->
  (forall y x z : X, rt_closure x y -> P x y -> R y z -> P x z) ->
   forall x y : X, rt_closure x y -> P x y.
Proof.
  intros ? ? ? ? ? ?. 
  apply rtc_ind_with_trans. 
        auto.
        intros. apply (H0 x0). 
                apply rtc_refl. 
                auto.
                auto.
        intros ? ? ? HClo HP HClo' _.
               induction HClo'.
                         auto. 
                         apply IHHClo'. 
                         apply rtc_Last with x1.
                               auto.
                               auto. 
                               apply (@H0 x1);auto.
        auto.
Qed. 

End Reflexive_Transitive_Closure.

Add Parametric Relation T R : T (@rt_closure T R)
    reflexivity  proved by (@rtc_refl _ _)
    transitivity proved by (@rtc_Trans' _ _)
    as RTC_Rel.

Hint Resolve rtc_refl rtc_step rtc_R' : rt_closure.
Hint Constructors rt_closure.

Notation multistep := (rt_closure tm step).
Notation "t1 '~~>*' t2" := (multistep t1 t2) (at level 40).

Notation multimostep := (rt_closure configuration step_mo).
Notation "t '~~~>*' t'" := (multimostep t t') (at level 40).  

Tactic Notation "RTClosure_Ind_1" constr(name) := Case_aux RTClosure_Ind_1 name.
Tactic Notation "RTClosure_Ind_2" constr(name) := Case_aux RTClosure_Ind_2 name.
Tactic Notation "RTClosure_Ind_3" constr(name) := Case_aux RTClosure_Ind_3 name.

(** Two tactics for using [rtc_ind_with_trans] and [rtc_ind_steps_last] *)

Tactic Notation "R*" "induction" hyp(H) "using" "trans" :=
          (induction H using rtc_ind_with_trans; [Case_aux RTClosure_Ind_1 "Base" 
                                                 |Case_aux RTClosure_Ind_2 "Step" 
                                                 |Case_aux RTClosure_Ind_3 "Transitivity"]). 
 
Tactic Notation "R*" "induction" hyp(H) "using" "steps" :=  
          (induction H using rtc_ind_steps_last; [Case_aux RTClosure_Ind_1 "Base" 
                                                 |Case_aux RTClosure_Ind_2 "Step" ]).

Theorem rtc_mstep_refl : forall (t:tm),
        t ~~>* t.
Proof. eauto. Defined.

Theorem rtc_mstep_refl_mo : forall (co:configuration),
        co ~~~>* co.
Proof. eauto. Defined.

Hint Resolve rtc_mstep_refl rtc_mstep_refl_mo : core.

Theorem rtc_R : forall (t:tm) (t':tm),
        t ~~>  t' ->
        t ~~>* t'.
Proof.
 intros;eapply (rtc_R' _ _ _ _) in H; eauto.
Qed.

Theorem rtc_R_mo : forall co co',
        co ~~~>  co' ->
        co ~~~>* co'.
Proof.
 intros; eapply (rtc_R' _ _ _ _) in H. eauto.
Qed.

Theorem rtc_Trans : forall (t:tm) (t':tm) (t'':tm),
        t  ~~>* t'  ->
        t' ~~>* t'' ->
        t  ~~>* t''.
Proof.
  intros; eapply rtc_Trans'; eauto.
Qed.

Theorem rtc_Trans_mo : forall co co' co'',
        co  ~~~>* co'  ->
        co' ~~~>* co'' ->
        co  ~~~>* co''.
Proof.
  intros; eapply rtc_Trans';eauto.
Qed.

Ltac rtc_Trans := eapply rtc_Trans || eapply rtc_Trans_mo || fail "This is not the case you're looking for..." .

Hint Extern 1 (?X ~~>* ?Z) =>
     match goal with 
     | [ H: multistep X ?Y, H': multistep ?Y Z |- _] => apply (rtc_Trans X Y Z H H')
     end.

Hint Extern 1 (?X ~~~>* ?Z) =>
     match goal with 
     | [ H: multimostep X ?Y, H': multimostep ?Y Z |- _] => apply (rtc_Trans_mo X Y Z H H')
     end.
(* An invariant on stores needed for the proof of the Progress theorem. *)

Inductive same_length {A:Type} : list A -> list A -> Prop :=
  | same_length_nil  : same_length nil nil
  | same_length_cons : forall x y l1 l2, 
                       same_length l1 l2 -> same_length (x::l1) (y::l2).

Lemma same_length_refl : forall {A:Type} (l:list A), same_length l l.
Proof.
  induction l; constructor; auto.
Qed.

Hint Resolve same_length_refl.

Lemma same_length_trans : forall {A:Type} (l1 l2 l3:list A), same_length l1 l2 -> same_length l2 l3 -> same_length l1 l3.
Proof.
  induction l1; intros.
  Case "nil".
    inversion H; subst. inversion H0; subst. constructor.
  Case "cons".
    inversion H; subst. inversion H0; subst.
    constructor.
    eapply IHl1; eauto.
Qed.

Lemma step_mo_same_length : forall t t' Sto Sto',
                              (t,Sto) ~~~> (t',Sto') ->
                              same_length Sto Sto'.
Proof.
  induction t; intros t' Sto Sto' Hstep; inversion Hstep; subst; try (apply same_length_refl).
  eapply IHt1; eauto.
  constructor; eapply IHt; eauto.
  eapply IHt; eauto.
  eapply IHt; eauto.
  constructor; apply same_length_refl.
  pose proof (IHt1 _ _ _ H6) as H; inversion H; auto.
  eapply IHt; eauto.
Qed.

Lemma step_mo_same_length_star : forall t t' Sto Sto',
                                   (t,Sto) ~~~>* (t',Sto') ->
                                   same_length Sto Sto'.
Proof.
  intros t t' Sto Sto' Hstep.
  remember (t,Sto) as co.
  remember (t',Sto') as co'.
  generalize dependent Sto'.
  generalize dependent Sto.
  generalize dependent t'.
  generalize dependent t.
  induction Hstep; intros; subst.
  Case "0".
    inversion Heqco'; subst.
    apply same_length_refl.
  Case "n".
    destruct y as (tMid,StoMid).
    pose proof (IHHstep _ _ _ eq_refl _ eq_refl).
    pose proof (step_mo_same_length _ _ _ _ H).
    eapply same_length_trans; eauto.
Qed.


(** Some lemmas about values and done configurations. *)
Lemma value_not_step : forall v, value v -> forall t, ~(v ~~> t).
Proof.
  induction v; intros Hval t';
    try (intro Hstep; inversion Hstep; fail);
    try (inversion Hval; fail);
     inversion Hval; subst;
    intro Hstep; inversion Hstep; subst; hide hyp Hval;
    match goal with
    | [H1: (value ?v), H2: (?v ~~> ?t) |- _ ] => contradict H2; auto
    end.
Qed.

Lemma step_not_value : forall t, (exists t', t ~~> t') -> ~ value t.
Proof.
  intros t [t' Hstep] Hval; contradict Hstep; apply value_not_step; auto.
Qed.

Theorem done_not_step : forall co, done_mo co -> forall co', ~co ~~~> co'.
Proof.
  intros co Hdone.
  inversion Hdone; subst; intros co' Hstep;
  inversion Hstep; subst;
  inversion H3; subst;
  match goal with
  | [H1: (value ?v), H2: (?v ~~> ?t) |- _ ] => contradict H2;apply value_not_step; auto
  end.
Qed.


(** *** Determinism *)
Theorem step_deterministic' : forall t u v,
                                ((t ~~> u -> t ~~> v -> u = v)
                                 /\ forall Sto Sto' Sto'',
                                      ((t,Sto) ~~~> (u,Sto') ->
                                       (t,Sto) ~~~> (v,Sto'') ->
                                       (u = v /\ Sto' = Sto''))).
Proof.
  induction t; intros u v.

  Case "tvar".
    split; intros.
    SCase "lc".
      inversion H.
    SCase "mo".
      inversion H; subst.
      inversion H2.

  Case "tapp".
    assert (tapp t1 t2 ~~> u -> tapp t1 t2 ~~> v -> u = v).
      intros Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          auto.
          inversion H3.
          contradict H4; apply value_not_step; auto.
        inversion Hstep2; subst.
          inversion H2.
          assert (t1' = t1'0). apply IHt1. assumption. assumption.
          subst. reflexivity.
          contradict H2; apply value_not_step; auto.
        inversion Hstep2; subst.
          contradict H3; apply value_not_step; auto.
          contradict H4; apply value_not_step; auto.
          assert (t2' = t2'0). apply IHt2. assumption. assumption.
          subst. reflexivity.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          split; auto.

  Case "tabs".
    split.
    SCase "lc".
      intro H; inversion H.
    SCase "mo".
      intros ? ? ? H; inversion H; subst.
      inversion H1.

  Case "tunit".
    split.
    SCase "lc".
      intro H; inversion H.
    SCase "mo".
      intros ? ? ? H; inversion H; subst.
      inversion H1.

  Case "tpair".
    assert (tpair t1 t2 ~~> u -> tpair t1 t2 ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0). apply IHt1. assumption. assumption.
          subst. reflexivity.
          contradict H2; apply value_not_step; auto.
        inversion HstepV; subst.
          contradict H4; apply value_not_step; auto.
          assert (t2'=t2'0). apply IHt2. assumption. assumption.
          subst. reflexivity.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          split; auto.

  Case "tproj".
    assert (tproj t1 t2 ~~> u -> tproj t1 t2 ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0). apply IHt1; auto. subst; auto.
          contradict H2; apply value_not_step; auto.
          contradict H2; apply value_not_step; repeat constructor; auto.
        inversion HstepV; subst.
          contradict H4; apply value_not_step; auto.
          assert (t2'=t2'0). apply IHt2. assumption. assumption. subst; auto.
          contradict H3; apply value_not_step; auto.
        inversion HstepV; subst.
          contradict H5; apply value_not_step; repeat constructor; auto.
          contradict H6; apply value_not_step; auto.
          auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          split; auto.

  Case "tinl".
    rename t into T.
    rename t0 into t.
    assert (tinl T t ~~> u -> tinl T t ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0). apply IHt. assumption. assumption.
          subst; reflexivity.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          split; auto.

  Case "tinr".
    rename t into T.
    rename t0 into t.
    assert (tinr T t ~~> u -> tinr T t ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0). apply IHt. assumption. assumption.
          subst; reflexivity.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          split; auto.

  Case "tcase".
    assert (tcase t1 t2 t3 ~~> u -> tcase t1 t2 t3 ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0). apply IHt1. assumption. assumption. subst; reflexivity.
          assert (value (tinl T v1)). constructor. assumption.
          contradict H3; apply value_not_step; auto.
          assert (value (tinr T v1)). constructor. assumption.
          contradict H3; apply value_not_step; auto.
       inversion HstepV; subst.
          assert (value (tinl T v1)). constructor. assumption.
          contradict H4; apply value_not_step; auto.
          reflexivity.
       inversion HstepV; subst.
          assert (value (tinr T v1)). constructor. assumption.
          contradict H4; apply value_not_step; auto.
          reflexivity.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          split; auto.

  Case "treturn".
    assert (treturn m t ~~> u -> treturn m t ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0) by (apply IHt; auto). subst; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros Sto Sto' Sto'' HstepU HstepV.
        inversion HstepU; subst.
          inversion HstepV; subst.
            split; auto.

  Case "tbind".
    assert (tbind t1 t2 t3 ~~> u -> tbind t1 t2 t3 ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0) by (apply IHt1; auto). subst; auto.
          contradict H3; apply value_not_step; auto.
        inversion HstepV; subst.
          contradict H5; apply value_not_step; auto.
          assert (t2'=t2'0) by (apply IHt2; auto). subst; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros Sto Sto' Sto'' Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion H1; subst.
          inversion Hstep2; subst.
            inversion H2; subst; auto.
            contradict H5; apply value_not_step; repeat constructor; auto.
            contradict H5; apply value_not_step; repeat constructor; auto.
            contradict H1; apply value_not_step; repeat constructor; auto.
          inversion Hstep2; subst.
            split; auto.
            contradict H1; apply value_not_step; repeat constructor; auto.
            contradict H1; apply value_not_step; repeat constructor; auto.
            contradict H1; apply value_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.

          pose proof (IHt1 t1' t1'0).
          destruct H0.
          pose proof (H1 Sto Sto' Sto'').
          pose proof (H2 H8 H11).
          destruct H5; subst; auto.

          contradict H8; apply done_not_step; repeat constructor; auto.
          contradict H8; apply done_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.
          contradict H10; apply done_not_step; repeat constructor; auto.
          auto.
        inversion Hstep2; subst.  
          contradict H1; apply value_not_step; repeat constructor; auto.
          contradict H10; apply done_not_step; repeat constructor; auto.
          auto.

  Case "tlift".
    assert (tlift m t ~~> u -> tlift m t ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0) by (apply IHt; auto); subst; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          auto.
          inversion H1; subst. contradict H5; apply value_not_step; auto.
          inversion H1; subst. contradict H5; apply value_not_step; auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.
          pose proof (IHt t' t'0). destruct H0. pose proof (H1 _ _ _ H6 H11). destruct H4; subst; auto.
          contradict H6; apply done_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.
          pose proof (IHt t' t'0). destruct H0. pose proof (H1 _ _ _ H6 H10). destruct H4; subst; auto.
          contradict H6; apply done_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H2; apply value_not_step; repeat constructor; auto.
          contradict H9; apply done_not_step; repeat constructor; auto.
          auto.
        inversion Hstep2; subst.
          contradict H2; apply value_not_step; repeat constructor; auto.
          contradict H9; apply done_not_step; repeat constructor; auto.
          auto.

  Case "televate".
    assert (televate s t ~~> u -> televate s t ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0) by (apply IHt; auto); subst; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.
          pose proof (IHt t1' t1'0). destruct H0. pose proof (H1 _ _ _ H6 H8). destruct H4; subst; auto.
          contradict H6; apply done_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H2; apply value_not_step; repeat constructor; auto.
          contradict H7; apply done_not_step; repeat constructor; auto.
          auto.

  Case "tget".  
    assert (tget s ~~> u -> tget s ~~> v -> u = v).
      intro.
      inversion H.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion H1.
        inversion Hstep2; subst.
          inversion H1.
          auto.

  Case "tput".
    assert (tput s t ~~> u -> tput s t ~~> v -> u = v).
      intros HstepU HstepV.
        inversion HstepU; subst.
          inversion HstepV; subst.
            assert (t1'=t1'0) by (apply IHt; auto); subst; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H2; apply value_not_step; repeat constructor; auto.
          auto.

  Case "trunst".
    assert (trunst t1 t2 ~~> u -> trunst t1 t2 ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0) by (apply IHt1; auto); subst; auto.
          contradict H2; apply value_not_step; auto.
        inversion HstepV; subst.
          contradict H4; apply value_not_step; auto.
          assert (t2'=t2'0) by (apply IHt2; auto); subst; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.
          pose proof (IHt1 t1' t1'0). destruct H0. pose proof (H1 _ _ _ H7 H10). destruct H2. inversion H3; subst; auto.
          contradict H7; apply done_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.
          contradict H9; apply done_not_step; repeat constructor; auto.
          auto.

  Case "trunid".
    assert (trunid t ~~> u -> trunid t ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0) by (apply IHt; auto); subst; auto.
          contradict H0; apply value_not_step; auto.
          contradict H0; apply value_not_step; repeat constructor; auto.
        inversion HstepV; subst.
          contradict H2; apply value_not_step; auto.
          pose proof (IHt t1' t1'0). destruct H. pose proof (H4 _ _ _ H1 H3). destruct H5; subst; auto.
          contradict H1; apply done_not_step; repeat constructor; auto.
        inversion HstepV; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.
          contradict H2; apply done_not_step; repeat constructor; auto.
          auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          auto.

  Case "tpause".
    assert (tpause m t t0 ~~> u -> tpause m t t0 ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0) by (apply IHt; auto); subst; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          auto.

  Case "tunfold".
    rename m into M.
    rename t1 into TA.
    rename t2 into TB.
    rename t3 into t1.
    rename t4 into t2.

    assert (tunfold M TA TB t1 t2 ~~> u -> tunfold M TA TB t1 t2 ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0) by (apply IHt1; auto); subst; auto.
          contradict H5; apply value_not_step; auto.
        inversion HstepV; subst.
          contradict H7; apply value_not_step; auto.
          assert (t2'=t2'0) by (apply IHt2; auto); subst; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.
          auto.

  Case "trunre".
    rename t into T.
    rename t0 into t.

    assert (trunre T t ~~> u -> trunre T t ~~> v -> u = v).
      intros HstepU HstepV.
      inversion HstepU; subst.
        inversion HstepV; subst.
          assert (t1'=t1'0) by (apply IHt; auto); subst; auto.

    split.

    SCase "lc".
      auto.

    SCase "mo".
      intros ? ? ? Hstep1 Hstep2.
      inversion Hstep1; subst.
        inversion Hstep2; subst.
          auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
          contradict H1; apply value_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H1; apply value_not_step; repeat constructor; auto.
          pose proof (IHt t1' t1'0). destruct H0. pose proof (H1 _ _ _ H6 H8). destruct H4; subst; auto.
          contradict H6; apply done_not_step; repeat constructor; auto.
          contradict H6; apply done_not_step; repeat constructor; auto.
        inversion Hstep2; subst.
          contradict H2; apply value_not_step; repeat constructor; auto.
          contradict H7; apply done_not_step; repeat constructor; auto.
          auto.
        inversion Hstep2; subst.
          contradict H2; apply value_not_step; repeat constructor; auto.
          contradict H7; apply done_not_step; repeat constructor; auto.
          auto.
Qed.

Theorem step_deterministic : forall t u v, t ~~> u -> t ~~> v -> u = v.
Proof.
  intros.
  pose proof (step_deterministic' t u v).
  destruct H1; auto.
Qed.

Theorem step_deterministic_mo : forall co1 co2 co3, co1 ~~~> co2 -> co1 ~~~> co3 -> co2 = co3.
Proof.
  intros.
  destruct co1 as [t1 Sto1].
  destruct co2 as [t2 Sto2].
  destruct co3 as [t3 Sto3].
  pose proof (step_deterministic' t1 t2 t3).
  destruct H1.
  pose proof (H2 Sto1 Sto2 Sto3).
  pose proof (H3 H H0).
  destruct H4; subst; auto.
Qed.



(* ###################################################################### *)
(** *** Lemmas concerning how values/done congfigurations 
        interact with the step and multistep reduction relations. *)

Lemma step_monad_second : forall t t' t'' Sto Sto',
                            t ~~> t' ->
                            value t'' ->
                            (t,Sto) ~~~>* (t'',Sto') ->
                            (t',Sto) ~~~>* (t'',Sto').
Proof.
  intros.
  generalize dependent t'.
  remember (t,Sto) as co1.
  remember (t'',Sto') as co2.
  generalize dependent Sto'.
  generalize dependent Sto.
  generalize dependent t''.
  generalize dependent t.
  induction H1; intros; subst.
  * inversion Heqco2; subst.
    contradict H; apply value_not_step; auto.
  * assert (y=(t',Sto)).
      assert ((t,Sto) ~~~> (t',Sto)).
        apply STM_LC; auto.
      eapply step_deterministic_mo; eauto.
    subst.
    eapply rtc_Trans'.
      eauto.
      eapply rtc_refl; eauto.
Qed.      

Lemma step_value_second : forall t t' v,
                            value v ->
                            t ~~> t' ->
                            t ~~>* v ->
                            t' ~~>* v.
Proof.
  intros t t' v Hvalv Hsteptt' Hsteptv.
  generalize dependent t'.
  generalize dependent Hvalv.
  induction Hsteptv.
    intros Hvalx t' Hstepx. contradict Hvalx; eapply step_not_value; eauto.
    intros Hvalz t' Hstepx. assert (y=t') by (eapply step_deterministic; eauto); subst; assumption.
Qed.

Lemma value_unique : forall t v v',
                       value v ->
                       value v' ->
                       t ~~>* v ->
                       t ~~>* v' ->
                       v = v'.
Proof.
  intros t v v' Hdonev Hdonev' Hsteptv.
  generalize dependent v'.
  induction Hsteptv.
  Case "0".
    intros v' Hdonev' Hstepxv'.
    inversion Hstepxv'.
    SCase "0".
      auto.
    SCase "n".
      subst.
      contradict H; eapply value_not_step; eauto.
  Case "n".
    intros v' Hdonev' Hstepxv'.
    assert (y ~~>* v').
      eapply step_value_second; eauto.
    eapply IHHsteptv; eauto.
Qed.

Lemma step_same_value : forall t t' v,
                          value v ->
                          t ~~>* t' ->
                          t ~~>* v ->
                          t' ~~>* v.
Proof.
  intros t t' v Hvalv Hstept'. generalize dependent v. induction Hstept'.
    auto.
    intros v Hvalv Hstepxv.
    apply IHHstept'. assumption. eapply step_value_second; eauto.
Qed.

Lemma step_done_second : forall co co' co'',
                           done_mo co'' ->
                           co ~~~> co' ->
                           co ~~~>* co'' ->
                           co' ~~~>* co''.
Proof.
  intros co co' co'' Hdoneco'' Hstepcoco' Hstepcoco''.
  generalize dependent co'.
  generalize dependent Hdoneco''.
  induction Hstepcoco''.
    intros Hdonex co' Hstepx. contradict Hstepx; eapply done_not_step; eauto.
    intros Hdonez co' Hstepx. assert (y=co') by (eapply step_deterministic_mo; eauto); subst; assumption.
Qed.

Lemma step_same_done : forall co co' co'',
                         done_mo co'' ->
                         co ~~~>* co' ->
                         co ~~~>* co'' ->
                         co' ~~~>* co''.
Proof.
  intros co co' co'' Hdoneco'' Hstepcoco'. generalize dependent co''. induction Hstepcoco'.
    auto.
    intros co'' Hdoneco'' Hstepxco''.
    apply IHHstepcoco'. assumption. eapply step_done_second; eauto.
Qed.

Lemma done_unique : forall co co' co'',
                      done_mo co' ->
                      done_mo co'' ->
                      co ~~~>* co' ->
                      co ~~~>* co'' ->
                      co' = co''.
Proof.
  intros co co' co'' Hdoneco' Hdoneco'' Hstepcoco'.
  generalize dependent co''.
  induction Hstepcoco'.
  Case "0".
    intros co'' Hdoneco'' Hstepxco''.
    inversion Hstepxco''.
    SCase "0".
      auto.
    SCase "n".
      subst.
      contradict H; eapply done_not_step; eauto.
  Case "n".
    intros co'' Hdoneco'' Hstepxco''.
    assert (y ~~~>* co'').
      eapply step_done_second; eauto.
    eapply IHHstepcoco'; eauto.
Qed.

Lemma done_multistep_only_self : forall co, done_mo co -> forall co', co ~~~>* co' -> co = co'.
Proof.
  intros co Hdone co' Hstep.
  generalize dependent Hdone.
  induction Hstep; intros.
    auto.
    contradict H; apply done_not_step; auto.
Qed.


Lemma nonval_step_step_mo : forall t t' Sto Sto',
                              ~ value t ->
                              (t,Sto) ~~~> (t',Sto') ->
                              (t ~~> t' /\ Sto = Sto').
Proof.
  intros t t' Sto Sto' Hnval Hstep.
  remember (t,Sto) as co; remember (t',Sto') as co'.
  induction Hstep; inversion Heqco; inversion Heqco'; subst;
  try (contradict Hnval; repeat constructor; auto; fail).
    split; auto.
Qed.
  
Lemma step_mo_not_value_step : forall t t' Sto Sto',
                                 (t,Sto) ~~~> (t',Sto') ->
                                 ~ value t ->
                                 t ~~> t'.
Proof.
  induction t; intros ? ? ? Hstep Hnval; inversion Hstep; subst;
    try auto;
    try (contradict Hnval; repeat constructor; auto; fail).
Qed.

Lemma step_mo_value_not_step : forall t t' Sto Sto',
       (t,Sto) ~~~> (t',Sto') ->
                     value t ->
                     ~ (t ~~> t').
Proof.
  induction t; intros ? ? ? Hstep Hval; inversion Hstep; subst;
   (match goal with 
       | [H1: value ?t |- ~ (?t ~~> ?t')] => apply (value_not_step t H1 t')
       end).
Qed. 

Lemma done_val : forall t Sto, 
      done_mo (t,Sto) -> value t.
Proof.
  intros ? ? Hdone; inversion Hdone; subst; repeat constructor; auto.
Qed.

Lemma step_mo_pure : forall t t' Sto t'' Sto',
                        t ~~> t' ->
                        (t,Sto) ~~~> (t'',Sto') ->
                        (t' = t'' /\ Sto = Sto').
Proof.
  intros t t' Sto t'' Sto' Hstep Hstepmo; inversion Hstepmo; subst;
  try (match goal with
       | [H1: (?t ~~> ?t') |- _ ] => let HV := fresh in 
                                         assert (HV: value t) by (repeat constructor; auto);
                                         contradict H1; apply value_not_step;auto
       end).
  split; auto.
  eapply step_deterministic; eauto.
Qed.


(* ###################################################################### *)
(** *** Congruence lemmas on multistep *)

Lemma Congruence_App1 : forall t1 t1' t2,
      (t1 ~~>* t1') ->
      (tapp t1 t2) ~~>* (tapp t1' t2).
Proof.
 intros t1 t1' t2 STM. induction STM.
    apply rtc_refl.
    eapply rtc_step.
      apply ST_App1. apply H. auto.
Qed.

Lemma Congruence_App2 : forall v t t', 
          value v -> 
      (t ~~>* t') -> 
      (tapp v t) ~~>* (tapp v t').
Proof.
  intros v t t' V STM. induction STM. 
   apply rtc_refl.
   eapply rtc_step.
     apply ST_App2; eauto.  auto.
Qed.

Lemma Congruence_Pair1 : forall t1 t1' t2,
      (t1 ~~>* t1') -> 
      (tpair t1 t2) ~~>* (tpair t1' t2).
Proof.
  intros t1 t1' t2 STM. induction STM.
    apply rtc_refl.
    eapply rtc_step.
      apply ST_Pair1. apply H. auto.
Qed.

Lemma Congruence_Pair2 : forall v1 t2 t2',
  value v1 -> t2 ~~>* t2' -> (tpair v1 t2) ~~>* (tpair v1 t2').
Proof.
  intros v1 t2 t2' V STM. induction STM.
    apply rtc_refl.
    eapply rtc_step.
      apply ST_Pair2. auto. apply H. auto.
Qed.

Lemma Congruence_Proj1 : forall t1 t1' t2,
      (t1 ~~>* t1') -> 
      (tproj t1 t2) ~~>* (tproj t1' t2).
Proof.
  intros t1 t1' t2 STM. induction STM.
    apply rtc_refl.
    eapply rtc_step.
      apply ST_Proj1. apply H. auto.
Qed.

Lemma Congruence_Proj2 : forall v1 t2 t2',
  value v1 -> t2 ~~>* t2' -> (tproj v1 t2) ~~>* (tproj v1 t2').
Proof.
  intros v1 t2 t2' V STM. induction STM.
    apply rtc_refl.
    eapply rtc_step.
      apply ST_Proj2. auto. apply H. auto.
Qed.

Lemma Congruence_Tinl : forall T t1 t2,
      t1 ~~>* t2 -> (tinl T t1) ~~>* (tinl T t2).
Proof.
  intros t1 t2 HTy STM. induction STM.
    apply rtc_refl.
    eapply rtc_step.
      apply ST_InL. auto. apply H. auto.
Qed.

Lemma Congruence_Tinr : forall T t1 t2,
      t1 ~~>* t2 -> (tinr T t1) ~~>* (tinr T t2).
Proof.
  intros t1 t2 HTy STM. induction STM.
    apply rtc_refl.
    eapply rtc_step.
      apply ST_InR. auto. apply H. auto.
Qed.

Lemma Congruence_Tcase : forall t1 t1' t2 t3,
      t1 ~~>* t1' -> tcase t1 t2 t3 ~~>* tcase t1' t2 t3.
Proof.
  intros t1 t1' t2 t3 STM. induction STM.
    apply rtc_refl. eapply rtc_step.
    apply ST_Case. apply H. apply IHSTM.
Qed.

Lemma Congruence_ST_CaseL  : forall T v1 t2 t3,
      value v1 ->
      tcase (tinl T v1) t2 t3 ~~>* tapp t2 v1.
Proof.
  intros v1 T t2 t3 STM. 
   pose proof (ST_CaseL T v1 t2 t3 STM).
   apply rtc_R in H. assumption.
Qed.

Lemma Congruence_ST_CaseR  : forall T v1 t2 t3,
      value v1 ->
      tcase (tinr T v1) t2 t3 ~~>* tapp t3 v1.
Proof.
  intros v1 T t2 t3 STM. 
   pose proof (ST_CaseR T v1 t2 t3 STM).
   apply rtc_R in H. assumption.
Qed.

Lemma Congruence_LC : forall t1 t1' Sto,
      t1 ~~>* t1' -> (t1,Sto) ~~~>* (t1',Sto).
Proof.
  intros t1 t1' Sto STM.
  induction STM. apply rtc_refl. eapply rtc_step.
  eapply STM_LC in H. eauto. apply IHSTM.
Qed.

Lemma step_val_done : forall t vt Sto t' Sto',
                        t ~~>* vt ->
                        value vt ->
                        (t,Sto) ~~~>* (t',Sto') ->
                        done_mo (t',Sto') ->
                        (vt,Sto) ~~~>* (t',Sto').
Proof.
  intros.
  assert ((t,Sto) ~~~>* (vt,Sto)).
    eapply Congruence_LC;eauto.
  eapply step_same_done; eauto.
Qed.

Lemma Congruence_LC_Ret : forall M t1 t1',
      t1 ~~>* t1' -> treturn M t1 ~~>* treturn M t1'.
Proof.
  intros.
  induction H; eauto.
Qed.  

Lemma Congruence_Treturn : forall t1 t1' M Sto,
      t1 ~~>* t1' -> (treturn M t1,Sto) ~~~>* (treturn M t1',Sto).
Proof.
  intros t1 t1' mo sto STM.
  induction STM. apply rtc_refl. eapply rtc_step.
  eapply ST_Return in H. apply STM_LC. apply H. apply IHSTM.
Qed.

Lemma Congruence_Tbind1  : forall T t1 t1' t2,
     t1 ~~>* t1' ->
     tbind T t1 t2 ~~>* tbind T t1' t2.
Proof.
 intros.
 induction H; eauto.
Qed.

Lemma Congruence_Tbind2  : forall T t1 t2 t2',
     value t1 ->
     t2 ~~>* t2' ->
     tbind T t1 t2 ~~>* tbind T t1 t2'.
Proof.
 intros.
 induction H0; eauto.
Qed.

Lemma Congruence_Mo_Bind1 : forall T t1 Sto t1' Sto' t2,
  value t2 ->
  (t1,Sto) ~~~>* (t1',Sto') ->
  (tbind T t1 t2,Sto) ~~~>* (tbind T t1' t2,Sto').
Proof.
  intros ? ? ? ? ? ? Hval Hstep.
  remember (t1,Sto) as co.
  remember (t1',Sto') as co'.
  generalize dependent Sto'.
  generalize dependent t1'.
  generalize dependent Sto.
  generalize dependent t1.
  induction Hstep; intros; subst.
  Case "0-step".
    inversion Heqco'; subst.
    apply rtc_refl.
  Case "n-step".
    destruct y as (t1Mid,StoMid).
    pose proof (IHHstep _ _ eq_refl _ _ eq_refl).
    eapply rtc_step; eauto.
    decide_value t1.
    SCase "t1 val".
      eapply STM_Bind1; eauto.
    SCase "t1 nonval".
      pose proof (nonval_step_step_mo _ _ _ _ Hnotval H).
      destruct H1; subst.
      eapply STM_LC.
      eapply ST_Bind1; auto.
Qed.

Lemma Congruence_BindRet : forall T (v1 v2 : tm) (M : Mo) (Sto : store),
      value v1 ->
      value v2 ->
      (tbind T (treturn M v1) v2, Sto) ~~~>* (tapp v2 v1, Sto).
Proof.
   intros; apply rtc_R_mo; apply STM_BindRet; auto.
Qed.

Lemma Congruence_LiftRetSt : forall T b M v Sto,
      value v ->
      (tlift (MNonReact (MStateT T b M)) (treturn (MNonReact M) v),Sto) ~~~>* (treturn (MNonReact (MStateT T b M)) v,Sto).
Proof.
  intros; apply rtc_R_mo; apply STM_LiftRetSt; auto.
Qed.


Lemma Congruence_LiftRetRe : forall v TI TO SM Sto,
      value v ->
      (tlift (MReactT TI TO SM) (treturn (MNonReact SM) v),Sto)
                         ~~~>* (treturn (MReactT TI TO SM) v,Sto).
Proof.
   intros; apply rtc_R_mo; apply STM_LiftRetRe; auto.
Qed.


Lemma Congruence_LiftSt : 
      forall (t t' : tm) (Sto Sto' : store) (s : tm) (TS : Ty) b (SM : SMo),
       value t ->
       (t, Sto) ~~~>* (t', Sto') ->
       (tlift (MNonReact (MStateT TS b SM)) t, s :: Sto) ~~~>*
       (tlift (MNonReact (MStateT TS b SM)) t', s :: Sto').
Proof. intros ? ? ? ? ? ? ? ? HVal HStep. 
       remember (t,Sto)   as co;
       remember (t',Sto') as co'.
       generalize dependent t' Sto' t Sto.
       R* induction HStep using steps; intros; subst.
       inv Heqco.
       destruct y as (u,S). 
       decide_value u. 
  Case "value u".
       pose proof (IHHStep _ _ eq_refl t HVal Sto eq_refl).
       eapply rtc_Trans_mo.
              eapply H0. 
       eapply rtc_R_mo.
              eapply STM_LiftSt; auto.
 Case "~ value u". 
      pose proof (IHHStep _ _ eq_refl t HVal Sto eq_refl).
      destruct (nonval_step_step_mo _ _ _ _ Hnotval H); subst. 
      eapply rtc_Trans_mo.
             eapply H0.
      eapply Congruence_LC.
             eapply rtc_R. 
             eapply ST_Lift.
             auto.
Qed.

Lemma Congruence_LiftRe : forall t t' Sto Sto' TI TO SM,
      value t ->
      (t,Sto) ~~~>* (t',Sto') ->
      (tlift (MReactT TI TO SM) t,Sto) ~~~>*
                        (tlift (MReactT TI TO SM) t',Sto').
Proof.
       intros ? ? ? ? ? ? ? HVal HStep. 
       remember (t,Sto)   as co;
       remember (t',Sto') as co'.
       generalize dependent t' Sto' t Sto.
       R* induction HStep using steps; intros; subst.
       inv Heqco.
       destruct y as (u,S). 
       decide_value u. 
  Case "value u".
       pose proof (IHHStep _ _ eq_refl t HVal Sto eq_refl).
       eapply rtc_Trans_mo.
              eapply H0. 
       eapply rtc_R_mo.
              eapply STM_LiftRe; auto.
  Case "~ value u". 
      pose proof (IHHStep _ _ eq_refl t HVal Sto eq_refl).
      destruct (nonval_step_step_mo _ _ _ _ Hnotval H); subst. 
      eapply rtc_Trans_mo.
             eapply H0.
      eapply Congruence_LC.
             eapply rtc_R. 
             eapply ST_Lift.
             auto.
Qed.

Lemma Congruence_Lift : forall M t1 t1',
                  t1 ~~>* t1' ->
                  tlift M t1 ~~>* tlift M t1'.
Proof. 
 intros.
 induction H; eauto.
Qed.

Lemma Congruence_STM_Elevate : forall (SM : SMo) (t t' : tm) (Sto Sto' : store),
                  value t ->
                  (t, Sto) ~~~>* (t', Sto') ->
                  (televate SM t, Sto) ~~~>* (televate SM t', Sto').
Proof.
 intros ? ? ? ? ? HVal HStep. 
       remember (t,Sto)   as co;
       remember (t',Sto') as co'.
       generalize dependent t' Sto' t Sto.
       R* induction HStep using steps; intros; subst.
       inv Heqco.
       destruct y as (u,S). 
       decide_value u. 
  Case "value u".
       pose proof (IHHStep _ _ eq_refl t HVal Sto eq_refl).
       eapply rtc_Trans_mo.
              eapply H0. 
       eapply rtc_R_mo.
              eapply STM_Elevate; auto.
  Case "~ value u". 
      pose proof (IHHStep _ _ eq_refl t HVal Sto eq_refl).
      destruct (nonval_step_step_mo _ _ _ _ Hnotval H); subst. 
      eapply rtc_Trans_mo.
             eapply H0.
      eapply Congruence_LC.
             eapply rtc_R. 
             eapply ST_Elevate.
             auto.
Qed.

Lemma Congruence_STM_ElevateRet : forall (SM SM' : SMo) (v : tm) (Sto : store),
                     value v ->
                     (televate SM' (treturn (MNonReact SM) v), Sto) ~~~>*
                     (treturn (MNonReact SM') v, Sto).
Proof.
   intros; apply rtc_R_mo; apply STM_ElevateRet; auto.
Qed.

Lemma Congruence_ST_Elevate : forall (SM : SMo) (t1 t1' : tm),
      t1 ~~>* t1' -> 
      televate SM t1 ~~>* televate SM t1'.
Proof.
 intros.
 induction H; eauto.
Qed.

Lemma Congruence_ST_Put : forall (SM : SMo) (t1 t1' : tm),
           t1 ~~>* t1' -> 
           tput SM t1 ~~>* tput SM t1'.
Proof.
 intros.
 induction H; eauto.
Qed.

Lemma Congruence_PutE : forall SM t Sto t',
      t ~~>* t' ->
      (tput SM t,Sto) ~~~>* (tput SM t',Sto).
Proof.
  intros SM t Sto t' STM.
  induction STM. apply rtc_refl. eapply rtc_step.
  apply STM_LC. eapply ST_Put in H. apply H. auto.
Qed. 

Lemma Congruence_RunId   : forall t t',
                    t ~~>* t' ->
                    trunid t ~~>* trunid t'.
Proof.
 intros.
 induction H; eauto.
Qed.

Lemma Congruence_RunIdMo : forall t t',
                             (t,nil) ~~~>* (t',nil) ->
                             trunid t ~~>* trunid t'.
Proof.
  intros t t' Hstep.
  remember (t,nil) as co.
  remember (t',nil) as co'.
  generalize dependent t'.
  generalize dependent t.
  induction Hstep; intros; subst.
  Case "0".
    inversion Heqco'; subst.
    apply rtc_refl.
  Case "n".
    destruct y as (tMid,StoMid).
    pose proof (step_mo_same_length _ _ _ _ H).
    inversion H0; subst.
    pose proof (IHHstep _ eq_refl _ eq_refl).
    eapply rtc_step; eauto.
    decide_value t.
    SCase "t is a value".
      eapply ST_RunIdMo; eauto.
    SSCase "t is not a value".
      pose proof (step_mo_not_value_step _ _ _ _ H Hnotval).
      constructor; auto.
Qed.

Lemma Congruence_RunSt1   : forall t1 t1' t2,
                    t1 ~~>* t1' ->
                    trunst t1 t2 ~~>* trunst t1' t2.
Proof.
 intros.
 induction H; eauto.
Qed.

Lemma Congruence_RunSt2   : forall t1 t2 t2',
                    value t1 ->
                    t2 ~~>* t2' ->
                    trunst t1 t2 ~~>* trunst t1 t2'.
Proof.
 intros.
 induction H0; eauto.
Qed. 

Lemma step_mo_still_values : forall t t' Sto Sto',
                               store_all_values Sto ->
                               (t,Sto) ~~~> (t',Sto') ->
                               store_all_values Sto'.
Proof.
  intros ? ? ? ? Hav Hstep.
  remember (t,Sto) as co.
  remember (t',Sto') as co'.
  generalize dependent t Sto t' Sto'.
  induction Hstep; intros; subst;
    inversion Heqco; subst; inversion Heqco'; subst; eauto.
    Case "tlift".
      inversion Hav; subst.
      pose proof (IHHstep _ _ H3 eq_refl _ _ eq_refl).
      constructor; auto.
    Case "treturn".
      inversion Hav; subst.
      constructor; auto.
    Case "trunst".
      assert (store_all_values (s::Sto0)) by (constructor; auto).
      pose proof (IHHstep _ _ H1 eq_refl _ _ eq_refl).
      inversion H2; auto.
Qed.

Lemma Congruence_STM_RunSt : forall (t t' s s' : tm) (Sto Sto' : store),
      value s ->
      store_all_values Sto ->
      (t,(s :: Sto)) ~~~>* (t',(s':: Sto')) ->
      (trunst t s, Sto) ~~~>* (trunst t' s', Sto').
Proof.
  intros.
  remember (t,s::Sto) as co.
  remember (t',s'::Sto') as co'.
  generalize dependent t s Sto t' s' Sto'.
  induction H1; intros; subst.
  Case "0-step".
    inversion Heqco'; auto.
  Case "n-step".
    destruct y as (tMid,sStoMid).
    pose proof (step_mo_same_length _ _ _ _ H).
    inversion H3; subst.
    rename l2 into StoMid.
    rename y into sMid.
    decide_value t.
    SCase "t is a value".
      assert (store_all_values (s::Sto)) by (constructor; auto).
      pose proof (step_mo_still_values _ _ _ _ H4 H).
      inversion H5; subst.
      pose proof (STM_RunSt _ _ _ _ _ _ Hval H0 H).
      eapply rtc_step; eauto.
    SCase "t is not a value".
      inversion H; subst; try (contradict Hnotval; repeat constructor; auto; fail).
      eapply rtc_step. eapply STM_LC. eapply ST_RunSt1; eauto.
      eauto.
Qed.


Lemma Congruence_STM_RunStRet : forall (t1 s : tm) (Sto : store) (TS : Ty) b
                     (SM : SMo),
                   value t1 ->
                   value s ->
                   (trunst (treturn (MNonReact (MStateT TS b SM)) t1) s, Sto) ~~~>*
                   (treturn (MNonReact SM) (tpair t1 s), Sto).
Proof.
intros; apply rtc_R_mo; apply STM_RunStRet; auto.
Qed.

Lemma Congruence_ST_Pause : forall M T t1 t1',
                   t1 ~~>* t1' ->
                   tpause M T t1 ~~>* tpause M T t1'.
Proof.
 intros.
 induction H; eauto.
Qed.

Lemma Congruence_ST_RunRe : forall (T : Ty) (t1 t1' : tm),
               t1 ~~>* t1' -> trunre T t1 ~~>* trunre T t1'.
Proof.
 intros.
 induction H; eauto.
Qed.

Lemma Congruence_ST_Unfold1 : forall M TA TB t1 t1' t2,
                   t1 ~~>* t1' ->
                   tunfold M TA TB t1 t2 ~~>* tunfold M TA TB t1' t2.
Proof.
 intros.
 induction H; eauto.
Qed.

Lemma Congruence_ST_Unfold2 : forall M TA TB t1 t2 t2',
                   value t1 ->
                   t2 ~~>* t2' ->
                   tunfold M TA TB t1 t2 ~~>* tunfold M TA TB t1 t2'.
Proof.
 intros.
 induction H0; eauto.
Qed.

Lemma Congruence_STM_Unfold     : forall t1 t2 Sto TI TO SM TA TB,
                       value t1 ->
                       value t2 ->
                       (tunfold (MReactT TI TO SM) TA TB t1 t2,Sto) ~~~>*
                       (tbind TA
                              (tlift (MReactT TI TO SM) (tapp t2 t1))
                              (tabs 0 (TSum TA (TProd TO (TArrow TI TB)))
                                (tcase (tvar 0)
                                   (tabs 1 TA (treturn (MReactT TI TO SM) (tvar 1)))
                                   (tabs 1 (TProd TO (TArrow TI TB))
                                      (tproj
                                         (tvar 1)
                                         (tabs 2 TO
                                           (tabs 3 (TArrow TI TB)
                                             (tpause (MReactT TI TO SM) TA
                                                  (treturn (MNonReact SM)
                                                    (tpair (tvar 2) 
                                                           (tabs 4 TI
                                                             (tunfold (MReactT TI TO SM) TA TB
                                                               (tapp (tvar 3) (tvar 4))
                                                               t2))))))))))),Sto).
Proof.
intros; apply rtc_R_mo; apply STM_Unfold; auto.
Qed.

Lemma Congruence_STM_PauseBind  : forall t1 t2 Sto TI TO SM T1 T2,
                       value t1 ->
                       value t2 ->
                       (tbind T2 (tpause (MReactT TI TO SM) T1 t1) t2,Sto) ~~~>*
                         (tpause
                            (MReactT TI TO SM)
                            T2
                            (tbind
                               (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T2)))
                               t1
                               (tabs 0 (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T1)))
                                 (tproj
                                   (tvar 0)
                                   (tabs 1 TO
                                     (tabs 2 (TArrow TI (TMonadic (MReactT TI TO SM) T1))
                                       (treturn (MNonReact SM)
                                          (tpair
                                            (tvar 1)
                                            (tabs 3 TI
                                              (tbind
                                                T2
                                                (tapp (tvar 2) (tvar 3))
                                                t2))))))))),Sto).
Proof.
intros; apply rtc_R_mo; apply STM_PauseBind; auto.
Qed.

Lemma Congruence_AppAbs_Mo : forall x T t v Sto,
           value v ->
           ((tapp (tabs x T t) v),Sto) ~~~>* ([x:=v]t,Sto).
Proof.
  intros X T t v Sto Hval.
  apply (Congruence_LC _ _ Sto (rtc_R _ _ (ST_AppAbs X T t v Hval))).
Qed.

Lemma Congruence_Unfold1 : forall TI TO SM TA TB t1 t1' t2,
                                  t1 ~~>* t1' ->
                                  tunfold (MReactT TI TO SM) TA TB t1 t2 ~~>* tunfold (MReactT TI TO SM) TA TB t1' t2.
Proof. intros; apply Congruence_ST_Unfold1; auto. Qed. 

Lemma Congruence_Unfold2 : forall TI TO SM TA TB t1 t2 t2',
                                  value t1 ->
                                  t2 ~~>* t2' ->
                                  tunfold (MReactT TI TO SM) TA TB t1 t2 ~~>* tunfold (MReactT TI TO SM) TA TB t1 t2'.
Proof. intros; eapply Congruence_ST_Unfold2; eauto. Qed. 


Lemma Congruence_STM_RunRe      : forall T t1 t1' Sto Sto',
                       value t1 ->
                       (t1,Sto) ~~~>* (t1',Sto') ->
                       (trunre T t1,Sto) ~~~>* (trunre T t1',Sto').
Proof.
 intros ? ? ? ? ? HVal HStep. 
       remember (t1,Sto)   as co;
       remember (t1',Sto') as co'.
       generalize dependent t1' Sto' t1 Sto.
       R* induction HStep using steps; intros; subst.
       inv Heqco.
       destruct y as (u,S). 
       decide_value u. 
  Case "value u".
       pose proof (IHHStep _ _ eq_refl t1 HVal Sto eq_refl).
       eapply rtc_Trans_mo.
              eapply H0. 
       eapply rtc_R_mo.
              eapply STM_RunRe; auto.
  Case "~ value u". 
      pose proof (IHHStep _ _ eq_refl t1 HVal Sto eq_refl).
      destruct (nonval_step_step_mo _ _ _ _ Hnotval H); subst. 
      eapply rtc_Trans_mo.
             eapply H0.
      eapply Congruence_LC.
             eapply rtc_R. 
             eapply ST_RunRe.
             auto.
Qed.

Lemma Congruence_STM_RunReRet   : forall v TI TO SM TA Sto,
                       value v ->
                       (trunre TA (treturn (MReactT TI TO SM) v),Sto) ~~~>*
                          (treturn (MNonReact SM) (tinl (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) TA))) v),Sto).
Proof. intros; apply rtc_R_mo; apply STM_RunReRet; auto. Qed.


Lemma Congruence_STM_RunRePause : forall v TI TO SM TA Sto,
                       value v ->
                       (trunre TA (tpause (MReactT TI TO SM) TA v),Sto) ~~~>*
                          (tbind (TSum TA (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) TA))))
                                 v
                                 (tabs 0 (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) TA)))
                                       (treturn (MNonReact SM) (tinr TA (tvar 0)))),Sto).
Proof. intros; apply rtc_R_mo; apply STM_RunRePause; auto. Qed.

(* ###################################################################### *)
(** *** Injectivity Lemmas for some of the term constructors. *)

Theorem tinl_stays_tinl : forall t T t',
                            tinl T t ~~>* t' ->
                            exists t'', t' = tinl T t''.
Proof.
  intros t T t' Hstep.
  remember (tinl T t) as x.
  generalize dependent T.
  generalize dependent t.
  induction Hstep; intros; subst x.
  Case "0".
    exists t; auto.
  Case "n".
    inversion H; subst.
    eapply IHHstep; eauto.
Qed.

Theorem tinr_stays_tinr : forall t T t',
                            tinr T t ~~>* t' ->
                            exists t'', t' = tinr T t''.
Proof.
  intros t T t' Hstep.
  remember (tinr T t) as x.
  generalize dependent T.
  generalize dependent t.
  induction Hstep; intros; subst x.
  Case "0".
    exists t; auto.
  Case "n".
    inversion H; subst.
    eapply IHHstep; eauto.
Qed.


Lemma treturn_step_inj : forall Mo t1 t2 Sto1 Sto2,
                           (treturn Mo t1,Sto1) ~~~> (treturn Mo t2,Sto2) ->
                           t1 ~~> t2 /\ Sto1 = Sto2.
Proof.
  intros Mo t1 t2 Sto1 Sto2 Hstep.
  inv Hstep. split.
  Case "lc here".
    inv H0.
  Case "lc in".
    auto.
Qed.

Lemma treturn_step_inj_star : forall Mo t1 t2 Sto1 Sto2,
                                (treturn Mo t1,Sto1) ~~~>* (treturn Mo t2,Sto2) ->
                                t1 ~~>* t2 /\ Sto1 = Sto2.
Proof.
  intros Mo t1 t2 Sto1 Sto2 Hstep.
  remember (treturn Mo t1,Sto1) as x.
  remember (treturn Mo t2,Sto2) as y.
  generalize dependent Sto2.
  generalize dependent Sto1.
  generalize dependent t2.
  generalize dependent t1.
  generalize dependent Mo.
  induction Hstep; intros; subst.
  Case "0".
    inversion Heqy; subst.
    split; auto.
  Case "n".
    destruct y.
    inversion H; subst.
    inversion H1.
    subst.
    pose proof (IHHstep Mo t1' t2 s eq_refl Sto2 eq_refl).
    destruct H0.
    split; auto.
    eapply rtc_step; eauto.
Qed.

Lemma step_return_no_change_store : forall t t' Mo Sto Sto',
                                      (treturn Mo t,Sto) ~~~>* (t',Sto') -> Sto = Sto'.
Proof.
  intros t t' Mo Sto Sto' Hstep.
  remember (treturn Mo t,Sto) as x.
  remember (t',Sto') as y.
  generalize dependent Sto'.
  generalize dependent Sto.
  generalize dependent Mo.
  generalize dependent t'.
  generalize dependent t.
  induction Hstep; intros; subst.
  Case "0".
    inversion Heqy; auto.
  Case "n".
    destruct y.
    inversion H; subst.
    inversion H1. subst.
    eapply IHHstep; eauto.
Qed.

Theorem pair_step_inj_l : forall t1 t2 v1 u2,
                            value v1 ->
                            tpair t1 t2 ~~>* tpair v1 u2 ->
                            t1 ~~>* v1.
Proof.
  intros t1 t2 v1 u2 Hvalv1 Hstep.
  generalize dependent Hvalv1.
  remember (tpair t1 t2) as x.
  remember (tpair v1 u2) as y.
  generalize dependent u2.
  generalize dependent v1.
  generalize dependent t2.
  generalize dependent t1.
  induction Hstep; intros; subst.
  Case "0".
    inversion Heqy; subst.
    apply rtc_refl.
  Case "n".
    inversion H; subst.
    SCase "l".
      pose proof (IHHstep t1' t2 eq_refl v1 u2 eq_refl Hvalv1).
      eapply rtc_step; eauto.
    SCase "r".
      eapply IHHstep; eauto.
Qed.

Theorem pair_step_inj_r : forall t1 t2 u1 v2,
                            value v2 ->
                            tpair t1 t2 ~~>* tpair u1 v2 ->
                            t2 ~~>* v2.
Proof.
  intros t1 t2 u1 v2 Hvalv2 Hstep.
  generalize dependent Hvalv2.
  remember (tpair t1 t2) as x.
  remember (tpair u1 v2) as y.
  generalize dependent v2.
  generalize dependent u1.
  generalize dependent t2.
  generalize dependent t1.
  induction Hstep; intros; subst.
  Case "0".
    inversion Heqy; subst.
    apply rtc_refl.
  Case "n".
    inversion H; subst.
    SCase "l".
      eapply IHHstep; eauto.
    SCase "r".
      pose proof (IHHstep t1 t2' eq_refl u1 v2 eq_refl Hvalv2).
      eapply rtc_step; eauto.
Qed.

Theorem tinl_step_inj : forall t T v,
                          value v ->
                          tinl T t ~~>* tinl T v ->
                          t ~~>* v.
Proof.
  intros t T v Hvalv Hstep.
  remember (tinl T t) as x.
  remember (tinl T v) as y.
  generalize dependent v.
  generalize dependent T.
  generalize dependent t.
  induction Hstep; intros; subst.
  Case "0".
    inversion Heqy; subst.
    apply rtc_refl.
  Case "n".
    inversion H; subst.
    pose proof (IHHstep t1' T eq_refl v Hvalv eq_refl).
    eapply rtc_step; eauto.
Qed.

Theorem tinr_step_inj : forall t T v,
                          value v ->
                          tinr T t ~~>* tinr T v ->
                          t ~~>* v.
Proof.
  intros t T v Hvalv Hstep.
  remember (tinr T t) as x.
  remember (tinr T v) as y.
  generalize dependent v.
  generalize dependent T.
  generalize dependent t.
  induction Hstep; intros; subst.
  Case "0".
    inversion Heqy; subst.
    apply rtc_refl.
  Case "n".
    inversion H; subst.
    pose proof (IHHstep t1' T eq_refl v Hvalv eq_refl).
    eapply rtc_step; eauto.
Qed. 

Lemma step_value_eq : forall v v',
                        value v   ->
                        value v'  ->
                        v ~~>* v' -> v = v'.
Proof. intros ? ? HValv HValv' HStep.
       assert (v ~~>* v) by eauto. 
       eapply value_unique;eauto.
Qed.
