Require Import Lib.
Require Import ReWire_Base.
Require Import ReWire_Subst.
Require Export Relations.Relation_Definitions.
Require Import Coq.Classes.EquivDec. 
Require Import Coq.Logic.FunctionalExtensionality.

(** Here's what a naive implementation of Fibonacci looks like *)

Inductive Bit :=
  | Zero : Bit
  | One : Bit.

(* Source code:

data Bit        = Zero | One

*)

Inductive W8 :=
| _W8 (* The '_' is there because constructors can't have the same names as their types *) : Bit -> Bit -> Bit -> Bit -> Bit -> Bit -> Bit -> Bit -> W8.

(* Source code: 

data W8         = W8 Bit Bit Bit Bit Bit Bit Bit Bit

*)

Definition plusW8 := fun (x : W8) (y : W8) => x (* replace x with user-defined input. *).
(* Source code:

plusW8 :: W8 -> W8 -> W8
{-# INLINE plusW8 #-}
plusW8 = nativeVhdl "plusW8" plusW8

*)

Definition zeroW8 :=  _W8 Zero Zero Zero Zero Zero Zero Zero Zero.
(* Source code:

zeroW8 :: W8
zeroW8 = W8 Zero Zero Zero Zero Zero Zero Zero Zero

 *)

Definition oneW8  :=  _W8 Zero Zero Zero Zero Zero Zero Zero One. 
(* Source code: 

oneW8 :: W8
oneW8 = W8 Zero Zero Zero Zero Zero Zero Zero One

 *)

(* The following functions need to occur in reverse order because order of definitions matter in COQ. *)


Fixpoint do_fac (n : nat) (m : nat) :=
  match n with
    | 0      => m
    | (S o)  => do_fac o (m * n)
  end.

Definition tr_factorial n := do_fac n 1.

Fixpoint do_fib (n res pre : nat) {struct n} : nat :=
  match n with
    |   0     => res
    | (S m)   => do_fib m (res + pre) res
 end.

Definition tr_fibonacci n := do_fib n 0 1.


Eval compute in tr_fibonacci 30.


(* In the following encoding, we use SUMs of UNITs to encoding bit values.
   By convention, we use the left as 1 and the right as 0. *)

Definition One  : tm := tinl TNil tunit.
Definition Zero : tm := tinr TNil tunit.
Definition Bit  : Ty := (TSum TNil TNil).

Notation "(| a ',' b ',' c ',' d ',' e ',' f ',' g ',' h |)" := (tpair (tpair (tpair (tpair (tpair (tpair (tpair a b) c) d) e) f) g) h) (at level 40, no associativity).

Check (| Zero, Zero, Zero, Zero, Zero, Zero, Zero, Zero |).
Check (tpair (tpair (tpair (tpair (tpair (tpair (tpair Zero Zero) Zero) Zero) Zero) Zero) Zero) Zero).
(*                                                                   (((((((0, 0), 0), 0),0),0),0),0) *)

Definition full_adder 


Definition W8 : Ty := TProd (TProd (TProd (TProd (TProd (TProd (TProd Bit Bit) Bit) Bit) Bit) Bit) Bit) Bit.


Lemma One_is_Value : value One.
Proof.
 constructor;auto.
Qed.
 
Lemma Zero_is_Value : value Zero.
Proof.
  constructor;auto.
Qed.

Lemma One_is_WT : empty |-- One \in Bit.
Proof.
do 2 constructor.
Qed.

Lemma Zero_is_WT : empty |-- Zero \in Bit.
Proof.
  do 2 constructor.
Qed.

Definition fourwider :=
  \(0:Bit) (\(1:Bit) (\(2:Bit) (\(3:Bit) (tpair (tpair (tpair (tvar 3) (tvar 2)) (tvar 1)) (tvar 0))))).

Function four_bit_helper (a1 a2 : tm) (carryin : tm) :=
  match (a1,a2,carryin) with
    | (tinr TNil tunit,tinr TNil tunit,tinr TNil tunit) => (tinr TNil tunit,tinr TNil tunit)
    | (tinr TNil tunit,tinr TNil tunit,tinl TNil tunit) => (tinl TNil tunit,tinr TNil tunit)
    | (tinr TNil tunit,tinl TNil tunit,tinr TNil tunit) => (tinl TNil tunit,tinr TNil tunit)
    | (tinr TNil tunit,tinl TNil tunit,tinl TNil tunit) => (tinr TNil tunit,tinl TNil tunit)
    | (tinl TNil tunit,tinr TNil tunit,tinr TNil tunit) => (tinl TNil tunit,tinr TNil tunit)
    | (tinl TNil tunit,tinr TNil tunit,tinl TNil tunit) => (tinr TNil tunit,tinl TNil tunit)
    | (tinl TNil tunit,tinl TNil tunit,tinr TNil tunit) => (tinr TNil tunit,tinl TNil tunit)
    | (tinl TNil tunit,tinl TNil tunit,tinl TNil tunit) => (tinl TNil tunit,tinl TNil tunit)
    | (_,_,_)                                           => (tinr TNil tunit,tinr TNil tunit)
  end.
                                                              

Function four_bit_addition (t t' : tm) :=
  match t with
    | (tpair (tpair (tpair a1 b1) c1) d1) => 
        (match t' with
           | (tpair (tpair (tpair a2 b2) c2) d2) =>
             let d := (four_bit_helper d1 d2 (tinr TNil tunit)) in
             let c := (four_bit_helper c1 c2 (snd d)) in 
             let b := (four_bit_helper b1 b2 (snd c)) in 
             let a := (four_bit_helper b1 b2 (snd b)) in
             tpair (tpair (tpair (fst a) (fst b)) (fst c)) (fst d) 
          | _ => (tpair (tpair (tpair Zero Zero) Zero) Zero)
        end)
    | _ => (tpair (tpair (tpair Zero Zero) Zero) Zero)
  end.

Lemma four_bit_addition_test : four_bit_addition (tpair (tpair (tpair Zero Zero) Zero) Zero) (tpair (tpair (tpair Zero Zero) Zero) Zero) = (tpair (tpair (tpair Zero Zero) Zero) Zero).
Proof. simpl.
       cbv.
       auto.
Qed.

Lemma four_bit_addition_one : four_bit_addition (tpair (tpair (tpair Zero Zero) Zero) One) (tpair (tpair (tpair Zero Zero) Zero) Zero) = (tpair (tpair (tpair Zero Zero) Zero) One).
Proof. simpl.
       cbv.
       auto.
Qed.

Lemma four_bit_addition_m : four_bit_addition (tpair (tpair (tpair Zero Zero) One) Zero) (tpair (tpair (tpair Zero Zero) Zero) One) = (tpair (tpair (tpair Zero Zero) One) One).
Proof. simpl.
       cbv.
       auto.
Qed.

Lemma adams_issue : [0 := ((tpair (tpair (tpair Zero Zero) One) Zero))](four_bit_addition (tvar 0) (tvar 0)) =
                    four_bit_addition (tpair (tpair (tpair Zero Zero) One) Zero) (tpair (tpair (tpair Zero Zero) One) Zero).
Proof.
  eval
  auto.
Qed.
  
Lemma two_plus_two_plus_two :
  let k := (four_bit_addition (tpair (tpair (tpair Zero Zero) One) Zero) (tpair (tpair (tpair Zero Zero) One) Zero)) in
  four_bit_addition k (tpair (tpair (tpair Zero Zero) One) Zero) = tpair (tpair (tpair Zero One) One) Zero.  
Proof. cbv beta. 
  simpl.
  cbv.
  auto.
  


Lemma W8_Zero_WT : empty |-- W8_Zero \in W8.
Proof. constructor; repeat (constructor; auto).
Qed.