Require Import Lib.
Require Import ReWire_Base.
Require Import ReWire_Subst.
Require Import ReWire_Step.
Require Import ReWire_Progress.
Require Import ReWire_Preservation.
Require Import ReWire_Normalization.

(** Equational Theory *)

CoInductive along2_react : (store -> store -> Prop) -> 
                           (tm -> tm -> Prop) -> 
                           (tm -> tm -> Prop) -> 
                           (tm -> tm -> Prop) -> 
                                configuration -> 
                                configuration -> 
                                Prop :=
   | eq_return        : forall (PS: store -> store -> Prop)
                               (PI PO PR: tm -> tm -> Prop)
                               TI TO SM t1 Sto1 t1' Sto1' t2 Sto2 t2' Sto2',
                       (t1,Sto1) ~~~>* (treturn (MReactT TI TO SM) t1',Sto1') ->
                       (t2,Sto2) ~~~>* (treturn (MReactT TI TO SM) t2',Sto2') ->
                       value t1'      ->
                       value t2'      ->
                       PR t1' t2'     ->
                       PS Sto1' Sto2' ->
                       along2_react PS PI PO PR (t1,Sto1) (t2,Sto2)
   | eq_pause  :  forall (PS: store -> store -> Prop)
                            (PI PO PR: tm -> tm -> Prop)
                            TI TO SM T t1 Sto1 t1' Sto1' t2 Sto2 t2' Sto2' vl vl' vr vr' Sto1'' Sto2'',
                       (t1,Sto1)   ~~~>* (tpause (MReactT TI TO SM) T t1',Sto1')         ->
                       (t2,Sto2)   ~~~>* (tpause (MReactT TI TO SM) T t2',Sto2')         ->
                       value t1'        ->
                       value t2'        ->
                       value vl         ->
                       value vl'        ->
                       value vr         ->
                       value vr'        ->
                       (t1',Sto1') ~~~>* (treturn (MNonReact SM) (tpair vl vr)  ,Sto1'') ->
                       (t2',Sto2') ~~~>* (treturn (MNonReact SM) (tpair vl' vr'),Sto2'') ->
                       PO vl vl'        ->
                       PS Sto1' Sto2'   ->
                       PS Sto1'' Sto2'' ->
                       (forall t3 t3' Sto3 Sto3',
                          PI t3 t3'     ->
                          PS Sto3 Sto3' ->
                          along2_react PS PI PO PR (tapp vr t3,Sto3) (tapp vr' t3',Sto3')) ->
                       along2_react PS PI PO PR (t1,Sto1) (t2,Sto2).

Fixpoint term_eq (T:Ty) (t t': tm) {struct T} : Prop :=
     match T with
     | TArrow T1 T2 => \empty |-- t  \in T /\
                       \empty |-- t' \in T /\
                       forall t1 t1',
                         (term_eq T1 t1 t1' ->
                          exists t2 t2',
                           (tapp t t1   ~~>* t2  /\
                             tapp t' t1' ~~>* t2' /\
                             term_eq T2 t2 t2'))
     | TProd T1 T2  => \empty |-- t  \in T /\
                       \empty |-- t' \in T /\
                       exists t1 t2 t1' t2',
                         (t  ~~>* tpair t1 t2   /\
                          t' ~~>* tpair t1' t2' /\
                          term_eq T1 t1 t1'     /\
                          term_eq T2 t2 t2')
     | TSum T1 T2   => \empty |-- t  \in T /\
                       \empty |-- t' \in T /\
                       ((exists t1 t1',
                          (t  ~~>* tinl T2 t1  /\
                           t' ~~>* tinl T2 t1' /\
                           term_eq T1 t1 t1'))
                          \/
                        (exists t2 t2',
                          (t  ~~>* tinr T1 t2  /\
                           t' ~~>* tinr T1 t2' /\
                           term_eq T2 t2 t2')))
     | TNil         => \empty |-- t  \in T /\
                       \empty |-- t' \in T
     | TMonadic M T' => \empty |-- t  \in T /\
                        \empty |-- t' \in T /\
                        forall Sto1 Sto2,
                          store_eq M Sto1 Sto2 ->
                          match M with
                          | MNonReact SM     => exists t1' Sto1' t2' Sto2', 
                                                  ((t, Sto1) ~~~>* (treturn (MNonReact SM) t1',Sto1') /\
                                                   (t',Sto2) ~~~>* (treturn (MNonReact SM) t2',Sto2') /\
                                                   value t1' /\ value t2' /\
                                                   store_eq M Sto1' Sto2' /\
                                                   term_eq T' t1' t2')
                          | MReactT TI TO SM => exists t1' Sto1' t2' Sto2',
                                                  ((t, Sto1) ~~~>* (t1',Sto1') /\
                                                   (t',Sto2) ~~~>* (t2',Sto2') /\
                                                   value t1' /\ value t2' /\
                                                   along2_react
                                                     (store_eq M) (term_eq TI) (term_eq TO) (term_eq T')
                                                     (t1',Sto1') (t2',Sto2'))
                          end
     end
with store_eq (M:Mo) (Sto1 Sto2:store) {struct M} : Prop :=
  store_matches_mo Sto1 M /\ store_all_values Sto1 /\ 
  store_matches_mo Sto2 M /\ store_all_values Sto2 /\ 
  match M with
  | MNonReact SM     => sm_eq SM Sto1 Sto2
  | MReactT TI TO SM => sm_eq SM Sto1 Sto2
  end
with sm_eq (SM:SMo) (Sto1 Sto2:store) {struct SM} : Prop :=
   match SM with
   | MIdentity       => True
   | MStateT T b SM' => exists t1 t2 Sto1' Sto2',
                          term_eq T t1 t2 /\
                          sm_eq SM' Sto1' Sto2'  /\
                          Sto1 = t1::Sto1' /\
                          Sto2 = t2::Sto2'
   end.

Notation "x =[ T ]= y" := (term_eq T x y) 
                       (at level 70, T at level 99, format "x =[ T ]= y").

Lemma store_eq_all_values_l : forall {M:Mo} {Sto1 Sto2:store},
                              store_eq M Sto1 Sto2 -> store_all_values Sto1.
Proof.
   intros; destruct M; simpl in H; destruct H as [? [? _]]; auto.
Qed.

Lemma store_eq_all_values_r : forall {M:Mo} {Sto1 Sto2:store},
                              store_eq M Sto1 Sto2 -> store_all_values Sto2.
Proof.
   intros; destruct M; simpl in H; destruct H as [_ [_ [_ [? _]]]]; auto.
Qed.

Lemma eq_typable_empty_l : forall {T} {t} {t'}, 
               term_eq T t t' -> 
      \empty |-- t \in T.
Proof. 
   intros; destruct T; simpl in H; try (destruct H as [? [? ?]]; auto).
           destruct H; auto.
Qed.

Lemma eq_typable_empty_r : forall {T} {t} {t'}, 
               term_eq T t t' -> 
      \empty |-- t' \in T.
Proof. 
   intros; destruct T; simpl in H; try (destruct H as [? [? ?]]; auto).
           destruct H; auto.
Qed. 

Lemma store_eq_stack : forall E T SM t1 t2 Sto1 Sto2,
                         store_eq (MNonReact (MStateT T E SM)) (t1::Sto1) (t2::Sto2) ->
                         store_eq (MNonReact SM) Sto1 Sto2.
Proof.                         
  intros.
  destruct H as [? [? [? [? ?]]]].
  simpl in H3.
  destruct H3 as [t [t' [Sto1' [Sto2' [? [? [? ?]]]]]]].
  repeat (split;auto).
         inv H.
         inv H0.
         inv H1.
         inv H2.
         inv H5.
         inv H6.
Qed.

Lemma sm_eq_lp : forall SM SM' Sto Sto', 
      smo_less_permissive SM SM' -> 
              sm_eq SM  Sto Sto' -> 
              sm_eq SM' Sto Sto'.
Proof.
  induction SM; intros.
  inv H.
  destruct H0 as [t1 [t2 [Sto1 [Sto2 [Htermeq [Hstoeq [Hid1 Hid2]]]]]]].  
  subst.
  inv H;
  simpl;
    exists t1, t2, Sto1, Sto2;
          repeat (split;auto). 
Qed.


Lemma store_eq_lp : forall SM SM' Sto Sto', 
      smo_less_permissive SM SM' -> 
              store_eq (MNonReact SM)  Sto Sto' -> 
              store_eq (MNonReact SM') Sto Sto'.
Proof.
  induction SM; intros.
  inv H.
  destruct H0 as [? [? [? [? ?]]]].
  pose proof (sm_eq_lp (MStateT t e SM) SM' Sto Sto' H H4).
  pose proof (less_permissive_store_matches (MStateT t e SM) SM' Sto H H0).
  pose proof (less_permissive_store_matches (MStateT t e SM) SM' Sto' H H2).       
  simpl. repeat split; auto.
Qed.

Lemma sm_eq_mp : forall SM SM' Sto Sto', 
      smo_less_permissive SM' SM -> 
              sm_eq SM Sto Sto' -> 
              sm_eq SM' Sto Sto'.
Proof.
  induction SM; intros.
  inv H.
  destruct H0 as [t1 [t2 [Sto1 [Sto2 [Htermeq [Hstoeq [Hid1 Hid2]]]]]]].  
  subst.
  inv H;
  simpl;
    exists t1, t2, Sto1, Sto2;
          repeat (split;auto). 
Qed.


Lemma store_eq_mp : forall SM SM' Sto Sto', 
      smo_less_permissive SM' SM -> 
              store_eq (MNonReact SM)  Sto Sto' -> 
              store_eq (MNonReact SM') Sto Sto'.
Proof.
  induction SM; intros.
  inv H.
  destruct H0 as [? [? [? [? ?]]]].
  pose proof (sm_eq_mp (MStateT t e SM) SM' Sto Sto' H H4).
  pose proof (more_permissive_store_matches (MStateT t e SM) SM' Sto H H0).
  pose proof (more_permissive_store_matches (MStateT t e SM) SM' Sto' H H2).       
  simpl. repeat split; auto.
Qed.

Lemma store_eq_non_react : forall Sto Sto' TI TO SM,
                             store_eq (MNonReact SM) Sto Sto' ->
                             store_eq (MReactT TI TO SM) Sto Sto'.
Proof.
 intros.
 simpl.
 destruct H as [? [? [? [? ?]]]].
 repeat split; auto; try (constructor;auto).
Qed.

(* ###################################################################### *)
(** *** Lemmas concerning properties of the along2_react coinductive relation; as well as,
        the term_eq typed equality relation. *)

(* ###################################################################### *)
(** *** Symmetry Lemmas *)

Lemma along2_react_provisional_sym :
  forall (PS : store -> store -> Prop) (PI PO PR : tm -> tm -> Prop),
    (forall Sto1 Sto2, PS Sto1 Sto2 -> PS Sto2 Sto1) ->
    (forall t1 t2, PI t1 t2 -> PI t2 t1) ->
    (forall t1 t2, PO t1 t2 -> PO t2 t1) ->
    (forall t1 t2, PR t1 t2 -> PR t2 t1) ->
    forall co1 co2,
      along2_react PS PI PO PR co1 co2 -> along2_react PS PI PO PR co2 co1.
Proof.
  cofix; intros.
  inversion H3; subst.
    eapply eq_return; eauto.
    eapply eq_pause. 
           eapply H5. 
           eapply H4. 
           auto. auto. 
           eapply H9. 
           eapply H8. 
           eapply H11. 
           eapply H10.
           eapply H13.
           eapply H12.
            apply H1; auto.
            apply H;  auto.
            apply H;  auto.
            auto.
Qed.

(** Symmetry for term_eq, store_eq and sm_eq: *)

Theorem eq_sym :
  (forall T t1 t2, term_eq T t1 t2 -> term_eq T t2 t1) /\
  (forall M Sto1 Sto2, store_eq M Sto1 Sto2 -> store_eq M Sto2 Sto1) /\
  (forall SM Sto1 Sto2, sm_eq SM Sto1 Sto2 -> sm_eq SM Sto2 Sto1).
Proof.
  apply Ty_Mo_SMo_mutind_better; intros.
  Case "TArrow".
    rename H into IHT1.
    rename H0 into IHT2.
    destruct H1 as [? [? ?]].
    split; auto.
    split; auto.
    intros.
    pose proof (IHT1 _ _ H2).
    pose proof (H1 _ _ H3).
    destruct H4 as [? [? [? [? ?]]]].
    pose proof (IHT2 _ _ H6).
    exists x0, x.
    auto.
  Case "TProd".
    rename H into IHT1.
    rename H0 into IHT2.
    destruct H1 as [? [? ?]].
    split; auto.
    split; auto.
    destruct H1 as [t1L [t1R [t2L [t2R [? [? [? ?]]]]]]].
    exists t2L, t2R, t1L, t1R; auto.
  Case "TSum".
    rename H into IHT1.
    rename H0 into IHT2.
    destruct H1 as [? [? ?]].
    split; auto.
    split; auto.
    destruct H1 as [ [t1' [t2' [? [? ?]]]] | [t1' [t2' [? [? ?]]]] ].
    SCase "going left".
      left; exists t2', t1'; auto.
    SCase "going right".
      right; exists t2', t1'; auto.
  Case "TNil".
    destruct H as [? ?].
    unfold term_eq; auto.
  Case "TMonadic".
    destruct H as [IHM IHMonio].
    rename H0 into IHT.
    destruct H1 as [? [? ?]].
    split; auto.
    split; auto.
    intros. 
    destruct M. 
   SCase "MReactT".
      rename t into TI.
      rename t0 into TO. apply IHM in H2. 
            destruct (H1 _ _ H2) as [t2' [Sto2' [t1' [Sto1' [? [? [? [? ?]]]]]]]].
        exists t1', Sto1', t2', Sto2'.
          split. auto. 
          split. auto.
          split. auto.
          split. auto. inversion IHMonio.
          eapply along2_react_provisional_sym;auto. 
    SCase "MNonReact".
      apply IHM in H2.
      destruct (H1 _ _ H2) as [t2' [Sto2' [t1' [Sto1' [? [? [? [? [? ?]]]]]]]]].
        exists t1', Sto1', t2', Sto2'.
          split. auto. 
          split. auto.
          split. auto.
          split. auto.
          split. apply IHM; auto. 
                 apply IHT; auto.
  Case "MReactT".
    unfold store_eq. fold sm_eq.
    destruct H2 as [? [? [? [? ?]]]]. 
    apply H1 in H6.
    auto.
  Case "MNonReact".
    unfold store_eq. fold sm_eq.
    destruct H0 as [? [? [? [? ?]]]]. 
    apply H in H4.
    auto.
  Case "MIdentity".
    auto.
  Case "MStateT".
    unfold sm_eq; fold sm_eq; fold store_eq; fold term_eq.
    unfold sm_eq in H1; fold sm_eq in H1; fold store_eq in H1; fold term_eq in H1.
    destruct H1 as [t1 [t2 [Sto1' [Sto2' [? [? [?]]]]]]].
    pose proof (H0 _ _ H2).
    pose proof (H _ _ H1).
    do 4 eexists; repeat split; eauto.
Qed.

Corollary term_eq_sym : forall T t1 t2, term_eq T t1 t2 -> term_eq T t2 t1.
  apply eq_sym.
Qed.

Corollary store_eq_sym : forall M Sto1 Sto2, store_eq M Sto1 Sto2 -> store_eq M Sto2 Sto1.
  apply eq_sym.
Qed.

Corollary sm_eq_sym : forall SM Sto1 Sto2, sm_eq SM Sto1 Sto2 -> sm_eq SM Sto2 Sto1.
  apply eq_sym.
Qed. 

(** Given that term_eq, store_eq and sm_eq are symmetric, 
    symmetry of along2_react is easily established. *)

Corollary along2_react_sym : forall M TI TO TR co1 co2,
      along2_react (store_eq M) (term_eq TI) (term_eq TO) (term_eq TR) co1 co2 -> 
      along2_react (store_eq M) (term_eq TI) (term_eq TO) (term_eq TR) co2 co1.
Proof.
  intros.
  eapply along2_react_provisional_sym.
  eapply store_eq_sym.
  eapply term_eq_sym.
  eapply term_eq_sym.
  eapply term_eq_sym.
  auto.
Qed.


(* ###################################################################### *)
(** *** Lemmas concerning preservation of term_eq w.r.t. to single and 
        multistep reduction *)
     
Lemma _step_preserves_term_eq' :
  forall T,
   (forall t t' t'', t ~~> t' -> term_eq T t t'' -> term_eq T t' t'') ->
   forall t t' t'', t ~~>* t' -> term_eq T t t'' -> term_eq T t' t''.
Proof.
  intros T Hs t t' t'' Hstep Heq.
  generalize dependent Heq.
  generalize dependent t''.
  generalize dependent T.
  induction Hstep; intros T Hsingle t'' Hstept.
    assumption.
    eapply IHHstep.
      intros. eapply Hsingle. eauto. assumption.
      eapply Hsingle; eauto.
Qed.

Theorem step_preserves_eq : forall T t t' t'', 
      t ~~> t' -> term_eq T t t'' -> term_eq T t' t''.
Proof.
  induction T; intros t t' t'' Hstep Heq; 
  try (destruct Heq as [? [? ?]]); simpl; [ | | | |destruct m]. 
  Case "TArrow".
       repeat split;[apply (preservation _ _ H _ Hstep)|auto|intros].
       destruct (H1 _ _ H2) as [t2 [t2' [? [? ?]]]].
       destruct (normalization _ _ (eq_typable_empty_l H5)) as [vt2 [? ?]].
       destruct (normalization _ _ (eq_typable_empty_r H5)) as [vt2' [? ?]].
       pose proof (rtc_Trans _ _ _ H3 H6); pose proof (rtc_Trans _ _ _ H4 H8).
       pose proof (ST_App1 _ _ t1 Hstep).
       pose proof (step_value_second _ _ _ H7 H12 H10).
       exists vt2; exists vt2'.
       pose proof (_step_preserves_term_eq' _ IHT2).
       pose proof (H14 _ _ t2' H6 H5). 
       pose proof (term_eq_sym _ _ _ H15). 
       pose proof (term_eq_sym _ _ _ (H14 _ _ vt2 H8 H16)). 
       repeat split; auto.
  Case "TProd".
       repeat split;[apply (preservation _ _ H _ Hstep)|auto|]. 
       destruct H1 as [t1 [t2 [t1' [t2' [? [? [? ?]]]]]]].
       destruct (normalization _ _ (eq_typable_empty_l H3)) as [v1 [? ?]];
       destruct (normalization _ _ (eq_typable_empty_r H3)) as [v1' [? ?]];
       destruct (normalization _ _ (eq_typable_empty_l H4)) as [v2 [? ?]];
       destruct (normalization _ _ (eq_typable_empty_r H4)) as [v2' [? ?]].
       assert (t' ~~>* tpair v1 v2).
               assert (tpair t1 t2 ~~>* tpair v1 v2) by (
               pose proof (Congruence_Pair1 _ _ t2 H5);
               pose proof (Congruence_Pair2 _ _ _ H6 H9); eauto).
               eapply step_value_second. 
               constructor; auto.
               eapply Hstep.
               eauto.
       assert (tpair t1' t2' ~~>* tpair v1' v2') by (
               pose proof (Congruence_Pair1 _ _ t2' H7);
               pose proof (Congruence_Pair2 _ _ _ H8 H11); eauto).
       pose proof (_step_preserves_term_eq' _ IHT1);
       pose proof (_step_preserves_term_eq' _ IHT2).
       exists v1; exists v2; exists v1'; exists v2'. 
              repeat split. eauto. eauto.
              eapply H15. eauto.
              apply term_eq_sym.
              eapply H15. eauto.
              apply term_eq_sym.
              assumption.
              eapply H16. eauto.
              apply term_eq_sym.
              eapply H16. eauto.
              apply term_eq_sym.
              assumption.
  Case "TSum".
       repeat split;[apply (preservation _ _ H _ Hstep)|auto|]. 
       pose proof (_step_preserves_term_eq' _ IHT1) as IHT1';
       pose proof (_step_preserves_term_eq' _ IHT2) as IHT2'.
       destruct H1.
       SCase "left".
              left. 
              destruct H1 as [t1 [t1' [? [? ?]]]]. 
              destruct (normalization _ _ (eq_typable_empty_l H3)) as [v1 [? ?]];
              destruct (normalization _ _ (eq_typable_empty_r H3)) as [v1' [? ?]].
              exists v1; exists v1'; repeat split.
                     eapply step_value_second.
                     constructor; auto.
                     eauto.
                     eapply rtc_Trans; eauto.
                     eapply Congruence_Tinl; eauto.
                     eapply rtc_Trans; eauto.
                     eapply Congruence_Tinl; eauto. 
                     eapply IHT1'. eauto.
                     apply term_eq_sym.
                     eapply IHT1'. eauto.
                     apply term_eq_sym.
                     assumption.
       SCase "right".
              right.
              destruct H1 as [t2 [t2' [? [? ?]]]]. 
              destruct (normalization _ _ (eq_typable_empty_l H3)) as [v2 [? ?]];
              destruct (normalization _ _ (eq_typable_empty_r H3)) as [v2' [? ?]].
              exists v2; exists v2'; repeat split.
                     eapply step_value_second.
                     constructor; auto.
                     eauto.
                     eapply rtc_Trans; eauto.
                     eapply Congruence_Tinr; eauto.
                     eapply rtc_Trans; eauto.
                     eapply Congruence_Tinr; eauto. 
                     eapply IHT2'. eauto.
                     apply term_eq_sym.
                     eapply IHT2'. eauto.
                     apply term_eq_sym.
                     assumption.
  Case "TNil".
       destruct Heq. repeat split;[apply (preservation _ _ H _ Hstep)|auto]. 
  Case "MReactT".
        repeat split;[apply (preservation _ _ H _ Hstep)|auto|intros].
        destruct (H1 _ _ H2) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? ?]]]]]]]]. 
        pose proof (step_monad_second t t' t1' Sto1 Sto1' Hstep H5 H3).
        exists t1'; exists Sto1'; exists t2'; exists Sto2'; repeat split.
               auto. auto. auto. auto. auto.
  Case "MNonReact". 
        repeat split; [preservation;auto|auto|intros].
        destruct (H1 _ _ H2) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? [? ?]]]]]]]]]. 
        assert (value ((treturn (MNonReact s) t1'))) by (constructor; auto).
        pose proof (step_monad_second t t' _ Sto1 Sto1' Hstep H9 H3).
        destruct H7 as [? [? [? [? ?]]]].
        exists t1'; exists Sto1'; exists t2'; exists Sto2'; repeat split; auto.
Qed.

Corollary step_preserves_eq' : forall T t t' t'', 
          t ~~>* t' -> term_eq T t t'' -> term_eq T t' t''.
Proof.
  intros T t t' t'' Hstept HTt.
  induction Hstept; intros.
  Case "0".
    assumption.
  Case "n".
    pose proof (step_preserves_eq T x y _ H HTt).
    eapply IHHstept; eauto.
Qed.

Theorem step_preserves_eq_rev : forall T t t' t'', 
           \empty |-- t \in T ->
                     t ~~> t' -> 
             term_eq T t' t'' -> 
             term_eq T t t''.
Proof.
  induction T; intros t t' t'' HTt Hstep Heq;
               try (destruct Heq as [HTt' [HTt'' H_All]]);
               try (destruct Heq as [HTt' HTt'']); simpl;
               try (split; [auto|split;[auto|]]); try (split;auto).
  Case "TArrow".
    intros;pose proof (H_All _ _ H) as H'; destruct H' as [t2 [t3 [? [? ?]]]]. 
     destruct (normalization _ _ (eq_typable_empty_l H2)) as [vt2 [? ?]].
     destruct (normalization _ _ (eq_typable_empty_r H2)) as [vt2' [? ?]].
       pose proof (rtc_Trans _ _ _ H0 H3); pose proof (rtc_Trans _ _ _ H1 H5).
       pose proof (Congruence_App1 _ _ t1 (rtc_R _ _ Hstep)).
       exists vt2; exists vt2'.
              repeat split.
              eauto. eauto.
              eapply step_preserves_eq'. 
                     eapply H3.
                     apply term_eq_sym. 
              eapply step_preserves_eq'. 
                     eapply H5.
                     apply term_eq_sym.
              assumption.
   Case "TProd".
       destruct H_All as [t1 [t2 [t1' [t2' [? [? [? ?]]]]]]].
       destruct (normalization _ _ (eq_typable_empty_l H1)) as [v1 [? ?]];
       destruct (normalization _ _ (eq_typable_empty_r H1)) as [v1' [? ?]];
       destruct (normalization _ _ (eq_typable_empty_l H2)) as [v2 [? ?]];
       destruct (normalization _ _ (eq_typable_empty_r H2)) as [v2' [? ?]].
       assert (t ~~>* tpair v1 v2).
               assert (tpair t1 t2 ~~>* tpair v1 v2) by (
               pose proof (Congruence_Pair1 _ _ t2 H3);
               pose proof (Congruence_Pair2 _ _ _ H4 H7); eauto).
               eapply rtc_Trans. eapply rtc_R; eauto. 
               auto.
       assert (tpair t1' t2' ~~>* tpair v1' v2') by (
               pose proof (Congruence_Pair1 _ _ t2' H5);
               pose proof (Congruence_Pair2 _ _ _ H6 H9); eauto).
       exists v1; exists v2; exists v1'; exists v2'. 
              repeat split.
              auto. auto.
              eapply step_preserves_eq'. 
                     eapply H3.
                     apply term_eq_sym. 
              eapply step_preserves_eq'. 
                     eapply H5.
                     apply term_eq_sym.
              assumption.
              eapply step_preserves_eq'. 
                     eapply H7.
                     apply term_eq_sym. 
              eapply step_preserves_eq'. 
                     eapply H9.
                     apply term_eq_sym.
              assumption.
  Case "TSum".
       destruct H_All.
       SCase "left".
              left. 
              destruct H as [t1 [t1' [? [? ?]]]]. 
              destruct (normalization _ _ (eq_typable_empty_l H1)) as [v1 [? ?]];
              destruct (normalization _ _ (eq_typable_empty_r H1)) as [v1' [? ?]].
              exists v1; exists v1'; repeat split.
                     eapply rtc_Trans.
                     eapply rtc_R; eauto.
                     eapply rtc_Trans ; eauto.
                     eapply Congruence_Tinl;eauto.
                     eapply rtc_Trans; eauto.
                     eapply Congruence_Tinl; eauto.
                     eapply step_preserves_eq'. 
                     eapply H2.
                     apply term_eq_sym. 
                     eapply step_preserves_eq'. 
                     eapply H4.
                     apply term_eq_sym.
              assumption.
       SCase "right".
              right.
              destruct H as [t1 [t1' [? [? ?]]]]. 
              destruct (normalization _ _ (eq_typable_empty_l H1)) as [v1 [? ?]];
              destruct (normalization _ _ (eq_typable_empty_r H1)) as [v1' [? ?]].
              exists v1; exists v1'; repeat split.
                     eapply rtc_Trans.
                     eapply rtc_R; eauto.
                     eapply rtc_Trans ; eauto.
                     eapply Congruence_Tinr;eauto.
                     eapply rtc_Trans; eauto.
                     eapply Congruence_Tinr; eauto.
                     eapply step_preserves_eq'. 
                     eapply H2.
                     apply term_eq_sym. 
                     eapply step_preserves_eq'. 
                     eapply H4.
                     apply term_eq_sym.
              assumption.
  Case "TMonadic".
       intros. destruct m.
  SCase "MReactT".
        destruct (H_All _ _ H) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? ?]]]]]]]]. 
        pose proof (Congruence_LC _ _ Sto1 (rtc_R _ _ Hstep)).
        exists t1'; exists Sto1'; exists t2'; exists Sto2'; repeat split.
               auto. auto. auto. auto. auto.
  SCase "MNonReact".
        destruct (H_All _ _ H) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? [? ?]]]]]]]]]. 
        pose proof (Congruence_LC _ _ Sto1 (rtc_R _ _ Hstep)).
        destruct H4 as [? [? [? [? ?]]]].
        exists t1'; exists Sto1'; exists t2'; exists Sto2'; repeat split; auto.
Qed.

Corollary step_preserves_eq_rev' : forall T t t' t'', 
           \empty |-- t \in T ->
                     t ~~>* t' -> 
             term_eq T t' t'' -> 
             term_eq T t t''.
Proof.
  intros. induction H0.
  Case "Base case".
       assumption.
  Case "Inductive Step".
      assert (\empty |-- y \in T).
             eapply preservation;eauto.
      pose proof (IHrt_closure H3 H1).
      eapply step_preserves_eq_rev;eauto.
Qed.

Lemma eq_square : forall T t1 t1' t2 t2', 
      term_eq T t1 t2 -> 
          t1 ~~>* t1' -> 
          t2 ~~>* t2' -> 
      term_eq T t1' t2'.
Proof.
  intros T t1 t1' t2 t2' Heq12 Hstep1 Hstep2.
        assert (term_eq T t1' t2).
               eapply step_preserves_eq';eauto.
        apply term_eq_sym in H.
        assert (term_eq T t2' t1').
               eapply step_preserves_eq';eauto.
        apply term_eq_sym in H0.
  assumption.
Qed.

Lemma eq_square_rev : forall T t1 t1' t2 t2', 
      \empty |-- t1 \in T ->
      \empty |-- t2 \in T ->
      t1 ~~>* t1' ->
      t2 ~~>* t2' ->
      term_eq T t1' t2' ->
      term_eq T t1 t2.
Proof.
  intros T t1 t1' t2 t2' HTt1 HTt2 Hstep1 Hstep2 Heq12'.
        assert (term_eq T t1 t2').
               eapply step_preserves_eq_rev';eauto.
        apply term_eq_sym in H.
        assert (term_eq T t2 t1).
               eapply step_preserves_eq_rev'; eauto.
        apply term_eq_sym in H0.
  assumption.
Qed.
 
Theorem along2_react_provisional_trans
                               : forall (PS : store -> store -> Prop) (PI PO PR : tm -> tm -> Prop),
                                 (forall Sto1 Sto2 Sto3,
                                    PS Sto1 Sto2 -> PS Sto2 Sto3 -> PS Sto1 Sto3) ->
                                 (forall t1 t2 t3,
                                    PI t1 t2 -> PI t2 t3 -> PI t1 t3) ->
                                  (forall t1 t2 t3,
                                    PO t1 t2 -> PO t2 t3 -> PO t1 t3) ->
                                 (forall t1 t2 t3,
                                    PR t1 t2 -> PR t2 t3 -> PR t1 t3) ->
                                 (forall Sto1 Sto2,
                                    PS Sto1 Sto2 -> PS Sto2 Sto1) ->
                                 (forall t1 t2,
                                    PI t1 t2 -> PI t2 t1) ->
                                 (forall t1 t2,
                                    PO t1 t2 -> PO t2 t1) ->
                                 (forall t1 Sto1 t2 Sto2 t3 Sto3,
                                    along2_react PS PI PO PR (t1,Sto1) (t2,Sto2) ->
                                    along2_react PS PI PO PR (t2,Sto2) (t3,Sto3) ->
                                    along2_react PS PI PO PR (t1,Sto1) (t3,Sto3)).
Proof.
  cofix COIND; intros.
  inversion H6; subst.
  Case "t1 going to treturn".
    inversion H7; subst.
    SCase "t2 going to treturn".
      eapply eq_return.
        eapply H12.

        assert (done_mo (treturn (MReactT TI TO SM) t2',Sto2')) by (constructor; auto).
        assert (done_mo (treturn (MReactT TI0 TO0 SM0) t1'0,Sto1'0)) by (constructor; auto).
        pose proof (done_unique _ _ _ H8 H9 H13 H14).
        inversion H10; subst.
        eapply H15; eauto.
        auto.
        auto.
        
        assert (done_mo (treturn (MReactT TI TO SM) t2',Sto2')) by (constructor; auto).
        assert (done_mo (treturn (MReactT TI0 TO0 SM0) t1'0,Sto1'0)) by (constructor; auto).
        pose proof (done_unique _ _ _ H8 H9 H13 H14).
        inversion H10; subst.
        eapply H2; eauto.
        assert (done_mo (treturn (MReactT TI TO SM) t2',Sto2')) by (constructor; auto).
        assert (done_mo (treturn (MReactT TI0 TO0 SM0) t1'0,Sto1'0)) by (constructor; auto).
        pose proof (done_unique _ _ _ H8 H9 H13 H14).
        inversion H10; subst.
        eapply H; eauto.
    SCase "t2 going to pause".
      assert (done_mo (treturn (MReactT TI TO SM) t2', Sto2')) by (constructor; auto).
      assert (done_mo (tpause (MReactT TI0 TO0 SM0) T t1'0, Sto1'0)) by (constructor; auto).
         pose proof (done_unique _ _ _ H8 H9 H13 H14).
        inversion H10; subst.
  Case "t1 going to tpause".
    inversion H7; subst.
    SCase "t2 going to treturn".
      assert (done_mo (tpause (MReactT TI TO SM) T t2', Sto2')) by (constructor; auto).
      assert (done_mo (treturn (MReactT TI0 TO0 SM0) t1'0, Sto1'0)) by (constructor; auto).
       pose proof (done_unique _ _ _ H8 H9 H13 H22).
      inversion H10.
    SCase "t2 going to pause".
      eapply eq_pause.
        eapply H12.

        assert (done_mo (tpause (MReactT TI TO SM) T t2', Sto2')) by (constructor; auto).
        assert (done_mo (tpause (MReactT TI0 TO0 SM0) T0 t1'0, Sto1'0)) by (constructor; auto).
        pose proof (done_unique _ _ _ H8 H9 H13 H22).
        inversion H10; subst. 
        eapply H23.

        auto.

        auto.

        assert (done_mo (tpause (MReactT TI TO SM) T t2', Sto2')) by (constructor; auto).
        assert (done_mo (tpause (MReactT TI0 TO0 SM0) T0 t1'0, Sto1'0)) by (constructor; auto).
         pose proof (done_unique _ _ _ H8 H9 H13 H22).
         inversion H10; subst.
        eapply H16. 
        eapply H31.
        eapply H18. 
        eapply H33. 
        eapply H20. 
        assert (done_mo (tpause (MReactT TI TO SM) T t2', Sto2')) by (constructor; auto).
        assert (done_mo (tpause (MReactT TI0 TO0 SM0) T0 t1'0, Sto1'0)) by (constructor; auto).
         pose proof (done_unique _ _ _ H8 H9 H13 H22).
         inversion H10; subst.
        eapply H35. 
        assert (done_mo (tpause (MReactT TI TO SM) T t2', Sto2')) by (constructor; auto).
        assert (done_mo (tpause (MReactT TI0 TO0 SM0) T0 t1'0, Sto1'0)) by (constructor; auto).
         pose proof (done_unique _ _ _ H8 H9 H13 H22).
         inversion H10; subst. 
        assert (done_mo (treturn (MNonReact SM0) (tpair vl0 vr0), Sto1''0)) by (constructor; auto).
        assert (done_mo (treturn (MNonReact SM0) (tpair vl' vr'), Sto2'')) by (constructor; auto).
         pose proof (done_unique _ _ _ H11 H36 H34 H21).
         inv H37. 
        eapply H1;auto. 
        eapply H26. auto.
        assert (done_mo (tpause (MReactT TI TO SM) T t2', Sto2')) by (constructor; auto).
        assert (done_mo (tpause (MReactT TI0 TO0 SM0) T0 t1'0, Sto1'0)) by (constructor; auto).
         pose proof (done_unique _ _ _ H8 H9 H13 H22).
         inversion H10; subst.
        eapply H. eapply H27. auto. 
       assert (done_mo (tpause (MReactT TI TO SM) T t2', Sto2')) by (constructor; auto).
        assert (done_mo (tpause (MReactT TI0 TO0 SM0) T0 t1'0, Sto1'0)) by (constructor; auto).
         pose proof (done_unique _ _ _ H8 H9 H13 H22).
         inversion H10; subst.
        eapply H. eauto. 
        assert (done_mo (treturn (MNonReact SM0) (tpair vl0 vr0), Sto1''0)) by (constructor; auto).
        assert (done_mo (treturn (MNonReact SM0) (tpair vl' vr'), Sto2'')) by (constructor; auto).
         pose proof (done_unique _ _ _ H11 H36 H34 H21).
         inversion H37; subst.
         auto.
        intros.
        assert (done_mo (tpause (MReactT TI TO SM) T t2', Sto2')) by (constructor; auto).
        assert (done_mo (tpause (MReactT TI0 TO0 SM0) T0 t1'0, Sto1'0)) by (constructor; auto).
         pose proof (done_unique _ _ _ H10 H11 H13 H22).
         inversion H36; subst. 
        assert (done_mo (treturn (MNonReact SM0) (tpair vl0 vr0), Sto1''0)) by (constructor; auto).
        assert (done_mo (treturn (MNonReact SM0) (tpair vl' vr'), Sto2'')) by (constructor; auto).
         pose proof (done_unique _ _ _ H37 H38 H34 H21).
         inversion H39; subst.
        pose proof (H0 _ _ _ (H4 _ _ H8) H8). 
        pose proof (H  _ _ _ (H3 _ _ H9) H9).
        pose proof (H29 _ _ _ _ H8 H9).
        pose proof (H43 _ _ _ _ H44 H45). 
        eapply COIND; auto. 
        eapply H46. auto.
Qed.
       

Lemma step_preserves_along2_react :
  forall t1 Sto1 t1' Sto1' t2 Sto2 t2' Sto2' (PS: store -> store -> Prop) PI PO PR,
    (t1,Sto1) ~~~> (t1',Sto1') ->
    (t2,Sto2) ~~~> (t2',Sto2') ->
    along2_react PS PI PO PR (t1',Sto1') (t2',Sto2') ->
    along2_react PS PI PO PR (t1,  Sto1) (t2,  Sto2).
Proof.
  intros  t1 Sto1 t1' Sto1' t2 Sto2 t2' Sto2' PS PI PO PR Hstep1 Hstep2 Heq_react.
  inversion Heq_react; subst.
  Case "to return".
    eapply eq_return; eauto.
  Case "to pause".
    eapply eq_pause. 
    eapply rtc_Trans_mo. 
    eapply rtc_R_mo.
    eauto. 
    eapply H3. 
    eapply rtc_Trans_mo. 
    eapply rtc_R_mo.
    eauto. 
    eapply H4.
    auto.
    auto. 
    eapply H7. 
    eapply H8.
    eapply H9.
    eapply H10.
    eapply H11.
    eapply H12.
    auto.
    auto.
    auto.
    auto.
Defined.

Lemma step_preserves_along2_react': forall t1 Sto1 t1' Sto1' t2 Sto2 t2' Sto2' SM TI TO TR,
      (t1,Sto1) ~~~> (t1',Sto1') ->
      (t2,Sto2) ~~~> (t2',Sto2') ->
      along2_react (store_eq SM) (term_eq TI) (term_eq TO) (term_eq TR) (t1,Sto1) (t2,Sto2) ->
      along2_react (store_eq SM) (term_eq TI) (term_eq TO) (term_eq TR) (t1',Sto1') (t2',Sto2').
Proof.
   intros  t1 Sto1 t1' Sto1' t2 Sto2 t2' Sto2' SM TI TO TR Hstep1 Hstep2 Heq_react.
   inversion Heq_react; subst.
   Case "going to return".
        eapply eq_return.   
        assert (done_mo ((treturn (MReactT TI0 TO0 SM0) t1'0, Sto1'0))) by (constructor;auto).
        eapply (step_done_second (t1,Sto1) (t1',Sto1')  _ H Hstep1 H3).
        assert (done_mo ((treturn (MReactT TI0 TO0 SM0) t2'0, Sto2'0))) by (constructor;auto).
        eapply (step_done_second (t2,Sto2) (t2',Sto2')  _ H Hstep2 H4).
        auto. auto. auto. auto.
   Case "going to pause".
        eapply eq_pause.
        assert (done_mo ((tpause (MReactT TI0 TO0 SM0) T t1'0, Sto1'0))) by (constructor;auto).
        eapply (step_done_second (t1,Sto1) (t1',Sto1')  _ H Hstep1 H3).
        assert (done_mo ((tpause (MReactT TI0 TO0 SM0) T t2'0, Sto2'0))) by (constructor;auto).
        eapply (step_done_second (t2,Sto2) (t2',Sto2')  _ H Hstep2 H4).
        auto. auto. 
        eapply H7. 
        eapply H8. 
        eapply H9.
        eapply H10.
        eapply H11. 
        eapply H12.
        auto. auto. auto. auto.
Defined.

Lemma step_preserves_along2_react_star : forall (PS: store -> store -> Prop) PI PO PR,
  forall t1 Sto1 t1' Sto1' t2 Sto2 t2' Sto2', 
    (t1,Sto1) ~~~>* (t1',Sto1') ->
    (t2,Sto2) ~~~>* (t2',Sto2') ->
    along2_react PS PI PO PR (t1',Sto1') (t2',Sto2') ->
    along2_react PS PI PO PR (t1,Sto1) (t2,Sto2).
Proof.
  intros. 
  inversion H1; subst.
  Case "going to return".
       eapply eq_return.
         eapply rtc_Trans_mo.
           eauto.
           eauto.
         eapply rtc_Trans_mo.
           eauto.
           eauto.
         auto.
         auto.
         auto.
         auto.
  Case "going to pause".
       eapply eq_pause.
         eapply rtc_Trans_mo.
           eauto.
           eauto.
         eapply rtc_Trans_mo.
           eauto.
           eauto.
         auto.
         auto. 
         eapply H10.
         eapply H11.
         eapply H12.
         eapply H13.
         eapply H14.
         eapply H15.
         auto.
         auto.
         auto.
         auto.
Defined.
  
Lemma step_preserves_along2_react_star': 
      forall t1 Sto1 t1' Sto1' t2 Sto2 t2' Sto2' SM TI TO TR,
            (t1,Sto1) ~~~>* (t1',Sto1') ->
            (t2,Sto2) ~~~>* (t2',Sto2') ->
            along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) (t1,Sto1) (t2, Sto2) ->
            along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) (t1',Sto1') (t2',Sto2').
Proof.
  intros.
  inversion H1; subst.
  Case "going to return".
       assert (HS1: (t1',Sto1') ~~~>* (treturn (MReactT TI0 TO0 SM0) t1'0, Sto1'0)).
              eapply step_same_done.
              constructor;auto.
              eassumption.
              auto.
       assert (HS2: (t2', Sto2') ~~~>* (treturn (MReactT TI0 TO0 SM0) t2'0, Sto2'0)).
              eapply step_same_done.
              constructor;auto.
              eassumption.
              auto.
       eapply eq_return; eauto.
  Case "going to pause".
       assert (HS1: (t1',Sto1') ~~~>* (tpause (MReactT TI0 TO0 SM0) T t1'0, Sto1'0)).
              eapply step_same_done.
              constructor;auto.
              eassumption.
              auto.
       assert (HS2: (t2', Sto2') ~~~>* (tpause (MReactT TI0 TO0 SM0) T t2'0, Sto2'0)).
              eapply step_same_done.
              constructor;auto.
              eassumption.
              auto.
       eapply eq_pause.  
       eapply HS1. 
       eapply HS2. 
       auto.
       auto. 
       eapply H10.
       eapply H11.
       eapply H12.
       eapply H13.
       eapply H14. 
       eapply H15.
       auto. 
       auto.
       auto.
       auto.
Defined.

Lemma step_along2_react: forall t1 Sto1 t2 Sto2 t3 Sto3 SM TI TO TR,
      (t1,Sto1) ~~~>* (t2,Sto2) ->
      along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) (t1,Sto1) (t3,Sto3) ->
      along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) (t2,Sto2) (t3,Sto3).
Proof.       
  intros. 
  assert ((t3,Sto3) ~~~>* (t3,Sto3)) by (eauto).
  apply (step_preserves_along2_react_star' t1 Sto1 t2 Sto2 t3 Sto3 t3 Sto3 SM TI TO TR H H1 H0).
Defined.

Lemma step_along2_react_rev : forall t1 Sto1 t2 Sto2 t3 Sto3 SM TI TO TR,
      (t1,Sto1) ~~~>* (t2,Sto2) ->
      along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) (t2,Sto2) (t3,Sto3) ->
      along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) (t1,Sto1) (t3,Sto3).
Proof.       
  intros. 
  assert ((t3,Sto3) ~~~>* (t3,Sto3)) by (eauto).
  apply (step_preserves_along2_react_star _ _ _ _ t1 Sto1 t2 Sto2 t3 Sto3 t3 Sto3 H H1 H0).
Defined.


(** *)

Lemma TPair_Term_eq : forall TL TR tl tr tl' tr',
        (term_eq TL tl tl' /\  term_eq TR tr tr') <->
        term_eq (TProd TL TR) (tpair tl tr) (tpair tl' tr').
Proof.
  intros TL TR tl tr tl' tr'.  
  split.
  Case "->".
    intros. 
    destruct H as [HeqTL HeqTR]. 
    repeat split;[ constructor; [ eapply eq_typable_empty_l in HeqTL;eauto |
                                  eapply eq_typable_empty_l in HeqTR;eauto ]| 
                   constructor; [ eapply eq_typable_empty_r in HeqTL;eauto |
                                  eapply eq_typable_empty_r in HeqTR;eauto ]|
                 ].
    exists tl, tr, tl', tr'.
         repeat split; [ eapply rtc_refl;eauto |
                         eapply rtc_refl;eauto |
                         auto |
                         auto ].
  Case "<-".
    intros HProd; destruct HProd as [HTtpair [HTtpair' H_Exists]];
                  destruct H_Exists as [t1 [t2 [t1' [t2'[Hsteplr [Hsteplr' [HeqTLR HeqTLR']]]]]]]. 
    destruct (normalization _ _ (eq_typable_empty_l HeqTLR)) as [vt1 [? ?]].
    destruct (normalization _ _ (eq_typable_empty_r HeqTLR)) as [vt1' [? ?]].
    destruct (normalization _ _ (eq_typable_empty_l HeqTLR')) as [vt2 [? ?]].
    destruct (normalization _ _ (eq_typable_empty_r HeqTLR')) as [vt2' [? ?]].
    assert (term_eq TL vt1 vt1') by (eapply eq_square;eauto).    
    assert (term_eq TR vt2 vt2') by (eapply eq_square;eauto). 
    assert (tpair t1 t2 ~~>* tpair vt1 vt2).
           pose proof (Congruence_Pair1 t1 vt1 t2 H).
           pose proof (Congruence_Pair2 vt1 t2 vt2 H0 H3).
           eauto.
    assert (tpair t1' t2' ~~>* tpair vt1' vt2').
           pose proof (Congruence_Pair1 t1' vt1' t2' H1).
           pose proof (Congruence_Pair2 vt1' t2' vt2' H2 H5).
           eauto.
    assert (tpair tl tr ~~>* tpair vt1 vt2) by (eapply rtc_Trans;eauto).
    assert (tpair tl' tr' ~~>* tpair vt1' vt2') by (eapply rtc_Trans;eauto).
    assert (value (tpair vt1 vt2)) by (constructor;auto).
    assert (value (tpair vt1' vt2')) by (constructor;auto).
    assert (HTtpair12: \empty |-- tpair t1 t2 \in TProd TL TR).
           constructor; [ eapply eq_typable_empty_l in HeqTLR;eauto |
                          eapply eq_typable_empty_l in HeqTLR';eauto].
    assert (HTtpair12': \empty |-- tpair t1' t2' \in TProd TL TR).
           constructor; [ eapply eq_typable_empty_r in HeqTLR;eauto |
                          eapply eq_typable_empty_r in HeqTLR';eauto].
    assert (HTl:  \empty |-- tl \in TL).
          inversion HTtpair;subst. 
          auto.
    assert (HTl': \empty |-- tl' \in TL).
          inversion HTtpair';subst. 
          auto.
    assert (HTr:  \empty |-- tr \in TR).
          inversion HTtpair;subst. 
          auto.
    assert (HTr': \empty |-- tr' \in TR).
          inversion HTtpair';subst. 
          auto.
    destruct (normalization _ _ HTl)  as [vL [? ?]].
    destruct (normalization _ _ HTl') as [vL'[? ?]].
    destruct (normalization _ _ HTr)  as [vR [? ?]].
    destruct (normalization _ _ HTr') as [vR'[? ?]].
    assert (tpair tl tr ~~>* tpair vL vR).
           pose proof (Congruence_Pair1 tl vL tr H15).
           pose proof (Congruence_Pair2 vL tr vR H16 H19). 
           eauto.
    assert (tpair tl' tr' ~~>* tpair vL' vR').
           pose proof (Congruence_Pair1 tl' vL' tr' H17).
           pose proof (Congruence_Pair2 vL' tr' vR' H18 H21).
           eauto.
    assert (value (tpair vL vR)) by (constructor;auto).
    assert (value (tpair vL' vR')) by (constructor;auto).
    assert ((tpair vL vR) = (tpair vt1 vt2)) by (eapply value_unique;eauto).
    assert ((tpair vL' vR') = (tpair vt1' vt2')) by (eapply value_unique;eauto).
    inversion H27; inversion H28; subst.
    split.
      eapply eq_square_rev.
        auto.
        auto.
        eassumption.
        eassumption.
        assumption.
     eapply eq_square_rev.
        auto.
        auto.
        eassumption.
        eassumption.
        assumption.
Qed.

(* ###################################################################### *)
(** *** Transitivity Lemmas *)

Lemma eq_trans :
  (forall T t t' t'', term_eq T t t' -> term_eq T t' t'' -> term_eq T t t'') /\
  (forall M Sto1 Sto2 Sto3, store_eq M Sto1 Sto2 -> store_eq M Sto2 Sto3 -> store_eq M Sto1 Sto3) /\
  (forall SM Sto1 Sto2 Sto3, sm_eq SM Sto1 Sto2 -> sm_eq SM Sto2 Sto3 -> sm_eq SM Sto1 Sto3).
Proof.
  apply Ty_Mo_SMo_mutind_better; intros;
  try (rename H into IHT1; rename H0 into IHT2;
       rename H1 into H; rename H2 into H0);
  try (remember H as H'; remember H0 as H0'; 
  clear HeqH'; clear HeqH0').
  Case "TArrow".
    repeat split;
    try (pose proof (eq_typable_empty_l H); eauto); 
    try (pose proof (eq_typable_empty_r H); eauto); 
    try (pose proof (eq_typable_empty_r H0); eauto);
    try (apply (eq_typable_empty_l H1));
    try (apply (eq_typable_empty_r H2)).
    destruct H  as [? [? H_1_2]];
    destruct H0 as [? [HTt3 H_2_3]];
    intros targ1 targ3 Hteqarg13.
    assert (HTapp1: \empty |-- tapp t targ1 \in T2). 
           econstructor; eauto. apply (eq_typable_empty_l Hteqarg13).
    assert (HTapp3: \empty |-- tapp t'' targ3 \in T2).
           econstructor; eauto. apply (eq_typable_empty_r Hteqarg13).
    pose proof (normalization _ _ HTapp1) as [va1 [Hstepva1 Hvalva1]].
    pose proof (normalization _ _ HTapp3) as [va2 [Hstepva2 Hvalva2]].
    pose proof (normalization _ _ H1) as [vt1 [Hstept1 Hvalvt1]].
    pose proof (normalization _ _ H2) as [vt2 [Hstept2 Hvalvt2]].
    pose proof (normalization _ _ H3) as [vt3 [Hstept3 Hvalvt3]].
    assert (HeqVT12: term_eq (TArrow T1 T2) vt1 vt2).
           pose proof (eq_square (TArrow T1 T2) t vt1 t' vt2 H' Hstept1 Hstept2).
           auto.
    assert (HeqVT23: term_eq (TArrow T1 T2) vt2 vt3).
           pose proof (eq_square (TArrow T1 T2) t' vt2 t'' vt3 H0' Hstept2 Hstept3).
           auto.
    pose proof (normalization _ _ (eq_typable_empty_l Hteqarg13)) as [varg1 [Hstepvarg1 Hvalvarg1]].
    pose proof (normalization _ _ (eq_typable_empty_r Hteqarg13)) as [varg3 [Hstepvarg3 Hvalvarg3]].
    assert (HeqVTarg13: term_eq T1 varg1 varg3) by (eapply eq_square;eauto).
    assert (H_ST1: tapp t targ1 ~~>* tapp vt1 varg1).
           pose proof (Congruence_App1 _ vt1 targ1 Hstept1) as H_ST1a.
           pose proof (Congruence_App2 _ _ _ Hvalvt1 Hstepvarg1) as H_ST1a'. eauto.
    assert (H_ST3: tapp t'' targ3 ~~>* tapp vt3 varg3).
           pose proof (Congruence_App1 _ vt3 targ3 Hstept3) as H_ST3a.
           pose proof (Congruence_App2 _ _ _ Hvalvt3 Hstepvarg3) as H_ST3a'.
           pose proof (rtc_Trans _ _ _ H_ST3a H_ST3a'). eauto.
    assert (HTtapp_vt1: \empty |-- tapp vt1 varg1 \in T2).
            assert (\empty |-- vt1 \in TArrow T1 T2) by (eapply eq_typable_empty_l;eauto).
            assert (\empty |-- varg1 \in T1) by (eapply eq_typable_empty_l;eauto).
            eapply T_App;eauto. 
    assert (HTtapp_vt3: \empty |-- tapp vt3 varg3 \in T2).
            assert (\empty |-- vt3 \in TArrow T1 T2) by (eapply eq_typable_empty_r;eauto).
            assert (\empty |-- varg3 \in T1) by (eapply eq_typable_empty_r;eauto).
            eapply T_App;eauto.
    pose proof (normalization _ _ HTtapp_vt1) as [vappvv1 [Hstepvappvv1 Hvalvapvv1]].
    pose proof (normalization _ _ HTtapp_vt3) as [vappvv3 [Hstepvappvv3 Hvalvapvv3]].
    pose proof (rtc_Trans _ _ _ H_ST1 Hstepvappvv1).
    pose proof (rtc_Trans _ _ _ H_ST3 Hstepvappvv3).
    exists vappvv1, vappvv3.
          repeat (split; auto).
    destruct HeqVT12 as [HTvt1 [HTvt2 Hvt_1_2]];
    destruct HeqVT23 as [_ [HTvt3 Hvt_2_3]].
    pose proof (Hvt_2_3 _ _ HeqVTarg13) as [t2 [t3 [Happt2 [Happt3 HeqT2t2t3]]]].
    destruct (normalization _ _ (eq_typable_empty_l HeqT2t2t3)) as [val_t2 [Hstepval_t2 Hval_t2]].
    destruct (normalization _ _ (eq_typable_empty_r HeqT2t2t3)) as [val_t3 [Hstepval_t3 Hval_t3]].
    pose proof (rtc_Trans _ _ _ Happt2 Hstepval_t2).
    pose proof (rtc_Trans _ _ _ Happt3 Hstepval_t3).
    assert (HeqValt2t3: term_eq T2 val_t2 val_t3).
           eapply eq_square.
           apply HeqT2t2t3.
           auto. auto.
    assert (HeqVarg1: term_eq T1 varg1 varg1).
           pose proof (step_preserves_eq' _ _ _ _ Hstepvarg1 Hteqarg13).
           pose proof (term_eq_sym _ _ _ H9).
           eapply IHT1. eauto. auto.
    pose proof (Hvt_1_2 _ _ HeqVarg1).
    destruct H9 as [t2' [t3' [Happt2' [Happt3' HeqT2t2t3']]]].
    destruct (normalization _ _ (eq_typable_empty_l HeqT2t2t3')) as [val_t2' [Hstepval_t2' Hval_t2']].
    destruct (normalization _ _ (eq_typable_empty_r HeqT2t2t3')) as [val_t3' [Hstepval_t3' Hval_t3']].
    pose proof (rtc_Trans _ _ _ Happt2' Hstepval_t2').
    pose proof (rtc_Trans _ _ _ Happt3' Hstepval_t3').
    assert (HeqValt2t3': term_eq T2 val_t2' val_t3').
           eapply eq_square.
           apply HeqT2t2t3'.
           auto. auto.
    assert (Hcoqeq: val_t2 = val_t3') by (eapply value_unique;eauto).
    rewrite Hcoqeq in HeqValt2t3.
    assert (term_eq T2 val_t2' val_t3).
           eapply IHT2.
           apply HeqValt2t3'.
           auto.
    pose proof (rtc_Trans _ _ _ H_ST1 H9).
    pose proof (rtc_Trans _ _ _ H_ST3 H8).
    assert (val_t2' = vappvv1) by (eapply value_unique;eauto).
    assert (val_t3  = vappvv3) by (eapply value_unique;eauto).
    rewrite <- H14.
    rewrite <- H15.
    assumption.
  Case "TProd".
    repeat split;
    try (pose proof (eq_typable_empty_l H); eauto); 
    try (pose proof (eq_typable_empty_r H); eauto); 
    try (pose proof (eq_typable_empty_r H0); eauto);
    try (apply (eq_typable_empty_l H1));
    try (apply (eq_typable_empty_r H2)).
    destruct H as [HTt1 [HTt2 [t1L [t1R [t2L [t2R [Hstept1 [Hstept2 [Heq12L Heq12R]]]]]]]]];
    destruct H0 as [_ [HTt3 [t2L' [t2R' [t3L [t3R [Hstept2' [Hstept3 [Heq2'3L Heq2'3R]]]]]]]]].
    pose proof (normalization _ _ H1) as [vt1 [Hsteptv1 Hvalvtv1]].
    assert (\empty |-- vt1 \in TProd T1 T2) as HTvt1.
      eapply preservation_star. eassumption. eapply HTt1.
    pose proof (canonical_forms_pair _ _ _ HTvt1 Hvalvtv1) as [vt1L [vt1R]] ; subst.
    pose proof (normalization _ _ H2) as [vt2 [Hsteptv2 Hvalvtv2]].
    assert (\empty |-- vt2 \in TProd T1 T2) as HTvt2.
      eapply preservation_star. eassumption. eapply HTt2.
    pose proof (canonical_forms_pair _ _ _ HTvt2 Hvalvtv2) as [vt2L [vt2R]]; subst.
    pose proof (normalization _ _ H3) as [vt3 [Hsteptv3 Hvalvtv3]].
    assert (\empty |-- vt3 \in TProd T1 T2) as HTvt3.
      eapply preservation_star. eassumption. eapply HTt3.
    pose proof (canonical_forms_pair _ _ _ HTvt3 Hvalvtv3) as [vt3L [vt3R]]; subst.
    assert (HTvtprod: term_eq (TProd T1 T2) (tpair vt1L vt1R) (tpair vt2L vt2R)).
           eapply eq_square.
           apply H'.
           auto.
           auto.
    assert (HTvtprod': term_eq (TProd T1 T2) (tpair vt2L vt2R) (tpair vt3L vt3R)).
           eapply eq_square.
           apply H0'.
           auto.
           auto.
    pose proof (TPair_Term_eq T1 T2 vt1L vt1R vt2L vt2R). 
    destruct H. 
    pose proof (H0 HTvtprod) as [HTeq12L HTeq12R].
    pose proof (TPair_Term_eq T1 T2 vt2L vt2R vt3L vt3R). 
    destruct H4. 
    pose proof (H5 HTvtprod') as [HTeq23L HTeq23R].
    clear H; clear H0; clear H4; clear H5.
    pose proof (IHT1 _ _ _ HTeq12L HTeq23L) as HTeq13L.
    pose proof (IHT2 _ _ _ HTeq12R HTeq23R) as HTeq13R.
    exists vt1L, vt1R, vt3L, vt3R.
          repeat (split;auto).
  Case "TSum".
    repeat split;
    try (pose proof (eq_typable_empty_l H); eauto); 
    try (pose proof (eq_typable_empty_r H); eauto); 
    try (pose proof (eq_typable_empty_r H0); eauto);
    try (apply (eq_typable_empty_l H1));
    try (apply (eq_typable_empty_r H2)).
    rename t into t1; rename t' into t2; rename t'' into t3;
    destruct H as [HTt1 [HTt2 H_1_2]];
    destruct H0 as [_ [HTt3 H_2_3]].
    pose proof (normalization _ _ H1) as [val_t1 [Hs_val1 Hval_valt1]].
    pose proof (normalization _ _ H2) as [val_t2 [Hs_val2 Hval_valt2]].
    pose proof (normalization _ _ H3) as [val_t3 [Hs_val3 Hval_valt3]].
    assert (HTsum_val12: term_eq (TSum T1 T2) val_t1 val_t2).
           eapply eq_square. 
           apply H'.
           auto.
           auto.
    assert (HTsum_val23: term_eq (TSum T1 T2) val_t2 val_t3).
           eapply eq_square. 
           apply H0'.
           auto.
           auto.
    destruct HTsum_val12 as [_ [_ H_v12]].
    destruct HTsum_val23 as [_ [_ H_v23]].
    clear H1;clear H2;clear H3.
    destruct H_v12.
    SCase "1 and 2 going left".
      destruct H_v23.
      SSCase "2 and 3 going left".
        left.
        destruct H as [tL1 [tL2 [Hstept1 [Hstept2 Heq12]]]].
        destruct H0 as [tL2' [tL3 [Hstept2' [Hstept3 Heq2'3]]]].

        pose proof (normalization _ _ (eq_typable_empty_l Heq12)) as [vtL1 [HvsL1 H_vtL1]].
        pose proof (Congruence_Tinl T2 tL1 vtL1 HvsL1).
        pose proof (rtc_Trans _ _ _ Hstept1 H).
        pose proof (rtc_Trans _ _ _ Hs_val1 H0).

        pose proof (normalization _ _ (eq_typable_empty_r Heq12)) as [vtL2 [HvsL2 H_vtL2]].
        pose proof (Congruence_Tinl T2 tL2 vtL2 HvsL2).
        pose proof (rtc_Trans _ _ _ Hstept2 H2).
        pose proof (rtc_Trans _ _ _ Hs_val2 H3).

        pose proof (normalization _ _ (eq_typable_empty_l Heq2'3)) as [vtL2' [HvsL2' H_vtL2']].
        pose proof (Congruence_Tinl T2 tL2' vtL2' HvsL2').
        pose proof (rtc_Trans _ _ _ Hstept2' H5).
        pose proof (rtc_Trans _ _ _ Hs_val2 H6).
       
        pose proof (normalization _ _ (eq_typable_empty_r Heq2'3)) as [vtL3 [HvsL3 H_vtL3]].
        pose proof (Congruence_Tinl T2 tL3 vtL3 HvsL3).
        pose proof (rtc_Trans _ _ _ Hstept3 H8).
        pose proof (rtc_Trans _ _ _ Hs_val3 H9).
        assert (Hcoq2: (tinl T2 vtL2) = (tinl T2 vtL2')).
               eapply value_unique;eauto.
        inversion Hcoq2;subst.
        assert (He1: term_eq T1 vtL2' vtL3).
               eapply eq_square.
               apply Heq2'3.
               auto.
               auto.
        assert (He1': term_eq T1 vtL1 vtL2').
               eapply eq_square.
               apply Heq12.
               auto.
               auto.
        pose proof (IHT1 _ _ _ He1' He1).
        exists vtL1, vtL3.
               repeat (split;auto).
   SSCase "2 and 3 going right (contra)".
        destruct H as [t1L [t2L [? [? ?]]]].
        destruct H0 as [t2R [t3R [? [? ?]]]].
        assert (HTt1L: \empty |-- tinl T2 t1L \in TSum T1 T2). 
               econstructor.
                  eapply eq_typable_empty_l;eauto.
        assert (HTt2L: \empty |-- tinl T2 t2L \in TSum T1 T2). 
               econstructor.
                  eapply eq_typable_empty_r;eauto.
        assert (HTt2R: \empty |-- tinr T1 t2R \in TSum T1 T2). 
               econstructor.
                  eapply eq_typable_empty_l;eauto.
        assert (HTt3R: \empty |-- tinr T1 t3R \in TSum T1 T2). 
               econstructor.
                  eapply eq_typable_empty_r;eauto.
        pose proof (normalization _ _ HTt1L) as [? [? ?]].
        pose proof (normalization _ _ HTt2L) as [? [? ?]].
        pose proof (normalization _ _ HTt2R) as [? [? ?]].
        pose proof (normalization _ _ HTt3R) as [? [? ?]].
        pose proof (tinl_stays_tinl _ _ _ H5) as [?]. subst x.
        pose proof (tinl_stays_tinl _ _ _ H7) as [?]. subst x0.
        pose proof (tinr_stays_tinr _ _ _ H9) as [?]. subst x1.
        pose proof (tinr_stays_tinr _ _ _ H11) as [?]. subst x2.
        assert (t2 ~~>* tinl T2 x).
               pose proof (rtc_Trans _ _ _ H1 H7). 
               pose proof (rtc_Trans _ _ _ Hs_val2 H13).
               assumption.
        assert (t2 ~~>* tinr T1 x0). 
               pose proof (rtc_Trans _ _ _ H0 H9). 
               pose proof (rtc_Trans _ _ _ Hs_val2 H14).
               assumption.
        pose proof (value_unique _ _ _ H8 H10 H13 H14).
        inversion H15.
 SCase "1 and 2 going right".
      destruct H_v23.
      SSCase "2 and 3 going left (contra)".
        destruct H as [t1R [t2R [? [? ?]]]].
        destruct H0 as [t2L [t3L [? [? ?]]]].
         assert (HTt1R: \empty |-- tinr T1 t1R \in TSum T1 T2). 
               econstructor.
                  eapply eq_typable_empty_l in H2;eauto.
        assert (HTt2R: \empty |-- tinr T1 t2R \in TSum T1 T2). 
               econstructor.
                  eapply eq_typable_empty_r;eauto.
        assert (HTt2L: \empty |-- tinl T2 t2L \in TSum T1 T2). 
               econstructor.
                  eapply eq_typable_empty_l;eauto.
        assert (HTt3L: \empty |-- tinl T2 t3L \in TSum T1 T2). 
               econstructor.
                  eapply eq_typable_empty_r;eauto.
        pose proof (normalization _ _ HTt1R) as [? [? ?]].
        pose proof (normalization _ _ HTt2R) as [? [? ?]].
        pose proof (normalization _ _ HTt2L) as [? [? ?]].
        pose proof (normalization _ _ HTt3L) as [? [? ?]].
        pose proof (tinr_stays_tinr _ _ _ H5) as [?]. subst x.
        pose proof (tinr_stays_tinr _ _ _ H7) as [?]. subst x0.
        pose proof (tinl_stays_tinl _ _ _ H9) as [?]. subst x1.
        pose proof (tinl_stays_tinl _ _ _ H11) as [?]. subst x2.
        assert (t2 ~~>* tinr T1 x).
               pose proof (rtc_Trans _ _ _ H1 H7). 
               pose proof (rtc_Trans _ _ _ Hs_val2 H13).
               assumption.
        assert (t2 ~~>* tinl T2 x0). 
               pose proof (rtc_Trans _ _ _ H0 H9). 
               pose proof (rtc_Trans _ _ _ Hs_val2 H14).
               assumption.
        pose proof (value_unique _ _ _ H8 H10 H13 H14).
        inversion H15.
      SSCase "2 and 3 going right".
        right.
        destruct H as [tR1 [tR2 [Hstept1 [Hstept2 Heq12]]]].
        destruct H0 as [tR2' [tR3 [Hstept2' [Hstept3 Heq2'3]]]].
        pose proof (normalization _ _ (eq_typable_empty_l Heq12)) as [vtR1 [HvsR1 H_vtR1]].
        pose proof (Congruence_Tinr T1 tR1 vtR1 HvsR1).
        pose proof (rtc_Trans _ _ _ Hstept1 H).
        pose proof (rtc_Trans _ _ _ Hs_val1 H0).

        pose proof (normalization _ _ (eq_typable_empty_r Heq12)) as [vtR2 [HvsR2 H_vtR2]].
        pose proof (Congruence_Tinr T1 tR2 vtR2 HvsR2).
        pose proof (rtc_Trans _ _ _ Hstept2 H2).
        pose proof (rtc_Trans _ _ _ Hs_val2 H3).

        pose proof (normalization _ _ (eq_typable_empty_l Heq2'3)) as [vtR2' [HvsR2' H_vtR2']].
        pose proof (Congruence_Tinr T1 tR2' vtR2' HvsR2').
        pose proof (rtc_Trans _ _ _ Hstept2' H5).
        pose proof (rtc_Trans _ _ _ Hs_val2 H6).
       
        pose proof (normalization _ _ (eq_typable_empty_r Heq2'3)) as [vtR3 [HvsR3 H_vtR3]].
        pose proof (Congruence_Tinr T1 tR3 vtR3 HvsR3).
        pose proof (rtc_Trans _ _ _ Hstept3 H8).
        pose proof (rtc_Trans _ _ _ Hs_val3 H9).
        assert (Hcoq2: (tinr T1 vtR2) = (tinr T1 vtR2')).
               eapply value_unique;eauto.
        inversion Hcoq2;subst.
        assert (He1: term_eq T2 vtR2' vtR3).
               eapply eq_square.
               apply Heq2'3.
               auto.
               auto.
        assert (He1': term_eq T2 vtR1 vtR2').
               eapply eq_square.
               apply Heq12.
               auto.
               auto.
        pose proof (IHT2 _ _ _ He1' He1).
        exists vtR1, vtR3.
               repeat (split;auto).
  Case "TNil".
       split.
       apply (eq_typable_empty_l H).
       apply (eq_typable_empty_r H0).
  Case "TMonadic".
       clear H'; clear H0'.
       destruct H as [? [? ?]].
       destruct H0 as [? [? ?]].
       destruct M; intros.
       SCase "MReactT".
             rename t0 into TI; rename t1 into TO; rename s into SM.  
             destruct IHT1 as [IHSto IHT1b].
             destruct IHT1b.
             rename H6 into IHTO; rename H5 into IHTI.
             split. auto. 
             split. auto. 
             intros Sto1 Sto3 HStoeq. 
             pose proof (store_eq_sym _ _ _ HStoeq).
             pose proof (IHSto _ _ _ HStoeq H5).
             pose proof (IHSto _ _ _ H5 HStoeq).
             pose proof (H2 _ _ H6) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? ?]]]]]]]]. 
             pose proof (H4 _ _ HStoeq) as [t2'' [Sto2'' [t3' [Sto3' [? [? [? [? ?]]]]]]]].
             exists t1', Sto1', t3', Sto3'.
                    repeat split; auto. 
             inv H12.
             SSCase "going to return".
             inv H17.
             SSSCase "both to return".
                assert (HT1: (t', Sto1) ~~~>* (treturn (MReactT TI1 TO1 SM1) t1'1, Sto1'1)) by (eauto).
                assert (HT2: (t', Sto1) ~~~>* (treturn (MReactT TI0 TO0 SM0) t2'0, Sto2'0)) by (eauto).
                assert (HT3: done_mo (treturn (MReactT TI0 TO0 SM0) t2'0, Sto2'0)) by (constructor;auto).
                assert (HT4: done_mo ((treturn (MReactT TI1 TO1 SM1) t1'1, Sto1'1))) by (constructor;auto).
                pose proof (done_unique _ _ _ HT3 HT4 HT2 HT1) as HT5.
                inv HT5.
                eapply eq_return; eauto.
             SSSCase "other to pause".
                assert (HT1: (t', Sto1) ~~~>* (tpause (MReactT TI1 TO1 SM1) T0 t1'1, Sto1'1)) by (eauto).
                assert (HT2: (t', Sto1) ~~~>* ((treturn (MReactT TI0 TO0 SM0) t2'0, Sto2'0))) by (eauto).
                assert (HT3: done_mo ((treturn (MReactT TI0 TO0 SM0) t2'0, Sto2'0))) by (constructor;auto).
                assert (HT4: done_mo ((tpause (MReactT TI1 TO1 SM1) T0 t1'1, Sto1'1))) by (constructor;auto).
                pose proof (done_unique _ _ _ HT3 HT4 HT2 HT1) as HT5.
                inv HT5.
             SSCase "going to pause".
             inv H17.
             SSSCase "other to return".
                assert (HT1: (t', Sto1) ~~~>* (tpause (MReactT TI0 TO0 SM0) T0 t2'0, Sto2'0)) by (eauto).
                assert (HT2: (t', Sto1) ~~~>* (treturn (MReactT TI1 TO1 SM1) t1'1, Sto1'1)) by (eauto).
                assert (HT3: done_mo ((tpause (MReactT TI0 TO0 SM0) T0 t2'0, Sto2'0))) by (constructor;auto).
                assert (HT4: done_mo (treturn (MReactT TI1 TO1 SM1) t1'1, Sto1'1)) by (constructor;auto).
                pose proof (done_unique _ _ _ HT3 HT4 HT1 HT2) as HT5.
                inv HT5.
             SSSCase "both to pause".
                assert (HT1: (t', Sto1) ~~~>* (tpause (MReactT TI1 TO1 SM1) T1 t1'1, Sto1'1)) by (eauto).
                assert (HT2: (t', Sto1) ~~~>* (tpause (MReactT TI0 TO0 SM0) T0 t2'0, Sto2'0)) by (eauto).
                assert (HT3: done_mo ((tpause (MReactT TI1 TO1 SM1) T1 t1'1, Sto1'1))) by (constructor;auto).
                assert (HT4: done_mo ((tpause (MReactT TI0 TO0 SM0) T0 t2'0, Sto2'0))) by (constructor;auto).
                pose proof (done_unique _ _ _ HT3 HT4 HT1 HT2) as HT5.
                inv HT5.
                eapply eq_pause. 
                eapply H22.
                eapply H33. 
                auto.
                auto. Focus 5. eapply H30.
                      Focus 5. eapply H45.
                      auto. auto. auto. auto. 
                eapply IHTO. eapply H36. 
                assert (HTO1: done_mo (treturn (MNonReact SM0) (tpair vl0 vr0), Sto1''0)) by (constructor;auto). 
                assert (HTO2: done_mo (treturn (MNonReact SM0) (tpair vl' vr'), Sto2''0)) by (constructor;auto).
                pose proof (done_unique _ _ _ HTO1 HTO2 H44 H31) as HTO3.
                inv HTO3. 
                eapply IHSto. eauto. auto. 
                assert (HTO1: done_mo (treturn (MNonReact SM0) (tpair vl0 vr0), Sto1''0)) by (constructor;auto). 
                assert (HTO2: done_mo (treturn (MNonReact SM0) (tpair vl' vr'), Sto2''0)) by (constructor;auto).
                pose proof (done_unique _ _ _ HTO1 HTO2 H44 H31) as HTO3.
                inv HTO3.
                eapply IHSto. eauto. auto.
             SSSSCase "tapp clause". 
                intros. 
                eapply along2_react_provisional_trans. 
                       auto. auto. auto. auto. apply store_eq_sym. apply term_eq_sym. apply term_eq_sym.
                pose proof (IHTI _ _ _ H18 (term_eq_sym _ _ _ H18)). 
                pose proof (IHSto _ _ _ H19 (store_eq_sym _ _ _ H19)).
                pose proof (H39 _ _ _ _ H20 H21).
                pose proof (H53 _ _ _ _ H18 H19).
                assert (HTO1: done_mo (treturn (MNonReact SM0) (tpair vl0 vr0), Sto1''0)) by (constructor;auto). 
                assert (HTO2: done_mo (treturn (MNonReact SM0) (tpair vl' vr'), Sto2''0)) by (constructor;auto).
                pose proof (done_unique _ _ _ HTO1 HTO2 H44 H31) as HTO3.
                inv HTO3.
                apply (H39 _ _ _ _ H20 H21). 
                assert (HTO1: done_mo (treturn (MNonReact SM0) (tpair vl0 vr0), Sto1''0)) by (constructor;auto). 
                assert (HTO2: done_mo (treturn (MNonReact SM0) (tpair vl' vr'), Sto2''0)) by (constructor;auto).
                pose proof (done_unique _ _ _ HTO1 HTO2 H44 H31) as HTO3.
                inv HTO3.
       SCase "MNonReact".
             rename s into SM.   
             destruct IHT1 as [IHSto IHT1b].
             destruct IHT1b.
             split. auto. 
             split. auto. 
             intros Sto1 Sto3 HStoeq. 
             pose proof (store_eq_sym _ _ _ HStoeq).
             pose proof (IHSto _ _ _ HStoeq H5).
             pose proof (IHSto _ _ _ H5 HStoeq).
             destruct (H2 _ _ H6) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? [? ?]]]]]]]]].
             destruct (H4 _ _ HStoeq) as [t1'' [Sto1'' [t3' [Sto3' [? [? [? [? [? ?]]]]]]]]].
             assert (HD1: done_mo (treturn (MNonReact SM) t1'', Sto1'')) by (constructor;auto).
             assert (HD2: done_mo (treturn (MNonReact SM) t2', Sto2'  )) by (constructor;auto).
             pose proof (done_unique _ _ _ HD1 HD2 H14 H9) as HD3.
             inv HD3. 
             generalize H18 as H18'; intros H18'.
             destruct H18 as [? [? [? [? ?]]]].
             exists t1', Sto1', t3', Sto3'.
                    repeat split; auto.
                    eapply IHSto. eapply H12.   
                           apply store_eq_sym. auto. 
                           eapply store_eq_all_values_l;eauto.
                    pose proof (IHSto _ _ _ H12 H18').
                    destruct H24 as [? [? [? [? ?]]]].
                           auto.
                    eapply IHT2. eapply H13. auto.
    Case "Store_eq MReactT".
       clear H'; clear H0'.
       destruct H0 as [? [? [? [? ?]]]].
       destruct H3 as [? [? [? [? ?]]]].
       simpl. split; auto. split; auto.
              split; auto. split; auto.
              eapply H; eauto.
    Case "Store_eq MNonReact".
       clear H'; clear H0'.
       destruct H0 as [? [? [? [? ?]]]].
       destruct H1 as [? [? [? [? ?]]]].
       simpl. split; auto. split; auto.
              split; auto. split; auto.
              eapply H; eauto.
    Case "sm_eq MIdentity".
       simpl. auto. 
    Case "sm_eq (MStateT T B SM)".
       clear H'; clear H0'. 
       destruct H as [t1 [t2 [Sto1' [Sto2' [? [? [? ?]]]]]]]. 
       destruct H0 as [t3 [t4 [Sto3' [Sto4' [? [? [? ?]]]]]]]. 
       assert (t2 :: Sto2' = t3 :: Sto3').
              rewrite <- H5.
              rewrite -> H3.
              reflexivity.
       inversion H7. 
       simpl. exists t1, t4, Sto1', Sto4'.
              subst.
              repeat split; auto.
              eapply IHT1. eapply H. apply H0.
              eapply IHT2. eapply H1. apply H4.
Qed.

Corollary term_eq_trans : forall T t t' t'',
          term_eq T t t' ->
          term_eq T t' t'' ->
          term_eq T t t''.
Proof.
  intros. eapply eq_trans.
          eapply H.
          auto.
Qed.

Corollary store_eq_trans : forall M Sto1 Sto2 Sto3,
          store_eq M Sto1 Sto2 ->
          store_eq M Sto2 Sto3 ->
          store_eq M Sto1 Sto3.
Proof.
  intros. eapply eq_trans.
          eapply H.
          auto.
Qed.

Corollary sm_eq_trans : forall SM Sto1 Sto2 Sto3, 
          sm_eq SM Sto1 Sto2 -> 
          sm_eq SM Sto2 Sto3 -> 
          sm_eq SM Sto1 Sto3.
Proof.
  intros. eapply eq_trans.
          eapply H.
          auto.
Qed.

(** Given the transitivity of term_eq, store_eq and sm_eq, 
    transitivity of along2_react is easily established. *)

Lemma along2_react_trans : forall M TI TO TR co1 co2 co3,
      along2_react (store_eq M) (term_eq TI) (term_eq TO) (term_eq TR) co1 co2 -> 
      along2_react (store_eq M) (term_eq TI) (term_eq TO) (term_eq TR) co2 co3 ->
      along2_react (store_eq M) (term_eq TI) (term_eq TO) (term_eq TR) co1 co3.
Proof. intros. 
       destruct co1 as (t1,Sto1);
       destruct co2 as (t2,Sto2);
       destruct co3 as (t3,Sto3).
       eapply along2_react_provisional_trans;intros.
              eapply store_eq_trans;eauto.
              eapply term_eq_trans;eauto.
              eapply term_eq_trans;eauto.
              eapply term_eq_trans;eauto.
              eapply store_eq_sym;eauto.
              eapply term_eq_sym;eauto.
              eapply term_eq_sym;eauto.
              eapply H.
              eapply H0.
Defined.

(** *** A relation is [quasi-reflexive] A relation ~ on a set S is called 
        quasi-reflexive if every element that is related to some element 
        is also related to itself, formally: if y∈S, x~y -> x~x ∧ y~y. *)

Theorem along2_react_quasi_refl : forall SM TI TO TR co1 co2,
      along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) co1 co2 -> 
      along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) co1 co1 /\
      along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) co2 co2.
Proof. intros. split.
               eapply along2_react_trans.
               eapply H.
               eapply along2_react_sym.
               auto.
               eapply along2_react_trans.
               eapply along2_react_sym.
               eapply H.
               auto.
Defined.


(* ###################################################################### *)
(** *** Definition and properties pertaining to instantiations for term_eq. This 
        constitutes a natural extension of the approach used to establish the 
        msubst_R lemma in the proof of strong normalization. Thus, although it 
        is not used by the authors of SF, the approach developed here remains 
        faithful to the machinery borrowed from SF. *)

Inductive eq_instantiation :  tass -> env -> env -> Prop :=
  | eq_i_nil  : eq_instantiation nil nil nil
  | eq_i_cons : forall x T t1 t2 c e1 e2,
(*                  value v -> *)
                  term_eq T t1 t2 ->
                  eq_instantiation c e1 e2 -> 
                  eq_instantiation ((x,T)::c) ((x,t1)::e1) ((x,t2)::e2).

Lemma store_eq_matches_l : forall {M} {Sto} {Sto'},
                             store_eq M Sto Sto' -> store_matches_mo Sto M.
Proof.
   intros. destruct M; destruct H as [?[? ?]]; auto.
Qed.

Lemma store_eq_matches_r : forall {M} {Sto} {Sto'},
                             store_eq M Sto Sto' -> store_matches_mo Sto' M.
Proof.
  intros. destruct M; destruct H as [? [? [? ?]]]; auto. 
Qed.

Lemma eq_instantiation_domains_match_l: forall {c} {eL} {eR}, 
      eq_instantiation c eL eR -> 
      forall {x} {T}, lookup x c = Some T -> exists t, lookup x eL = Some t.
Proof.
  intros c eL eR V. induction V; intros x0 T0 C.
    inversion C. 
    simpl in *.
    destruct (eq_id_dec x x0); eauto. 
Qed.

Lemma eq_instantiation_domains_match_r: forall {c} {eL} {eR}, 
      eq_instantiation c eL eR -> 
      forall {x} {T}, lookup x c = Some T -> exists t, lookup x eR = Some t.
Proof.
  intros c eL eR V. induction V; intros x0 T0 C.
    inversion C. 
    simpl in *.
    destruct (eq_id_dec x x0); eauto. 
Qed.

Lemma eq_instantiation_env_closed_l : forall c eL eR,
      eq_instantiation c eL eR -> closed_env eL.
Proof.
  intros c eL eR V; induction V; intros. 
    econstructor. 
    unfold closed_env. fold closed_env. 
    split.  eapply typable_empty_closed. eapply eq_typable_empty_l. eauto.
        auto.
Qed.

Lemma eq_instantiation_env_closed_r : forall c eL eR,
      eq_instantiation c eL eR -> closed_env eR.
Proof.
  intros c eL eR V; induction V; intros. 
    econstructor. 
    unfold closed_env. fold closed_env. 
    split.  eapply typable_empty_closed. eapply eq_typable_empty_r. eauto.
        auto.
Qed.

Lemma eq_instantiation_eq : forall c eL eR, 
      eq_instantiation c eL eR -> 
                        forall x tL tR T, 
      lookup x c = Some T -> 
      lookup x eL = Some tL -> 
      lookup x eR = Some tR -> 
                  term_eq T tL tR.
Proof.
  intros c eL eR V. induction V; intros x' tL' tR' T' G EL ER. 
    inversion EL. inversion ER.
    unfold lookup in *.  destruct (eq_id_dec x x'). 
      inversion G; inversion EL; inversion ER; subst.  auto.
      eauto. 
Qed.

Lemma eq_instantiation_drop : forall c envL envR,
  eq_instantiation c envL envR -> forall x, eq_instantiation (drop x c) (drop x envL) (drop x envR).
Proof.
  intros c eL eR V. induction V.  
    intros.  simpl.  constructor.  
    intros. unfold drop. destruct (eq_id_dec x x0); auto. constructor; eauto. 
Qed.

Lemma msubst_preserves_typing_l : forall c eL eR,
     eq_instantiation c eL eR -> 
     forall Gamma t S, (mextend Gamma c) |-- t \in S -> 
     Gamma |-- (msubst eL t) \in S.
Proof.
    induction 1; intros.  
    simpl in H. simpl. auto. 
    simpl in H1.  simpl. 
    apply IHeq_instantiation.
    eapply subst_preserves_typing; eauto. 
    apply (eq_typable_empty_l H).  
Qed.

Lemma msubst_preserves_typing_r : forall c eL eR,
     eq_instantiation c eL eR -> 
     forall Gamma t S, (mextend Gamma c) |-- t \in S -> 
     Gamma |-- (msubst eR t) \in S.
Proof.
    induction 1; intros.  
    simpl in H. simpl. auto. 
    simpl in H1.  simpl. 
    apply IHeq_instantiation.
    eapply subst_preserves_typing; eauto. 
    apply (eq_typable_empty_r H).  
Qed.

Lemma TCase_Term_eq : forall TL TR TRes t1 t2 t3 t1' t2' t3',
                 term_eq (TSum TL TR) t1 t1' ->
             term_eq (TArrow TL TRes) t2 t2' ->
             term_eq (TArrow TR TRes) t3 t3' -> 
       term_eq TRes (tcase t1 t2 t3) (tcase t1' t2' t3').
Proof with eauto.
  intros TL TR TRes t1 t2 t3 t1' t2' t3' HTSum HTArrL HTArrR.
  assert (HLWT: empty |-- tcase t1 t2 t3 \in TRes) by (
         econstructor;  eapply eq_typable_empty_l;eauto).
  assert (HRWT: \empty |-- tcase t1' t2' t3' \in TRes) by ( 
         econstructor;   eapply eq_typable_empty_r;eauto).
   pose proof (normalization _ _ HLWT) as HLhalts.
    destruct HLhalts as [vtcase [HvtStep Htcaseval]].
   pose proof (normalization _ _ HRWT) as HRhalts.
    destruct HRhalts as [vtcase' [HvtStep' Htcaseval']].
  remember HTSum. clear Heqt. rename t into HTSum'.
  remember HTArrL. clear Heqt. rename t into HTArrL'.
  remember HTArrR. clear Heqt. rename t into HTArrR'.
  simpl in HTSum'. destruct HTSum' as [Ht1WT [Ht1'WT HextCond]].
                  destruct HextCond.
  SCase "Leftval". 
                  destruct H as [tL [tL'[HtL [HtL' HteqLL']]]].
  pose proof (eq_square (TSum TL TR) t1 (tinl TR tL) t1' (tinl TR tL') HTSum HtL HtL') as HLvEq.
   assert (HtlTL:  \empty |-- tL  \in TL) by (eapply eq_typable_empty_l in HteqLL'; eauto).
   assert (Htl'TL: \empty |-- tL' \in TL) by (eapply eq_typable_empty_r in HteqLL'; eauto).
   pose proof (normalization _ _ HtlTL) as HtLhalts.
   destruct HtLhalts as [vtL [HtLStep HtLval]].
   pose proof (normalization _ _ Htl'TL) as HtL'halts.
   destruct HtL'halts as [vtL' [HtL'Step HtL'val]].
   assert (HvtlTL : \empty |-- vtL \in TL) by (pose proof (preservation_star _ _ _ HtLStep HtlTL);eauto).
   assert (Hvtl'TL : \empty |-- vtL' \in TL) by (pose proof (preservation_star _ _ _ HtL'Step Htl'TL) ;eauto).
   assert (Hvtleqvtl': term_eq TL vtL vtL') by (eapply eq_square;eauto).
   destruct HTArrL' as [Ht2WT [Ht2'WT HteqTarrL]]. 
   pose proof (HteqTarrL vtL vtL' Hvtleqvtl').
   destruct H as [x [x0 [HtappX [HtappX0 HtermeqTRes]]]]. 
   assert (tcase t1 t2 t3 ~~>* tcase (tinl TR tL) t2 t3) by (pose proof (Congruence_Tcase t1 (tinl TR tL) t2 t3 HtL); eauto).
   assert (tcase (tinl TR tL) t2 t3 ~~>* tcase (tinl TR vtL) t2 t3).
      eapply Congruence_Tinl in HtLStep; eapply Congruence_Tcase;eauto.
   assert (tcase t1 t2 t3 ~~>* tcase (tinl TR vtL) t2 t3).
      eapply rtc_Trans;eauto.
   assert (tcase (tinl TR vtL) t2 t3 ~~>* tapp t2 vtL) by (eapply Congruence_ST_CaseL; eauto).
   assert (tcase t1 t2 t3 ~~>* tapp t2 vtL) by (eapply rtc_Trans;eauto).
   assert (tcase t1 t2 t3 ~~>* x) by (eapply rtc_Trans;eauto).
   assert (tcase t1' t2' t3' ~~>* tcase (tinl TR tL') t2' t3') by (pose proof (Congruence_Tcase t1' (tinl TR tL') t2' t3' HtL'); eauto).
   assert (tcase (tinl TR tL') t2' t3' ~~>* tcase (tinl TR vtL') t2' t3').
      eapply Congruence_Tinl in HtL'Step; eapply Congruence_Tcase;eauto.
   assert (tcase t1' t2' t3' ~~>* tcase (tinl TR vtL') t2' t3').
      eapply rtc_Trans;eauto.
   assert (tcase (tinl TR vtL') t2' t3' ~~>* tapp t2' vtL') by (eapply Congruence_ST_CaseL; eauto).
   assert (tcase t1' t2' t3' ~~>* tapp t2' vtL') by (eapply rtc_Trans;eauto).
   assert (tcase t1' t2' t3' ~~>* x0) by (eapply rtc_Trans;eauto).
   eapply eq_square_rev; eauto.
SCase "Rightval". 
  destruct H as [tR [tR'[HtR [HtR' HteqRR']]]].
   pose proof (eq_square (TSum TL TR) t1 (tinr TL tR) t1' (tinr TL tR') HTSum HtR HtR') as HRvEq.
   assert (HtrTR:  \empty |-- tR  \in TR) by (eapply eq_typable_empty_l in HteqRR'; eauto).
   assert (Htr'TR: \empty |-- tR' \in TR) by (eapply eq_typable_empty_r in HteqRR'; eauto).
   pose proof (normalization _ _ HtrTR) as HtRhalts.
   destruct HtRhalts as [vtR [HtRStep HtRval]].
   pose proof (normalization _ _ Htr'TR) as HtR'halts.
   destruct HtR'halts as [vtR' [HtR'Step HtR'val]].
   assert (HvtRTR : \empty |-- vtR \in TR) by (pose proof (preservation_star _ _ _ HtRStep HtrTR) ;eauto).
   assert (HvtR'TR : \empty |-- vtR' \in TR) by (pose proof (preservation_star _ _ _ HtR'Step Htr'TR) ;eauto).
   assert (HvtReqvtR': term_eq TR vtR vtR') by (eapply eq_square;eauto).
   destruct HTArrR' as [Ht2WT [Ht2'WT HteqTarrR]]. 
   pose proof (HteqTarrR vtR vtR' HvtReqvtR').
   destruct H as [x [x0 [HtappX [HtappX0 HtermeqTRes]]]]. 
   assert (tcase t1 t2 t3 ~~>* tcase (tinr TL tR) t2 t3) by (pose proof (Congruence_Tcase t1 (tinr TL tR) t2 t3 HtR); eauto).
   assert (tcase (tinr TL tR) t2 t3 ~~>* tcase (tinr TL vtR) t2 t3).
      pose proof (Congruence_Tinr TL _ _ HtRStep). eapply Congruence_Tcase;eauto.
   assert (tcase t1 t2 t3 ~~>* tcase (tinr TL vtR) t2 t3).
      eapply rtc_Trans;eauto.
   assert (tcase (tinr TL vtR) t2 t3 ~~>* tapp t3 vtR) by (eapply Congruence_ST_CaseR; eauto).
   assert (tcase t1 t2 t3 ~~>* tapp t3 vtR) by (eapply rtc_Trans;eauto).
   assert (tcase t1 t2 t3 ~~>* x) by (eapply rtc_Trans;eauto).
   assert (tcase t1' t2' t3' ~~>* tcase (tinr TL tR') t2' t3') by (pose proof (Congruence_Tcase t1' (tinr TL tR') t2' t3' HtR'); eauto).
   assert (tcase (tinr TL tR') t2' t3' ~~>* tcase (tinr TL vtR') t2' t3').
      eapply Congruence_Tinr in HtR'Step; eapply Congruence_Tcase;eauto.
   assert (tcase t1' t2' t3' ~~>* tcase (tinr TL vtR') t2' t3').
      eapply rtc_Trans;eauto.
   assert (tcase (tinr TL vtR') t2' t3' ~~>* tapp t3' vtR') by (eapply Congruence_ST_CaseR; eauto).
   assert (tcase t1' t2' t3' ~~>* tapp t3' vtR') by (eapply rtc_Trans;eauto).
   assert (tcase t1' t2' t3' ~~>* x0) by (eapply rtc_Trans;eauto).
   eapply eq_square_rev; eauto.
Qed.

Lemma TApp_Term_eq : forall T1 T2 t1 t1' t2 t2',
      term_eq (TArrow T1 T2) t1 t1' ->
      term_eq T1 t2 t2' ->
      term_eq T2 (tapp t1 t2) (tapp t1' t2').
Proof.
intros T1 T2 t1 t1' t2 t2' HeqTArr12 HeqT1t2t2'. 
assert (HTappWT1: \empty |-- tapp t1 t2 \in T2).
       econstructor. eapply eq_typable_empty_l; eauto.
                     eapply eq_typable_empty_l; eauto.
assert (HTappWT2: \empty |-- tapp t1' t2' \in T2).
       econstructor. eapply eq_typable_empty_r; eauto.
                     eapply eq_typable_empty_r; eauto.
 destruct (normalization _ _ HTappWT1) as [tv1 [HStep1 Htv1]]. 
 destruct (normalization _ _ HTappWT2) as [tv2 [HStep2 Htv2]].
 pose proof (preservation_star _ _ _ HStep1 HTappWT1) as Htv1WT.
 pose proof (preservation_star _ _ _ HStep2 HTappWT2) as Htv2WT.
 simpl in HeqTArr12.
  assert (Ht2WT1: \empty |-- t2 \in T1) by (eapply eq_typable_empty_l;eauto). 
  assert (Ht2'WT1:\empty |-- t2' \in T1) by (eapply eq_typable_empty_r;eauto).
 destruct HeqTArr12. destruct H0.
 pose proof (H1 t2 t2' HeqT1t2t2').
 destruct H2 as [t12 [t1'2' [HTappt1t2 [HTappt1't2' Htermeq]]]].
 eapply eq_square_rev;eauto.
Qed.

Lemma TReturn_Term_eq : forall T Mo t1 t2,
      term_eq T t1 t2 ->
      term_eq (TMonadic Mo T) (treturn Mo t1) (treturn Mo t2).
Proof.
 intros T Mo t1 t2 Htermeq.
  assert (\empty |-- treturn Mo t1 \in TMonadic Mo T).
         assert (\empty |-- t1 \in T) by (eapply eq_typable_empty_l;eauto).
         eapply T_Return;eauto.
  assert (\empty |-- treturn Mo t2 \in TMonadic Mo T).
         assert (\empty |-- t2 \in T) by (eapply eq_typable_empty_r;eauto).
         eapply T_Return;eauto. 
       destruct (normalization _ _ (eq_typable_empty_l Htermeq)) as [v1 [? ?]].
       destruct (normalization _ _ (eq_typable_empty_r Htermeq)) as [v2 [? ?]].
  destruct Mo.
  Case "MReactT".
       split. auto.
       split. auto.
       intros. simpl in H5.
       pose proof (Congruence_Treturn _ _ (MReactT t t0 s) Sto1 H1).
       pose proof (Congruence_Treturn _ _ (MReactT t t0 s) Sto2 H3).
       exists (treturn (MReactT t t0 s) v1), Sto1, (treturn (MReactT t t0 s) v2), Sto2. 
               repeat (split;eauto).
               eapply eq_return; eauto. 
               eapply eq_square; eauto.
  Case "MNonReact".
       split. auto.
       split. auto.
       intros.
       pose proof (Congruence_Treturn _ _ (MNonReact s) Sto1 H1).
       pose proof (Congruence_Treturn _ _ (MNonReact s) Sto2 H3).
       exists v1, Sto1, v2, Sto2. 
               repeat (split;eauto).
               eapply eq_square; eauto.
Qed.

(* ###################################################################### *)
(** *** The following lemmas arise out of an unavoidable use of the COFIX tactic, that
        is required in order establish that term_eq behaves correctly w.r.t. 
        the cases of our tbind and tunfold constructors.  *)

Lemma tbind_along2 : forall t1 Sto1 t2 Sto2 T1 T2 TI TO SM t3 t4 Sto1' t3' t4' Sto2',
       value t3 -> value t4 -> value t3' -> value t4' ->
       (t1,Sto1) ~~~>* (tbind T2 t3  t4 ,Sto1')                    ->
       (t2,Sto2) ~~~>* (tbind T2 t3' t4',Sto2')                    ->
       term_eq (TMonadic (MReactT TI TO SM) T1) t3 t3'             ->
       term_eq (TArrow T1 (TMonadic (MReactT TI TO SM) T2)) t4 t4' ->
                          store_eq  (MReactT TI TO SM) Sto1' Sto2' ->
       along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq T2) (t1,Sto1) (t2,Sto2).
Proof. cofix;intros. 
       generalize H5 as H5'; intro H5'.
       generalize H6 as H6'; intro H6'.
       destruct H5' as [? [? ?]]. 
       destruct H6' as [? [? ?]]. 
       destruct (H10 _ _ H7) as [t1' [Sto1'0 [t2' [Sto2'0 [? [? [? [? ?]]]]]]]].
       clear H10.
       inversion H18; subst. 
 Case "going to return".
       assert (TI = TI0 /\ TO = TO0 /\ SM = SM0).
                assert ((t1',Sto1'0) |> TMonadic (MReactT TI TO SM) T1). 
                       eapply preservation_mo_star;eauto. 
                       constructor. auto. destruct H7 as [? [? ?]]. auto.
                       eapply store_eq_matches_l; eauto.
                pose proof (preservation_mo_star _ _ _ _ _ H22 H10).
                inversion H19. inversion H26. auto. 
       destruct H10 as [HRe1 [HRe2 HRe3]]. 
             rewrite <- HRe1 in *; clear HRe1;
             rewrite <- HRe2 in *; clear HRe2;
             rewrite <- HRe3 in *; clear HRe3. 
       assert ((t1, Sto1) ~~~>* (tapp t4 t1'0, Sto1'1)).
              eapply rtc_Trans_mo. 
                eapply H3. 
              eapply rtc_Trans_mo.
              eapply Congruence_Mo_Bind1;eauto.
              eapply rtc_Trans_mo.
              eapply Congruence_Mo_Bind1;eauto.
              eapply Congruence_BindRet;eauto. 
       assert ((t2, Sto2) ~~~>* (tapp t4' t2'0, Sto2'1)).
              eapply rtc_Trans_mo. 
                eapply H4. 
              eapply rtc_Trans_mo.
              eapply Congruence_Mo_Bind1;eauto.
              eapply rtc_Trans_mo.
              eapply Congruence_Mo_Bind1;eauto.
              eapply Congruence_BindRet;eauto. 
       destruct (H13 _ _ H30) as [t5 [t5' [? [? ?]]]].
       destruct H24 as [? [? ?]]. 
       destruct (H26 _ _ H31) as [? [? [? [? [? [? [? [? ?]]]]]]]].
       eapply step_preserves_along2_react_star. 
              eapply H10.
              eapply H19.
       eapply step_preserves_along2_react_star. 
              eapply Congruence_LC. eapply H20.
              eapply Congruence_LC. eapply H21.
       eapply step_preserves_along2_react_star. 
              eapply H27.
              eapply H32.
       auto.
 Case "going to pause". 
       assert (TI = TI0 /\ TO = TO0 /\ SM = SM0 /\ T = T1).
              assert ((t1',Sto1'0) |> TMonadic (MReactT TI TO SM) T1). 
                     eapply preservation_mo_star;eauto. 
                     constructor. auto. destruct H7 as [? [? ?]]. auto.
                     eapply store_eq_matches_l;eauto.
                pose proof (preservation_mo_star _ _ _ _ _ H22 H10).
                inversion H19. inversion H34. auto. 
       destruct H10 as [HRe1 [HRe2 [HRe3 HRe4]]]. 
             rewrite <- HRe1 in *; clear HRe1;
             rewrite <- HRe2 in *; clear HRe2;
             rewrite <- HRe3 in *; clear HRe3;
             rewrite <- HRe4 in *; clear HRe4.
    assert (HG1:
     (t1'0 >>=[TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T2))] 
     (\(0: TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))) 
           tproj (tvar 0) (\(1:TO) (\(2: TArrow TI (TMonadic (MReactT TI TO SM) T)) 
       treturn (MNonReact SM) (tpair (tvar 1) (\(3:TI) (tapp (tvar 2) (tvar 3) >>=[T2] t4)))))),Sto1'1) ~~~>*
      (treturn (MNonReact SM) (tpair vl (\(3:TI) (tapp vr (tvar 3) >>=[T2] t4))),Sto1'')). 
            eapply rtc_Trans_mo. 
              eapply Congruence_Mo_Bind1. 
                  constructor. 
                  eapply H30. 
       eapply rtc_Trans_mo. 
                eapply Congruence_BindRet. 
                  constructor;auto.
                  constructor. 
       eapply rtc_Trans_mo.
              eapply Congruence_AppAbs_Mo.
                constructor;auto. 
                simpl. 
                rewrite_subst. 
       eapply rtc_Trans_mo.
              eapply Congruence_LC.
              eapply rtc_R.
              eapply ST_Proj;eauto. 
       eapply rtc_Trans_mo.
              eapply Congruence_LC.
              eapply Congruence_App1.
              eapply rtc_R.
              eapply ST_AppAbs;eauto. 
              simpl. 
               
              rewrite_subst.
       eapply rtc_Trans_mo.
              eapply Congruence_LC.
              eapply rtc_R.
              eapply ST_AppAbs;eauto. 
              simpl.
              assert (\empty |-- vl \in TO) by (eapply eq_typable_empty_l;eauto).
              rewrite_subst. 
          eapply rtc_refl.
   assert (HG2:
     (t2'0 >>=[TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T2))] 
     (\(0: TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))) 
           tproj (tvar 0) (\(1:TO) (\(2: TArrow TI (TMonadic (MReactT TI TO SM) T)) 
       treturn (MNonReact SM) (tpair (tvar 1) (\(3:TI) (tapp (tvar 2) (tvar 3) >>=[T2] t4')))))),Sto2'1) ~~~>*
      (treturn (MNonReact SM) (tpair vl' (\(3:TI) (tapp vr' (tvar 3) >>=[T2] t4'))),Sto2'')). 
            eapply rtc_Trans_mo. 
              eapply Congruence_Mo_Bind1. 
                  constructor. 
                  eapply H31. 
       eapply rtc_Trans_mo. 
                eapply Congruence_BindRet. 
                  constructor;auto.
                  constructor. 
       eapply rtc_Trans_mo.
              eapply Congruence_AppAbs_Mo.
                constructor;auto. 
                simpl.
                 
                rewrite_subst. 
       eapply rtc_Trans_mo.
              eapply Congruence_LC.
              eapply rtc_R.
              eapply ST_Proj;eauto. 
       eapply rtc_Trans_mo.
              eapply Congruence_LC.
              eapply Congruence_App1.
              eapply rtc_R.
              eapply ST_AppAbs;eauto. 
              simpl. 
               
              rewrite_subst.
       eapply rtc_Trans_mo.
              eapply Congruence_LC.
              eapply rtc_R.
              eapply ST_AppAbs;eauto.
              simpl.
               
              assert (\empty |-- vl' \in TO) by (eapply eq_typable_empty_r;eauto).
              rewrite_subst. 
          eapply rtc_refl. 
       assert (HTt1: \empty |-- t1'0 \in TMonadic (MNonReact SM) (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
       assert ((t1',Sto1'0) |> TMonadic (MReactT TI TO SM) T). 
              eapply preservation_mo_star. eapply H14.
              constructor. auto. destruct H7 as [? [? ?]]. auto. 
              eapply store_eq_matches_l;eauto.
                pose proof (preservation_mo_star _ _ _ _ _ H22 H10).
                inversion H19. inversion H34. auto. 
       assert ((treturn (MNonReact SM) (tpair vl vr), Sto1'') |> TMonadic (MNonReact SM) (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
              eapply preservation_mo_star. eapply H30.
              constructor;auto. destruct H37 as [? [? [? [? ?]]]].
                                auto.
                                destruct H37 as [? [? [? [? ?]]]].
                                inv H10.
       assert (HTt2: \empty |-- t2'0 \in TMonadic (MNonReact SM) (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
       assert ((t2', Sto2'0) |> TMonadic (MReactT TI TO SM) T). 
              eapply preservation_mo_star. eapply H15.
              constructor. auto. destruct H7 as [? [? [? [? ?]]]]. auto. 
                                 destruct H7 as [? [? [? [? ?]]]]. auto.
                pose proof (preservation_mo_star _ _ _ _ _ H23 H19).
                inversion H20. inversion H35. auto.
       assert ((treturn (MNonReact SM) (tpair vl' vr'), Sto2'') |> TMonadic (MNonReact SM) (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
              eapply preservation_mo_star. eapply H31.
              constructor;auto.  destruct H37 as [? [? [? [? ?]]]]. auto. 
                                 destruct H37 as [? [? [? [? ?]]]]. auto. inv H21.
       assert (HTvr: \empty |-- vr \in (TArrow TI (TMonadic (MReactT TI TO SM) T))).
              inversion H10. inversion H34. inversion H43. auto.
       assert (HTvr': \empty |-- vr' \in (TArrow TI (TMonadic (MReactT TI TO SM) T))).
              inversion H19. inversion H34. inversion H43. auto.
       eapply eq_pause. 
       eapply rtc_Trans_mo. 
              eapply H3.
              eapply rtc_Trans_mo. 
              eapply Congruence_Mo_Bind1.
              auto. eapply H14.  
              eapply rtc_Trans_mo. 
              eapply Congruence_Mo_Bind1.
              auto. eapply H22.
              eapply Congruence_STM_PauseBind. auto. auto.  
       eapply rtc_Trans_mo. 
              eapply H4.
              eapply rtc_Trans_mo. 
              eapply Congruence_Mo_Bind1.
              auto. eapply H15. 
              eapply rtc_Trans_mo. 
              eapply Congruence_Mo_Bind1.
              auto. eapply H23.
              eapply Congruence_STM_PauseBind. auto. auto.  
              constructor. 
              auto.
              auto. 
              constructor.
              auto. 
              constructor.
              eapply H26.
              eapply H27. 
               instantiate (1:= (\(3:TI) (tapp vr (tvar 3) >>=[T2] t4))); constructor.
              instantiate (1:= (\(3:TI) (tapp vr' (tvar 3) >>=[T2] t4'))); constructor. 
      eapply HG1. 
      eapply HG2. 
      auto. 
      auto.
      auto.
      intros. 
       destruct (normalization _ _ (eq_typable_empty_l H20)) as [tv1 [HStep1 Htv1]]. 
       destruct (normalization _ _ (eq_typable_empty_r H20)) as [tv2 [HStep2 Htv2]].
       assert (term_eq TI tv1 tv2) by (eapply eq_square;eauto).
       pose proof (H39 _ _ _ _ H32 H21).
      assert (HTapp1: \empty |-- tapp vr tv1 \in (TMonadic (MReactT TI TO SM) T)).
             econstructor. eauto. eapply eq_typable_empty_l;eauto.
      assert (HTapp2: \empty |-- tapp vr' tv2 \in (TMonadic (MReactT TI TO SM) T)).
             econstructor. eauto. eapply eq_typable_empty_r;eauto.
      destruct (normalization _ _ HTapp1) as [vf1 [? ?]]. 
      destruct (normalization _ _ HTapp2) as [vf2 [? ?]].
      assert ((tapp (\(3:TI) (tapp vr (tvar 3) >>=[T2] t4)) t0, Sto3) ~~~>* (vf1 >>=[T2] t4, Sto3)).
              eapply rtc_Trans_mo. eapply Congruence_LC. eapply Congruence_App2. constructor. eapply HStep1.
              eapply rtc_Trans_mo. eapply Congruence_AppAbs_Mo. auto. simpl.
               
              rewrite_subst. eapply Congruence_Mo_Bind1. auto.
                             eapply Congruence_LC. auto.
      assert ((tapp (\(3:TI) (tapp vr' (tvar 3) >>=[T2] t4')) t3'0, Sto3') ~~~>* (vf2 >>=[T2] t4', Sto3')).
              eapply rtc_Trans_mo. eapply Congruence_LC. eapply Congruence_App2. constructor. eapply HStep2.
              eapply rtc_Trans_mo. eapply Congruence_AppAbs_Mo. auto. simpl.
               
              rewrite_subst. eapply Congruence_Mo_Bind1. auto.
                             eapply Congruence_LC. auto.
       apply (tbind_along2 
             (tapp (\(3:TI) (tapp vr (tvar 3) >>=[T2] t4)) t0) Sto3
             (tapp (\(3:TI) (tapp vr' (tvar 3) >>=[T2] t4')) t3'0) Sto3' 
             T T2 TI TO SM vf1 t4 Sto3 vf2 t4' Sto3').
        auto. auto. auto. auto. auto. auto.  
        apply (eq_square (TMonadic (MReactT TI TO SM) T) (tapp vr tv1) vf1 (tapp vr' tv2) vf2).
        split. auto.
        split. auto.
        intros. pose proof (H39 _ _ _ _ H32 H44).
                exists vf1, Sto0, vf2, Sto4.
                       repeat split.
                       eapply Congruence_LC;auto.
                       eapply Congruence_LC;auto.
                       auto.
                       auto.
                       eapply step_preserves_along2_react_star'.
                       eapply Congruence_LC. eapply H34.
                       eapply Congruence_LC. eapply H40.
                       auto. auto. auto. auto. auto.
Qed.


Lemma tunfold_along2 : forall t1 Sto1 t2 Sto2 TA TB TI TO SM t3 t4 Sto1' t3' t4' Sto2',
                         ((t1,Sto1) ~~~>* (tunfold (MReactT TI TO SM) TA TB t3 t4,Sto1')) ->
                         ((t2,Sto2) ~~~>* (tunfold (MReactT TI TO SM) TA TB t3' t4',Sto2')) ->
                        term_eq TB t3 t3' ->
                        term_eq (TArrow TB (TMonadic (MNonReact SM) (TSum TA (TProd TO (TArrow TI TB))))) t4 t4' ->
                        store_eq  (MReactT TI TO SM) Sto1' Sto2' ->
                        along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TA) (t1,Sto1) (t2,Sto2).
Proof.
 cofix; intros t1 Sto1 t2 Sto2 TA TB TI TO SM t3 t4 Sto1' t3' t4' Sto2' Hstep1 Hstep2 HeqTB HeqTBArrow HStoeq.
 destruct (normalization _ _ (eq_typable_empty_l HeqTB))      as [v3  [? ?]].
 destruct (normalization _ _ (eq_typable_empty_r HeqTB))      as [v3' [? ?]].
 destruct (normalization _ _ (eq_typable_empty_l HeqTBArrow)) as [v4  [? ?]].
 destruct (normalization _ _ (eq_typable_empty_r HeqTBArrow)) as [v4' [? ?]].
 pose proof (rtc_Trans_mo _ _ _ Hstep1
            (rtc_Trans_mo _ _ _
            (Congruence_LC _ _ Sto1'
            (rtc_Trans _ _ _ (Congruence_Unfold1 TI TO SM TA TB _ _ t4 H)
                             (Congruence_Unfold2 TI TO SM TA TB v3 _ _ H0 H3)))
            (Congruence_STM_Unfold v3 v4 Sto1' TI TO SM TA TB H0 H4))).
 pose proof (rtc_Trans_mo _ _ _ Hstep2            
            (rtc_Trans_mo _ _ _
            (Congruence_LC _ _ Sto2'
            (rtc_Trans _ _ _ (Congruence_Unfold1 TI TO SM TA TB _ _ t4' H1)
                             (Congruence_Unfold2 TI TO SM TA TB v3' _ _ H2 H5)))
            (Congruence_STM_Unfold v3' v4' Sto2' TI TO SM TA TB H2 H6))).            
 assert (HeqTB' : term_eq TB v3 v3') by (eapply eq_square;eauto).
 assert (HeqTBArrow': term_eq (TArrow TB
                               (TMonadic (MNonReact SM)
                                (TSum TA (TProd TO (TArrow TI TB))))) v4 v4') by (eapply eq_square;eauto).
 generalize HeqTBArrow'; intro HeqFinal.
 destruct HeqTBArrow' as [? [? ?]].
 destruct (H11 _ _ HeqTB') as [t5 [t5' [? [? ?]]]].
 assert (HStoeq': store_eq (MNonReact SM) Sto1' Sto2').
        destruct HStoeq as [? [? [? [? ?]]]].
        split. inv H15.
        split. inv H16.
        split. inv H17. 
        split. inv H18.
        auto.
 destruct (normalization _ _ (eq_typable_empty_l H14)) as [v5  [HStep5  ?]].
 pose proof (rtc_Trans _ _ _ H12 HStep5).       
 destruct (normalization _ _ (eq_typable_empty_r H14)) as [v5' [HStep5' ?]].
 pose proof (rtc_Trans _ _ _ H13 HStep5').       
 assert (H14': term_eq (TMonadic (MNonReact SM) (TSum TA (TProd TO (TArrow TI TB)))) v5 v5') by (eapply eq_square;eauto).
 destruct H14' as [? [? ?]].
 destruct (H21 _ _ HStoeq') as [v6 [Sto6 [v6' [Sto6' [? [? [? [? [? ?]]]]]]]]].
 pose proof (rtc_Trans_mo _ _ _
            (Congruence_LC _ _ Sto1' (Congruence_Lift (MReactT TI TO SM) _ _ H16))
            (rtc_Trans_mo _ _ _ (Congruence_LiftRe _ _ _ _ TI TO SM H15 H22)
                                (Congruence_LiftRetRe _ TI TO SM Sto6 H24))).
 pose proof (rtc_Trans_mo _ _ _
            (Congruence_LC _ _ Sto2' (Congruence_Lift (MReactT TI TO SM) _ _ H18))
            (rtc_Trans_mo _ _ _ (Congruence_LiftRe _ _ _ _ TI TO SM H17 H23)
                                (Congruence_LiftRetRe _ TI TO SM Sto6' H25))).
 assert (HV1: value (tabs 0 (TSum TA (TProd TO (TArrow TI TB)))
                                (tcase (tvar 0)
                                   (tabs 1 TA (treturn (MReactT TI TO SM) (tvar 1)))
                                   (tabs 1 (TProd TO (TArrow TI TB))
                                      (tproj
                                         (tvar 1)
                                         (tabs 2 TO
                                           (tabs 3 (TArrow TI TB)
                                             (tpause (MReactT TI TO SM) TA
                                                  (treturn (MNonReact SM)
                                                    (tpair (tvar 2) 
                                                           (tabs 4 TI
                                                             (tunfold (MReactT TI TO SM) TA TB
                                                               (tapp (tvar 3) (tvar 4))
                                                               v4)))))))))))) by (constructor).
 (* assert (HV1': value ((treturn (MReactT TI TO SM) v6))) by (constructor;auto).*)
 assert (HV2: value (tabs 0 (TSum TA (TProd TO (TArrow TI TB)))
                                (tcase (tvar 0)
                                   (tabs 1 TA (treturn (MReactT TI TO SM) (tvar 1)))
                                   (tabs 1 (TProd TO (TArrow TI TB))
                                      (tproj
                                         (tvar 1)
                                         (tabs 2 TO
                                           (tabs 3 (TArrow TI TB)
                                             (tpause (MReactT TI TO SM) TA
                                                  (treturn (MNonReact SM)
                                                    (tpair (tvar 2) 
                                                           (tabs 4 TI
                                                             (tunfold (MReactT TI TO SM) TA TB
                                                               (tapp (tvar 3) (tvar 4))
                                                               v4')))))))))))) by (constructor).
(* assert (HV2': value ((treturn (MReactT TI TO SM) v6'))) by (constructor;auto).*)
 pose proof (rtc_Trans_mo _ _ _
            (rtc_Trans_mo _ _ _ H7
            (rtc_Trans_mo _ _ _ 
            (Congruence_Mo_Bind1 TA _ _ _ _ _ HV1 H28)
            (Congruence_BindRet TA _ _ (MReactT TI TO SM) Sto6 H24 HV1)))
            (Congruence_AppAbs_Mo 0 (TSum TA (TProd TO (TArrow TI TB))) _ _ Sto6 H24)).
 simpl in H30.
       
      rewrite_subst.
 pose proof (rtc_Trans_mo _ _ _
            (rtc_Trans_mo _ _ _ H8
            (rtc_Trans_mo _ _ _ 
            (Congruence_Mo_Bind1 TA _ _ _ _ _ HV2 H29)
            (Congruence_BindRet TA  _ _ (MReactT TI TO SM) Sto6' H25 HV2)))
            (Congruence_AppAbs_Mo 0 (TSum TA (TProd TO (TArrow TI TB))) _ _ Sto6' H25)).
 simpl in H31.
        
       rewrite_subst.
 Case "destructing term_eq (TSum TA (TProd TO (TArrow TI TB))) v6 v6'".
      destruct H27 as [? [? ?]].
      destruct H33 as [HSUM | HSUM].
  SCase "left".
        destruct HSUM as [t7 [t7' [HS6 [HS6' HeqTA]]]].
        destruct (normalization _ _ (eq_typable_empty_l HeqTA)) as [v7  [HS7  ?]].
        destruct (normalization _ _ (eq_typable_empty_r HeqTA)) as [v7' [HS7' ?]].
        pose proof (rtc_Trans _ _ _ HS6
                   (Congruence_Tinl (TProd TO (TArrow TI TB)) _ _ HS7)) as HS8.
        pose proof (rtc_Trans _ _ _ HS6'
                   (Congruence_Tinl (TProd TO (TArrow TI TB)) _ _ HS7')) as HS8'.
        pose proof (rtc_Trans_mo _ _ _ H30
                   (Congruence_LC _ _ Sto6
                   (rtc_Trans _ _ _
                   (rtc_Trans _ _ _
                   (Congruence_Tcase _ _
                                     (\(1:TA) treturn (MReactT TI TO SM) (tvar 1))
         (\(1:TProd TO (TArrow TI TB)) tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4))))))
                                     HS8)
                   (Congruence_ST_CaseL (TProd TO (TArrow TI TB)) _
                                        (\(1:TA) treturn (MReactT TI TO SM) (tvar 1))
         (\(1:TProd TO (TArrow TI TB)) tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4))))))
                                        H33))
                  (rtc_R _ _ (ST_AppAbs 1 TA (treturn (MReactT TI TO SM) (tvar 1)) _ H33))))).
        simpl in H35.  
        pose proof (rtc_Trans_mo _ _ _ H31
                   (Congruence_LC _ _ Sto6'
                   (rtc_Trans _ _ _
                   (rtc_Trans _ _ _
                   (Congruence_Tcase _ _
                                     (\(1:TA) treturn (MReactT TI TO SM) (tvar 1))
         (\(1:TProd TO (TArrow TI TB)) tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4'))))))
                                     HS8')
                   (Congruence_ST_CaseL (TProd TO (TArrow TI TB)) _
                                        (\(1:TA) treturn (MReactT TI TO SM) (tvar 1))
         (\(1:TProd TO (TArrow TI TB)) tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4'))))))
                                        H34))
                  (rtc_R _ _ (ST_AppAbs 1 TA (treturn (MReactT TI TO SM) (tvar 1)) _ H34))))).
        simpl in H36.  
        assert (HeqTA': term_eq TA v7 v7') by (eapply eq_square;eauto).
        destruct (TReturn_Term_eq TA (MReactT TI TO SM) _ _ HeqTA') as [? [? ?]].
        destruct (H39 _ _ (store_eq_non_react _ _ TI TO SM H26)) as [v8 [Sto8 [v8' [Sto8' [? [? [? [? ?]]]]]]]].
        pose proof (rtc_Trans_mo _ _ _ H35 H40).
        pose proof (rtc_Trans_mo _ _ _ H36 H41).
        apply (step_preserves_along2_react_star _ _ _ _ _ _ _ _ _ _ _ _ H45 H46 H44).  
  SCase "right". 
        destruct HSUM as [t7 [t7' [HS6 [HS6' HeqTProd]]]].
        destruct (normalization _ _ (eq_typable_empty_l HeqTProd)) as [v7  [HS7  ?]].
        destruct (normalization _ _ (eq_typable_empty_r HeqTProd)) as [v7' [HS7' ?]].
        pose proof (rtc_Trans _ _ _ HS6
                   (Congruence_Tinr TA _ _ HS7)) as HS8.
        pose proof (rtc_Trans _ _ _ HS6'
                   (Congruence_Tinr TA _ _ HS7')) as HS8'.
        assert (HeqTProd': term_eq (TProd TO (TArrow TI TB)) v7 v7') by (eapply eq_square;eauto).
   (*   pose proof (rtc_Trans_mo _ _ _ H30
                   (Congruence_LC _ _ Sto6 *)
        pose proof (rtc_Trans _ _ _
                   (rtc_Trans _ _ _
                   (Congruence_Tcase _ _
                                     (\(1:TA) treturn (MReactT TI TO SM) (tvar 1))
         (\(1:TProd TO (TArrow TI TB)) tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4))))))
                                     HS8)
                  (Congruence_ST_CaseR TA _
                                        (\(1:TA) treturn (MReactT TI TO SM) (tvar 1))
         (\(1:TProd TO (TArrow TI TB)) tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4))))))
                                        H33))
                   (rtc_R _ _
                   (ST_AppAbs 1 (TProd TO (TArrow TI TB))
                                  (tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4)))))) _ H33))).
          simpl in H35.   rewrite_subst.
           pose proof (rtc_Trans _ _ _ 
                      (rtc_Trans _ _ _
                      (Congruence_Tcase _ _
                                     (\(1:TA) treturn (MReactT TI TO SM) (tvar 1))
         (\(1:TProd TO (TArrow TI TB)) tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4'))))))
                                     HS8')
                  (Congruence_ST_CaseR TA _
                                        (\(1:TA) treturn (MReactT TI TO SM) (tvar 1))
         (\(1:TProd TO (TArrow TI TB)) tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4'))))))
                                        H34))
                  (rtc_R _ _ 
                  (ST_AppAbs 1 (TProd TO (TArrow TI TB))
                                 (tproj (tvar 1)
                                         (\(2:TO) (\(3:TArrow TI TB) 
                                                    tpause (MReactT TI TO SM)
                                                      TA
                                                      (treturn (MNonReact SM)
                                                         (tpair (tvar 2)
                                                            (\(4:TI) 
                                                              tunfold
                                                                (MReactT TI
                                                                   TO SM) TA
                                                                TB
                                                                (tapp
                                                                   (tvar 3)
                                                                   (tvar 4))
                                                                v4'))))))  _ H34))).
          simpl in H36.   rewrite_subst.
          destruct HeqTProd' as [? [? ?]].
          destruct H39 as [t8 [t9 [t8' [t9' [? [? [HeqTO HeqTArrow]]]]]]].
          destruct (normalization _ _ (eq_typable_empty_l HeqTO))     as [v8  [? ?]].
          destruct (normalization _ _ (eq_typable_empty_r HeqTO))     as [v8' [? ?]].
          destruct (normalization _ _ (eq_typable_empty_l HeqTArrow)) as [v9  [? ?]].
          destruct (normalization _ _ (eq_typable_empty_r HeqTArrow)) as [v9' [? ?]].
          assert (HeqTO': term_eq TO v8 v8') by (eapply eq_square; eauto).
          assert (HeqTArrow': term_eq (TArrow TI TB) v9 v9') by (eapply eq_square;eauto).
          pose proof (rtc_Trans _ _ _ H39
                     (rtc_Trans _ _ _ (Congruence_Pair1 _ _ t9 H41)
                                      (Congruence_Pair2 _ _ _ H42 H45))).
          pose proof (rtc_Trans _ _ _ H40
                     (rtc_Trans _ _ _ (Congruence_Pair1 _ _ t9' H43)
                                      (Congruence_Pair2 _ _ _ H44 H47))).
          assert (HV7: value (\(2:TO) (\(3:TArrow TI TB) tpause (MReactT TI TO SM) TA
                                      (treturn (MNonReact SM)
                                         (tpair (tvar 2)
                                            (\(4:TI) tunfold
                                                        (MReactT TI TO SM) TA
                                                        TB
                                                        (tapp (tvar 3)
                                                           (tvar 4)) v4)))))) by (constructor).
          assert (HV7': value (\(2:TO) (\(3:TArrow TI TB) tpause (MReactT TI TO SM) TA
                                      (treturn (MNonReact SM)
                                         (tpair (tvar 2)
                                            (\(4:TI) tunfold
                                                        (MReactT TI TO SM) TA
                                                        TB
                                                        (tapp (tvar 3)
                                                           (tvar 4)) v4')))))) by (constructor).
          pose proof (Congruence_Proj1 _ _ 
                      (\(2:TO) (\(3:TArrow TI TB) tpause (MReactT TI TO SM) TA
                                      (treturn (MNonReact SM)
                                         (tpair (tvar 2)
                                            (\(4:TI) tunfold
                                                        (MReactT TI TO SM) TA
                                                        TB
                                                        (tapp (tvar 3)
                                                           (tvar 4)) v4))))) H49).
          pose proof (rtc_R _ _ (ST_Proj _ _ _ H42 H46 HV7)).
          pose proof (ST_AppAbs 2 TO 
                                    (\(3:TArrow TI TB) tpause (MReactT TI TO SM) TA
                                      (treturn (MNonReact SM)
                                         (tpair (tvar 2)
                                            (\(4:TI) tunfold
                                                        (MReactT TI TO SM) TA
                                                        TB
                                                        (tapp (tvar 3)
                                                           (tvar 4)) v4)))) v8 H42).
          simpl in H53.   rewrite_subst.
          pose proof (rtc_R _ _ (ST_App1 _ _ v9 H53)). 
          pose proof (rtc_R _ _ 
                     (ST_AppAbs 3 (TArrow TI TB)
                                (tpause (MReactT TI TO SM) TA
                                      (treturn (MNonReact SM)
                                         (tpair v8
                                            (\(4:TI) tunfold
                                                        (MReactT TI TO SM) TA
                                                        TB
                                                        (tapp (tvar 3)
                                                           (tvar 4)) v4)))) v9 H46)).
          simpl in H55.   rewrite_subst.
          pose proof (rtc_Trans_mo _ _ _ H30
                                   (Congruence_LC _ _ Sto6
                                                  ((rtc_Trans _ _ _ H35
                                                              (rtc_Trans _ _ _ H51
                                                                         (rtc_Trans _ _ _ H52
                                                                                    (rtc_Trans _ _ _ H54 H55))))))).
           assert (([3 := v9]v8) = v8).
                 eapply subst_closed;eauto.
                 eapply typable_empty_closed.
                 eapply (eq_typable_empty_l HeqTO').
          rewrite -> H57 in H56. clear H57.
          clear H51 H52 H53 H54 H55.
          destruct HeqTArrow' as [? [? ?]].           
          pose proof (Congruence_Proj1 _ _ (\(2:TO) (\(3:TArrow TI TB) tpause (MReactT TI TO SM) TA
                                      (treturn (MNonReact SM)
                                         (tpair (tvar 2)
                                            (\(4:TI) tunfold
                                                        (MReactT TI TO SM) TA
                                                        TB
                                                        (tapp (tvar 3)
                                                           (tvar 4)) v4'))))) H50).
          pose proof (rtc_R _ _ (ST_Proj _ _ _ H44 H48 HV7')).
          pose proof (ST_AppAbs 2 TO (\(3:TArrow TI TB) tpause (MReactT TI TO SM) TA
                                      (treturn (MNonReact SM)
                                         (tpair (tvar 2)
                                            (\(4:TI) tunfold
                                                        (MReactT TI TO SM) TA
                                                        TB
                                                        (tapp (tvar 3)
                                                           (tvar 4)) v4'))))
                                     v8' H44).
          simpl in H57.   rewrite_subst.
          pose proof (rtc_R _ _ (ST_App1 _ _ v9' H57)). 
          pose proof (rtc_R _ _ 
                     (ST_AppAbs 3 (TArrow TI TB)
                                (tpause (MReactT TI TO SM) TA
                                      (treturn (MNonReact SM)
                                         (tpair v8'
                                            (\(4:TI) tunfold
                                                        (MReactT TI TO SM) TA
                                                        TB
                                                        (tapp (tvar 3)
                                                           (tvar 4)) v4'))))
                                 v9' H48)).
          simpl in H59.   rewrite_subst.
          assert (([3 := v9']v8') = v8').
                 eapply subst_closed;eauto.
                 eapply typable_empty_closed.
                 eapply (eq_typable_empty_r HeqTO').
          rewrite -> H60 in H59. clear H60.
          pose proof (rtc_Trans_mo _ _ _ H31
                                   (Congruence_LC _ _ Sto6'
                                                  (rtc_Trans _ _ _ H36
                                                             (rtc_Trans _ _ _ H54
                                                                        (rtc_Trans _ _ _ H55
                                                                                   (rtc_Trans _ _ _ H58 H59)))))).
          clear H54 H55 H57 H58 H59.
          eapply eq_pause.
                 eapply H56.
                 eapply H60.
                 constructor;auto.
                 constructor;auto. 
                 Focus 5. eapply rtc_refl. 
                 Focus 5. eapply rtc_refl. 
                          auto.
                          auto.
                          constructor.
                          constructor.
                          auto.
                 apply (store_eq_non_react Sto6 Sto6' TI TO SM H26).
                 apply (store_eq_non_react Sto6 Sto6' TI TO SM H26).
          intros.
          destruct (normalization _ _ (eq_typable_empty_l H54))  as [v10  [? ?]].
          destruct (normalization _ _ (eq_typable_empty_r H54))  as [v10' [? ?]].
          assert (HeqTI: term_eq TI v10 v10') by (eapply eq_square;eauto).
          destruct (H53 _ _ HeqTI) as [t11 [t11' [? [? HeqTB11]]]].
          assert (HV10: value (tabs 4 TI (tunfold (MReactT TI TO SM) TA TB (tapp v9 (tvar 4)) v4))) by (constructor).
          assert (HV10': value (tabs 4 TI (tunfold (MReactT TI TO SM) TA TB (tapp v9' (tvar 4)) v4'))) by (constructor).
          pose proof (rtc_Trans _ _ _
                                (Congruence_App2 _ _ _ HV10 H57)
                     (rtc_R _ _ (ST_AppAbs 4 TI
                                 (tunfold (MReactT TI TO SM) TA TB (tapp v9 (tvar 4)) v4) v10 H58))).
          simpl in H64.  
          assert (([4 := v10]v4) = v4 /\ (([4 := v10]v9) = v9)).
                 split; eapply subst_closed;eauto; eapply typable_empty_closed;eauto.
          destruct H65.
          rewrite -> H65 in H64. rewrite -> H66 in H64. clear H65 H66.
          pose proof (rtc_Trans _ _ _
                                (Congruence_App2 _ _ _ HV10' H59)
                     (rtc_R _ _ (ST_AppAbs 4 TI
                                 (tunfold (MReactT TI TO SM) TA TB (tapp v9' (tvar 4)) v4') v10' H61))).
          simpl in H65.
          assert (([4 := v10']v4') = v4' /\ (([4 := v10']v9') = v9')).
                 split; eapply subst_closed;eauto; eapply typable_empty_closed;eauto.
          destruct H66.
          rewrite -> H66 in H65. rewrite -> H67 in H65. clear H66 H67.
           
          pose proof (Congruence_LC _ _ Sto3 H64).
          pose proof (Congruence_LC _ _ Sto3' H65).
          eapply tunfold_along2.
                 eapply H66.
                 eapply H67.
                 eapply eq_square_rev.
                        econstructor;eauto.
                        apply (eq_typable_empty_l HeqTI).
                        econstructor;eauto.
                        apply (eq_typable_empty_r HeqTI).
                        eapply H62.
                        eapply H63.
                        auto.
                        auto.
                        auto.
Qed.


Lemma TBind_Term_eq' : forall Mo T1 T2 t1 t1' t2 t2',
                      value t1  -> 
                      value t1' -> 
                      value t2  -> 
                      value t2' ->
      term_eq (TArrow T1 (TMonadic Mo T2)) t1' t2' ->
      term_eq (TMonadic Mo T1) t1 t2 ->
      term_eq (TMonadic Mo T2) (tbind T2 t1 t1') (tbind T2 t2 t2').
Proof.
  intros Mo T1 T2 t1 t1' t2 t2' Hval1 Hval1' Hval2 Hval2' Heq1 Heq2.
         generalize dependent Heq1;
         generalize dependent Heq2;           (*
         generalize dependent (tbind t1 t1');
         generalize dependent (tbind t2 t2'); *)
         generalize dependent t1'; generalize dependent t2';
         generalize dependent t1 ; generalize dependent t2 ;
         generalize dependent T1;
         generalize dependent T2;
         induction Mo;
         [rename t into TI; rename t0 into TO; rename s into SM 
         |rename s into SM ]; intros.
  Case "MReactT". 
    generalize Heq2 as Heq2';intros.
    generalize Heq1 as Heq1';intros.
       destruct Heq1 as [? [? ?]].
       destruct Heq2 as [? [? ?]]. 
       split;[econstructor; [eapply H2
                            |eapply H ]
             |split;[econstructor; [eapply H3
                                   |eapply H0]
                    |intros]].
       exists (tbind T2 t1 t1'), Sto1, (tbind T2 t2 t2'), Sto2.
              repeat (split;eauto). 
              eapply tbind_along2. 
                     Focus 5. eapply rtc_refl. 
                     Focus 5. eapply rtc_refl.
              auto. auto. auto. auto. 
              eapply Heq2'.
              auto.
              auto.
   
 Case "MNonReact".
      destruct Heq2 as [? [? ?]];
      destruct Heq1 as [? [? ?]].
      split;[econstructor; [eapply H|eapply H2]|
            split;[econstructor; [eapply H0|eapply H3]|
            intros]].
       destruct (H1 _ _ H5) as [v1' [Sto1' [v2' [Sto2' [? [? [? [? [? ?]]]]]]]]].
       pose proof (Congruence_Mo_Bind1 T2 _ Sto1 _ Sto1' _ Hval1' H6).
       pose proof (Congruence_Mo_Bind1 T2 _ Sto2 _ Sto2' _ Hval2' H7).
       pose proof (rtc_R_mo _ _ (STM_BindRet _ _  T2 (MNonReact SM) Sto1' H8 Hval1')).
       pose proof (rtc_R_mo _ _ (STM_BindRet _ _  T2 (MNonReact SM) Sto2' H9 Hval2')).
       pose proof (rtc_Trans_mo _ _ _ H12 H14).
       pose proof (rtc_Trans_mo _ _ _ H13 H15).
       destruct (H4 _ _ H11) as [v3 [v3' [? [? ?]]]].
       destruct H20 as [? [? ?]].
       destruct (H22 _ _ H10) as [? [? [? [? [? [? [? [? [? ?]]]]]]]]].
       pose proof (rtc_Trans_mo _ _ _ H16 ((rtc_Trans_mo _ _ _ (Congruence_LC _ _ Sto1' H18) H23))).
       pose proof (rtc_Trans_mo _ _ _ H17 ((rtc_Trans_mo _ _ _ (Congruence_LC _ _ Sto2' H19) H24))).
       do 4 eexists.
          destruct H27 as [? [? [? [? ?]]]].
          repeat split.
          eapply H29.
          eapply H30.
          auto. 
          auto.
          auto.
          auto.
          auto.
          auto. 
          auto.
          auto.
Qed. 

Corollary TBind_Term_eq : forall Mo T1 T2 t1 t1' t2 t2',
          term_eq (TArrow T1 (TMonadic Mo T2)) t1' t2' ->
          term_eq (TMonadic Mo T1) t1 t2 ->
          term_eq (TMonadic Mo T2) (tbind T2 t1 t1') (tbind T2 t2 t2').
Proof.
  intros. 
  generalize H; intros.
  generalize H0; intros.
  destruct H as [? [? _]];  
  destruct H0 as [? [? _]].
  destruct (normalization _ _ H) as [v1' [? ?]];
  destruct (normalization _ _ H3) as [v2' [? ?]];
  destruct (normalization _ _ H0) as [v1 [? ?]];
  destruct (normalization _ _ H4) as [v2 [? ?]].
  assert (tbind T2 t1 t1' ~~>* tbind T2 v1 v1')  by ( 
         assert (tbind T2 t1 t1' ~~>* tbind T2 v1 t1') by (
                (eapply Congruence_Tbind1; auto));
         assert (tbind T2 v1 t1' ~~>* tbind T2 v1 v1') by (
                (eapply Congruence_Tbind2; auto)); eauto);
  assert (tbind T2 t2 t2' ~~>* tbind T2 v2 v2') by ( 
         assert (tbind T2 t2 t2' ~~>* tbind T2 v2 t2') by (
                (eapply Congruence_Tbind1; auto));
         assert (tbind T2 v2 t2' ~~>* tbind T2 v2 v2') by (
                (eapply Congruence_Tbind2; auto)); eauto).
  assert (term_eq (TArrow T1 (TMonadic Mo T2)) v1' v2') by (eapply eq_square; eauto).
  assert (term_eq (TMonadic Mo T1) v1 v2) by (eapply eq_square;eauto).
  eapply eq_square_rev; eauto. 
  eapply TBind_Term_eq';eauto. 
Qed.

Lemma TAbs_Term_eq : forall T1 T2 t1 t2 x,
      term_eq T2 t1 t2 ->
      term_eq (TArrow T1 T2) (tabs x T1 t1) (tabs x T1 t2).
Proof.
  intros T1 T2 t1 t2 x Heq12.
  pose proof (eq_typable_empty_l Heq12).
  eapply typable_empty_everywhere in H.
  pose proof (eq_typable_empty_r Heq12).
  eapply typable_empty_everywhere in H0.
  simpl; repeat split; [ constructor; apply H |
                         constructor; apply H0|
                        ].
  intros.
  destruct (normalization _ _ (eq_typable_empty_l H1)) as [val_t0 [Hstept0 Hvalt0]].
  destruct (normalization _ _ (eq_typable_empty_r H1)) as [val_t1' [Hstept1' Hvalt1']].
  assert (Htabst1_val: value (tabs x T1 t1)) by (constructor;auto).
  assert (Htabst2_val: value (tabs x T1 t2)) by (constructor;auto).
  assert (HGoalt1: tapp (tabs x T1 t1) t0 ~~>* t1).
         assert (Htapp1: tapp (tabs x T1 t1) t0 ~~>* tapp (tabs x T1 t1) val_t0).
                eapply Congruence_App2. 
                auto. auto.
         assert (Htapp1': tapp (tabs x T1 t1) val_t0 ~~>* [x := val_t0]t1).
                pose proof (ST_AppAbs x T1 t1 val_t0 Hvalt0).
                eapply rtc_R. auto.
         pose proof (rtc_Trans _ _ _ Htapp1 Htapp1') as Htapp1''.
         assert (Htapp1''': [x:= val_t0]t1 = t1).
                pose proof (typable_empty_closed t1 T2 (eq_typable_empty_l Heq12)).
                eapply subst_closed. 
                auto.
         rewrite Htapp1''' in Htapp1''. 
         assumption.
  assert (HGoalt2: tapp (tabs x T1 t2) t1' ~~>* t2).
         assert (Htapp2: tapp (tabs x T1 t2) t1' ~~>* tapp (tabs x T1 t2) val_t1').
                eapply Congruence_App2. 
                auto. auto.
         assert (Htapp2': tapp (tabs x T1 t2) val_t1' ~~>* [x := val_t1']t2).
                pose proof (ST_AppAbs x T1 t2 val_t1' Hvalt1').
                eapply rtc_R. auto.
         pose proof (rtc_Trans _ _ _ Htapp2 Htapp2') as Htapp2''.
         assert (Htapp2''': [x:= val_t1']t2 = t2).
                pose proof (typable_empty_closed t2 T2 (eq_typable_empty_r Heq12)).
                eapply subst_closed. 
                auto.
         rewrite Htapp2''' in Htapp2''. 
         assumption.
   exists t1, t2.
      repeat (split;auto).
Qed.

Lemma TPause_Term_eq : forall (t t' : tm) (TI TO T : Ty) (SM : SMo),   
 term_eq (TMonadic (MNonReact SM)
         (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))) t t' ->
 term_eq (TMonadic (MReactT TI TO SM) T) 
          (tpause (MReactT TI TO SM) T t)
          (tpause (MReactT TI TO SM) T t').
Proof.
  intros.
  destruct (normalization _ _ (eq_typable_empty_l H)) as [v  [? ?]];
  destruct (normalization _ _ (eq_typable_empty_r H)) as [v' [? ?]].
    assert (H0':term_eq (TMonadic (MNonReact SM)
  (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))) v v') by (eapply eq_square;eauto).
    split. constructor. apply (eq_typable_empty_l H).
    split. constructor. apply (eq_typable_empty_r H).
    intros.
    pose proof (Congruence_LC _ _  Sto1 
               (Congruence_ST_Pause (MReactT TI TO SM) T _ _ H0)).
    pose proof (Congruence_LC _ _  Sto2 
               (Congruence_ST_Pause (MReactT TI TO SM) T _ _ H2)).
    exists (tpause (MReactT TI TO SM) T v), Sto1, (tpause (MReactT TI TO SM) T v'), Sto2.
           split. auto.
           split. auto.
           split. (constructor;auto). 
           split. (constructor;auto).
    destruct H0' as [? [? ?]].
    assert (store_eq (MNonReact SM) Sto1 Sto2).
           destruct H4 as [? [? [? [? ?]]]]. 
           repeat split; auto. inv H4. inv H11.
    destruct (H9 _ _ H10) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? [? ?]]]]]]]]].
    destruct H16 as [? [? ?]]. 
    destruct H18 as [tl [tr [tl' [tr' [? [? [? ?]]]]]]]. 
    destruct (normalization _ _ (eq_typable_empty_l H20)) as [vl  [? ?]];
    destruct (normalization _ _ (eq_typable_empty_r H20)) as [vl' [? ?]];
    destruct (normalization _ _ (eq_typable_empty_l H21)) as [vr  [? ?]];
    destruct (normalization _ _ (eq_typable_empty_r H21)) as [vr' [? ?]]. 
    assert ((treturn (MNonReact SM) t1',Sto1') ~~~>* (treturn (MNonReact SM) (tpair vl vr),Sto1')).
           eapply Congruence_LC. eapply Congruence_LC_Ret.
           eapply rtc_Trans. eapply H18.
           eapply rtc_Trans. eapply Congruence_Pair1;eauto.
           eapply rtc_Trans. eapply Congruence_Pair2;eauto. 
           eauto.
    assert ((treturn (MNonReact SM) t2',Sto2') ~~~>* (treturn (MNonReact SM) (tpair vl' vr'),Sto2')).
           eapply Congruence_LC. eapply Congruence_LC_Ret.
           eapply rtc_Trans. eapply H19.
           eapply rtc_Trans. eapply Congruence_Pair1;eauto.
           eapply rtc_Trans. eapply Congruence_Pair2;eauto. 
           eauto.
    eapply eq_pause. 
     eapply rtc_refl.
     eapply rtc_refl.
     auto.
     auto.
     eapply H23. 
     eapply H25.
     eapply H27.
     eapply H29.
     eapply rtc_Trans_mo. eapply H11.
                          eapply H30.
     eapply rtc_Trans_mo. eapply H12.
                          eapply H31.
     eapply eq_square.
     eapply H20.
     auto.
     auto.
     auto.
     destruct H15 as [? [? [? [? ?]]]].
       repeat split; auto.
     intros.
     assert (term_eq (TArrow TI (TMonadic (MReactT TI TO SM) T)) vr vr') by (eapply eq_square;eauto).
     destruct H34 as [? [? ?]].
     destruct (H36 _ _ H32) as [t4 [t4' [? [? ?]]]].
     eapply step_preserves_along2_react_star. 
            eapply Congruence_LC;eauto.
            eapply Congruence_LC;eauto.
     destruct H39 as [? [? ?]].
     destruct (H41 _ _ H33) as [t1'0 [Sto1'0 [t2'0 [Sto2'0 [? [? [? [? ?]]]]]]]].
     eapply step_preserves_along2_react_star;eauto. 
Qed.
                                              
Lemma msubst_eq : forall c envL envR t T,
     (mextend empty c) |-- t \in T -> 
     eq_instantiation c envL envR -> 
     term_eq T (msubst envL t) (msubst envR t).
Proof.
  intros c envL envR t T HT V.
  generalize dependent envR.
  generalize dependent envL.
  remember (mextend empty c) as Gamma.
  assert (forall x, Gamma x = lookup x c).
    intros. rewrite HeqGamma. rewrite mextend_lookup. auto.
  clear HeqGamma.
  generalize dependent c.
  induction HT; intros.
  Case "T_Var".
    rewrite H0 in H.
    destruct (eq_instantiation_domains_match_l V H) as [tL PL].
    destruct (eq_instantiation_domains_match_r V H) as [tR PR].
    eapply eq_instantiation_eq; eauto.
      rewrite msubst_var. rewrite PL. auto. eapply eq_instantiation_env_closed_l; eauto.
      rewrite msubst_var. rewrite PR. auto. eapply eq_instantiation_env_closed_r; eauto.
  Case "T_Abs".
    repeat (rewrite msubst_abs).
    assert (WTL: empty |-- (tabs x T (msubst (drop x envL) t)) \in (TArrow T T')).
      eapply T_Abs. eapply msubst_preserves_typing_l. eapply eq_instantiation_drop; eauto.
      eapply context_invariance. apply HT.
      intros.
      unfold extend.
        rewrite mextend_drop.
        destruct (eq_id_dec x x0). auto.
        rewrite H.
        clear - c n. induction c.
          simpl. rewrite neq_id; auto.
          simpl. destruct a. unfold extend. destruct (eq_id_dec i x0); auto.
    assert (WTR: empty |-- (tabs x T (msubst (drop x envR) t)) \in (TArrow T T')).
      eapply T_Abs. eapply msubst_preserves_typing_r. eapply eq_instantiation_drop; eauto.
      eapply context_invariance. apply HT.
      intros.
      unfold extend.
        rewrite mextend_drop.
        destruct (eq_id_dec x x0). auto.
        rewrite H.
        clear - c n. induction c.
          simpl. rewrite neq_id; auto.
          simpl. destruct a. unfold extend. destruct (eq_id_dec i x0); auto.
    unfold term_eq. fold term_eq. repeat split.
      auto.
      auto.
      intros.
      assert (forall x0 : id, extend Gamma x T x0 = lookup x0 ((x,T)::c)).
        intros. unfold extend, lookup. destruct (eq_id_dec x x0); auto.
      pose proof (normalization _ _ (eq_typable_empty_l H0)) as [vt1 [Hstept1 Hvalvt1]].
      pose proof (normalization _ _ (eq_typable_empty_r H0)) as [vt1' [Hstept1' Hvalvt1']].
      assert (eq_instantiation ((x, T) :: c) ((x,vt1)::envL) ((x,vt1')::envR)). 
             constructor; auto.
             eapply eq_square; eauto.
      pose proof (IHHT ((x,T)::c) H1 ((x,vt1)::envL) ((x,vt1')::envR) H2).
      pose proof (preservation_star _ _ _ Hstept1 (eq_typable_empty_l H0)) as HTvt1.
      pose proof (preservation_star _ _ _ Hstept1' (eq_typable_empty_r H0)) as HTvt1'.
      pose proof (ST_AppAbs x T (msubst (drop x envL) t) _ Hvalvt1).
      pose proof (ST_AppAbs x T (msubst (drop x envR) t) _ Hvalvt1'). 
      assert (tapp (tabs x T (msubst (drop x envL) t)) t1 ~~>*
              tapp (tabs x T (msubst (drop x envL) t)) vt1).
        apply Congruence_App2. constructor. assumption.
      assert (tapp (tabs x T (msubst (drop x envR) t)) t1' ~~>*
              tapp (tabs x T (msubst (drop x envR) t)) vt1').
        apply Congruence_App2. constructor. assumption.
      assert (msubst ((x, vt1) :: envL) t = [x := vt1]msubst (drop x envL) t).
        simpl.
        apply subst_msubst.
        eapply typable_empty_closed; eauto.
        eapply eq_instantiation_env_closed_l; eauto.
      assert (msubst ((x, vt1') :: envR) t = [x := vt1']msubst (drop x envR) t).
        simpl.
        apply subst_msubst.
        eapply typable_empty_closed; eauto.
        eapply eq_instantiation_env_closed_r; eauto.
      rewrite <- H8 in H4.
      rewrite <- H9 in H5.
      exists (msubst ((x,vt1)::envL) t), (msubst ((x,vt1')::envR) t).
      split; [| split].
        eapply rtc_Trans. eapply H6. apply rtc_R. assumption.
        eapply rtc_Trans. eapply H7. apply rtc_R. assumption. 
        assumption.
  Case "T_App".
    repeat (rewrite msubst_app).
    pose proof (IHHT1 c H envL envR V) as P1.
    pose proof (IHHT2 c H envL envR V) as P2.
    pose proof (eq_typable_empty_l P2).
    pose proof (eq_typable_empty_r P2).
    destruct P1 as [HTarr1 [HTarr2 _]]. 
    eapply TApp_Term_eq;eauto. 
  Case "T_Nil".
    repeat (rewrite msubst_tunit); simpl; repeat split; constructor.
  Case "T_Prod".
    rewrite msubst_pair. rewrite msubst_pair.
    pose proof (IHHT1 c H envL envR V).
    pose proof (IHHT2 c H envL envR V).
    assert (term_eq T1 (msubst envL t1) (msubst envR t1) /\ term_eq T2 (msubst envL t2) (msubst envR t2)) by (split;auto).
    pose proof (TPair_Term_eq T1 T2 (msubst envL t1) (msubst envL t2) (msubst envR t1) (msubst envR t2)).
    rewrite -> H3 in H2. auto.
  Case "T_Proj".
    rewrite msubst_proj. rewrite msubst_proj.
    pose proof (IHHT1 c H envL envR V).
    pose proof (IHHT2 c H envL envR V).
    destruct (normalization _ _ (eq_typable_empty_l H0)) as [vLt [HstepvLt H_val_Lt]].
    destruct (normalization _ _ (eq_typable_empty_r H0)) as [vRt [HstepvRt H_val_Rt]].
    destruct (normalization _ _ (eq_typable_empty_l H1)) as [vLt' [HstepvLt' H_val_Lt']].
    destruct (normalization _ _ (eq_typable_empty_r H1)) as [vRt' [HstepvRt' H_val_Rt']].
    assert (H0': term_eq (TProd T1 T2) vLt vRt).
             eapply eq_square.
                    apply H0.
                    auto.
                    auto.
      assert (H1': term_eq (TArrow T1 (TArrow T2 T3)) vLt' vRt').
             eapply eq_square.
                    apply H1.
                    auto.
                    auto.
      simpl in H0'.
            destruct H0'.
            destruct H3.
            destruct H4 as [tL1 [tL2 [tR1 [tR2 [HstepLpair [HstepRpair [HeqT1 HeqT2]]]]]]].
      simpl in H1'.
            destruct H1'.
            destruct H5.
      destruct (normalization _ _ (eq_typable_empty_l HeqT1)) as [vtL1 [? ?]].
      destruct (normalization _ _ (eq_typable_empty_r HeqT1)) as [vtR1 [? ?]].
      destruct (normalization _ _ (eq_typable_empty_l HeqT2)) as [vtL2 [? ?]].
      destruct (normalization _ _ (eq_typable_empty_r HeqT2)) as [vtR2 [? ?]].
      assert (HeqT1': term_eq T1 vtL1 vtR1).
             eapply eq_square;eassumption.
      assert (HeqT2': term_eq T2 vtL2 vtR2).
             eapply eq_square;eassumption.
      pose proof (H6 vtL1 vtR1 HeqT1').
      destruct H15 as [t2 [t3 [Htapp_tL1 [Htapp_tR1 [HTt2 [HTt3 ?]]]]]].
      pose proof (H15 vtL2 vtR2 HeqT2').
      destruct H16 as [t2' [t3' [Htappt2' [Htappt3' HteqT3]]]].
      assert (HGoalL: tapp (tapp vLt' vtL1) vtL2 ~~>*  t2').
             assert (HGL: tapp (tapp vLt' vtL1) vtL2 ~~>*  tapp t2 vtL2).
                    pose proof (Congruence_App1 _ _ vtL2 Htapp_tL1).
                    auto.
             pose proof (rtc_Trans _ _ _ HGL Htappt2').
             auto.
      assert (HGoalR: tapp (tapp vRt' vtR1) vtR2 ~~>*  t3').
             assert (HGL: tapp (tapp vRt' vtR1) vtR2 ~~>*  tapp t3 vtR2).
                    pose proof (Congruence_App1 _ _ vtR2 Htapp_tR1).
                    auto.
             pose proof (rtc_Trans _ _ _ HGL Htappt3').
             auto.
      assert (Hval_Lpair: tpair tL1 tL2 ~~>* tpair vtL1 vtL2).
             pose proof (Congruence_Pair1 _ _ tL2 H7).
             pose proof (Congruence_Pair2 _ _ _ H8 H11).
             pose proof (rtc_Trans _ _ _ H16 H17). 
             auto.
      assert (Hval_Rpair: tpair tR1 tR2 ~~>* tpair vtR1 vtR2).
             pose proof (Congruence_Pair1 _ _ tR2 H9).
             pose proof (Congruence_Pair2 _ _ _ H10 H13).
             pose proof (rtc_Trans _ _ _ H16 H17). 
             auto.
      assert (HTL: (msubst envL t) ~~>* tpair vtL1 vtL2).
             eapply rtc_Trans;eauto.
             eapply rtc_Trans;eauto.
      assert (HTR: (msubst envR t) ~~>* tpair vtR1 vtR2).
             eapply rtc_Trans;eauto.
             eapply rtc_Trans;eauto.
      assert (Hv_Lpair: value (tpair vtL1 vtL2)) by (constructor;auto).
      assert (Hv_Rpair: value (tpair vtR1 vtR2)) by (constructor;auto).
      assert (Hproj_VL: tproj (msubst envL t) (msubst envL t') ~~>* tproj (tpair vtL1 vtL2) vLt').
             pose proof (Congruence_Proj1 _ _ (msubst envL t') HTL).
             pose proof (Congruence_Proj2 _ _ _ Hv_Lpair HstepvLt').
             pose proof (rtc_Trans _ _ _ H16 H17). 
             auto.
      assert (Hproj_VL': tproj (tpair vtL1 vtL2) vLt' ~~>* tapp (tapp vLt' vtL1) vtL2).
             pose proof (ST_Proj _ _ _ H8 H12 H_val_Lt').
             eapply rtc_R; eauto.
      assert (Hproj_VL'': tproj (msubst envL t) (msubst envL t') ~~>* tapp (tapp vLt' vtL1) vtL2).
             eapply rtc_Trans.
             apply Hproj_VL. 
             eauto.
      assert (Hproj_VR: tproj (msubst envR t) (msubst envR t') ~~>* tproj (tpair vtR1 vtR2) vRt').
             pose proof (Congruence_Proj1 _ _ (msubst envR t') HTR).
             pose proof (Congruence_Proj2 _ _ _ Hv_Rpair HstepvRt').
             pose proof (rtc_Trans _ _ _ H16 H17). 
             auto.
      assert (Hproj_VR': tproj (tpair vtR1 vtR2) vRt' ~~>* tapp (tapp vRt' vtR1) vtR2).
             pose proof (ST_Proj _ _ _ H10 H14 H_val_Rt').
             eapply rtc_R; eauto.
      assert (Hproj_VR'': tproj (msubst envR t) (msubst envR t') ~~>* tapp (tapp vRt' vtR1) vtR2).
             eapply rtc_Trans. 
             apply Hproj_VR. 
             eauto.
      assert (HGoalL': tproj (msubst envL t) (msubst envL t') ~~>* t2') by (eapply rtc_Trans; eassumption).
      assert (HGoalR': tproj (msubst envR t) (msubst envR t') ~~>* t3') by (eapply rtc_Trans; eassumption).
      eapply eq_square_rev. 
             econstructor. 
                eapply eq_typable_empty_l; eauto.
                eapply eq_typable_empty_l; eauto.
             econstructor. 
                eapply eq_typable_empty_r; eauto.
                eapply eq_typable_empty_r; eauto.
             apply HGoalL'.
             apply HGoalR'.
             auto.
    Case "T_tinl". 
         repeat (rewrite msubst_tinl);
         simpl; pose proof (IHHT c H envL envR V);
         repeat split; [ econstructor; eapply eq_typable_empty_l; eauto |
                         econstructor; eapply eq_typable_empty_r; eauto |
                         left; exists (msubst envL t), (msubst envR t); 
                         repeat (split; (try (eauto; apply rtc_refl)))].
   Case "T_tinr". 
        repeat (rewrite msubst_tinr);
        simpl; pose proof (IHHT c H envL envR V);
        repeat split; [ econstructor; eapply eq_typable_empty_l; eauto |
                        econstructor; eapply eq_typable_empty_r; eauto |
                        right; exists (msubst envL t), (msubst envR t); 
                        repeat (split; (try (eauto; apply rtc_refl)))].
  Case "T_Res".  
       repeat (rewrite msubst_tcase);
                            pose proof (IHHT1 c H envL envR V);
                            pose proof (IHHT2 c H envL envR V);
                            pose proof (IHHT3 c H envL envR V);
                            pose proof (eq_typable_empty_l H0);
                            pose proof (eq_typable_empty_r H0);
                            assert (\empty |-- tcase (msubst envL ts) (msubst envL tl) (msubst envL tr) \in Tres) by (eapply eq_typable_empty_l in H1; eapply eq_typable_empty_l in H2;eapply T_Case;eauto);
                            assert (\empty |-- tcase (msubst envR ts) (msubst envR tl) (msubst envR tr) \in Tres) by (eapply eq_typable_empty_r in H1; eapply eq_typable_empty_r in H2;eapply T_Case;eauto); 
                            eapply TCase_Term_eq;eauto. 
Case "treturn". 
   repeat (rewrite msubst_return).
   pose proof (IHHT c H envL envR V).
   eapply TReturn_Term_eq;eauto.
Case "tbind".
  repeat (rewrite msubst_bind).
  pose proof (IHHT1 c H envL envR V).
  pose proof (IHHT2 c H envL envR V).
  eapply TBind_Term_eq; eauto. 
Case "tlift_st". 
  repeat (rewrite msubst_lift).
  pose proof (IHHT c H envL envR V).
  destruct (normalization _ _ (eq_typable_empty_l H0)) as [vt  [? ?]];
  destruct (normalization _ _ (eq_typable_empty_r H0)) as [vt' [? ?]].
  assert (H0': term_eq (TMonadic (MNonReact SM) T') vt vt') by (eapply eq_square;eauto).
  split.
  constructor.
  apply (eq_typable_empty_l H0).
  split.
  constructor.
  apply (eq_typable_empty_r H0).
  intros.
  destruct H5 as [? [? [? [? [t1 [t2 [Sto1' [Sto2' [? [? [? ?]]]]]]]]]]].
  destruct H0' as [? [? ?]].
  assert (store_eq (MNonReact SM) Sto1' Sto2').
         subst.
         repeat split; auto. inv H5.  inv H6. 
                             inv H7.  inv H8.
  destruct (H15 _ _ H16) as [v3 [Sto3 [v3' [Sto3' [? [? [? [? [? ?]]]]]]]]].
  pose proof (Congruence_LiftSt _ _ _ _ t1 T E SM H2 H17) as HL1.
  pose proof (Congruence_LiftRetSt T E SM v3 (t1 :: Sto3) H19) as HL2.
  pose proof (Congruence_LC _ _ (t1 :: Sto1') (Congruence_Lift (MNonReact (MStateT T E SM)) _ _ H1)) as HL3. 
  pose proof (rtc_Trans_mo _ _ _ (rtc_Trans_mo _ _ _ HL3 HL1) HL2) as HL4.
  pose proof (Congruence_LiftSt _ _ _ _ t2 T E SM H4 H18) as HL1'.
  pose proof (Congruence_LiftRetSt T E SM v3' (t2 :: Sto3') H20) as HL2'.
  pose proof (Congruence_LC _ _ (t2 :: Sto2') (Congruence_Lift (MNonReact (MStateT T E SM)) _ _ H3)) as HL3'. 
  pose proof (rtc_Trans_mo _ _ _ (rtc_Trans_mo _ _ _ HL3' HL1') HL2') as HL4'.
  subst.
  destruct H21 as [? [? [? [? ?]]]].
  do 4 eexists.
     split. eapply HL4.
     split. eapply HL4'.
     repeat split; auto.
     constructor.
     apply (eq_typable_empty_l H9).
     auto. inv H6. 
     constructor. auto. auto. 
     auto. constructor.
     apply (eq_typable_empty_r H9).
     auto. 
     inv H8. constructor. 
     auto. auto.
     simpl.
     exists t1, t2, Sto3, Sto3'.
     repeat split; auto.
Case "tlift_re".
  repeat (rewrite msubst_lift).
  pose proof (IHHT c H envL envR V).
  destruct (normalization _ _ (eq_typable_empty_l H0)) as [vt  [? ?]];
  destruct (normalization _ _ (eq_typable_empty_r H0)) as [vt' [? ?]].
  assert (H0': term_eq (TMonadic (MNonReact SM) T) vt vt') by (eapply eq_square;eauto).
  split.
  constructor.
  apply (eq_typable_empty_l H0).
  split.
  constructor.
  apply (eq_typable_empty_r H0).
  intros.
  pose proof (Congruence_LC _ _ Sto1 (Congruence_Lift (MReactT TI TO SM) _ _ H1)).
  pose proof (Congruence_LC _ _ Sto2 (Congruence_Lift (MReactT TI TO SM) _ _ H3)).
  assert (store_eq (MNonReact SM) Sto1 Sto2).
         destruct H5 as [? [? [? [? ?]]]].
         repeat split; auto. inv H5. inv H9.    
  destruct H0' as [? [? ?]].
  destruct (H11 _ _ H8) as [v3 [Sto3 [v3' [Sto3' [? [? [? [? [? ?]]]]]]]]].
  pose proof (Congruence_LiftRe _ _ _ _ TI TO SM H2 H12) as HL1.
  pose proof (Congruence_LiftRetRe v3 TI TO SM Sto3 H14) as HL2.
  pose proof (rtc_Trans_mo _ _ _ (rtc_Trans_mo _ _ _ H6 HL1) HL2) as HL4.
  pose proof (Congruence_LiftRe _ _ _ _ TI TO SM H4 H13) as HL1'.
  pose proof (Congruence_LiftRetRe v3' TI TO SM Sto3' H15) as HL2'.
  pose proof (rtc_Trans_mo _ _ _ (rtc_Trans_mo _ _ _ H7 HL1') HL2') as HL4'.
  pose proof (TReturn_Term_eq T (MReactT TI TO SM) v3 v3' H17).
  destruct H18 as [?[? ?]].
  assert (store_eq (MReactT TI TO SM) Sto3 Sto3').
         destruct H16 as [? [? [? [? ?]]]].
         repeat split; auto.
  destruct (H20 _ _ H21) as [? [? [? [? [? [? [? [? ?]]]]]]]]. 
  pose proof (rtc_Trans_mo _ _ _ HL4 H22).
  pose proof (rtc_Trans_mo _ _ _ HL4' H23).
  do 4 eexists.
    repeat split; eauto.
Case "televate".
  repeat (rewrite msubst_elevate).
  pose proof (IHHT c H0 envL envR V).
  destruct (normalization _ _ (eq_typable_empty_l H1)) as [vt  [? ?]];
  destruct (normalization _ _ (eq_typable_empty_r H1)) as [vt' [? ?]].
  assert (H0': term_eq (TMonadic (MNonReact SM) T) vt vt') by (eapply eq_square;eauto).
  split. econstructor;eauto. apply (eq_typable_empty_l H1).
  split. econstructor;eauto. apply (eq_typable_empty_r H1).
  intros.
  destruct H0' as [?[? ?]].
  destruct (H9 _ _ (store_eq_mp _ _ _ _ H H6)) as [t1' [Sto1' [t2' [Sto2' [? [? [? [? [? ?]]]]]]]]].
  pose proof (Congruence_LC _ _ Sto1 (Congruence_ST_Elevate SM' _ _ H2)) as HS1.
  pose proof (Congruence_STM_Elevate SM' _ _ _ _ H3 H10) as HS2. 
  pose proof (Congruence_STM_ElevateRet SM SM' _ Sto1' H12) as HS3.
  pose proof (rtc_Trans_mo _ _ _ HS1 (rtc_Trans_mo _ _ _ HS2 HS3)) as HS4.
  pose proof (Congruence_LC _ _ Sto2 (Congruence_ST_Elevate SM' _ _ H4)) as HS1'.
  pose proof (Congruence_STM_Elevate SM' _ _ _ _ H5 H11) as HS2'. 
  pose proof (Congruence_STM_ElevateRet SM SM' _ Sto2' H13) as HS3'.
  pose proof (rtc_Trans_mo _ _ _ HS1' (rtc_Trans_mo _ _ _ HS2' HS3')) as HS4'.
  pose proof (store_eq_lp _ _ _ _ H H14). 
  exists t1', Sto1', t2', Sto2'.
        split; auto.
Case "tget".
  repeat (rewrite msubst_get).
  split.
  constructor;auto.
  split. constructor;auto.
  intros.
  destruct H1 as [? [? [? [? ?]]]].
  destruct H5 as [t1 [t2 [Sto1' [Sto2' [Htermeq [Hstoeq [Hid1 Hid2]]]]]]].  
  destruct (normalization _ _ (eq_typable_empty_l Htermeq)) as [v1 [? ?]].
  destruct (normalization _ _ (eq_typable_empty_r Htermeq)) as [v2 [? ?]].
  assert (term_eq T v1 v2) by (eapply eq_square;eauto).
  pose proof (rtc_Trans_mo _ _ _
             (rtc_R_mo _ _ (STM_Get (MStateT T E SM) Sto1' t1))
             (Congruence_Treturn _ _  (MNonReact (MStateT T E SM)) (t1 :: Sto1') H5)). 
  pose proof (rtc_Trans_mo _ _ _
             (rtc_R_mo _ _ (STM_Get (MStateT T E SM) Sto2' t2)) 
             (Congruence_Treturn _ _  (MNonReact (MStateT T E SM)) (t2 :: Sto2') H7)).
  assert (store_eq (MNonReact (MStateT T E SM)) Sto1 Sto2).
         simpl. repeat split; auto.
         exists t1, t2, Sto1', Sto2'. repeat split; eauto.
  exists v1, (t1 :: Sto1'), v2, (t2 :: Sto2').
  subst.
  split; auto.
Case "tput".
  repeat (rewrite msubst_tput).
  pose proof (IHHT c H0 envL envR V).
  split. constructor. auto.
                      apply (eq_typable_empty_l H1).
  split. constructor. auto.
                      apply (eq_typable_empty_r H1).
  intros.
    destruct (normalization _ _ (eq_typable_empty_l H1)) as [v1 [? ?]].
    destruct (normalization _ _ (eq_typable_empty_r H1)) as [v2 [? ?]].
    assert (H0': term_eq T v1 v2) by (eapply eq_square;eauto).
    destruct H2 as [? [? [? [? ?]]]].
    destruct H10 as [t1 [t2 [Sto1' [Sto2' [Htermeq [Hstoeq [Hid1 Hid2]]]]]]]. 
    pose proof (rtc_Trans_mo _ _ _
         (Congruence_LC _ _ (t1 :: Sto1')
         (Congruence_ST_Put (MStateT T E SM) _ _ H3))
         (rtc_R_mo _ _ (STM_Put v1 (MStateT T E SM) Sto1' t1 H4))).
    pose proof (rtc_Trans_mo _ _ _
         (Congruence_LC _ _ (t2 :: Sto2')
         (Congruence_ST_Put (MStateT T E SM) _ _ H5))
         (rtc_R_mo _ _ (STM_Put v2 (MStateT T E SM) Sto2' t2 H6))).
    exists tunit, (v1 :: Sto1'), tunit, (v2 :: Sto2').
    subst.
    split; auto.
    split; auto.
    split. constructor.
    split. constructor.
    split. 
    split. constructor.
           apply (eq_typable_empty_l H0').
           inv H2.
    split. constructor. auto. 
           inv H7. 
    split. constructor.
           apply (eq_typable_empty_r H0').
           inv H8.
    split. constructor. auto. 
           inv H9.
    simpl.
    exists v1, v2, Sto1', Sto2'.
    auto.
    simpl.
    split; constructor.
 Case "trunst".
      repeat (rewrite msubst_trunst).
      pose proof (IHHT1 c H envL envR V).
      pose proof (IHHT2 c H envL envR V).
      destruct (normalization _ _ (eq_typable_empty_l H0)) as [v1  [? ?]];
      destruct (normalization _ _ (eq_typable_empty_r H0)) as [v1' [? ?]].
      destruct (normalization _ _ (eq_typable_empty_l H1)) as [v2  [? ?]];
      destruct (normalization _ _ (eq_typable_empty_r H1)) as [v2' [? ?]].
      assert (HO': term_eq (TMonadic (MNonReact (MStateT TS E SM)) T) v1 v1') by (eapply eq_square;eauto).
      assert (H1': term_eq TS v2 v2') by (eapply eq_square;eauto).
      split. econstructor; [apply (eq_typable_empty_l H0)|
                           apply (eq_typable_empty_l H1)].
      split. econstructor; [apply (eq_typable_empty_r H0)|
                           apply (eq_typable_empty_r H1)].
      intros.
      destruct HO' as [? [? ?]]. 
      pose proof (rtc_Trans _ _ _ 
                 (Congruence_RunSt1 _ _ (msubst envL t2) H2)
                 (Congruence_RunSt2 _ _ _ H3 H6)).
      pose proof (rtc_Trans _ _ _ 
                 (Congruence_RunSt1 _ _ (msubst envR t2) H4)
                 (Congruence_RunSt2 _ _ _ H5 H8)).
      assert (HStoeq: store_eq (MNonReact (MStateT TS E SM)) (v2 :: Sto1) (v2' :: Sto2)).
             destruct H10 as [? [? [? [? ?]]]].
             split. constructor. apply (eq_typable_empty_l H1').
             auto.
             split. constructor. auto. auto.
             split. constructor. apply (eq_typable_empty_r H1').
             auto.
             split. constructor;auto.
             auto.
             simpl.
             exists v2, v2', Sto1, Sto2.
             auto.
      destruct (H13 _ _ HStoeq) as [t3 [Sto3 [t3' [Sto3' [? [? [? [? [? ?]]]]]]]]].
      simpl in H20.
      destruct H20 as [? [? [? [? [t4 [t4' [Sto4 [Sto4' [Htermeq [Hstoeq [Hid1 Hid2]]]]]]]]]]].
      subst.
      destruct (normalization _ _ (eq_typable_empty_l Htermeq)) as [v4  [? ?]];
      destruct (normalization _ _ (eq_typable_empty_r Htermeq)) as [v4' [? ?]].
      assert (Htermeq': term_eq TS v4 v4') by (eapply eq_square;eauto).
      assert (HV3: value (treturn (MNonReact (MStateT TS E SM)) t3)) by (constructor;auto).
      assert (HV3': value (treturn (MNonReact (MStateT TS E SM)) t3')) by (constructor;auto).
      pose proof (rtc_Trans_mo _ _ _
                 (rtc_Trans_mo _ _ _
                 (Congruence_LC _ _ Sto1 H14)
                 (rtc_Trans_mo _ _ _
                 (Congruence_STM_RunSt v1 _ v2 t4 Sto1 Sto4 H7 (store_eq_all_values_l H10) H16)
                 (Congruence_LC _ _  Sto4 (Congruence_RunSt2 _ _ _ HV3 H25)))) 
                 (Congruence_STM_RunStRet t3 v4 Sto4 TS E SM H18 H26)) as HS1.
      pose proof (rtc_Trans_mo _ _ _
                 (rtc_Trans_mo _ _ _
                 (Congruence_LC _ _ Sto2 H15)
                 (rtc_Trans_mo _ _ _
                 (Congruence_STM_RunSt v1' _ v2' t4' Sto2 Sto4' H9 (store_eq_all_values_r H10) H17)
                 (Congruence_LC _ _  Sto4' (Congruence_RunSt2 _ _ _ HV3' H27))))  
                 (Congruence_STM_RunStRet t3' v4' Sto4' TS E SM H19 H28)) as HS1'.
      assert (term_eq (TProd T TS) (tpair t3 v4) (tpair t3' v4')) by (
                rewrite <- (TPair_Term_eq T TS t3 v4 t3' v4'); split; auto).
      exists (tpair t3 v4), Sto4, (tpair t3' v4'), Sto4'.
             repeat (split; auto). 
             inv H20. inv H22.
             inv H23. inv H24.
 Case "trunid".
    repeat (rewrite msubst_trunid).
    pose proof (IHHT c H envL envR V).
    destruct (normalization _ _ (eq_typable_empty_l H0)) as [v1  [? ?]];
    destruct (normalization _ _ (eq_typable_empty_r H0)) as [v1' [? ?]].
    assert (H0':term_eq (TMonadic (MNonReact MIdentity) T) v1 v1') by (eapply eq_square;eauto).
    destruct H0' as [? [? ?]].
    assert (store_eq (MNonReact MIdentity) nil nil).
           repeat (split;auto).
                  constructor.
                  constructor.
    destruct (H7 _ _ H8) as [t1 [Sto1 [t2 [Sto2 [? [? [? [? [? ?]]]]]]]]].
    destruct H13 as [? [? [? [? ?]]]].
    inv H13. inv H16. 
    pose proof (rtc_Trans _ _ _
               (rtc_Trans _ _ _
               (Congruence_RunId _ _ H1)
               (Congruence_RunIdMo _ _ H9))
               (rtc_R _ _ (ST_RunIdRet (MNonReact MIdentity) _ H11))).
    pose proof (rtc_Trans _ _ _
               (rtc_Trans _ _ _
               (Congruence_RunId _ _ H3)
               (Congruence_RunIdMo _ _ H10))
               (rtc_R _ _ (ST_RunIdRet (MNonReact MIdentity) _ H12))).
    eapply eq_square_rev. 
           constructor. apply (eq_typable_empty_l H0).
           constructor. apply (eq_typable_empty_r H0).
           eapply H19.
           eapply H20.
           auto.
 Case "tpause".
    repeat (rewrite msubst_tpause).
    pose proof (IHHT c H envL envR V). 
    eapply TPause_Term_eq;eauto.
 Case "tunfold".
    repeat (rewrite msubst_tunfold).
    pose proof (IHHT1 c H envL envR V).
    pose proof (IHHT2 c H envL envR V).
    destruct (normalization _ _ (eq_typable_empty_l H0)) as [v1  [? ?]];
    destruct (normalization _ _ (eq_typable_empty_r H0)) as [v1' [? ?]].
    destruct (normalization _ _ (eq_typable_empty_l H1)) as [v2  [? ?]];
    destruct (normalization _ _ (eq_typable_empty_r H1)) as [v2' [? ?]].
    pose proof (rtc_Trans _ _ _
               (Congruence_ST_Unfold1 (MReactT TI TO SM) TA TB _ _ (msubst envL t2) H2)
               (Congruence_ST_Unfold2 (MReactT TI TO SM) TA TB _ _ _ H3 H6)).
    pose proof (rtc_Trans _ _ _
               (Congruence_ST_Unfold1 (MReactT TI TO SM) TA TB _ _ (msubst envR t2) H4)
               (Congruence_ST_Unfold2 (MReactT TI TO SM) TA TB _ _ _ H5 H8)).
    assert (H0' : term_eq TB v1 v1') by (eapply eq_square;eauto).
    assert (H1' : term_eq (TArrow TB
                            (TMonadic (MNonReact SM) (TSum TA (TProd TO (TArrow TI TB))))) v2 v2') by (eapply eq_square;eauto).     
    split. constructor; [apply (eq_typable_empty_l H0) |
                         apply (eq_typable_empty_l H1) ].
    split. constructor; [apply (eq_typable_empty_r H0) |
                         apply (eq_typable_empty_r H1) ].
    intros. pose proof (Congruence_LC _ _ Sto1 H10).
            pose proof (Congruence_LC _ _ Sto2 H11).
    exists (tunfold (MReactT TI TO SM) TA TB v1 v2), Sto1, (tunfold (MReactT TI TO SM) TA TB v1' v2'), Sto2.
           split. auto.
           split. auto.
           split. (constructor;auto).
           split. (constructor;auto).
           eapply tunfold_along2;eauto.
 Case "trunre". 
      repeat (rewrite msubst_trunre).
      pose proof (IHHT c H envL envR V).
      destruct (normalization _ _ (eq_typable_empty_l H0)) as [v  [? ?]];
      destruct (normalization _ _ (eq_typable_empty_r H0)) as [v' [? ?]].
      assert (H0' : term_eq (TMonadic (MReactT TI TO SM) T) v v') by (eapply eq_square;eauto).
      destruct H0' as [? [? ?]]. 
      split. constructor; apply (eq_typable_empty_l H0).
      split. constructor; apply (eq_typable_empty_r H0).
      intros.
      pose proof (store_eq_non_react Sto1 Sto2 TI TO SM H8).
      destruct (H7 _ _ H9) as [v1 [Sto1' [v2 [Sto2' [? [? [? [? ?]]]]]]]].
      assert (Halong: along2_react (store_eq (MReactT TI TO SM)) 
                          (term_eq TI) 
                          (term_eq TO)
                          (term_eq T) (v,Sto1) (v',Sto2)) by ( 
                          eapply step_preserves_along2_react_star;eauto). 
      assert (HTv : (v, Sto1) |> TMonadic (MReactT TI TO SM) T) by (destruct H9 as [? [? ?]];constructor; auto).
      assert (HTv': (v',Sto2) |> TMonadic (MReactT TI TO SM) T) by (destruct H9 as [? [? [? [? ?]]]];constructor; auto).
      inv Halong.
  SCase "going to return".
        assert ((TI0 = TI) /\ (TO0 = TO) /\ (SM0 = SM)).
               pose proof (preservation_mo_star _ _ _ _ _ H19 HTv).
               inv H15. inv H22.  
        destruct H15 as [? [? ?]];subst.
        pose proof (rtc_Trans_mo _ _ _
                   (rtc_Trans_mo _ _ _
                   (Congruence_LC _ _ Sto1 (Congruence_ST_RunRe T _ _ H1))
                   (Congruence_STM_RunRe T _ _ Sto1 Sto1'0 H2 H19))
                   (Congruence_STM_RunReRet _ TI TO SM T Sto1'0 H25)).
        pose proof (rtc_Trans_mo _ _ _
                   (rtc_Trans_mo _ _ _
                   (Congruence_LC _ _ Sto2 (Congruence_ST_RunRe T _ _ H3))
                   (Congruence_STM_RunRe T _ _ Sto2 Sto2'0 H4 H20))
                   (Congruence_STM_RunReRet _ TI TO SM T Sto2'0 H26)).      
        exists (tinl (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))) t1'), Sto1'0,
               (tinl (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T))) t2'), Sto2'0.
               split. auto.
               split. auto.
               split. (constructor;auto).
               split. (constructor;auto).
               split.
               destruct H28 as [? [? [? [? ?]]]].
               split. inv H17.
               split. auto. 
               split. inv H21.
               split. auto.
               auto.
               split. constructor; apply (eq_typable_empty_l H27).
               split. constructor; apply (eq_typable_empty_r H27).
               left. exists t1', t2'.
                     split. eauto.
                     split. eauto.
                     auto.
  SCase "going to pause". 
       assert (TI = TI0 /\ TO = TO0 /\ SM = SM0 /\ T = T0).
              pose proof (preservation_mo_star _ _ _ _ _ H19 HTv).
              inv H15. inv H30. 
       destruct H15 as [HRe1 [HRe2 [HRe3 HRe4]]];
             rewrite <- HRe1 in *; clear HRe1;
             rewrite <- HRe2 in *; clear HRe2;
             rewrite <- HRe3 in *; clear HRe3;
             rewrite <- HRe4 in *; clear HRe4.
        do 4 eexists. 
           split. 
           eapply rtc_Trans_mo.
                  eapply Congruence_LC.
                  eapply Congruence_ST_RunRe. 
                  eauto. 
           eapply rtc_Trans_mo. 
                  eapply Congruence_STM_RunRe.
                  auto.
                  eapply H19.
           eapply rtc_Trans_mo.
                  eapply Congruence_STM_RunRePause.
                  auto.
           eapply rtc_Trans_mo. 
                  eapply Congruence_Mo_Bind1. 
                         constructor. 
                  eapply H27. 
           eapply rtc_Trans_mo. 
                  eapply Congruence_BindRet. 
                         constructor;auto.
                  constructor. 
           eapply rtc_Trans_mo.
                  eapply Congruence_AppAbs_Mo.
                         constructor;auto. 
                         simpl. 
                         rewrite_subst.
                         eapply rtc_refl.
           split. 
           eapply rtc_Trans_mo.
                  eapply Congruence_LC.
                         eapply Congruence_ST_RunRe. 
                         eauto. 
           eapply rtc_Trans_mo. 
                  eapply Congruence_STM_RunRe.
                  auto.
                  eapply H20.
           eapply rtc_Trans_mo.
           eapply Congruence_STM_RunRePause.
                  auto.
           eapply rtc_Trans_mo. 
                  eapply Congruence_Mo_Bind1. 
                         constructor. 
                  eapply H28. 
           eapply rtc_Trans_mo. 
                  eapply Congruence_BindRet. 
                         constructor;auto.
                  constructor. 
           eapply rtc_Trans_mo.
                  eapply Congruence_AppAbs_Mo.
                         constructor;auto. 
                  simpl. 
                  rewrite_subst.
                  eapply rtc_refl.
                   
       split. constructor;auto.
       split. constructor;auto.
       split. destruct H35 as [? [? [? [? ?]]]].
              repeat (split;auto). 
              inv H15. inv H17. 
       assert ((treturn (MNonReact SM) (tpair vl vr), Sto1'') |> TMonadic (MNonReact SM)
                         (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
                         destruct H34 as [? [? [? [? ?]]]].
                         eapply preservation_mo_star. eapply H27.
                         constructor.
                         pose proof (preservation_mo_star _ _ _ _ _ H19 HTv).
                         inversion H15. inversion H30. inversion H43.
                         auto. auto.
                         inv H15.
       assert ((treturn (MNonReact SM) (tpair vl' vr'), Sto2'') |> TMonadic (MNonReact SM)
                         (TProd TO (TArrow TI (TMonadic (MReactT TI TO SM) T)))).
                         destruct H34 as [? [? [? [? ?]]]].
                         eapply preservation_mo_star. eapply H28.
                         constructor.
                         pose proof (preservation_mo_star _ _ _ _ _ H20 HTv').
                         inversion H16. inversion H31. inversion H44. 
                         auto. auto. 
                         inv H18.
       split. constructor. inversion H15. inv H31.
       split. constructor. inversion H16. inv H31. 
       right. do 2 eexists.
                 split. eapply rtc_refl.
                 split. eapply rtc_refl.
       split. inversion H15. inv H31.
       split. inversion H16. inv H31.
              do 4 eexists. 
                 split. eapply rtc_refl.
                 split. eapply rtc_refl.
                 split. auto. 
                 split. inversion H15. inversion H31. inv H40.
                 split. inversion H16. inversion H31. inv H40.
              intros. 
              do 2 eexists.
                 split. eapply rtc_refl.
                 split. eapply rtc_refl. 
              assert (HT1: \empty |-- tapp vr t1 \in TMonadic (MReactT TI TO SM) T).
                     econstructor.
                          inversion H15. inversion H32. inv H41. 
                          eauto.
                          eapply eq_typable_empty_l;eauto. 
              assert (HT2: \empty |-- tapp vr' t1'0 \in TMonadic (MReactT TI TO SM) T).
                     econstructor.
                          inversion H16. inversion H32. inv H41. 
                          eauto.
                          eapply eq_typable_empty_r;eauto.
             destruct (normalization _ _ HT1) as [vapp1 [? ?]];
             destruct (normalization _ _ HT2) as [vapp2 [? ?]].
             do 2 (split;auto); intros.
                  pose proof (H36 _ _ _ _ H17 H32).
             do 4 eexists. 
                  split. eapply Congruence_LC; eauto.
                  split. eapply Congruence_LC; eauto.
                  split. auto.
                  split. auto.
                  eapply step_preserves_along2_react_star'.
                         eapply Congruence_LC; eauto.
                         eapply Congruence_LC; eauto.
                  auto. 
Qed.

Corollary term_eq_refl : forall T t,
          \empty |-- t \in T ->
          term_eq T t t.
Proof.
  intros T t HTt.
  replace t with (msubst nil t).
  assert (mextend \empty nil = \empty) as H; auto; rewrite <- H in HTt; clear H.
  eapply msubst_eq.
         eassumption.
         constructor.
         auto.   
Qed.

Corollary store_eq_refl : forall Sto M,
          store_all_values Sto   ->
          store_matches_mo Sto M -> store_eq M Sto Sto.
Proof.
  intros Sto M HSAV HStoM. generalize dependent M Sto.
  induction M; intros; simpl; repeat split; auto;
  generalize dependent s Sto; induction s; intros; try (simpl;auto;fail);
  inv HStoM.
Case "MReactT". 
     inv H1.
     pose proof (term_eq_refl _ _ H3).
     simpl.
     exists t2, t2, Sto0, Sto0.
            repeat (split;auto).
            eapply IHs.
            inv HSAV.
            constructor; auto.
Case "MNonReact".
     pose proof (term_eq_refl _ _ H2).
     simpl.
     exists t0, t0, Sto0, Sto0.
            repeat (split;auto).
            eapply IHs.
            inv HSAV.
            auto.
Qed.

Corollary sm_eq_refl : forall Sto SM,
          store_all_values Sto                ->
          store_matches_mo Sto (MNonReact SM) ->  sm_eq SM Sto Sto.              
Proof. intros; destruct (store_eq_refl _ _ H H0) as [? [? [? [? ?]]]]; auto. Qed.

Theorem along2_refl : forall TI TO TR SM co, 
                        co |> TMonadic (MReactT TI TO SM) TR ->
                        along2_react (store_eq (MReactT TI TO SM)) (term_eq TI) (term_eq TO) (term_eq TR) co co.
Proof. intros ? ? ? ? (t,Sto) HTco.
       inv HTco.
       destruct (term_eq_refl _ _ H3) as [_ [_ ?]].
       destruct (H _ _ (store_eq_refl _ _ H4 H5)) as [? [? [? [? [? [? [? [? ?]]]]]]]].
       eapply step_preserves_along2_react_star.
              eapply H0.
              eapply H1.
              auto.
Qed.

(** Work on rewrite stuff *)


Theorem term_eq_extensionality : forall (X Y : Ty) (f g : tm),
          \empty |-- f \in TArrow X Y ->
          \empty |-- g \in TArrow X Y ->
   (forall x : tm, \empty |-- x \in X ->
             tapp f x =[Y]= tapp g x) ->  f =[TArrow X Y]= g.
Proof.
  intros.
  repeat (split;auto).
  intros.
  pose proof (H1 _ (eq_typable_empty_l H2)).
  pose proof (H1 _ (eq_typable_empty_r H2)).
  assert (tapp f t1=[Y]=tapp f t1').
         eapply TApp_Term_eq. 
                eapply term_eq_refl;eauto.
                auto.
  assert (tapp f t1=[Y]=tapp g t1') by (eapply term_eq_trans;eauto).
  do 2 eexists;repeat (split;auto).
Qed.  

Theorem wt_coqeq : forall (T : Ty) (t1 t2 : tm),
        \empty |-- t1 \in T -> \empty |-- t2 \in T -> 
        t1 = t2 -> t1 =[T]= t2.
Proof. intros. subst. eapply term_eq_refl;eauto. Qed.


Theorem multistep_eq : forall T t t',
        \empty |-- t \in T ->
                 t ~~>* t' ->
               term_eq T t t'.
Proof.
  intros.
  apply term_eq_sym.
  eapply step_preserves_eq'.
  eapply H0.
  eapply term_eq_refl;auto.
Qed.


Require Import Relations.
Require Import Morphisms.
Require Import Setoid.

Instance term_eq_symmetric : forall T : Ty, (@Symmetric _ (term_eq T)).
Proof. red; apply term_eq_sym.  
Defined. 

Instance term_eq_transitive : forall T : Ty, (@Transitive _ (term_eq T)).
Proof. red; apply term_eq_trans.
Defined.

Instance term_eq_PER : forall T : Ty, (@PER _ (term_eq T)).
Proof. split; [apply term_eq_symmetric|apply term_eq_transitive].
Defined.

Add Parametric Relation (T:Ty) : tm (term_eq T)
  symmetry proved by symmetry
  transitivity proved by transitivity
  as TermEq_Rel.

Instance store_eq_symmetric : forall M : Mo, (@Symmetric _ (store_eq M)).
Proof. red; apply store_eq_sym.  
Defined. 

Instance store_eq_transitive : forall M : Mo, (@Transitive _ (store_eq M)).
Proof. red; apply store_eq_trans.
Defined.

Instance store_eq_PER : forall M : Mo, (@PER _ (store_eq M)).
Proof. split; [apply store_eq_symmetric|apply store_eq_transitive].
Defined.

Add Parametric Relation (M:Mo) : store (store_eq M)
  symmetry proved by symmetry
  transitivity proved by transitivity
  as StoreEq_Rel.

Instance sm_eq_symmetric : forall SM : SMo, (@Symmetric _ (sm_eq SM)).
Proof. red; apply sm_eq_sym.  
Defined. 

Instance sm_eq_transitive : forall SM : SMo, (@Transitive _ (sm_eq SM)).
Proof. red; apply sm_eq_trans.
Defined.

Instance sm_eq_PER : forall SM : SMo, (@PER _ (sm_eq SM)).
Proof. split; [apply sm_eq_symmetric|apply sm_eq_transitive].
Defined.

Add Parametric Relation (SM:SMo) : store (sm_eq SM)
  symmetry proved by symmetry
  transitivity proved by transitivity
  as SmEq_Rel.

Instance along2_symmetric :  forall M TI TO TR, (@Symmetric _
      (@along2_react (store_eq M) (term_eq TI) (term_eq TO) (term_eq TR))).
Proof. red; apply along2_react_sym. 
Defined.

Instance along2_transitive :  forall M TI TO TR, (@Transitive _
      (@along2_react (store_eq M) (term_eq TI) (term_eq TO) (term_eq TR))).
Proof. red; apply along2_react_trans. 
Defined.

Instance along2_PER :  forall M TI TO TR, (@PER _
      (@along2_react (store_eq M) (term_eq TI) (term_eq TO) (term_eq TR))).
Proof. split; [apply along2_symmetric |apply along2_transitive]. 
Defined.

Add Parametric Relation (M : Mo) (TI TO TR : Ty) : configuration (along2_react (store_eq M) (term_eq TI) (term_eq TO) (term_eq TR))
  symmetry proved by symmetry
  transitivity proved by transitivity
  as Along2Eq_Rel.


Hint Resolve term_eq_refl store_eq_refl sm_eq_refl along2_refl : core.